package blas;

import lp.Complex;
/**
 * Basic Linear Algebra Subroutines
 * @author Linpack, adapted by Marat
 */
public class Blas 
{
//------------------------------------------------------------------------------------------------
//	Assorted  routines adapted from Linpack follow
//	Two overloaded versions are available for most routines.  The first uses Complex variables, 
//	while the second used a pair of real (double) variables.  The second method is much faster
//	for some things.
//------------------------------------------------------------------------------------------------
	public static double cabs1(Complex x)
	{
		return (Math.abs(x.r) + Math.abs(x.i));
	}

	public static double cabs1(double xr, double xi)
	{
		return (Math.abs(xr) + Math.abs(xi));
	}

//-------------------------------------------------------------------------------------------------
	public static double cabs(Complex x)
	{
		return (Math.sqrt(x.r * x.r + x.i * x.i));
	}

	public static double cabs(double xr, double xi)
	{
		return (Math.sqrt(xr * xr + xi * xi));
	}
//-------------------------------------------------------------------------------------------------

	public static Complex csign(Complex x1, Complex x2)
	{
		return (Complex.mult(cabs(x1), Complex.div(x2, cabs(x2))));
	}

//-------------------------------------------------------------------------------------------------
// This is a modification of the LINPACK subroutine
// DSCAL.  In the LINPACK listing DSCAL is attributed to Jack Dongarra
// with a date of 3/11/78.

/// <summary>
/// This method scales a portion of a column of a matrix by a constant.  
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="a">The constant</param>
/// <param name="x">The matrix: x[][]</param>
/// <param name="begin">The starting row</param>
/// <param name="j">The id of the column</param>
	public static void colscal(int nrow, Complex a, Complex[][] x, int begin, int j)
	{
		int i,  end;

		if (nrow <= 0) return;

		end = begin + nrow - 1;

		for (i = begin; i <= end; i++)
		{			 
			x[i][j] = Complex.mult(x[i][j], a);
		}
		return;
	}

/// <summary>
/// This method scales a portion of a column of a matrix by a constant.  
/// </summary>
/// <param name="nrow">the number of rows involved</param>
/// <param name="ar">real part of the constant</param>
/// <param name="ai">imaginary part of the constant</param>
/// <param name="xr">rear array</param>
/// <param name="xi">imaginary array</param>
/// <param name="begin">the starting row</param>
/// <param name="j">the id of the column</param>
	public static void colscal(int nrow, double ar, double ai, double[][] xr, double[][] xi, 
			int begin, int j)
	{
		int i, end;
		double tempr = 0.0, tempi = 0.0;

		if (nrow <= 0) return;

		end = begin + nrow - 1;

		for (i = begin; i <= end; i++)
		{
			tempr = (xr[i][j] * ar) - (xi[i][j] * ai);
			tempi = (xr[i][j] * ai) + (xi[i][j] * ar);
			xr[i][j] = tempr;
			xi[i][j] = tempi;
		}
		return;
	}       

//-------------------------------------------------------------------------------------------------
// This is a modification of the LINPACK subroutine
// DSCAL.  In the LINPACK listing DSCAL is attributed to Jack Dongarra
// with a date of 3/11/78.

/// <summary>
/// This method scales a portion of a vector by a constant 
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="a">The constant</param>
/// <param name="x">The vector x[]</param>
/// <param name="begin">the row to begin</param>
	public static void dscalp (int nrow, Complex a, Complex[] x, int begin) 
	{
		int i, end;

		if (nrow <= 0) return;

		end = begin + nrow - 1;

		for (i = begin; i <= end; i++) 
		{
			x[i] = Complex.mult(x[i], a);
		}
		return;
	}

/// <summary>
/// This method scales a portion of a vector by a constant
/// </summary>
/// <param name="nrow">the number of rows involved</param>
/// <param name="ar">real part of the constant</param>
/// <param name="ai">imaginary part of the constant</param>
/// <param name="xr">real part of the vector</param>
/// <param name="xi">imaginary part of the vector</param>
/// <param name="begin">the row to begin</param>
	public static void dscalp(int nrow, double ar, double ai, double[] xr, double[] xi, 
			int begin)
	{

		int i, end;
		double prod_r, prod_i;

		if (nrow <= 0) return;

		end = begin + nrow - 1;

		for (i = begin; i <= end; i++)
		{
			prod_r = ar * xr[i] - ai * xi[i];
			prod_i = ar * xi[i] + ai * xr[i];
			xr[i] = prod_r;        xi[i] = prod_i;
		}
		return;
	}        

//-------------------------------------------------------------------------------------------------
// This is a modification of the LINPACK function
// dnrm2.
// In the LINPACK listing dnrm2 is attributed to C.L. Lawson
// with a date of January 8, 1978.  The routine below is based
// on a more recent dnrm2 version that is attributed in LAPACK
// documentation to Sven Hammarling.

/// <summary>
/// This method calculates the Euclidean norm of a portion of a column of a matrix.
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="x">The matrix x[][]</param>
/// <param name="begin">The starting row</param>
/// <param name="j">The id of the column</param>
/// <returns>The Euclidean norm</returns>
	public static double colnrm2 (int nrow, Complex[][] x, int begin, int j) 
	{
		double absxij, norm, scale, ssq, fac;
		int i, end;

		if (nrow < 1) 
		{
			norm = 0.0;
		}
		else if (nrow == 1)
		{
			norm = Math.abs(x[begin][j].r); // handle imag part ???
		} 
		else 
		{
			scale = 0.0;
			ssq = 1.0;

			end = begin + nrow - 1;

			for (i = begin; i <= end; i++) 
			{
				if (x[i][j].r != 0.0) 
				{
					absxij = Math.abs(x[i][j].r);

					if (scale < absxij) 
					{
						fac = scale/absxij;
						ssq = 1.0 + ssq*fac*fac;
						scale = absxij;
					} 
					else 
					{
						fac = absxij/scale;
						ssq += fac * fac;
					}
				}

				if (x[i][j].i != 0.0)
				{
					absxij = Math.abs(x[i][j].i);

					if (scale < absxij)
					{
						fac = scale / absxij;
						ssq = 1.0 + ssq * fac * fac;
						scale = absxij;
					}
					else
					{
						fac = absxij / scale;
						ssq += fac * fac;
					}
				}
			}
			norm = scale * Math.sqrt(ssq);
		}
		return norm;
	}

/// <summary>
/// This method calculates the Euclidean norm of a portion of a column of a matrix.
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="xr">The real part of the vector</param>
/// <param name="xi">The imaginary part of the vector</param>
/// <param name="begin">The starting row</param>
/// <param name="j">The id of the column</param>
/// <returns>The Euclidean norm</returns>
	public static double colnrm2(int nrow, double[][] xr, double[][] xi, 
			int begin, int j)
	{
		double absxij, norm, scale, ssq, fac;
		int i, end;

		if (nrow < 1)
		{
			norm = 0.0;
		}
		else if (nrow == 1)
		{
			norm = Math.abs(xr[begin][j]); // handle imag part ???
		}
		else
		{
			scale = 0.0;
			ssq = 1.0;

			end = begin + nrow - 1;

			for (i = begin; i <= end; i++)
			{
				if (xr[i][j] != 0.0)
				{
					absxij = Math.abs(xr[i][j]);

					if (scale < absxij)
					{
						fac = scale / absxij;
						ssq = 1.0 + ssq * fac * fac;
						scale = absxij;
					}
					else
					{
						fac = absxij / scale;
						ssq += fac * fac;
					}
				}

				if (xi[i][j] != 0.0)
				{
					absxij = Math.abs(xi[i][j]);

					if (scale < absxij)
					{
						fac = scale / absxij;
						ssq = 1.0 + ssq * fac * fac;
						scale = absxij;
					}
					else
					{
						fac = absxij / scale;
						ssq += fac * fac;
					}
				}
			}
			norm = scale * Math.sqrt(ssq);
		}
		return norm;
	} 

//-------------------------------------------------------------------------------------------------
// This is a modification of the LINPACK function dnrm2.
// In the LINPACK listing dnrm2 is attributed to C.L. Lawson
// with a date of January 8, 1978.  The routine below is based
// on a more recent dnrm2 version that is attributed in LAPACK
// documentation to Sven Hammarling.

/// <summary>
///  This method calculates the Euclidean norm of a portion of a vector x[ ]
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="x">vector x[]</param>
/// <param name="begin">The starting row</param>
/// <returns>The Euclidean norm</returns>
       public static double cnrm2p (int nrow, Complex[] x, int begin) 
       {
          double absxi, norm, scale, ssq, fac;
          int i, end;

          if (nrow < 1) 
	      {
             norm = 0.0;
          } 
	      else if (nrow == 1) 
	      {
             norm = Math.abs(x[begin].r);		// handle imag part ???
          } 
	      else 
	      {
             scale = 0.0;
             ssq = 1.0;

             end = begin + nrow - 1;

             for (i = begin; i <= end; i++) 
		     {
                if (x[i].r != 0.0) 
			    {
                   absxi = Math.abs(x[i].r);

                   if (scale < absxi) 
			       {
                      fac = scale/absxi;
                      ssq = 1.0 + ssq*fac*fac;
                      scale = absxi;
                   } 
			       else 
			       {
                      fac = absxi/scale;
                      ssq += fac*fac;
                   }
                }
			    if (x[i].i != 0.0)
			    {
				    absxi = Math.abs(x[i].i);

				    if (scale < absxi)
				    {
					    fac = scale / absxi;
					    ssq = 1.0 + ssq * fac * fac;
					    scale = absxi;
				    }
				    else
				    {
					    fac = absxi / scale;
					    ssq += fac * fac;
				    }
			    }
             }
             norm = scale * Math.sqrt(ssq);
          }
          return norm;
       }

/// <summary>
///  This method calculates the Euclidean norm of a portion of a vector x[ ]
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="xr">real part of vector x[]</param>
/// <param name="xi">imaginary part of vector x[]</param>
/// <param name="begin">The starting row</param>
/// <returns>The Euclidean norm</returns>
       public static double cnrm2p(int nrow, double[] xr, double[] xi, int begin)
       {
    	   double absxi, norm, scale, ssq, fac;
    	   int i, end;

    	   if (nrow < 1)
    	   {
    		   norm = 0.0;
    	   }
    	   else if (nrow == 1)
    	   {
    		   norm = Math.abs(xr[begin]);		// handle imag part ???
    	   }
    	   else
    	   {
    		   scale = 0.0;
    		   ssq = 1.0;

    		   end = begin + nrow - 1;

    		   for (i = begin; i <= end; i++)
    		   {
    			   if (xr[i] != 0.0)
    			   {
    				   absxi = Math.abs(xr[i]);

    				   if (scale < absxi)
    				   {
    					   fac = scale / absxi;
    					   ssq = 1.0 + ssq * fac * fac;
    					   scale = absxi;
    				   }
    				   else
    				   {
    					   fac = absxi / scale;
    					   ssq += fac * fac;
    				   }
    			   }
    			   if (xi[i] != 0.0)
    			   {
    				   absxi = Math.abs(xi[i]);

    				   if (scale < absxi)
    				   {
    					   fac = scale / absxi;
    					   ssq = 1.0 + ssq * fac * fac;
    					   scale = absxi;
    				   }
    				   else
    				   {
    					   fac = absxi / scale;
    					   ssq += fac * fac;
    				   }
    			   }
    		   }
    		   norm = scale * Math.sqrt(ssq);
    	   }
    	   return norm;
       } 

//-------------------------------------------------------------------------------------------------  
// This is a modification of the LINPACK function
// DDOT.  In the LINPACK listing DDOT is attributed to Jack Dongarra
// with a date of 3/11/78.

/// <summary>
/// This method calculates the dot product of portions of two columns of a matrix. 
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="x">The matrix: x[][]</param>
/// <param name="begin">The starting row</param>
/// <param name="j1">The id of the first column</param>
/// <param name="j2">The id of the second column</param>
/// <returns>the dot product</returns>
       public static Complex coldot (int nrow, Complex[][] x, int begin,
    		   int j1, int j2)
       {
    	   Complex coldot_ret = new Complex(0.0, 0.0);
    	   int i,  end;

    	   if (nrow <= 0) return coldot_ret;

    	   end = begin + nrow - 1;

    	   for (i = begin; i <= end; i++) 
    	   {
    		   //coldot_ret += Complex.conjf(x[i][j1]) * x[i][j2];
    		   coldot_ret = Complex.add(coldot_ret, Complex.mult(Complex.conjf(x[i][j1]), x[i][j2]));
    	   }
    	   return coldot_ret;
       }

       public static Complex coldot(int nrow, double[][] xr, double[][] xi, int begin,
    		   int j1, int j2)
       {
    	   Complex coldot_ret = new Complex(0.0, 0.0);
    	   int i, end;

    	   if (nrow <= 0) return coldot_ret;

    	   end = begin + nrow - 1;

    	   for (i = begin; i <= end; i++)
    	   {
    		   double xcr = xr[i][j1]; double xci = -xi[i][j1];  // complex conj of lhs

    		   double tr, ti;

    		   tr = xcr * xr[i][j2] - xci * xi[i][j2];
    		   ti = xcr * xi[i][j2] + xci * xr[i][j2];
    		   //coldot_ret += new Complex(tr, ti);
    		   coldot_ret = Complex.add(coldot_ret, new Complex(tr, ti));
    	   }
    	   return coldot_ret;
       } 

//-------------------------------------------------------------------------------------------------
// This is a modification of the LINPACK subroutine
// DAXPY.  In the LINPACK listing DAXPY is attributed to Jack Dongarra
// with a date of 3/11/78.

/// <summary>
///  This method multiplies a constant times a portion of a column
/// of a matrix and adds the product to the corresponding portion
/// of another column of the matrix --- a portion of col2 is 
/// replaced by the corresponding portion of a*col1 + col2.
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="a">The constant</param>
/// <param name="x">The matrix</param>
/// <param name="begin">The starting row</param>
/// <param name="j1">The id of col1</param>
/// <param name="j2">The id of col2</param>
       public static void colaxpy (int nrow, Complex a, Complex[][] x, int begin,
    		   int j1, int j2) 
       {
    	   int i, end;

    	   if (nrow <= 0) return;
    	   if ((Math.abs(a.r) + Math.abs(a.i)) == 0.0) return;

    	   end = begin + nrow - 1;

    	   for (i = begin; i <= end; i++) 
    	   {
    		   //x[i][j2] += a*x[i][j1];
    		   x[i][j2] = Complex.add(x[i][j2], Complex.mult(a, x[i][j1]));
    	   }

    	   return;
       }


/// <summary>
///  This method multiplies a constant times a portion of a column
/// of a matrix and adds the product to the corresponding portion
/// of another column of the matrix --- a portion of col2 is 
/// replaced by the corresponding portion of a*col1 + col2.
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="ar">Real part of the constant</param>
/// <param name="ai">Imaginary part of the constant</param>
/// <param name="xr">Real part of the matrix</param>
/// <param name="xi">Imaginary part of the matrix</param>
/// <param name="begin">The starting row</param>
/// <param name="j1">The id of col1</param>
/// <param name="j2">The id of col2</param>
       public static void colaxpy(int nrow, double ar, double ai, double[][] xr, double[][] xi, 
    		   int begin, int j1, int j2)
       {
    	   int i, end;

    	   if (nrow <= 0) return;
    	   if ((Math.abs(ar) + Math.abs(ai)) == 0.0) return;

    	   end = begin + nrow - 1;

    	   for (i = begin; i <= end; i++)
    	   {
    		   double prod_r, prod_i;

    		   prod_r = ar * xr[i][j1] - ai * xi[i][j1];
    		   prod_i = ar * xi[i][j1] + ai * xr[i][j1];

    		   xr[i][j2] = xr[i][j2] + prod_r;
    		   xi[i][j2] = xi[i][j2] + prod_i;
    	   }

    	   return;
       } 
//-------------------------------------------------------------------------------------------------
// This is a modification of the LINPACK function
// DSWAP.  In the LINPACK listing DSWAP is attributed to Jack Dongarra
// with a date of 3/11/78. 

/// <summary>
/// This method interchanges two columns of a matrix
/// </summary>
/// <param name="n">The number of rows of the matrix</param>
/// <param name="x">The matrix</param>
/// <param name="j1">The id of the first column</param>
/// <param name="j2">The id of the second column</param>
       public static void colswap (int n, Complex[][] x, int j1, int j2) 
       {
    	   Complex temp;
    	   int i;

    	   if (n <= 0) return;

    	   for (i = 0; i < n; i++)
    	   {
    		   temp = x[i][j1];
    		   x[i][j1] = x[i][j2];
    		   x[i][j2] = temp;
    	   }

    	   return;
       }

/// <summary>
/// This method interchanges two columns of a matrix
/// </summary>
/// <param name="n">The number of rows of the matrix</param>
/// <param name="xr">Real part of the matrix</param>
/// <param name="xi">Imaginary part of the</param>
/// <param name="j1">The id of the first column</param>
/// <param name="j2">The id of the second column</param>
       public static void colswap(int n, double[][] xr, double[][] xi, int j1, int j2)
       {
    	   double tempr, tempi;
    	   int i;

    	   if (n <= 0) return;

    	   for (i = 0; i < n; i++)
    	   {
    		   tempr = xr[i][j1];
    		   tempi = xi[i][j1];

    		   xr[i][j1] = xr[i][j2];
    		   xi[i][j1] = xi[i][j2];

    		   xr[i][j2] = tempr;
    		   xi[i][j2] = tempi;
    	   }

    	   return;
       }
//-------------------------------------------------------------------------------------------------
// This is a translation from FORTRAN to C# of the LINPACK subroutine
// DROTG.  In the LINPACK listing DROTG is attributed to Jack Dongarra
// with a date of 3/11/78.

/// <summary>
/// This method constructs a Givens plane rotation.
/// </summary>
/// <param name="rotvec">Contains the a,b,c,s values.</param>
       public static void rotg (double[] rotvec) 
       {
    	   double a, b, c, s, roe, scale, r, z, ra, rb;

    	   a = rotvec[0];
    	   b = rotvec[1];

    	   roe = b;

    	   if (Math.abs(a) > Math.abs(b)) roe = a;

    	   scale = Math.abs(a) + Math.abs(b);

    	   if (scale != 0.0) 
    	   {
    		   ra = a/scale;
    		   rb = b/scale;
    		   r = scale*Math.sqrt(ra*ra + rb*rb);
    		   r = sign(1.0, roe)*r;
    		   c = a/r;
    		   s = b/r;
    		   z = 1.0;
    		   if (Math.abs(a) > Math.abs(b)) z = s;
    		   if ((Math.abs(b) >= Math.abs(a)) && (c != 0.0)) z = 1.0/c;
    	   } 
    	   else 
    	   {
    		   c = 1.0;
    		   s = 0.0;
    		   r = 0.0;
    		   z = 0.0;
    	   }

    	   a = r;
    	   b = z;

    	   rotvec[0] = a;
    	   rotvec[1] = b;
    	   rotvec[2] = c;
    	   rotvec[3] = s;

    	   return;
       } 

//-------------------------------------------------------------------------------------------------
// This is a modification of the LINPACK function
// DROT.  In the LINPACK listing DROT is attributed to Jack Dongarra
// with a date of 3/11/78. 

/// <summary>
/// This method "applies a plane rotation."
/// </summary>
/// <param name="n">The order of x[ , ]</param>
/// <param name="x">The matrix</param>
/// <param name="j1">The id of the first column</param>
/// <param name="j2">The id of the second column</param>
/// <param name="c">"cos"</param>
/// <param name="s">"sin"</param>
       public static void colrot (int n, Complex[][] x,
    		   int j1, int j2, double c, double s) 
       {
    	   Complex temp;
    	   int i;

    	   if (n <= 0) return;

    	   for (i = 0; i < n; i++) 
    	   {
    		   //temp = c * x[i][j1] + s * x [i][j2];
    		   //x[i][j2] = c * x [i][j2] - s * x [i][j1];
    		   //x[i][j1] = temp;

    		   temp = Complex.add(Complex.mult(c, x[i][j1]), Complex.mult(s, x[i][j2]));
    		   x[i][j2] = Complex.sub(Complex.mult(c, x[i][j2]), Complex.mult(s, x[i][j1]));
    		   x[i][j1].r = temp.r; 	x[i][j1].i = temp.i; 
    	   } 
    	   return;
       }

       public static void colrot(int n, double[][] xr,double[][] xi,
    		   int j1, int j2, double c, double s)
       {
    	   double tempr, tempi;
    	   int i;

    	   if (n <= 0) return;

    	   for (i = 0; i < n; i++)
    	   {


    		   tempr = c * xr[i][j1] + s * xr[i][j2];
    		   tempi = c * xi[i][j1] + s * xi[i][j2];

    		   xr[i][j2] = c * xr[i][j2] - s * xr[i][j1];
    		   xi[i][j2] = c * xi[i][j2] - s * xi[i][j1];

    		   xr[i][j1] = tempr; xi[i][j1] = tempi;
    	   }
    	   return;
       } 

//---------------------------------------------------------------------------------------------
// This is a modification of the LINPACK subroutine
// DAXPY.  In the LINPACK listing DAXPY is attributed to Jack Dongarra
// with a date of 3/11/78.

/// <summary>
/// This method multiplies a constant times a portion of a column
/// of a matrix x[ , ] and adds the product to the corresponding portion
/// of a vector y[ ] --- a portion of y[ ] is replaced by the corresponding
/// portion of ax[ ,j] + y[ ].
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="a">The constant</param>
/// <param name="x">The matrix</param>
/// <param name="y">The vector</param>
/// <param name="begin">The starting row</param>
/// <param name="j">The id of the column of the x matrix</param>
       public static void colvaxpy (int nrow, Complex a, Complex[][] x, Complex[] y,
    		   int begin, int j) 
       {
    	   int i, end;

    	   if (nrow <= 0) return;
    	   if (a.r == 0.0 && a.i == 0.0) return;


    	   end = begin + nrow - 1;

    	   for (i = begin; i <= end; i++) 
    	   {
    		   //y[i] += a * x[i][j];
    		   y[i] = Complex.add(y[i], Complex.mult(a, x[i][j]));
    	   }
    	   return;
       }

/// <summary>
/// This method multiplies a constant times a portion of a column
/// of a matrix x[ , ] and adds the product to the corresponding portion
/// of a vector y[ ] --- a portion of y[ ] is replaced by the corresponding
/// portion of ax[ ,j] + y[ ].
/// </summary>
/// <param name="nrow"></param>
/// <param name="ar">Real part of constant</param>
/// <param name="ai">Imaginary part of constant</param>
/// <param name="xr">Real part of matrix</param>
/// <param name="xi">Imaginary part of matrix</param>
/// <param name="yr">Real part of vector</param>
/// <param name="yi">Imaginary part of vector</param>
/// <param name="begin">The starting row</param>
/// <param name="j">The id of the column of x[][]</param>
       public static void colvaxpy(int nrow, double ar, double ai, 
    		   double[][] xr, double[][] xi, double[] yr,double[] yi,
    		   int begin, int j)
       {
    	   int i, end;
    	   double tempr = 0.0, tempi = 0.0;

    	   if (nrow <= 0) return;
    	   if (ar == 0.0 && ai == 0.0) return;


    	   end = begin + nrow - 1;

    	   for (i = begin; i <= end; i++)
    	   {
    		   tempr = ar * xr[i][j] - ai * xi[i][j];
    		   tempi = ar * xi[i][j] + ai * xr[i][j];
    		   yr[i] = yr[i] + tempr;
    		   yi[i] = yi[i] + tempi;
    	   }
    	   return;
       } 

//-------------------------------------------------------------------------------------------------
// This is a modification of the LINPACK subroutine
// DAXPY.  In the LINPACK listing DAXPY is attributed to Jack Dongarra
// with a date of 3/11/78.

// Translated to Java and modified by Steve Verrill, March 3, 1997.
// Translated to C# for .NET, Kirk Marat, 2008
       
/// <summary>
/// This method multiplies a constant times a portion of a vector y[ ]
/// and adds the product to the corresponding portion
/// of a column of a matrix x[ ][ ] --- a portion of column j of x[ , ] 
/// is replaced by the corresponding
/// portion of ay[ ] + x[ ][j].
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="a">The constant</param>
/// <param name="y">The vector</param>
/// <param name="x">The matrix</param>
/// <param name="begin">The starting row</param>
/// <param name="j">The id of the column of the x matrix</param>
       public static void colvraxpy (int nrow, Complex a, Complex[] y, Complex[][] x,
						    int begin, int j) 
       {
	      int i, end;

	      if (nrow <= 0) return;
	      if (a.i == 0.0 && a.r == 0.0) return;

	      end = begin + nrow - 1;

	      for (i = begin; i <= end; i++)
	      {
		     //x[i,j] += a * y[i];
	    	  x[i][j] = Complex.add(x[i][j], Complex.mult(a, y[i]));
	      }
	      return;
       }

/// <summary>
/// This method multiplies a constant times a portion of a vector y[ ]
/// and adds the product to the corresponding portion
/// of a column of a matrix x[ ][ ] --- a portion of column j of x[ , ] 
/// is replaced by the corresponding
/// portion of ay[ ] + x[ ][j].
/// </summary>
/// <param name="nrow">The number of rows involved</param>
/// <param name="ar">Real part of the constant</param>
/// <param name="ai">Imaginary part of the constant</param>
/// <param name="yr">Real part of the vector</param>
/// <param name="yi">Imaginary part of the Vector</param>
/// <param name="xr">Real part of the matrix x</param>
/// <param name="xi">Imaginary part of the matrix x</param>
/// <param name="begin">The starting row</param>
/// <param name="j">The id of the column of the x matrix</param>
       public static void colvraxpy(int nrow, double ar, double ai, double[] yr,double[] yi,
    		   double[][] xr, double[][] xi,
    		   int begin, int j)
       {
    	   int i, end;
    	   double tempr, tempi;

    	   if (nrow <= 0) return;
    	   if (ai == 0.0 && ar == 0.0) return;

    	   end = begin + nrow - 1;

    	   for (i = begin; i <= end; i++)
    	   {
    		   tempr = ar * yr[i] - ai * yi[i];
    		   tempi = ar * yi[i] + ai * yr[i];
    		   xr[i][j] = xr[i][j] + tempr;
    		   xi[i][j] = xi[i][j] + tempi;
    	   }
    	   return;
       } 

//-------------------------------------------------------------------------------------------------
//	duplication of FORTRAN SIGN() function

       public static double sign(double x, double y)
       {
    	   if (y >= 0.0)
    		   return Math.abs(x);
    	   else
    		   return -Math.abs(x);
       }

 
//--------------------------
}	// end class
