package decon;

import spinworks.Util;
import spinworks.data_par;
import spinworks.global_par;

/**
 * Deconvolution (band fitting) utilities for SpinWorks Java
 * @author marat
 */
public class DeconUtil 
{
	public static double M_LN2 = 0.693147181;	// ln(2.0)
/**
 * calculates a synthetic spectrum from a decon_peaks object. 
 * Handles mixed Lorentz/Gauss lineshape
 * @param gp workspace parameters object
 * @param q a data par object for the current spectrum
 */
	public static void genspec(global_par gp,  data_par q)
	{
		int i, j;							// indices 
		int k;								// the peak reconstruction range in pts 
		int start, end;						// starting and indices in the array
		double wl, wg, deltav, width;


		if (gp.calc_array == null) return;

		for (i = 0; i < q.SI; i++)			// zero the calculated spectrum array
			gp.calc_array[i] = 0;


		for (i = 0; i < gp.n_decon_peaks; i++ )
		{
			width = gp.decon_peaks[i].width;

			double eta = 1.0 - (gp.decon_peaks[i].gauss / 100.0);

			try
			{
				k = 60 * (int)Math.round(width / q.hzperpt);	// 60 * linewidth for each peak 			
				start = (int)Math.round((q.offset - gp.decon_peaks[i].freq)
						/ q.hzperpt) - k;
			}
			catch (Exception ex)
			{
				Util.show_and_log("genspec(): " + ex.getMessage());
				return;
			}

			wl = 1.0 / (width * width);
			wg = 1.0 / (width * width / 4);

			if (start < 0)
				start = 0;

			end = start + 2 * k;
			if (end >= q.SI)
				end = q.SI - 1;

			for (j = start; j <= end; j++)
			{
				deltav = (q.offset - j * q.hzperpt) - gp.decon_peaks[i].freq;

				double loren = eta * 
						(gp.decon_peaks[i].intensity
								/ (1.0 + 4.0 * deltav * deltav * wl));

				double gauss = (1-eta) * 
						(gp.decon_peaks[i].intensity
								* Math.exp(-M_LN2 * deltav * deltav * wg));

				gp.calc_array[j] = gp.calc_array[j] + loren + gauss;
			}				
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * calculates a synthetic spectrum of a single decon peak and add to stack
 * @param n index of the peak to draw
 * @param gp workspace parameters object
 * @param q a data par object for the current spectrum
 */
	public static void genspec(int n, global_par gp, data_par q)
	{
		int j;							    // indices 
		int k;								// the peak reconstruction range in pts 
		int start, end;
		double wl, wg, deltav, width;

		if (gp.calc_array == null) return;

		for (int i = 0; i < q.SI; i++)		// zero the calculated spectrum array
			gp.stack_trace[n].data[i] = 0.0;

		width = gp.decon_peaks[n].width;

		double eta = 1.0 - (gp.decon_peaks[n].gauss / 100.0);

		try
		{
			k = 60 * (int)Math.round(width / q.hzperpt);	// 40 * linewidth for each peak 
		}
		catch (Exception ex)
		{
			Util.show_and_log("genspec(): " + ex.getMessage());
			return;
		}

		wl = 1.0 / (width * width);
		wg = 1.0 / (width * width / 4.0);

		try
		{
			start = (int)Math.round((q.offset - gp.decon_peaks[n].freq)
					/ q.hzperpt) - k;
		}
		catch (Exception ex)
		{
			Util.show_and_log("genspec(): " + ex.getMessage());
			return;
		}

		if (start < 0)
			start = 0;

		end = start + 2 * k;
		if (end >= q.SI)
			end = q.SI - 1;

		for (j = start; j <= end; j++)
		{
			deltav = (q.offset - j * q.hzperpt) - gp.decon_peaks[n].freq;

			double loren = eta *
					(gp.decon_peaks[n].intensity
							/ (1.0 + 4.0 * deltav * deltav * wl));

			double gauss = (1 - eta) *
					(gp.decon_peaks[n].intensity
							* Math.exp(-M_LN2 * deltav * deltav * wg));

			gp.stack_trace[n].data[j] = loren + gauss;
		}
	}

//-------------------------------------------------------------------------------------------------
//  Optimize the parameters of a deconvolution peak, simple-minded direct method.  A more 
//  refined method (Simplex, or partial devivates) is under development

/**
 * Optimize the frequency (minimize RMS deviation) of a single deconvolution peak
 * @param n index of the peak to be optimized
 * @param iter maximum number of iterations
 * @param f1 starting frequency for RMS calculation
 * @param f2 ending frequency for RMS calculation
 * @param step step size
 * @param gp a SpinWorks global parameter object
 * @param q status data parameters for the current peak
 */
	public static void optimize_freq(int n, int iter, double f1, double f2, double step,  
			global_par gp, data_par q)
	{
		double rms, last_rms;
		double a = 1.0;

		last_rms = Util.rms_dev(f1, f2, gp, q, false);      // get initial rms deviation

		for (int i = 1; i <= iter; i++)
		{
			gp.decon_peaks[n].freq = gp.decon_peaks[n].freq + a * step;
			genspec(gp, q);     //generate spectrum with new value
			rms = Util.rms_dev(f1, f2, gp, q, false);
			if (rms > last_rms) a = -a / 2.0;       // if rms increasing, reverse and dec. step
			last_rms = rms;                         // update rms
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * Optimize the amplitude (minimize RMS deviation) of a single deconvolution peak
 * @param n index of the peak to be optimized
 * @param iter maximum number of iterations
 * @param f1 starting frequency for RMS calculation
 * @param f2 ending frequency for RMS calculation
 * @param step step size
 * @param gp a SpinWorks global parameter object
 * @param q status data parameters for the current peak
 */
	public static void optimize_amp(int n, int iter, double f1, double f2, double step,
			global_par gp, data_par q)
	{
		double rms, last_rms;
		double a = 1.0;

		last_rms = Util.rms_dev(f1, f2, gp, q, false);      // get initial rms deviation

		for (int i = 1; i <= iter; i++)
		{
			gp.decon_peaks[n].intensity = gp.decon_peaks[n].intensity + a * step;
			genspec(gp, q);     //generate spectrum with new value
			rms = Util.rms_dev(f1, f2, gp, q, false);
			if (rms > last_rms) a = -a / 2.0;       // if rms increasing, reverse and dec. step
			last_rms = rms;                         // update rms
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * Optimize the width (minimize RMS deviation) of a single deconvolution peak
 * @param n index of the peak to be optimized
 * @param iter maximum number of iterations
 * @param f1 starting frequency for RMS calculation
 * @param f2 ending frequency for RMS calculation
 * @param step step size
 * @param gp a SpinWorks global parameter object
 * @param q status data parameters for the current peak
 */
	public static void optimize_width(int n, int iter, double f1, double f2, double step,
			global_par gp, data_par q)
	{
		double rms, last_rms;
		double a = 1.0;

		last_rms = Util.rms_dev(f1, f2, gp, q, false);      // get initial rms deviation

		for (int i = 1; i <= iter; i++)
		{
			gp.decon_peaks[n].width = gp.decon_peaks[n].width + a * step;
			genspec(gp, q);     //generate spectrum with new value
			rms = Util.rms_dev(f1, f2, gp, q, false);
			if (rms > last_rms) a = -a / 2.0;       // if rms increasing, reverse and dec. step
			last_rms = rms;                         // update rms
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * Optimize the gaussian fraction (minimize RMS deviation) of a single deconvolution peak
 * @param n index of the peak to be optimized
 * @param iter maximum number of iterations
 * @param f1 starting frequency for RMS calculation
 * @param f2 ending frequency for RMS calculation
 * @param step step size
 * @param gp a SpinWorks global parameter object
 * @param q status data parameters for the current peak
 */
	public static void optimize_gauss(int n, int iter, double f1, double f2, double step,
			global_par gp, data_par q)
	{
		double rms, last_rms;
		double a = 1.0;

		last_rms = Util.rms_dev(f1, f2, gp, q, false);      // get initial rms deviation

		if (gp.decon_peaks[n].gauss >= 100.0) a = -a;

		for (int i = 1; i <= iter; i++)
		{
			gp.decon_peaks[n].gauss = gp.decon_peaks[n].gauss + a * step;

			if (gp.decon_peaks[n].gauss <= 0.0 || gp.decon_peaks[n].gauss >= 100.0)
			{
				break;
			}
			genspec(gp, q);     //generate spectrum with new value
			rms = Util.rms_dev(f1, f2, gp, q, false);
			if (rms > last_rms) a = -a / 2.0;       // if rms increasing, reverse and dec. step
			last_rms = rms;                         // update rms
		}
	}
	
//-------------------------------------------------------------------------------------------------
}	// end class
