package lp;

import blas.Blas;
import spinworks.SW_main;
import spinworks.Util;
/**
 * Complex Singular Value Decomposition
 * @author marat, adapted from numerous public domain sources
 */
public class Csvd 
{	
//-------------------------------------------------------------------------------------------------
//  version with real and complex parts of the matrices split (for SPEED!)
//-------------------------------------------------------------------------------------------------
       //   x[n,p]     The matrix to be decomposed
       //   n			The number of rows of X  (n > p for Linear prediction)
       //   p			The number of columns of X
       //   s[p+1]		The singular values of X
       //   e[p]		See the documentation in the code
       //   u[n,p]		The left singular vectors
       //   v[p,p]		The right singular vectors
       //   job		    See the documentation in the code
       //   info		is no longer used
       //	 tol		convergence tolerence
       //   return:	true if successful, false if no convergence

/// <summary>
/// Complex singular value decomposition with less arithmetic using complex variables (faster)
/// </summary>
/// <param name="x">The matrix to be decomposed [n,p]</param>
/// <param name="n">The number of rows of X  (n > p for Linear prediction)</param>
/// <param name="p">The number of columns of X</param>
/// <param name="s">The singular values of X [p+1]</param>
/// <param name="e">See the documentation in the code [p]</param>
/// <param name="u">The left singular vectors [n,p]</param>
/// <param name="v">The right singular vectors [p,p]</param>
/// <param name="job">Job control.  See the documentation in the code</param>
/// <param name="tol">convergence tolerence</param>
/// <returns>true if successful, false if no convergence</returns>
       public static boolean csvdc_cs_split(Complex[][] x, int n, int p, Complex[] s,
                                    Complex[] e, Complex[][] u, Complex[][] v,
                                    int job, double tol)
       {
           if (SW_main.gp.log_on)
           {
               String col = String.format(" %d ", SW_main.current_column);
               Util.log_error("*** starting csvdc_cs_split, col = " + col);
           }

           double b, c, cs, f, g, scale, shift,
                  sn, t1, test, ztest;

           Complex t, r;
           Complex temp = new Complex(0.0, 0.0);
           double tr, ti;

           double tempr, tempi;

           double[] rotvec = new double[4];

           int i, iter, j, jobu, k, kase, kk, l, ll, lls, lm1, lp1, ls, lu,
               m, maxit, mm, mp1, nct, ncu, nrt;

           int lsm1, km1, mm1;

           boolean wantu, wantv;

           double[][] xr;
           double[][] xi;
           try
           {
                xr = new double[n][p];
                xi = new double[n][p];
           }
           catch (Exception ex)
           {
               Util.log_error(ex.getMessage() + "xr, ir");
               return false;
           }
          

           for(i = 0; i < n; i++)       // copy complex matrix into real/imag matrices
               for (j = 0; j < p; j++)
               {
                   xr[i][j] = x[i][ j].r;    xi[i][ j] = x[i][j].i;
               }

           double[] sr = new double[p + n];
           double[] si = new double[p + n];

           double[] er = new double[p + n];
           double[] ei = new double[p + n ];

           double[][] ur = new double[n][p];
           double[][] ui = new double[n][p];

           double[][] vr = new double[p][p];
           double[][] vi = new double[p][p];

           double[] workr = new double[n];      // temporary work array
           double[] worki = new double[n];

           ls = 0;

           //   set the maximum number of iterations
           maxit = 30;

           //   determine what is to be computed
           wantu = false;
           wantv = false;

           jobu = (job % 100) / 10;

           ncu = n;

           
           if (jobu > 1) ncu = Math.min(n, p);
           if (jobu != 0) wantu = true;
           if ((job % 10) != 0) wantv = true;

//   reduce x to bidiagonal form, storing the diagonal elements
//   in s and the super-diagonal elements in e
           nct = Math.min(n - 1, p);
           nrt = Math.max(0, Math.min(p - 2, n));
           lu = Math.max(nct, nrt);

           for (l = 1; l <= lu; l++)		// do 160	<< original FORTRAN line numbers
           {
               lm1 = l - 1;
               lp1 = l + 1;

               if (l <= nct)
               {

//   compute the transformation for the l-th column and
//   place the l-th diagonal in s[lm1]
                   sr[lm1] = Blas.colnrm2(n - l + 1, xr, xi, lm1, lm1);     si[lm1] = 0.0;

                   if (Blas.cabs1(sr[lm1], si[lm1]) != 0.0)
                   {                
                       if (Blas.cabs1(xr[lm1][lm1], xi[lm1][lm1]) != 0.0)
                       {
                           temp = Blas.csign(Complex.cmplx(sr[lm1], si[lm1]),
                               Complex.cmplx(xr[lm1][ lm1], xi[lm1][lm1]));
                           sr[lm1] = temp.r;     si[lm1] = temp.i;
                       }

                       double denom = sr[lm1] * sr[lm1] + si[lm1] * si[lm1];
                       tempr = sr[lm1] / denom;
                       tempi = - si[lm1] / denom;

                       Blas.colscal(n - l + 1, tempr, tempi, xr, xi, lm1, lm1);

                       xr[lm1][lm1] = xr[lm1][lm1] + 1.0;
                   }  // 10
                   sr[lm1] = -sr[lm1]; si[lm1] = -si[lm1];
               }   // 20              

               if (p >= lp1)
               {
                   for (j = l; j < p; j++)			// do 40
                   {
                       if ((l <= nct) && (Blas.cabs1(sr[lm1], si[lm1]) != 0.0))
                       {
                           //   apply the transformation
                           t = Complex.div(Complex.minus(Blas.coldot(n - l + 1, xr, xi, lm1, lm1, j)),
                                 new Complex(xr[lm1][lm1], xi[lm1][lm1])); 

                           tr = t.r;    ti = t.i;

                           Blas.colaxpy(n - l + 1, tr, ti, xr, xi, lm1, lm1, j);
                       }

//   place the l-th row of x into e for the
//   subsequent calculation of the row transformation
                       er[j] = xr[lm1][j];      ei[j] = -xi[lm1][j];
                   }	// 40
               }		// 50

               if (wantu && (l <= nct))
               {
//   place the transformation in u for subsequent 
//   back multiplication

                   for (i = lm1; i < n; i++)	// do 60
                   {
                       ur[i][lm1] = xr[i][lm1];
                       ui[i][lm1] = xi[i][lm1];
                   }							// 60
               }

               if (l <= nrt)
               {
//   compute the l-th row transformation and place the
//   l-th super-diagonal in e[lm1]
                   er[lm1] = Blas.cnrm2p(p - l, er, ei, l);          ei[lm1] = 0.0;

                   if (Blas.cabs1(er[lm1], ei[lm1]) != 0.0)
                   {
                       if (Blas.cabs1(er[l], ei[l]) != 0.0)
                       {
                           temp = new Complex(0.0, 0.0);
                           temp = Blas.csign(Complex.cmplx(er[lm1], ei[lm1]),
                               Complex.cmplx(er[l], ei[l]));
                           er[lm1] = temp.r;    ei[lm1] = temp.i;
                       }

                       double denom = er[lm1] * er[lm1] + ei[lm1] * ei[lm1];
                       tempr = er[lm1] / denom;
                       tempi = -ei[lm1] / denom;

                       Blas.dscalp(p - l, tempr, tempi, er, ei, l);
                       er[l] = er[l] + 1.0;
                   } // 80
  
                   er[lm1] = -er[lm1];

                   if ((lp1 <= n) && (Blas.cabs1(er[lm1], ei[lm1]) != 0.0))
                   {
                       //   apply the transformation

                       for (i = l; i < n; i++)
                       {
                           workr[i] = 0.0; worki[i] = 0.0;
                       }  // 90

                       for (j = l; j < p; j++)
                       {
                           Blas.colvaxpy(n - l, er[j], ei[j], xr, xi,
                               workr, worki, l, j);             
                       }  // 100

                       for (j = l; j < p; j++)
                       {
                           Complex ej = new Complex(er[j], ei[j]);
                           Complex el = new Complex(er[l], ei[l]);                    

                           temp = Complex.conjf(Complex.div(Complex.minus(ej) , el));                        

                           Blas.colvraxpy(n - l, temp.r, temp.i, 
                               workr, worki, xr, xi, l, j);
                       }  // 110
                   }  //120

                   if (wantv)
                   {
//   place the transformation in v for subsequent
//   back multiplication

                       for (i = l; i < p; i++)
                       {
                           vr[i][lm1] = er[i];
                           vi[i][lm1] = ei[i];
                       } // 130
                   }     // 140
               }		 // 150			 
           }			 // 160 (do 160)

//   set up the final bidiagonal matrix of order m

           m = Math.min(p, n + 1);

           if (nct < p) { sr[nct] = xr[nct][nct];       si[nct] = xi[nct][nct]; }
           if (n < m) { sr[m - 1] = 0.0;                si[m - 1] = 0.0; }
           if (nrt + 1 < m) {er[nrt] = xr[nrt][m-1];    ei[nrt] = xi[nrt][m-1]; }

           er[m - 1] = 0.0;     ei[m - 1] = 0.0;

//   if required, generate u
           if (wantu)  // goto 300
           {
               for (j = nct; j < ncu; j++)
               {
                   try
                   {
                       for (i = 0; i < n; i++)
                       {
                           ur[i][j] = 0.0; ui[i][j] = 0.0;
                       }		// 180
                       //ur[i, j] = 1.0; ui[i, j] = 0.0;
                       ur[j][j] = 1.0; ui[j][j] = 0.0;
                   }
                   catch (Exception ex)
                   {
                       Util.log_error(ex.getMessage() + "generating u");
                       return false;
                   }
               }  // 190

               for (ll = 1; ll <= nct; ll++)  // do 280
               {
                   l = nct - ll + 1;
                   lm1 = l - 1;

                   if(Blas.cabs1(sr[lm1], si[lm1]) != 0.0)
                   {
                       for (j = l; j < ncu; j++)
                       {
                           t = Complex.div(Complex.minus(Blas.coldot(n - l + 1, ur, ui, lm1, lm1, j)),
                               new Complex(ur[lm1][lm1], ui[lm1][lm1])) ;
                           
                           Blas.colaxpy(n - l + 1, t.r, t.i, ur, ui, lm1, lm1, j);
                       } // 210

                       Blas.colscal(n - l + 1, -1.0, 0.0, ur, ui, lm1, lm1);
                       ur[lm1][lm1] = ur[lm1][lm1] + 1.0;
                       
                       for (i = 0; i < lm1; i++)  // do 230
                       {
                           ur[i][lm1] = 0.0;        ui[i][lm1] = 0.0;
                       }
                   }
                   else  // 250
                   {
                       for (i = 0; i < n; i++)
                       {
                           ur[i][lm1] = 0.0;        ui[i][lm1] = 0.0;
                       }  // 260
                       ur[lm1][lm1] = 1.0;          ui[lm1][lm1] = 0.0;
                   }
               }  // 280
           }  // 300

//   if it is required, generate v
           if (wantv)  // goto 350
           {
               for (ll = 1; ll <= p; ll++)
               {
                   l = p - ll + 1;
                   lm1 = l - 1;
 
                   if ((l <= nrt) && (Blas.cabs1(er[lm1], ei[lm1]) != 0.0)) 
                   {
                       for (j = l; j < p; j++)
                       {
                           t = Complex.div(Complex.minus(Blas.coldot(p - l, vr, vi, l, lm1, j)), 
                               Complex.cmplx(vr[l][lm1], vi[l][lm1]));
                           
                           Blas.colaxpy(p - l, t.r, t.i, vr, vi, l, lm1, j);

                       } // 310
                   }   // 320

                   for (i = 0; i < p; i++)
                   {
                       vr[i][lm1] = 0.0;    vi[i][lm1] = 0.0;
                   }		// 330
                   vr[lm1][lm1] = 1.0;      vi[lm1][lm1] = 0.0;
               }   // 340
           }	 // 350  

//	transform S and E so that they are real
           for (i = 1; i <= m; i++) // do 380
           {
               int im1 = i - 1;
               if (Blas.cabs1(sr[im1], si[im1]) != 0.0)
               {
                   t = Complex.cmplx(Blas.cabs(sr[im1], si[im1]), 0.0);
                   r = Complex.div(Complex.cmplx(sr[im1], si[im1]) , t);
                   sr[im1] = t.r;           si[im1] = t.i;

                   if (i < m)
                   {
                       temp = Complex.div(Complex.cmplx(er[im1], ei[im1]) , r);
                       er[im1] = temp.r; ei[im1] = temp.i;
                   }

                   if (wantu)
                       Blas.colscal(n, r.r, r.i, ur, ui, 0, i - 1);

               }		// 360

               if (i == m) break;

               if (Blas.cabs1(er[im1], ei[im1]) != 0.0)
               {
                   t = Complex.cmplx(Blas.cabs(er[im1], ei[im1]), 0.0);
                   r = Complex.div(t , Complex.cmplx(er[im1], ei[im1]));
                   er[im1] = t.r;       ei[im1] = t.i;               

                   temp = Complex.mult(Complex.cmplx(sr[i], si[i]) , r);
                   sr[i] = temp.r;      si[i] = temp.i;

                   if (wantv)
                       Blas.colscal(p, r.r, r.i, vr, vi, 0, i);
               }	// 370
           }     // 380

//   main iteration loop for the singular values
           mm = m;
           iter = 0;

           while (true)	// infinite loop
           {
//   quit if all of the singular values have been found
               if (m == 0)
               {

                   for (i = 0; i < n; i++)
                       for (j = 0; j < p; j++)
                       {
                           x[i][j].r = xr[i][j]; x[i][j].i = xi[i][j];
                       }

                   for (i = 0; i < n; i++)
                       for (j = 0; j < p; j++)
                       {
                           u[i][j].r = ur[i][j]; u[i][j].i = ui[i][j];
                       }

                   for (i = 0; i < p; i++)
                       for (j = 0; j < p; j++)
                       {
                           v[i][j].r = vr[i][j]; v[i][j].i = vi[i][j];
                       }

                   for (i = 0; i < p; i++)
                   {
                       s[i].r = sr[i]; s[i].i = si[i];
                   }


                   for (i = 0; i < p; i++)
                   {
                       e[i].r = er[i]; e[i].i = ei[i];
                   }
                   return true;
               }

//   if too many iterations have been performed, 
//   set flag and return     
               if (iter >= maxit)
               {
                   return false;
               }

//This section of the program inspects for
//negligible elements in the s and e arrays.
//On completion the variables kase and l are 
//set as follows:

//   kase = 1     If s[m] and e[l-1] are negligible and l < m
//   kase = 2     If s[l] is negligible and l < m
//   kase = 3     If e[l-1] is negligible, l < m, and
//                s[l], ..., s[m] are not negligible (QR step)
//   kase = 4     If e[m-1] is negligible (convergence)

// The material above is for FORTRAN style indexing.  Subtract
// indices by 1 for C# style indexing.

               for (ll = 1; ll <= m; ll++)   // do 430
               {
                   l = m - ll;
                   lm1 = l - 1;

                   if (l == 0) break;

                   test = Blas.cabs(sr[lm1], si[lm1]) + Blas.cabs(sr[l], si[l]);
                   ztest = test + Blas.cabs(er[lm1], ei[lm1]);

                   if (Math.abs(ztest - test) < tol)
                   {
                       er[lm1] = 0.0;       ei[lm1] = 0.0;
                       break;
                   }
               }  // 430

               if (l == m - 1)
               {
                   kase = 4;
               }
               else  // 450
               {
                   lp1 = l + 1;
                   mp1 = m + 1;

                   for (lls = lp1; lls <= mp1; lls++)
                   {
                       ls = m - lls + lp1;
                       lsm1 = ls - 1;
                       if (ls == l) break;

                       test = 0.0;

                       if (ls != m) 
                           test += Blas.cabs(er[lsm1], ei[lsm1]);

                       if (ls != l + 1) 
                           test += Blas.cabs(er[lsm1 - 1], ei[lsm1 - 1]);

                       ztest = test + Blas.cabs(sr[lsm1], si[lsm1]);

                       if (Math.abs(ztest - test) < tol)
                       {
                           sr[lsm1] = 0.0;      si[lsm1] = 0.0;
                           break;
                       }  // 460
                   }  // 470

                   if (ls == l)
                   {
                       kase = 3;
                   }
                   else if (ls == m)
                   {
                       kase = 1;
                   }
                   else
                   {
                       kase = 2;
                       l = ls;
                   }
               }  // 520

               l++;
               lm1 = l - 1;
               mm1 = m - 1;

//   perform the task indicated by kase
               switch (kase)
               {
                   case 1:   // 530
                       f = er[m - 2];

                       er[m - 2] = 0.0;         ei[m - 2] = 0.0;

                       for (kk = l; kk <= mm1; kk++)   // do 550
                       {
                           k = (mm1) - kk + l;
                           km1 = k - 1;

                           t1 = sr[km1];

                           rotvec[0] = t1;
                           rotvec[1] = f;
                           Blas.rotg(rotvec);
                           t1 = rotvec[0];
                           f  = rotvec[1];
                           cs = rotvec[2];
                           sn = rotvec[3];

                           sr[km1] = t1;    si[km1] = 0.0;

                           if (k != l)
                           {
                               f = -sn * er[k - 2];

                               er[k - 2] = cs * er[k - 2];
                               ei[k - 2] = cs * ei[k - 2];
                           }
                           // 540
                           if (wantv) Blas.colrot(p, vr, vi, km1, mm1, cs, sn);  

                       } // 550

                       break;

                   case 2:  // 560
                       f = er[l - 2];
                       er[l - 2] = 0.0; ei[l - 2] = 0.0;

                       for (k = l; k <= m; k++) // do 570
                       {
                           km1 = k - 1;

                           t1 = sr[km1];

                           rotvec[0] = t1;
                           rotvec[1] = f;
                           Blas.rotg(rotvec);
                           t1 = rotvec[0];
                           f = rotvec[1];
                           cs = rotvec[2];
                           sn = rotvec[3];

                           sr[km1] = t1;        si[km1] = 0.0;

                           f = -sn * er[km1];
                           er[km1] = cs * er[km1];      ei[km1] = cs * ei[km1];

                           if (wantu) Blas.colrot(n, ur, ui, km1, l - 2, cs, sn);
                       } // 570

                       break;

                   case 3:
//   calculate the shift
                       scale = Math.max(Blas.cabs(sr[mm1], si[mm1]), 
                           Blas.cabs(sr[m - 2], si[m - 2]));
                       scale = Math.max(Blas.cabs(er[m - 2], ei[m - 2]), scale);
                       scale = Math.max(Blas.cabs(sr[lm1], si[lm1]), scale);
                       scale = Math.max(Blas.cabs(er[lm1], ei[lm1]), scale);

                       b = ((sr[m - 2] + sr[mm1]) * (sr[m - 2] - sr[mm1]) +
                             er[m - 2] * er[m - 2]) / (2.0 * scale * scale);

                       c = (sr[mm1] * er[m - 2]) * (sr[mm1] * er[m - 2]) 
                           / (scale * scale * scale * scale);

                       shift = 0.0;

                       if ((b != 0.0) || (c != 0.0))
                       {
                           shift = Math.sqrt(b * b + c);
                           if (b < 0.0) shift = -shift;
                           shift = c / (b + shift);
                       } // 590 

                       f = (sr[lm1] + sr[mm1]) * (sr[lm1] - sr[mm1])
                             / (scale * scale) - shift;

                       g = sr[lm1] * er[lm1]
                               / (scale * scale);

//   chase zeros

                       for (k = l; k <= mm1; k++)  // do 600
                       {
                           km1 = k - 1;
                           rotvec[0] = f;
                           rotvec[1] = g;
                           Blas.rotg(rotvec);
                           f = rotvec[0];
                           g = rotvec[1];
                           cs = rotvec[2];
                           sn = rotvec[3];

                           if (k != l) { er[k - 2] = f;     ei[k - 2] = 0.0; }

                           f = cs * sr[km1] + sn * er[km1];

                           er[km1] = cs * er[km1] - sn * sr[km1];
                           ei[km1] = cs * ei[km1] - sn * si[km1];

                           g = sn * sr[k];

                           sr[k] = cs * sr[k];
                           si[k] = cs * si[k];

                           if (wantv) Blas.colrot(p, vr, vi, km1, k, cs, sn);

                           rotvec[0] = f;
                           rotvec[1] = g;
                           Blas.rotg(rotvec);
                           f = rotvec[0];
                           g = rotvec[1];
                           cs = rotvec[2];
                           sn = rotvec[3];

                           sr[km1] = f;         si[km1] = 0.0;
                           f = cs * er[km1] + sn * sr[k];

                           sr[k] = -sn * er[km1] + cs * sr[k];
                           si[k] = -sn * ei[km1] + cs * si[k];

                           g = sn * er[k];

                           er[k] = cs * er[k];      ei[k] = cs * ei[k];

                           if (wantu && (k < n)) Blas.colrot(n, ur, ui, km1, k, cs, sn);
                       }  // 600

                       er[m - 2] = f;               ei[m - 2] = 0.0;

                       iter++;
                       break;

                   case 4:  // 610
//   make the singular value positive
                       if (sr[lm1] < 0.0)
                       {
                           sr[lm1] = -sr[lm1]; si[lm1] = -si[lm1];

                           if (wantv) Blas.colscal(p, -1.0, 0.0, vr, vi, 0, lm1);
                       }  // 620

//   order the singular value
                       while (l != mm)   // 630
                       {
                           if (sr[lm1] >= sr[l]) break;

                           t = Complex.cmplx(sr[lm1], si[lm1]);
                           sr[lm1] = sr[l];     si[lm1] = si[l];
                           sr[l] = t.r;     si[l] = t.i;

                           if (wantv && (l < p)) 
                               Blas.colswap(p, vr, vi, lm1, l);

                           if (wantu && (l < n)) 
                               Blas.colswap(n, ur, ui, lm1, l);

                           l++;
                       }		// 640

                       iter = 0;
                       m--;

                       break;
               }			// switch case
           }			// while (infinite loop)
       }      
//-------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------
/// <summary>
/// Given a csvd decomposition, csvd_fit calculates the least squares solution to the 
/// linear equation a x = d (d[2*m]). 
/// </summary>
       public static void csvd_fit(Complex[][] u, Complex[][] v, Complex[] w, Complex[] x, Complex[] d,
    		   int m, int n, int offset)
       {
    	   int i, j, k;
    	   double winv;
    	   Complex s = new Complex(0.0, 0.0);
    	   Complex z = new Complex(0.0, 0.0);
    	   Complex r = new Complex(0.0, 0.0);

    	   Complex[] t = new Complex[(n + 1)];     // temporary storage
    	   for (i = 0; i < n + 1; i++)
    		   t[i] = new Complex(0.0, 0.0);

    	   for (j = 0; j < n; j++)					
    	   {
    		   s.zero();							

    		   if (w[j].r != 0.0)					
    		   {
    			   for (i = 0; i < m; i++)			// loop over rows
    			   {
    				   r.r = u[i][j].r; r.i = u[i][j].i;
    				   r.conj();
    				   z = Complex.mult(r , d[i + offset]);	
    				   s.r = s.r + z.r;		s.i = s.i + z.i;
    			   }
    			   winv = 1.0 / w[j].r;
    			   s.r = s.r * winv; s.i = s.i * winv;	
    		   }
    		   t[j].r = s.r;
    		   t[j].i = s.i;
    	   }

    	   for (j = 0; j < n; j++)						// loop over columns
    	   {
    		   s.zero();

    		   for (k = 0; k < n; k++)					// loop over rows 
    		   {
    			   z = Complex.mult(v[j][k] , t[k]);		
    			   s.r = s.r + z.r;		s.i = s.i + z.i;		
    		   }

    		   x[j].r = s.r;	
    		   x[j].i = s.i;	
    	   }
       }

 //-------------------------------------------------------------------------------------------------
 /// <summary>
 /// eliminates singular values below significance
 /// </summary>
 /// <param name="w">array of singular values</param>
 /// <param name="cutoff">(0 to 1) determines which svd values are signficant. </param>
 /// <param name="n">number of singular values</param>
       public static void csvd_refine(Complex[] w, double cutoff, int n)
       {
    	   int j;
    	   double wmin, wmax;

    	   wmax = 0.0;

    	   for (j = 0; j < n; j++)
    	   {
    		   wmax = Math.max(wmax, w[j].r);
    	   }

    	   wmin = cutoff * wmax;

    	   for (j = 0; j < n; j++)
    	   {
    		   if (w[j].r < wmin)
    			   w[j].r = 0.0;
    	   }
       }
//-------------------------------------------------------------------------------------------------	       
}
