package lp;

import spinworks.SW_main;
import spinworks.data_par;
import spinworks.global_par;
import spinworks.Util;

/**
 * Linear prediction routines for SpinWorks Java,
 * adapted from numerous sources, including "Numerical Recipes in C"
 * @author marat, 
 * University of Manitoba
 */
public class Lp 
{
	private static double TOL1 = 2.0e-6; // originally machine epsilon
	private static double TOL2 = 1.0e-7; // originally machine EPSS
	private static int MAX_ITER = 80;
	private static int MT = 10;
	
//-------------------------------------------------------------------------------------------------
	/**
	 * Do Zhu-Bax LP method,  Java version
	 * @param dir DIRECT or INDIRECT
	 * @param gp A SpinWorks global parameter object
	 * @param p set parameters, direct dimension
	 * @param q status parameters, direct dimension
	 * @param p_ind set parameters, indirect dimension
	 * @param q_ind status parameters, indirect dimension
	 * @return true if OK, false on error
	 */
	public static boolean do_ZhuBax_LP(SW_main.direction dir, global_par gp,
			data_par p, data_par q, data_par p_ind, data_par q_ind)
	{
		float cutoff;
		int npts, length, nprd, n_total, size;
		int i, j;

		if (dir == SW_main.direction.DIRECT) // direct (acquisition) dimension
		{
			npts = q.TD / 2;
			length = p.lp_coef;
			nprd = p.lp_pred / 2; 
			cutoff = (float)p.lp_cutoff;
			size = q.SI;
		}
		else  // Indirect dimension Clean-up! only need to pass 1 p and q
		{
			if (p_ind.MC2 == SW_main.det_modes_IND.QF)
				npts = q_ind.TD;
			else
				npts = q_ind.TD / 2;

			length = p_ind.lp_coef;
			nprd = p_ind.lp_pred / 2;
			cutoff = (float)p_ind.lp_cutoff;
			size = q_ind.SI;
		}

		n_total = npts + Math.abs(nprd);

		Complex[] x = new Complex[n_total];
		for (i = 0; i < n_total; i++) 
			x[i] = new Complex(0.0, 0.0);

		j = 0;
		for (i = 0; i < npts && i < size; i++)
		{
			x[j].r = gp.proc_buffer[2 * i];
			x[j].i = gp.proc_buffer[2 * i + 1];
			j++;
		}

		if (Zhu_Bax_lp(x, npts, length, Math.abs(nprd), cutoff) == false)
		{
			Util.log_error("Zhu-Bax  LP routine returns error");
			return false;
		}

		if (dir == SW_main.direction.DIRECT) // direct (acquisition) dimension
		{
			q.td_eff = 2 * npts + 2 * nprd;

			if (q.td_eff > 2 * size)
				q.td_eff = 2 * size;

			for (i = 0; i < q.td_eff / 2; i++)
			{
				gp.proc_buffer[2 * i] = x[i].r;
				gp.proc_buffer[2 * i + 1] = x[i].i;
			}
		}
		else // indirect
		{
			q_ind.td_eff = 2 * npts + 2 * nprd;

			if (q_ind.td_eff > 2 * size)
				q_ind.td_eff = 2 * size;

			for (i = 0; i < q_ind.td_eff / 2; i++)
			{
				gp.proc_buffer[2 * i] = x[i].r;
				gp.proc_buffer[2 * i + 1] = x[i].i;
			}
		}

		return true;
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Do Zhu-Bax LP method, this version used only for acquisition dimension
	 * @param gp A SpinWorks global parameter object
	 * @param p set acquisition parameters
	 * @param q status acquisition parameters
	 * @return true if OK, false on error
	 */
	public static boolean do_ZhuBax_LP(global_par gp,
			data_par p, data_par q)
	{
		float cutoff;
		int npts, length, nprd, n_total, size;
		int i, j;

		npts = q.TD / 2;

		length = p.lp_coef;
		nprd = p.lp_pred / 2;
		cutoff = (float)p.lp_cutoff;
		size = q.SI;

		n_total = npts + Math.abs(nprd);

		Complex[] x = new Complex[n_total];
		for (i = 0; i < n_total; i++)
			x[i] = new Complex(0.0, 0.0);

		j = 0;
		for (i = 0; i < npts && i < size; i++)
		{
			x[j].r = gp.proc_buffer[2 * i];
			x[j].i = gp.proc_buffer[2 * i + 1];
			j++;
		}

		if (Zhu_Bax_lp(x, npts, length, Math.abs(nprd), cutoff) == false)
		{
			Util.log_error("Zhu-Bax LP routine returns error");
			return false;
		}

		q.td_eff = 2 * npts + 2 * nprd;

		if (q.td_eff > 2 * size)
			q.td_eff = 2 * size;

		for (i = 0; i < q.td_eff / 2; i++)
		{
			gp.proc_buffer[2 * i] = x[i].r;
			gp.proc_buffer[2 * i + 1] = x[i].i;
		}
		return true;
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Do backwards LP for F2 or 1D
	 * @param gp A SpinWorks global parameter object
	 * @param p set acquisition parameters
	 * @param q status acquisition parameters
	 */
	public static void do_back_lp_f2(global_par gp, data_par p, data_par q)
	{			
		int i;										// handy counter

		if (p.lp_input > q.TD / 2)
		{
			Util.show_and_log("LP Backwards: too many input points");
			return; // should return error code?
		}

		int n;

		if (p.SI < q.TD)
			n = p.SI;
		else
			n = q.TD;

		double[] data_real = new double[q.TD];

		for (i = 0; i < n; i++)
			data_real[i] = gp.proc_buffer[i];		// load data array

		lp_back(data_real, p.lp_coef, p.lp_input,	// do the linear prediction
				p.lp_pred, p.lp_cutoff);

		for (i = 0; i < n; i++)						// re-load processing buffer
			gp.proc_buffer[i] = data_real[i];		
	}
	

//-------------------------------------------------------------------------------------------------
	/**
	 * Zhu-Bax forward-backward LP routine, JMR 100 (1992) 202-207 
	 * @param x the input vector
	 * @param npoints number of points
	 * @param n number of coefficients
	 * @param npred number of predicted points
	 * @param cutoff convergence parameter
	 * @return true if OK, false on error
	 */
	private static boolean Zhu_Bax_lp(Complex[] x, int npoints, int n,
			int npred, double cutoff)
	{
		boolean polish = false;         // if polynomial roots should be "polished"
		// add as processing parameter??

		//	see Zhu and Bax, JMR 100 (1992) 202-207 
		int m = npoints - n - 1;		
		int  i, j;

		Complex[][] u;
		Complex[][] v;

		try
		{
			u = new Complex[m+1][ n];
			v = new Complex[n][ n];            
		}
		catch (Exception ex)
		{
			Util.show_and_log("LP: u, v, allocation error " + ex.getMessage());
			return false;
		}

		for (i = 0; i < n; i++)
			for (j = 0; j < n; j++)
				v[i][ j] = new Complex(0.0, 0.0);

		Complex[] w = new Complex[n];
		for (i = 0; i < n; i++)
			w[i] = new Complex(0.0, 0.0);

		Complex[] t1 = new Complex[(n + 1)];
		for (i = 0; i < n + 1; i++)
			t1[i] = new Complex(0.0, 0.0);

		Complex[] c = new Complex[n + 1];
		for (i = 0; i < n + 1; i++)
			c[i] = new Complex(0.0, 0.0);

		Complex[] d = new Complex[n + 1];
		for (i = 0; i < n + 1; i++)
			d[i] = new Complex(0.0, 0.0);

		Complex[] c_roots = new Complex[n];
		for (i = 0; i < n; i++)
			c_roots[i] = new Complex(0.0, 0.0);

		Complex[] d_roots = new Complex[n];
		for (i = 0; i < n; i++)
			d_roots[i] = new Complex(0.0, 0.0);

		Complex s = new Complex(0.0, 0.0);
		Complex z = new Complex(0.0, 0.0);

		//		load data into u array 
		for (i = 0; i < m; i++)
			for (j = 0; j < n; j++) 
			{
				u[i][ j] = new Complex(0.0, 0.0);		// ??? 			
				u[i][ j].r = x[i + j + 1].r;					
				u[i][ j].i = x[i + j + 1].i;
			}

		if (!Csvd.csvdc_cs_split(u, m, n, w, t1, u, v, 21, cutoff * SW_main.gp.raw_max))
		{
			Util.log_error("LP error forward_backward_lp(): Complex svd failed, no convergence");
			Complex.zero_vector(x, npoints, npred);
			return false;
		}

		Csvd.csvd_refine(w, cutoff, n);             // eliminate values below significance
		Csvd.csvd_fit(u, v, w, c, x, m, n, n+1); 

		c[n].r = 1.0;  c[n].i = 0.0;

		negate_vector(c, n);
		poly_roots(c, n, c_roots, polish, t1);                   
		reflect_roots(c_roots, n);		
		poly_coeffs(c, n, c_roots);

		// round 2 of fitting, as per Zhu and Bax
		Csvd.csvd_fit(u, v, w, d, x, m, n, 0);   
		d[n].r = 1.0; d[n].i = 0.0;				
		negate_vector(d, n);

		for (i = 0; i < n / 2 ; i++)		
		{
			Complex temp = new Complex(0.0, 0.0) ;
			temp.r = d[i].r;			temp.i = d[i].i;
			d[i].r = d[n - i - 1].r;	d[i].i = d[n - i - 1].i;
			d[n - i - 1].r = temp.r;	d[n - i - 1].i = temp.i;
		}

		poly_roots(d, n, d_roots, polish, t1);

		for (i = 0; i < n; i++)
			d_roots[i].conj();		

		reflect_roots(d_roots, n);			
		poly_coeffs(d, n, d_roots);

		//		find average of c and d 
		Complex.complex_add_vect(c, c, d, n);
		Complex.complex_scale_vect(c, 0.5, n);

		negate_vector(c, n);
		poly_roots(c, n, c_roots, polish, t1);
		reflect_roots(c_roots, n);
		poly_coeffs(c, n, c_roots);

		//		do the actual linear prediction
		for (i = 0; i < npred; i++)
		{
			s.zero();

			for (j = 0; j < n; j++)
			{
				z = Complex.mult(c[j] , x[npoints - n + i + j]);	
				s = Complex.add(s , z);
			}

			x[npoints + i].r = s.r;		
			x[npoints + i].i = s.i;		
		}

		return true;
	}

//-------------------------------------------------------------------------------------------------
	/**
	 *   backwards LP routine
	 * @param data the input data (fid)
	 * @param n_coef the number of coefficients
	 * @param n_input the number of input points to use
	 * @param n_points the number of points to predict
	 * @param cutoff convergence cutoff parameter (as fraction of most intense point)
	 */
	private static void lp_back(double[] data, int n_coef, int n_input, 
		int n_points, double cutoff)	
	{
		int j, k, m;
		double s;
		boolean converged;

        double[][] u = new double[n_input][ n_coef];
        double[][] v = new double[n_coef][n_coef];
        double[] x = new double[n_coef];
        double[] d = new double[n_input];
        double[] w = new double[n_coef];


        for (j = 0; j < n_input; j++)
            for (k = 0; k < n_coef; k++)
                u[j][ k] = data[j + k + n_points + 1];

        converged = Svd.svd(u, v, w, n_input, n_coef); 

		if (converged)
		{
            for (j = 0; j < n_input; j++)
				d[j] = data[j + n_points];

            //	Real svd is fine for the rather few prediced points in back LP.  The 90 deg
            //	phase shift between channels is predicted as well.
            Svd.svd_fit(u, v, w, x, d, cutoff, n_input, n_coef); 
		}
		else	// not converged 
		{
            for (int i = 0; i < n_coef; i++)
				x[i] = 0.0;
		}

//		do the actual prediction
        for (m = n_points - 1; m >= 0; m--)
		{
			s = 0;
            for (k = 0; k < n_coef; k++)        // summing over all coefficients
				s += x[k] * data[m + k + 1];

			data[m] = s;
		}		
	}

//-------------------------------------------------------------------------------------------------
//	Various support routines for Singular value decompossition follow
//  Adapted from "Numerical Recipes in C"
//-------------------------------------------------------------------------------------------------
	/**
	 * Negate a complex vector (real and imag)
	 * @param v the vector to negate
	 * @param n the number of elements
	 */
	private static void negate_vector(Complex[] v, int n)
	{
		for (int j = 0; j < n; j++)
		{ v[j].r = -v[j].r; v[j].i = -v[j].i; }
	}

//------------------------------------------------------------------------------------------------
	/**
	 * Reflect complex roots into unit circle if necessary
	 * @param z the complex roots of the polynomial
	 * @param n the number of elements in z
	 */
	private static void reflect_roots(Complex[] z, int n)
	{
		int i;
		double r;

		for (i = 0; i < n; i++)
		{
			r = (z[i].r * z[i].r) + (z[i].i * z[i].i);

			if (r > 1.0)
			{
				r = 1.0 / r;
				z[i].r = z[i].r * r;	z[i].i = z[i].i * r;
			}
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * find the complex coefficients of a polynomial given the complex roots
	 * @param c the coefficients of the polynomial
	 * @param n the number of coefficients
	 * @param c_roots the complex roots of the polynomial
	 */
	private static void poly_coeffs(Complex[] c, int n, Complex[] c_roots)
	{
		int i, j;
		Complex z = new Complex(0.0, 0.0);

		c[0].r = c_roots[0].r;		c[0].i = c_roots[0].i;

		for (i = 1; i < n; i++)
		{
			c[i] = Complex.add(c[i - 1] , c_roots[i]);

			for (j = i-1; j > 0; j--)
			{
				z = Complex.mult(c[j] , c_roots[i]);	
				c[j] = Complex.sub(c[j-1] , z);	
			}

			z = Complex.mult(c_roots[i] , c[0]);	
			c[0].r = z.r;		        c[0].i = z.i;
			c[0].r = c[0].r * -1.0;		c[0].i = c[0].i * -1.0;
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Finds the roots of a complex polynomial using Laguerre's method
	 * @param a
	 * @param n number of points
	 * @param roots the roots
	 * @param polish whether the roots should be "polished"
	 * @param w a temporary work vector (create locally???)
	 * @return true if converged, false if no convergence
	 */
	private static boolean poly_roots(Complex[] a, int n, Complex[] roots, 
			boolean polish, Complex[] w)
	{
		int i, j;
		Complex x = new Complex(0.0, 0.0);
		Complex b = new Complex(0.0, 0.0);
		Complex c = new Complex(0.0, 0.0);
		Complex d = new Complex(0.0, 0.0);

		Complex.complex_copy(w, a, n + 1);		// YES! n + 1 is right

		for (i = n-1; i >= 0; i--)
		{			
			x.zero();	

			if (!laguerre(w, i + 1, x))
			{					
				x.r = 1.0;		x.i = 0.0;	

				if (!laguerre(w, i + 1, x))
				{						
					x.r = 0.0;		x.i = 1.0;	

					if (!laguerre(w, i + 1, x))
						return false;
				}
			}

			if (Math.abs(x.i) <= (TOL1 * Math.abs(x.r))) 
				x.i = 0.0;

			roots[i].r = x.r;		roots[i].i = x.i;
			b.r = w[i + 1].r;		b.i = w[i + 1].i;     //?????

			for (j = i; j >= 0; j--)
			{
				c.i = w[j].i;		c.r = w[j].r;
				w[j].r = b.r;		w[j].i = b.i;
				d = Complex.mult(x , b);		
				b = Complex.add(d , c);	
			}	// for j
		}	// for i

		if (polish)
		{
			for (i = 0; i < n; i++)
			{
				if (!laguerre(a, n, roots[i]))
					return false;						
			}
		}

		for (i = 1; i < n; i++)
		{
			x.r = roots[i].r;		 x.i = roots[i].i;

			for (j = i-1; j >= 0; j--)
			{
				if (roots[j].r <= x.r) break;
				roots[j + 1].r = roots[j].r;	roots[j + 1].i = roots[j].i;
			}
			roots[j + 1].r = x.r;		roots[j + 1].i = x.i;
		}
		return  true;
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Find a root of a polynomial with complex coefficients a[] using Laguerre's method
	 * @param a the coefficients of the polynomial
	 * @param n the number of coefficients
	 * @param x the root
	 * @return true if root found, false if no convergence
	 */
	private static boolean laguerre(Complex[] a, int n, Complex x)
	{
		double[] frac = new double[] { 0.5, 0.25, 0.75, 0.13, 0.38, 0.62, 0.88, 1.0 };
		int i, j;
		double err, abx, abp, abm;

		Complex b = new Complex(0.0, 0.0);
		Complex d = new Complex(0.0, 0.0);
		Complex f = new Complex(0.0, 0.0);
		Complex g = new Complex(0.0, 0.0);
		Complex h = new Complex(0.0, 0.0);
		Complex g2 = new Complex(0.0, 0.0);
		Complex s = new Complex(0.0, 0.0);
		Complex gp = new Complex(0.0, 0.0);
		Complex gm = new Complex(0.0, 0.0);
		Complex dx = new Complex(0.0, 0.0);
		Complex x1 = new Complex(0.0, 0.0);
		Complex t = new Complex(0.0, 0.0);


		for (i = 0; i < MAX_ITER; i++)
		{
			b.r = a[n].r;		b.i = a[n].i;
			err = Complex.Abs(b);

			d.zero();
			f.zero();

			abx = Complex.Abs(x);

			for (j = n-1; j >= 0; j--)
			{
				t = Complex.mult(x , f);
				f = Complex.add(t , d);

				t = Complex.mult(x , d);	

				d = Complex.add(t , b);

				t = Complex.mult(x , b);	
				b = Complex.add(t , a[j]);

				err = Complex.Abs(b) + abx * err;
			}
			err *= TOL2;

			if (Complex.Abs(b) <= err)
				return true;

			g = Complex.div(d , b);
			g2 = Complex.mult(g , g);	
			f.r = f.r * 2.0;	 f.i = f.i * 2.0;
			t = Complex.div(f , b);
			h = Complex.sub(g2 , t);	
			h.r *= (double)n;	 h.i *= (double)n;	
			t = Complex.sub(h , g2);	
			t.r *= (double)(n-1);	t.i *= (double)(n-1);

			s = Complex.Sqrt(t);

			gp = Complex.add(g , s);	
			gm = Complex.sub(g , s);	

			abp = Complex.Abs(gp);
			abm = Complex.Abs(gm);

			if (abp < abm)
			{ gp.r = gm.r; gp.i = gm.i; }	

			if (Math.max(abp, abm) > 0.0)
			{
				t.r = (double)n; t.i = 0.0;
				dx = Complex.div(t , gp);	
			}
			else
			{
				dx.r = (1.0 + abx) * Math.cos((double) (i+1));
				dx.i = (1.0 + abx) * Math.sin((double) (i+1));
			}

			x1 = Complex.sub(x , dx);

			if (Complex.complex_equal(x1, x))
				return true;

			if (((i+1) % MT)!= 0)	
			{
				x.r = x1.r;		x.i = x1.i;		
			}
			else
			{
				j = i / MT;	
				dx.r = dx.r*frac[j]; dx.i = dx.i * frac[j];	
				x.r = x.r - dx.r;		x.i = x.i - dx.i;	
			}
		}
		return  false;
	}
//-------------------------------------------------------------------------------------------------	
}
