
package lp;

/// <copyright>
///	This adaptation of Numerical Recipies code Copyright (c) 2005-2015
///
///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2
///
///	All rights reserved.
///
///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is strictly prohibited without prior written authorization.
/// 
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// My employer (actually their lawyers) want this:
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.
///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

/**
 * Real Singular value decomposition for SpinWorks Java
 * @author marat, 
 * University of Manitoba
 */
public class Svd 
{
	public static int MAX_NITER = 30;

	/////////////////////////////////////////////////////////////////////////////////////////
	//svd calculation decomposes matrix a = u w v_transpose 
	//a[m][n], u[m][n], w[n][n], v[n][n], w is diagonal 

	//It is assumed that m >= n (this is not checked for). 
	//On input u stores the matrix, a, for which the svd is done. 
	//This is overwritten on output by the svd matrix u. 
	//Only the diagonal values of the matrix w are stored. 

	//based on the routine in "Numerical Recipes in C"
	//uses C type zero based indexing rather than the 1 based indexing used in 
	//"Numerical Recipes"

	public static boolean svd(double[][]u, double[][] v, double[] w,  int m, int n)
	{
		int i, iter, j, jj, k, l, nm = 0;
		boolean flag;
		double c, d, f, h, s, x, y, z;
		double anorm = 0.0, g = 0, scale = 0.0;
		double[] t = new double[n]; // temporary storage

		//Householder's reduction to bidiagonal form  
		for (i = 0; i < n; i++)
		{
			l = i + 1;
			t[i] = scale * g;
			g = s = scale = 0;

			for (k = i; k < m; k++)
				scale += Math.abs(u[k][i]);

			if (scale != 0)
			{
				for (k = i; k < m; k++)
				{
					u[k][i] /= scale;
					s += u[k][i] * u[k][i];
				}

				f = u[i][i];
				g = -sign(f) * Math.sqrt(s);
				h = f * g - s;
				u[i][i] = f - g;

				for (j = l; j < n; j++)
				{
					s = 0;
					for (k = i; k < m; k++)
						s += u[k][i] * u[k][j];

					f = s / h;
					for (k = i; k < m; k++)
						u[k][j] += f * u[k][i];
				}

				for (k = i; k < m; k++)
					u[k][i] *= scale;
			}

			w[i] = scale * g;
			g = s = scale = 0;

			for (k = l; k < n; k++)
				scale += Math.abs(u[i][k]);

			if (scale != 0)
			{
				for (k = l; k < n; k++)
				{
					u[i][k] /= scale;
					s += u[i][k] * u[i][k];
				}

				f = u[i][l];
				g = -sign(f) * Math.sqrt(s);
				h = f * g - s;
				u[i][l] = f - g;

				for (k = l; k < n; k++)
					t[k] = u[i][k] / h;

				for (j = l; j < m; j++)
				{
					s = 0;
					for (k = l; k < n; k++)
						s += u[j][k] * u[i][k];

					for (k = l; k < n; k++)
						u[j][k] += s * t[k];
				}

				for (k = l; k < n; k++)
					u[i][k] *= scale;
			}

			d = Math.abs(w[i]) + Math.abs(t[i]);
			anorm = Math.max(anorm, d);
		}

		//Accumulation of right-hand transformations  
		g = 0;
		for (i = n - 1; i >= 0; i--)
		{
			l = i + 1;

			if (g != 0)
			{
				for (j = l; j < n; j++)
					v[j][i] = (u[i][j] / u[i][l]) / g;

				for (j = l; j < n; j++)
				{
					s = 0;
					for (k = l; k < n; k++)
						s += u[i][k] * v[k][j];

					for (k = l; k < n; k++)
						v[k][j] += s * v[k][i];
				}
			}

			for (j = l; j < n; j++)
				v[i][j] = v[j][i] = 0;

			v[i][i] = 1;
			g = t[i];
		}

		//Accumulation of left-hand transformations  
		for (i = n - 1; i >= 0; i--)
		{
			l = i + 1;
			g = w[i];

			for (j = l; j < n; j++)
				u[i][j] = 0;

			if (g != 0)
			{
				g = 1 / g;

				for (j = l; j < n; j++)
				{
					s = 0;
					for (k = l; k < m; k++)
						s += u[k][i] * u[k][j];

					f = (s / u[i][i]) * g;
					for (k = i; k < m; k++)
						u[k][j] += f * u[k][i];
				}

				for (j = i; j < m; j++)
					u[j][i] *= g;
			}
			else
			{
				for (j = i; j < m; j++)
					u[j][i] = 0;
			}

			u[i][i] += 1;
		}

		//Diagonalization of the bidiagonal form  
		for (k = n - 1; k >= 0; k--)
		{
			for (iter = 0; iter < MAX_NITER; iter++)
			{
				flag = true;
				for (l = k; l >= 0; l--)
				{
					nm = l - 1;
					if ((Math.abs(t[l]) + anorm) == anorm)
					{
						flag = false;
						break;
					}

					if ((Math.abs(w[nm]) + anorm) == anorm)
						break;
				}

				if (flag)
				{
					s = 1;
					for (i = l; i <= k; i++)
					{
						f = s * t[i];
						if ((Math.abs(f) + anorm) != anorm)
						{
							g = w[i];
							h = hypotenous(f, g);
							w[i] = h;
							h = 1 / h;
							c = g * h;
							s = -f * h;

							for (j = 0; j < m; j++)
							{
								y = u[j][nm];
								z = u[j][i];
								u[j][nm] = y * c + z * s;
								u[j][i] = z * c - y * s;
							}
						}
					}
				}

				z = w[k];

				if (l == k)
				{
					if (z < 0)
					{
						w[k] = -z;
						for (j = 0; j < n; j++)
							v[j][k] = -v[j][k];
					}
					break;
				}
				else if (iter >= (MAX_NITER - 1))
				{
					return false;
				}

				//Shift from bottom 2 x 2 minor  
				x = w[l];
				nm = k - 1;
				y = w[nm];
				g = t[nm];
				h = t[k];
				f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2 * h * y);
				g = hypotenous(f, 1);
				f = ((x - z) * (x + z) + h * ((y / (f + (sign(f) * g))) - h)) / x;

				//Next QR transformation  
				c = s = 1;

				for (j = l; j <= nm; j++)
				{
					i = j + 1;
					g = t[i];
					y = w[i];
					h = s * g;
					g = c * g;
					z = hypotenous(f, h);
					t[j] = z;
					c = f / z;
					s = h / z;
					f = x * c + g * s;
					g = g * c - x * s;
					h = y * s;
					y = y * c;

					for (jj = 0; jj < n; jj++)
					{
						x = v[jj][j];
						z = v[jj][i];
						v[jj][j] = x * c + z * s;
						v[jj][i] = z * c - x * s;
					}

					z = hypotenous(f, h);
					w[j] = z;

					if (z != 0)
					{
						z = 1 / z;
						c = f * z;
						s = h * z;
					}

					f = c * g + s * y;
					x = c * y - s * g;

					for (jj = 0; jj < m; jj++)
					{
						y = u[jj][j];
						z = u[jj][i];
						u[jj][j] = y * c + z * s;
						u[jj][i] = z * c - y * s;
					}
				}

				t[l] = 0;
				t[k] = f;
				w[k] = x;
			}
		}
		return true;		
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	//Given an svd decomposition, svd_fit calculates the least 
	//squares solution to the linear equation a x = d (d[m]). 
	//cutoff (>0 and <1) determines which svd values are signficant. 


	public static void svd_fit(double[][] u, double[][] v, double[] w, double[] x,
			double[] d,  double cutoff, int m, int n)
	{
		int i, j, k;
		double s, wmin, wmax;
		double[] t = new double[n]; // temporary storage

		wmax = 0;

		for (j = 0; j < n; j++)
			wmax = Math.max(wmax, w[j]);

		wmin = cutoff * wmax;

		for (j = 0; j < n; j++)
			if (w[j] < wmin)
				w[j] = 0;

		for (j = 0; j < n; j++)
		{
			s = 0;
			if (w[j] != 0)
			{
				for (i = 0; i < m; i++)
					s += u[i][j] * d[i];

				s /= w[j];
			}

			t[j] = s;
		}

		for (j = 0; j < n; j++)
		{
			s = 0;
			for (k = 0; k < n; k++)
				s += v[j][k] * t[k];

			x[j] = s;
		}
	}

	//-------------------------------------------------------------------------------------------------
	/// <summary>
	/// determine sign of a
	/// </summary>
	/// <param name="a"></param>
	/// <returns>-1 if a -s negative, +1 if a is positive or zero</returns>
	private static double sign(double a)
	{
		if (a < 0.0) return -1.0;
		else return 1.0;
	}

	/// <summary>
	/// determines the hypotenous of a right angle triangle
	/// </summary>
	/// <param name="a">side a</param>
	/// <param name="b">side b</param>
	/// <returns>hypotenous</returns>
	private static double hypotenous(double a, double b)
	{
		return (Math.sqrt(a * a + b * b));
	}

}
