
package simulation;

/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;

import spinworks.SW_main;
import spinworks.data_par;
import spinworks.global_par;
import spinworks.Util;

/**
 * A class for reading data simulated with the Simpson program
 * @author marat
 */
public class Simpson 
{
	static int PARAM = 0;           // change to enum?  TO DO
	static int DATA = 1;
	static int SPECT = 2;			// type of data set, the two options are SPECT for spectra
	static int FID = 3;				// and FID for FID data
//-------------------------------------------------------------------------------------------------
	/**
	 * Read a spectrum file produced by the Simpson program
	 * @param fileName the full path to the file
	 * @param gp a SpinWorks global parameter object
	 * @param p a data_par object for the "set" parameters
	 * @param q a data_par object for the "status" parameters
	 * @return number of points read, negative on error
	 */
    public static int readSimpSpect(String fileName, global_par gp, data_par p, data_par q)
    {
        int mode = PARAM;						// start with parameters

        String instring;
        //char[] delim = new char[] { ' ', '=', ',' };
        String delim = "[ =,]+";
        String[] token;

        double  sw = 0.0, max = -1e10;
        int npoints = 0;

        int size = 16384;

        double[] reals, imags;
        reals = new double[1];
        imags = new double[1];

        int i = 0;						        // handy counters

//try to open the file 
        BufferedReader br;
        try { br = new BufferedReader(new FileReader(fileName)); }
        catch (Exception ex)
        {
        	Util.show_and_log("readSimpSpect(), cannot open: " + fileName, ex.getMessage());
        	return -1;
        }

        for(;;)
        {
        	try {instring = br.readLine();} 
			catch (IOException ex) 
			{
				ex.printStackTrace();
				Util.show_and_log("readSimpSpect(), read error: " + fileName, ex.getMessage());
				try { br.close(); } 
				catch (IOException e) { e.printStackTrace(); }
				return -2;
			}
        	if(instring == null) break;
        	
            if (instring.length() > 2)        // ignore any blank lines
            {
            	token = instring.split(delim);       

                if(token.length != 0)
                    if (token[0].equalsIgnoreCase("END")) break;	// end of file

                if (mode == PARAM)                      // reading parameters
                {
                    try
                    {
                        if (token[0].equalsIgnoreCase("NP"))
                        {
                            try { size = Integer.parseInt(token[1]); }
                            catch (Exception ex)
                            {
                                Util.show_and_log(String.format("error converting size value",
                                   ex.getMessage()));
                                return -3;
                            }

                            if (gp.data_array != null && size != q.SI)
                            {
                                Util.show_and_log(String.format(
                                    "Simpson spectrum size: %d != experimental size: %d", size, q.SI));
                                br.close();
                                return -3;
                            }
                            if (size > 8)
                            {
                                reals = new double[size];
                                imags = new double[size];
                            }
                            else
                            {
                                Util.show_and_log(String.format(
                                    "Simpson spectrum size: %d too small!", size));
                                br.close();
                                return -3;
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                    	Util.show_and_log(instring, ex.getMessage());
                    	return -4;
                    }

                    try
                    {
                        if (token[0].equalsIgnoreCase("SW"))
                        {
                            sw = Double.parseDouble(token[1]);
                        }
                    }
                    catch (Exception ex)
                    {
                    	Util.show_and_log(instring, ex.getMessage());
                    }
                }

                else if (mode == DATA)
                {
                    try
                    {
                        reals[npoints] = Double.parseDouble(token[0]);
                        imags[npoints] = Double.parseDouble(token[1]);
                        if (reals[npoints] > max) max = reals[npoints];
                        npoints++;
                    }
                    catch (Exception ex)
                    {
                    	Util.show_and_log(instring, ex.getMessage());
                    }
                }
                if (token[0].equalsIgnoreCase("DATA")) mode = DATA;
            }
        }   // while

        if (npoints != size)
        {
            Util.show_and_log(String.format(" points read %d != NP %d", npoints, size));
            try {	br.close();	} 
            catch (IOException ex) 
            {	ex.printStackTrace();	} 
        	return -5;
        }

        if (gp.data_array == null)
        {
            p.SI = size;   
            q.offset = sw;
            q.hzperpt = sw / (double)p.SI;
            q.SF = q.SFO1 - (sw / 2.0) / 1000000.0; p.SF = q.SF;
            q.SW = sw / q.SFO1;      p.SW = q.SW;
        }

        Util.alloc_calc(gp, p, q);
        double factor = 5.0e5 / max;

        for (i = 0; i < q.SI; i++)
        {
            gp.calc_array[i] = reals[i] * factor;
            gp.i_calc_array[i] = imags[i] * factor;
        }

        gp.calc_max = Math.abs(Util.dscan(gp.calc_array, q.SI, 0, q.SI));

        try {	br.close();	} 
        catch (IOException ex) {	ex.printStackTrace();	}     
        
        //System.out.println(String.format("points read = %d  calc_max = %f", npoints, gp.calc_max));
        return npoints;
    }

//-------------------------------------------------------------------------------------------------
    /**
	 * Read an FID file produced by the Simpson program
	 * @param fileName the full path to the file
	 * @param gp a SpinWorks global parameter object
	 * @param p a data_par object for the "set" parameters
	 * @param q a data_par object for the "status" parameters
	 * @return number of points read, negative on error
	 */
    public static int readSimpFID(String fileName, global_par gp, data_par p, data_par q)
    {
        int mode = PARAM;						// start with parameters

        String instring;
        //char[] delim = new char[] { ' ', '=', ',' };
        String delim = "[ =,]+";
        String[] token;

        double  sw = 0.0, max = 0.0;
        int npoints = 0;

        int size = 16384;

        double[] reals, imags;
        reals = new double[1];
        imags = new double[1];

        int i = 0;						        // handy counters

//try to open the file 
      //try to open the file 
        BufferedReader br;
        try { br = new BufferedReader(new FileReader(fileName)); }
        catch (Exception ex)
        {
        	Util.show_and_log("readSimpFID(), cannot open: " + fileName, ex.getMessage());
        	return -1;
        }

        for(;;)
        {
        	try {instring = br.readLine();} 
			catch (IOException ex) 
			{
				ex.printStackTrace();
				Util.show_and_log("readSimpFID(), read error: " + fileName, ex.getMessage());
				try { br.close(); } 
				catch (IOException e) { e.printStackTrace(); }
				return -2;
			}
        	if(instring == null) break;
        	
            if (instring.length() > 2)        // ignore any blank lines
            {
            	token = instring.split(delim);       

                if(token.length != 0)
                    if (token[0].equalsIgnoreCase("END")) break;	// end of file
                
                if (mode == PARAM)                      // reading parameters
                {
                    try
                    {
                        if (token[0].equalsIgnoreCase("NP"))
                        {
                            size = Integer.parseInt(token[1]);
                            if (size > 8)
                            {
                                reals = new double[size];
                                imags = new double[size];
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                    	Util.show_and_log( instring, ex.getMessage());
                    	try { br.close(); } 
        				catch (IOException e) { e.printStackTrace(); }
                    	return -3;
                    }

                    try
                    {
                        if (token[0].equalsIgnoreCase("SW"))
                        {
                            sw = Double.parseDouble(token[1]);
                        }
                    }
                    catch (Exception ex)
                    {
                        Util.show_and_log( instring, ex.getMessage());
                        try { br.close(); } 
        				catch (IOException e) { e.printStackTrace(); }
                        return -3;
                    }
                }

                else if (mode == DATA)
                {
                    try
                    {
                        imags[npoints] = Double.parseDouble(token[0]);
                        reals[npoints] = Double.parseDouble(token[1]);
                        if (Math.abs(reals[npoints]) > Math.abs(max)) max = Math.abs(reals[npoints]);
                        if (Math.abs(imags[npoints]) > Math.abs(max)) max = Math.abs(imags[npoints]);
                        npoints++;
                    }
                    catch (Exception ex)
                    {
                    	Util.show_and_log(instring, ex.getMessage());
                    	try { br.close(); } 
        				catch (IOException e) { e.printStackTrace(); }
                    	return -3;
                    }
                }
                if (token[0].equalsIgnoreCase("DATA")) 
                	mode = DATA;
            }
        }   // while

        if (npoints != size)
        {
            Util.show_and_log(String.format(" points read %d != NP %d", npoints, size));
            try {	br.close();} 
            catch (IOException e) 
            {	e.printStackTrace();	}
            return -4;
        }

        q.TD = 2 * size;
        q.hzperpt = sw / (double)p.SI;
        q.SW_Hz = sw;


        if (!q.simpson_referenced)
        {
            q.SF = q.SFO1; // -(sw / 2.0) / 1000000.0; p.SF = q.SF;
            p.SF = q.SF;

            q.offset = sw / 2.0;

            gp.freq1 = q.offset;
            gp.freq2 = -(q.offset);
        }

        q.SW = sw / q.SFO1; p.SW = q.SW;
                 
        while (p.SI < size)
        {
            p.SI = p.SI * 2;
        }
        q.SI = p.SI;
        
        Util.alloc_proc(gp, p, q);
        gp.fid = new double[q.TD];

        double factor = 5.0e6 / max;
        for (i = 0; i < size; i++)
        {
            gp.data_array[i] = reals[i] * factor;
            gp.i_data_array[i] = imags[i] * factor;
            gp.fid[2 * i] = gp.data_array[i];
            gp.fid[2 * i + 1] = gp.i_data_array[i];
        }

        double real_max = Math.abs(Util.dscan(gp.data_array, q.SI, 0, q.SI));
        double imag_max = Math.abs(Util.dscan(gp.i_data_array, q.SI, 0, q.SI));

        if (imag_max > real_max)
            gp.data_max = imag_max;
        else
            gp.data_max = real_max;

        gp.raw_max = gp.data_max;

        gp.fid_scale = (double)gp.dysize * 0.80 / gp.raw_max;

        q.dwell = 1.0 / sw;
        gp.t1 = 0.0;
        gp.t2 = (double)(q.TD / 2) * q.dwell;

        q.data_type = SW_main.data_types.SIMPSON;
  
        try {	br.close();	} 
        catch (IOException e) {	e.printStackTrace();	}
        return npoints;
    }
//-------------------------------------------------------------------------------------------------
}
