/// <copyright>
///	Copyright (c) 2006-2017

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

package spinworks;

public class AutoPhase 
{
	 private static int MAX_refPeaks = 256;
	 
//-------------------------------------------------------------------------------------------------
//	This routine seems to work well for 13C, but can get phase 180 degrees off (baseline flat, but 
//  peaks upside down
 /**
  * Alternate or second round of phasing, based on baseline flatness
  * @param gp
  * @param p
  * @param q
  */
	public static void blfAutoPhase(global_par gp, data_par p, data_par q)
	{
		double offset = 0.0, bl_quality, new_bl_quality;
		double noise;
		double angle = 1.0, phase_0 = 0.0;
		int iter;
		boolean reverse = false;

		for (int i = 1; i <= 3; i++)
		{
			angle = 1.0 / (double)i;

			noise = Util.noise(gp.data_array, q.SI, q.aq_mode);
			offset = AutoPhase.baselineOffset(gp, q);
			bl_quality = AutoPhase.baselineQuality(noise * 20, offset, gp, q);

			//Initial zero order correction
			for (iter = 1; iter <= 24; iter++)		// zero order
			{
				Fft.do_phase(gp.data_array, gp.i_data_array, q.SI, angle, 0.0);
				q.PHC0 = q.PHC0 + angle;

				noise = Util.noise(gp.data_array, q.SI, q.aq_mode);
				offset = AutoPhase.baselineOffset(gp, q);
				new_bl_quality = AutoPhase.baselineQuality(noise * 20, offset, gp, q);

				if (new_bl_quality < bl_quality)
				{
					if (reverse) break;
					reverse = true;
					angle = -angle;
				}
				bl_quality = new_bl_quality;
			}

			noise = Util.noise(gp.data_array, q.SI, q.aq_mode);
			offset = AutoPhase.baselineOffset(gp, q);
			bl_quality = AutoPhase.baselineQuality(noise * 20, offset, gp, q);

			angle = 3.0 / (double)i;

			for (iter = 1; iter <= 24; iter++)		// first order
			{
				phase_0 =
					-angle * ((double)(gp.pivot) / (double)(q.SI));

				Fft.do_phase(gp.data_array, gp.i_data_array, q.SI, phase_0, angle);
				q.PHC0 = q.PHC0 + phase_0;
				q.PHC1 = q.PHC1 + angle;

				noise = Util.noise(gp.data_array, q.SI, q.aq_mode);
				offset = AutoPhase.baselineOffset(gp, q);
				new_bl_quality = AutoPhase.baselineQuality(noise * 20, offset, gp, q);

				if (new_bl_quality < bl_quality)
				{
					if (reverse) break;
					reverse = true;
					angle = -angle;
				}
				bl_quality = new_bl_quality;
			}
		}

		p.PHC0 = q.PHC0;
		p.PHC1 = q.PHC1;
	}

//-------------------------------------------------------------------------------------------------
// linear regression method
	/**
	 * @param gp
	 * @param p
	 * @param q
	 */
	public static void lrAutoPhase(global_par gp, data_par p, data_par q)
	 {
		 int i, j, count = 0;
		 double[] reals, imags;
		
		 RefRegion[] refRegions = new RefRegion[MAX_refPeaks];
		 
		 p.PHC0 = 0.0; p.PHC1 = 0.0;

		 // allocate memory 
		 reals = new double[q.SI];
		 imags = new double[q.SI];
		 
		 for (i = 0; i < MAX_refPeaks; i++)
			 refRegions[i] = new RefRegion();		 
		 
// find an initial zero order by maximizing integral of whole spectrum		 
		 int iter;
		 double integ = 0.0, new_integ,  angle = 0.0;
		 double mult = 1.0;
		 double phase = 0.0;
		 double max_integ = 0.0;
		 
		 System.out.println("\n***********  zero order from total integral");
		 integ = totalInteg( gp.data_array, i, q.SI);
		 
		 for (iter = 1; iter <= 16; iter++)
		 {			
			 angle = (20.0 * mult);

			 if(Math.abs(angle) < 1.0)
				 if(angle < 0.0) 
					 angle = -1.0;
				 else
					 angle = 1.0;
						 
			 Fft.do_phase(gp.data_array, gp.i_data_array, q.SI, angle, 0.0);
			 phase = phase + angle;				 
			 new_integ = totalInteg(gp.data_array, i, q.SI);	
			 
			 if (new_integ > max_integ) max_integ = new_integ; 
			 
			 System.out.println(String.format("   integral = %f   phase = %f ", 
					 new_integ / 1.0e6, phase));

			 if(new_integ < integ)
			 {
				 if (iter == 1) mult = -mult;
				 else mult = mult * (-0.5);
			 }
			 integ = new_integ;
		 }	// for(iter...
		 
		 p.PHC0 = phase;
		 				 				 
		 int baselineFilter = 4;		// these values seem good for 13C
		 double baselineThresh = 4.0;
		 
		 if(q.SFO1 > 299.0)
		 {
			 baselineFilter = 4;		// these values seem better for 1H
			 baselineThresh = 8.0;
		 }
		 
		 boolean[] isPeak = new boolean[q.SI];
		 boxcarFilter(gp.data_array, gp.i_data_array, isPeak, baselineFilter, baselineThresh, q.SI);
		 count = findPeakRegions(isPeak, refRegions, q.SI);
				 
		 if(count < 2)
		 { 
			 reals = null; imags = null;	// not really necessaary
			 p.PHC0 = q.PHC0;
			 p.PHC1 = q.PHC1;
			 System.out.println("AutoPhase: only one peak found, only zero order performed");
			 return;
		 }

		 System.out.println();
		 
// Now find the phase for each peak		 
		 double[] x = new double[count];
		 double[] y = new double[count];
		 
		 for (i = 0; i < count; i++)	
		 {
			 integ = 0.0; new_integ = 0.0;  angle = 0.0;
			 mult = 1.0;
			 phase = 0.0;
			 
			 for(j = 0; j < q.SI; j++)
			 {
				 reals[j] = gp.data_array[j];
				 imags[j] = gp.i_data_array[j];
			 }
				
			 integ = regionInteg(refRegions, reals, i, q.SI);
			 for (iter = 1; iter <= 24; iter++)
			 {			
				 angle = (40.0 * mult);

				 if(Math.abs(angle) < 1.0)
					 if(angle < 0.0) 
						 angle = -1.0;
					 else
						 angle = 1.0;
							 
				 Fft.do_phase(reals, imags, q.SI, angle, 0.0);
				 phase = phase + angle;		
				 new_integ = regionInteg(refRegions, reals, i, q.SI);
				 
				 if(new_integ < integ)
				 {
					 if (iter == 1) mult = -mult;
					 else mult = mult * (-0.6);
				 }
				 integ = new_integ;				 			     					     			     
			 }	// for(iter...
			 			 
			 x[i] = ((double)(refRegions[i].end + refRegions[i].start) /2.0) / (double)q.SI;
			 y[i] = phase;	
			 
			 refRegions[i].phase = phase;	
			 
			 System.out.println(String.format("   peak %d  at  %.3f ppm  to %.3f      phase = %f ", i,
					 (q.offset - (double)refRegions[i].start * q.hzperpt)/q.SF,
					 (q.offset - (double)refRegions[i].end * q.hzperpt)/q.SF,
					  phase));
		 }	// for(i...   looping over all peaks
		 	 		 	 
		 RegResult fit = linearRegression(x, y, count);
		 
		 System.out.println(String.format(
				 "slope =  %.3f +/- %.3f  intercept = %.3f  +/- %.3f    R^2 = %.4f", 
				 fit.slope, fit.stdErrorSlope, fit.intercept, fit.stdErrorIcpt, fit.R2));		
		 
		// only do full autophase if values reasonable
		 if(Math.abs(fit.intercept) < 180.0 && Math.abs(fit.slope) < 360.0)	
		 {	 
			 Fft.do_phase(gp.data_array, gp.i_data_array, q.SI, fit.intercept, fit.slope);		 
			 p.PHC0 = q.PHC0 + fit.intercept;
			 p.PHC1 = q.PHC1 + fit.slope;
		 }
		 else
			 System.out.println("AutoPhase: linear regression out of range, only zero order performed");
		 
		 reals = null; imags = null;	// not really necessaary
		 p.PHC0 = q.PHC0;
		 p.PHC1 = q.PHC1;
	 }
//-------------------------------------------------------------------------------------------------
	@SuppressWarnings("unused")
/**
 * calculate a phase asymmetry parameter
 * @param ref_peak
 * @param n
 * @param gp
 * @param q
 * @return
 */
	private static double phaseAsym(RefPeak[] ref_peak, int n,
			global_par gp, data_par q)
	{
		if (ref_peak[n].index - ref_peak[n].width <= 0) return 1.0;
		if (ref_peak[n].index + ref_peak[n].width >= q.SI) return 1.0;

		double min 
		= Math.min(Math.abs((double)(gp.data_array[ref_peak[n].index - ref_peak[n].width]/2)),
				(double)(Math.abs(gp.data_array[ref_peak[n].index + ref_peak[n].width]/2)));

		double max = Math.max((double)(gp.data_array[ref_peak[n].index - ref_peak[n].width]/2),
				(double)(gp.data_array[ref_peak[n].index + ref_peak[n].width]/2));

		return (Math.abs(min / max));
	}
	
//-------------------------------------------------------------------------------------------------
	@SuppressWarnings("unused")
	/**
	 * Calculate a phase asymetry parameter
	 * @param ref_peak the reference peak array
	 * @param n the number of the reference peak
	 * @param reals the spectrum array (reals only)
	 * @param size number of points in the real spectrum (typically q.SI)
	 * @return the asymmetry parameter
	 */
	private static double phaseAsym(RefPeak[] ref_peak,  double[] reals, int n, int size)
	{
		if (ref_peak[n].index - ref_peak[n].width <= 0) return 1.0;
		if (ref_peak[n].index + ref_peak[n].width >= size) return 1.0;

		double min 
		= Math.min(Math.abs((double)(reals[ref_peak[n].index - ref_peak[n].width]/2)),
				(double)(Math.abs(reals[ref_peak[n].index + ref_peak[n].width]/2)));

		double max = Math.max((double)(reals[ref_peak[n].index - ref_peak[n].width]/2),
				(double)(reals[ref_peak[n].index + ref_peak[n].width]/2));

		return (Math.abs(min / max));
	}
//-------------------------------------------------------------------------------------------------
/**
 * calculate a baseline threshold
 * @param gp
 * @param q
 * @return
 */
	public static double baselineThresh(global_par gp, data_par q)
	{
		int start = 0, end = q.SI-1;
		double thresh_L = 0.0, thresh_R = 0.0;

		if (q.aq_mode == SW_main.det_modes.DSP)
		{
			start = q.SI / 32;
			end = q.SI - (q.SI / 32);
		}

		for (int i = start; i < start + q.SI / 32; i++ )
		{
			if (Math.abs(gp.data_array[i]) > thresh_L)
				thresh_L = Math.abs(gp.data_array[i]);
		}

		for (int i = end; i > end - q.SI / 32; i--)
		{
			if (Math.abs(gp.data_array[i]) > thresh_R)
				thresh_R = Math.abs(gp.data_array[i]);
		}
		return (thresh_L + thresh_R) / 2;
	}

//-------------------------------------------------------------------------------------------------
/**
 * calculate number of points beneath a certain threshold
 * @param thresh
 * @param offset
 * @param gp
 * @param q
 * @return
 */
	public static int baselineQuality(double thresh, double offset, global_par gp, data_par q)
	{
		int start = 0, end = q.SI - 1;
		

		if (q.aq_mode == SW_main.det_modes.DSP)
		{
			start = q.SI / 32;
			end = q.SI - (q.SI / 32);
		}

		int count = 0;

		for (int i = start + (q.SI / 32); i < end - (q.SI / 32); i++)
		{
			if ((gp.data_array[i] < offset + thresh) &&
				(gp.data_array[i] > offset - thresh))
				count++;
		}
		return count;
	}

//-------------------------------------------------------------------------------------------------
/**
 * calculate number of points beneath a certain threshold
 * @param thresh
 * @param offset
 * @param gp
 * @param q
 * @return
 */
	public static int baselineQuality(double[] array, double thresh, double offset, int start, int end)
	{
		int count = 0;

		for (int i = start; i < end; i++)
		{
			if ((array[i] < offset + thresh) &&
					(array[i] > offset - thresh))
				count++;
		}
		return count;
	}
		
//-------------------------------------------------------------------------------------------------
/**
 * calculate average baseline offset at the two ends
 * @param gp
 * @param q
 * @return
 */
	public static double baselineOffset(global_par gp, data_par q)
	{
		int start = 0, end = q.SI - 1;
		double offset_L = 0.0, offset_R = 0.0;

		if (q.aq_mode == SW_main.det_modes.DSP)
		{
			start = q.SI / 32;
			end = q.SI - (q.SI / 32);
		}

		for (int i = start; i < start + q.SI / 32; i++)
		{
				offset_L = offset_L + gp.data_array[i];
		}
		offset_L = offset_L / (q.SI / 32);

		for (int i = end; i > end - q.SI / 32; i--)
		{
				offset_R = offset_R + gp.data_array[i];
		}
		offset_R = offset_R / (q.SI / 32);

		return (offset_L + offset_R) / 2;
	}
 //-------------------------------------------------------------------------------------------------
 /**
  * calculate integrals for auto phase routine
  * @param ref_peak
  * @param n
  * @param gp
  * @param q
  * @return the integral of the peak at index n
  */
     public static double phaseInteg(RefPeak[] ref_peak, int n,
    		 global_par gp, data_par q)
     {
    	 int j, k, l;
    	 double integ = 0.0;

    	 k = ref_peak[n].index - ref_peak[n].width;
    	 if (k < 0) k = 0;
    	 l = ref_peak[n].index + ref_peak[n].width;
    	 if (l > q.SI) l = q.SI;

    	 for (j = k; j <= l; j++)
    		 integ = integ + gp.data_array[j];

    	 return integ;
     }
 
//-------------------------------------------------------------------------------------------------
 /**
  * calculate integrals for auto phase routine
  * @param ref_peak a reference peak list
  * @param reals the processed spectrum, real part
  * @param n the number in the peak list
  * @param size the size of the real data array (typically q.SI)
  * @return the integral value
  */
     public static double phaseInteg(RefPeak[] ref_peak, double[] reals, int n, int size)
     {
    	 int j, k, l;
    	 double integ = 0.0;

    	 k = ref_peak[n].index - ref_peak[n].width;
    	 if (k < 0) k = 0;
    	 l = ref_peak[n].index + ref_peak[n].width;
    	 if (l > size) l = size;

    	 for (j = k; j <= l; j++)
    		 integ = integ + reals[j];

    	 return integ;
     }
     
//-------------------------------------------------------------------------------------------------
 /**
  * calculate total spectral integral for auto phase routine
  * @param reals the processed spectrum, real part
  * @param n the number in the peak list
  * @param size the size of the real data array (typically q.SI)
  * @return the integral value
  */
     @SuppressWarnings("unused")
	public static double totalInteg(double[] reals, int n, int size)
     {
    	 int j;
    	 double integ = 0.0;
  	
    	 for (j = size / 16; j <= size - size / 16; j++)
    		 integ = integ + reals[j];

    	 return integ;
     }

//-------------------------------------------------------------------------------------------------
 /**
  * 
  * @param reals
  * @param size
  */
     public static int findPeakRegions(double[] array, RefRegion[] refRegion, double thresh, int size)
     {
    	 boolean onPeak = false;
    	 int j = 0;
    	    	 
    	 for(int i = size / 32; i < size - size / 32; i++)	// avoid ends of the spectrum
    	 {
    		 if(j >= MAX_refPeaks) return j;
    		 
    		 if(array[i] > thresh && !onPeak)
    		 {
    			 refRegion[j].start = i - 2;	// widen region a bit
    			 onPeak = true;
    		 }
    		 if(array[i] < thresh && onPeak)
    		 {
    			 refRegion[j].end = i + 2;		// widen region a bit
    			 onPeak = false;
    			 
    			 j++;   			 
    		 }
    	 }
    	 return j;
     }

//-------------------------------------------------------------------------------------------------
     public static int findPeakRegions(boolean[] array, RefRegion[] refRegion, int size)
     {
    	 boolean onPeak = false;
    	 int j = 0;
    	    	 
    	 for(int i = size / 32; i < size - size / 32; i++)	// avoid ends of the spectrum
    	 {
    		 if(j >= MAX_refPeaks) return j;
    		 
    		 if(array[i]  && !onPeak)
    		 {
    			 refRegion[j].start = i;
    			 onPeak = true;
    		 }
    		 if(!array[i] && onPeak)
    		 {
    			 refRegion[j].end = i;
    			 onPeak = false;

    			 j++;   			 
    		 }
    	 }
    	 return j;
     }
     
//-------------------------------------------------------------------------------------------------
 /**
  * find peak regions based on peak list
  * @param array
  * @param peaks
  * @param refRegion
  * @param thresh threshold, as a freaction of the peak intensity	
  * @param size size of the array typically q.SI
  * @param n the number of reference peaks
  * @return the number of reference regions found
  */
     public static int findPeakRegions(double[] array, RefPeak[] peaks, RefRegion[] refRegion, 
    		 double thresh, int size, int n)
     {
    	 int count = 0;
    	 int start = 0, end = 0;
    	 
    	 for(int i = 0; i < n; i++)		// loop over all found peaks
    	 {
    		 for(int j = peaks[i].index; j < size; j++)
    		 {
    			 if(array[j] < thresh * peaks[i].intensity)
    			 {
    				 end = j + 2;
    				 break;
    			 }
    		 }
    		 for(int j = peaks[i].index; j > 0; j--)
    		 {
    			 if(array[j] < thresh * peaks[i].intensity)
    			 {
    				 start = j - 2;
    				 break;
    			 }
    		 }
    		 if(i == 0)
    		 {
    			 refRegion[count].end = end;
    			 refRegion[count].start = start;
    			 count++;
    		 }
    		 else
    		 {
    			 if(start > refRegion[i-1].end)
    			 {
    				 refRegion[count].end = end;
        			 refRegion[count].start = start;
        			 count++;
    			 }
    		 }
    	 }  
    	 return count;
     }
     
//-------------------------------------------------------------------------------------------------
 /**
  * calculate integrals for auto phase routine
  * @param refRegion a reference region list
  * @param reals the processed spectrum, real part
  * @param n the number in the peak list
  * @param size the size of the real data array (typically q.SI)
  * @return the integral value
  */
     public static double regionInteg(RefRegion[] refRegion, double[] reals, int n, int size)
     {
    	 int j;
    	 double integ = 0.0;

    	 for (j = refRegion[n].start; j <= refRegion[n].end; j++)
    		 integ = integ + reals[j];

    	 return integ;
     }
     
//-------------------------------------------------------------------------------------------------
 /**
  * Linear regression routine
  * @param x array of x points
  * @param y array of y points
  * @param n number of points in the x and y arrays
  * @return a RegResult (regression result) object
  */
     public static RegResult linearRegression(double[] x, double[] y, int n)
     {
         // first pass: read in data, compute xbar and ybar
         double sumx = 0.0, sumy = 0.0, sumx2 = 0.0;
         for(int i = 0; i < n; i++) 
         {
             sumx  += x[i];
             sumx2 += x[i] * x[i];
             sumy  += y[i];
         }
         
         double xbar = sumx / n;
         double ybar = sumy / n;

         // second pass: compute summary statistics
         double xxbar = 0.0, yybar = 0.0, xybar = 0.0;
         for (int i = 0; i < n; i++) 
         {
             xxbar += (x[i] - xbar) * (x[i] - xbar);
             yybar += (y[i] - ybar) * (y[i] - ybar);
             xybar += (x[i] - xbar) * (y[i] - ybar);
         }
         double slope = xybar / xxbar;
         double intercept = ybar - slope * xbar;

         // analyze results
         int df = n - 2;
         double rss = 0.0;      // residual sum of squares
         double ssr = 0.0;      // regression sum of squares
         
         for (int i = 0; i < n; i++) 
         {
             double fit = slope*x[i] + intercept;
             rss += (fit - y[i]) * (fit - y[i]);
             ssr += (fit - ybar) * (fit - ybar);
         }
         
         double R2    = ssr / yybar;
         double svar  = rss / df;
         double svar1 = svar / xxbar;
         double svar0 = svar/n + xbar*xbar*svar1;
         svar0 = svar * sumx2 / (n * xxbar);
         
         RegResult result = new RegResult();
         result.slope = slope;
         result.intercept = intercept;
         result.R2 = R2;
         result.stdErrorSlope = Math.sqrt(svar1);
         result.stdErrorIcpt = Math.sqrt(svar0);
         return result;
     }
     
 //-------------------------------------------------------------------------------------------------
 /**
  * 
  * @param array
  * @param peaks
  * @param sens
  * @param thresh
  * @param gp
  * @param q
  * @return
  */
     public static int findPeaks(double[] array, RefPeak[] peaks, double sens, double mi, double noise,
    		 global_par gp, data_par q)
     {
    	 if (gp.data_array == null)
    	 {
    		 Util.show_and_log("No experimental data!");
    		 return 0;
    	 }

    	 int i, j;
    	 double isens;
    	 double frequency;

    	 isens = 20.0 * sens * noise;

    	 j = 0;

    	 for(i = 1; i < q.SI - 1; i++)
    	 {		 
    		 double intensity = array[i];

    		 if ((Math.abs(array[i]) - Math.abs(array[i-1])) > isens
    				 && (Math.abs(array[i]) - Math.abs(array[i+1])) > isens
    				 && intensity > mi)
    		 {
    			 frequency = q.offset - (double)(i) * q.hzperpt;

    			 peaks[j] = new RefPeak();

    			 if((frequency >= q.offset - (double)(i+1) * q.hzperpt) &&
    					 (frequency <= q.offset - (double)(i-1) * q.hzperpt))
    				 peaks[j].freq = frequency;

    			 peaks[j].intensity =
    					 array[i];
    			 
    			 peaks[j].index = i;   
    			 
    			 int k, start, end;
        		 
        		 for(k = i; k < q.SI; k ++)
        			 if (array[k] < peaks[j].intensity / 4.0)
        				 break;
        		 end = k;
        		 
        		 for(k = i; k > 0; k --)
        			 if (array[k] < peaks[j].intensity / 4.0)
        				 break;
        		 start = k;
        		 
        		 peaks[j].width = end - start;
    			 
    			 j++;
    		 }
    		     		  		 
    		 if(j >= MAX_refPeaks)  // too many peaks found 
    		 {
    			 Util.show_and_log("AutoPhase.findPeaks(): Too Many Peaks");
    			 break;
    		 }		 
    	 }  	    		     		    		 
    	 return j;
     }
 
//-------------------------------------------------------------------------------------------------
     public static void boxcarFilter(double[] real, double[] imag, boolean[] is_peak, 
    		 int filt, double bthresh, int n)
     {
    	 double sumr, sumi;
    	 
    	 // temp storage
    	 double[] tempr = new double[n];
    	 double[] tempi = new double[n];
    	 double[] res_array = new double[n];
    	 double[] i_res_array = new double[n];
    	 
    	 for (int i = filt; i < n - filt; i++)
 		{
 			sumr = 0.0; sumi = 0.0;
 			for (int j = -filt; j <= filt; j++)
 			{
 				sumr = sumr + real[i + j];
 				sumi = sumi + imag[i + j];
 			}
 			tempr[i] = sumr / (2 * filt + 1);
 			tempi[i] = sumi / (2 * filt + 1);
 		}
    	 
    	// now take the derivative (simple difference method)
 		// and calculate power spectrum
 		for (int i = filt; i < n - filt - 1; i++)
 		{
 			res_array[i] = tempr[i + 1] - tempr[i - 1];
 			i_res_array[i] = tempi[i + 1] - tempi[i - 1]; 	// derivative
 			res_array[i] = res_array[i] * res_array[i]
 					+ i_res_array[i] * i_res_array[i];     	// power spectrum
 		}
 		
 	// calculate initial threshold 
 		double max = Util.dscan(res_array, n, filt, n - filt);
 		double thresh = max / 50.0;

 		for (int i = 0; i < n; i++)
 			is_peak[i] = false;

 		double mean = 0.0;
 		double sd = 0.0;
 		
 		do  // loop, lowering threshold until reasonable peak areas are found
 		{
 			for (int i = 0; i < n; i++)
 			{
 				if (res_array[i] > thresh)
 					is_peak[i] = true;
 				else
 					is_peak[i] = false;
 			}

 			// find mean
 			sumr = 0.0; int j = 0;
 			for (int i = filt; i < n - filt; i++)
 			{
 				if (!is_peak[i])
 				{
 					sumr = sumr + res_array[i];
 					j++;
 				}
 			}
 			mean = sumr / (double)j;

 			// find standard deviation
 			sumr = 0.0; j = 0;
 			for (int i = filt; i < n - filt; i++)
 			{
 				if (!is_peak[i])
 				{
 					sumr = sumr + (mean - res_array[i]) * (mean - res_array[i]);
 					j++;
 				}
 			}
 			sd = Math.sqrt(sumr / (double)j);

 			thresh = thresh / 2.0;

 		} while (sd > bthresh * mean);      // end of loop, add other tests to 


 		is_peak[0] = is_peak[1] = false;
 		is_peak[n - 1] = is_peak[n - 2] = false;
 		
 		boolean[] temp = new boolean[n];
 		
 		//widen the peak windows a bit
 		for (int j = 0; j < 5; j++)     //  depends on size???
 		{
 			for (int i = 0; i < n; i++)
 				temp[i] = is_peak[i];

 			for (int i = 1; i < n - 2; i++)
 			{
 				if (!temp[i] && temp[i + 1])
 					is_peak[i] = true;

 				if (temp[i] && !temp[i + 1])
 					is_peak[i + 1] = true;
 			}
 		}
 				
 		//erosion filter
 		for (int i = 0; i < n; i++)
 			temp[i] = is_peak[i];
 		
 		int count = 0;
 		
 		for (int i = 2; i < n - 3; i++)
 		{
 			if (!temp[i] && temp[i - 1] && temp[i + 1])
 			{
 				is_peak[i] = true;
 				count++;
 			}
 			if (!temp[i] && !temp[i + 1] && temp[i - 1] && temp[i + 2])
 			{
 				is_peak[i] = is_peak[i + 1] = true;
 				count++;
 			}
 			if (!temp[i] && !temp[i + 1] && !temp[i + 2]
 					&& temp[i - 1] && temp[i + 3])
 			{
 				is_peak[i] = is_peak[i + 1] = is_peak[i + 2] = true;
 				count++;
 			}
 		}
 		
 		System.out.println(String.format("Autophase: erosion count = %d", count));		
     }
//-------------------------------------------------------------------------------------------------
     public static class RefPeak
     {
         public int index;			// index of peak in spectrum
         public double freq;		// frequency of the peak in Hz.
         public double intensity;	// intensity of the peak
         public int width;			// width of the peak on points
         public double phase;		// phase of the peak
     }
     
     public static class RefRegion
     {
         public int start;		// index of peak in spectrum
         public int end;		// width of the peak on points
         public double integ;	// integral, necessary?
         public double phase;	// phase of the peak
     }
     
     public static class RegResult
     {
    	 double slope;			// the slsope of the line
    	 double intercept;		// the y intercept on the line
    	 double R2;				// R^2 correlation coefficient
    	 double stdErrorSlope;	// standard error of slope
    	 double stdErrorIcpt;	// standard error of intercept
     }
     
//------------------------------------------------------------------------------------------------
}
