
/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

package spinworks;

public class BaseLine 
{
//-------------------------------------------------------------------------------------------------
	/**
	 * Simple-minded linear baseline correction 
	 * (Varian style "drift correction") to F2 or 1D spectrum
	 * @param array 1D data array
	 * @param n number of points
	 */
	public static void drift_corr(double[] array, int n)
	{
		int i, n_points;
		double sum;

		n_points = n / 16;							// use 1/16 of data at each end
		if (n_points < 4) n_points = 4;				// stupidly small data set
		double slope, left_offset, right_offset;

		sum = 0.0;
		for (i = 0; i < n_points; i++)
			sum = sum + array[i];
		left_offset = sum / (double)n_points;

		sum = 0.0;
		for (i = n - n_points; i < n; i++)
			sum = sum + array[i];
		right_offset = sum / (double)n_points;

		slope = (right_offset - left_offset) / (double)(n - n_points);

		for (i = 0; i < n; i++)
		{
			array[i] = array[i]
					- (int)left_offset
					- (int)((double)i * slope);
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Apply a simple-minded Varian style "drift correction" to a 2D column
	 * @param array the 2D data array (real, real
	 * @param col the column to correct
	 * @param n number of points in the colum
	 */
	public static void drift_corr(float[][] array, int col, int n)
	{
		int i, n_points;
		double slope, left_offset, right_offset, sum;

		n_points = n / 16;						// use 1/16 of data at each end
		if (n_points < 4) n_points = 4;			// stupidly small data set

		sum = 0.0;
		for (i = 0; i < n_points; i++)
			sum = sum + array[i] [col];
		left_offset = (double)sum / (double)n_points;

		sum = 0;
		for (i = n - n_points; i < n; i++)
			sum = sum + array[i][col];
		right_offset = (double)sum / (double)n_points;

		slope = (right_offset - left_offset) / (double)(n - n_points);


		for (i = 0; i < n; i++)
		{
			array[i][col] = array[i][col]
					- (int)left_offset
					- (int)((double)i * slope);
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * polynomial least-squares baseline correction 1D
	 * @param gp a global; parameter object
	 * @param q status acquisition parameters
	 * @return true if OK, false on convergence failure
	 */
	public static boolean bl_corr(global_par gp, data_par q)
	{
		int i,j;
		double correction;
		int point1, point2;
		int[] lista = new int[6];
		double[] coefficient = new double[6];
		double[][] covar = new double[gp.abs_degree][gp.abs_degree];

		if (gp.baseline_y[0] == 0.0)    // use old x values but find new y values
		{
			int range = q.SI / 1024;
			if (range < 8) range = 8;
			int count;
			double sum;

			for (i = 0; i < gp.n_bl_pts; i++)
			{
				try
				{
					point1 = (int) Math.round(gp.baseline_x[i]) - range;
					if (point1 < 0) point1 = 0;

					point2 = (int) Math.round(gp.baseline_x[i]) + range;
					if (point2 > q.SI - 1) point2 = q.SI - 1;
				}
				catch (Exception ex)
				{
					Util.show_and_log("bl_corr: point conversion  ", ex.getMessage());
					return false;
				}

				count = 0; sum = 0.0;

				for (j = point1; j <= point2; j++)
				{
					sum = sum + gp.data_array[j];
					count++;
				}
				gp.baseline_y[i] = sum / (double)count;
			}
		}

		if(gp.n_bl_pts < 3)				// more points please
		{
			Util.show_and_log("Too few points, ABS aborted", "SpinWorks 4");
			return false;
		}

		if (gp.abs_degree >= gp.n_bl_pts - 1)	// I'll try, but more points would be nice 
		{
			gp.abs_degree = gp.n_bl_pts - 2;
			Util.show_and_log("Not enough baseline points! ABS degree reduced!");
		}

		for (i = 0; i < gp.abs_degree; i++)   // re-set coefficients 
			coefficient[i] = 0.0;
		for (i = 0; i < gp.abs_degree; i++) 
			lista[i] = i;

		// now fit baseline points to a polynomial 			
		lfit(gp.baseline_x, gp.baseline_y, gp.baseline_sig, gp.n_bl_pts,
				coefficient, gp.abs_degree, lista, gp.abs_degree, covar);			 

		// end points for correction (display region)
		point1 = (int)((q.offset - gp.freq1) / q.hzperpt);
		point2 = (int)((q.offset - gp.freq2) / q.hzperpt);

		// check that points are within array
		if(point1 < 0) point1 = 0;
		if(point2 >= q.SI) point2 = q.SI;

		// now subtract polynomial from spectrum
		for(i=point1; i < point2; i++)
		{
			correction = 0.0;
			for(j=0; j < gp.abs_degree; j++)
			{			
				correction = correction + coefficient[j]*Math.pow((double)i, (double)j);
			}
			gp.data_array[i] = gp.data_array[i] - correction;
		}

		for (i = 0; i < gp.n_bl_pts; i++)
			gp.baseline_y[i] = 0.0;

		return true;
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * cubic spline baseline correction 1D
	 * @param gp a global; parameter object
	 * @param q status acquisition parameters
	 * @return true if OK, false on convergence failure
	 */
	public static boolean bl_corr_spline(global_par gp, data_par q)
	{
		int point1, point2;

		if (gp.n_bl_pts < 5)				// more points please
		{
			Util.show_and_log("Too few points, spline ABS aborted");
			return false;
		}

		if (gp.baseline_y[0] == 0.0) // use old x values but find new y values
		{
			int range = q.SI / 1024;
			if (range < 8) range = 8;
			double sum;
			int count;

			for (int i = 0; i < gp.n_bl_pts; i++)
			{
				try
				{
					point1 = (int) Math.round(gp.baseline_x[i]) - range;
					if (point1 < 0) point1 = 0;

					point2 = (int) Math.round(gp.baseline_x[i]) + range;
					if (point2 > q.SI - 1) point2 = q.SI - 1;
				}
				catch (Exception ex)
				{
					Util.show_and_log("bl_corr: point conversion ", ex.getMessage());
					return false;
				}

				count = 0; sum = 0.0;

				for (int j = point1; j <= point2; j++)
				{
					sum = sum + gp.data_array[j];
					count++;
				}
				gp.baseline_y[i] = sum / (double)count;
			}
		}

		double[] x = new double[gp.n_bl_pts + 1]; 
		double[] y = new double[gp.n_bl_pts + 1];
		double[] y2 = new double[gp.n_bl_pts + 1];

		// offset x and y arrays for spline, which still use unit offset (Fortran type) arrays
		// preserve original arrays!
		for (int i = 0; i < gp.n_bl_pts; i++)
		{
			x[i + 1] = gp.baseline_x[i];
			y[i + 1] = gp.baseline_y[i];
		}

		// end points for correction (display region)
		point1 = (int)((q.offset - gp.freq1) / q.hzperpt);
		point2 = (int)((q.offset - gp.freq2) / q.hzperpt);

		// check that points are within array
		if (point1 < 0) point1 = 0;
		if (point2 >= q.SI) point2 = q.SI;

		Cspline.spline(x, y, gp.n_bl_pts, 0.0, 0.0, y2);   // generate cubic spline

		for (int i = point1; i < point2; i++)               // subtract baseline from spectrum
		{
			gp.data_array[i] = gp.data_array[i]
					- Cspline.splint(x, y, y2, gp.n_bl_pts, (double)i);
		}

		for (int i = 0; i < gp.n_bl_pts; i++)
			gp.baseline_y[i] = 0.0;

		return true;
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Whittaker baseline smoothing routine
	 * @param array the real data array to be baseline corected
	 * @param i_array  the imaginary part of the data to be baseline corected
	 * @param n number of points
	 * @param filt baseline filter width in points (typically 2)
	 * @param bthresh baseline threshold (typically 1.5 to 2)
	 * @param lambda Lambda smoothing paramater (typically 50000 to 100000, or larger)
	 */
	public static void whittaker_bl(double[] array, double[] i_array, int n, int filt, 
			double bthresh, double lambda)
	{
		int i, j;
		double sumr, sumi;
		double mean = 0.0, sd = 0.0;
		double max = 0.0, thresh = 0.0;

		// temporary storage
		boolean[] is_peak = new boolean[n];
		double[] tempr = new double[n];
		double[] tempi = new double[n];
		double[] res_array = new double[n];
		double[] i_res_array = new double[n];      

		// first apply a smoothing filter (Simple boxcar smooth)
		for (i = filt; i < n - filt; i++)
		{
			sumr = 0.0; sumi = 0.0;
			for (j = -filt; j <= filt; j++)
			{
				sumr = sumr + array[i + j];
				sumi = sumi + i_array[i + j];
			}
			tempr[i] = sumr / (2 * filt + 1);
			tempi[i] = sumi / (2 * filt + 1);
		}

		// now take the derivative (simple difference method)
		// and calculate power spectrum
		for (i = filt; i < n - filt - 1; i++)
		{
			res_array[i] = tempr[i + 1] - tempr[i - 1];
			i_res_array[i] = tempi[i + 1] - tempi[i - 1]; 		// derivative
			res_array[i] = res_array[i] * res_array[i]
					+ i_res_array[i] * i_res_array[i];      	// power spectrum
		}

		// calculate initial threshold 
		max = Util.dscan(res_array, n, filt, n - filt);
		thresh = max / 50.0;

		for (i = 0; i < n; i++)
			is_peak[i] = false;

		do  // loop, lowering threshold until reasonable peak areas are found
		{

			for (i = 0; i < n; i++)
			{
				if (res_array[i] > thresh)
					is_peak[i] = true;
				else
					is_peak[i] = false;
			}

			// find mean
			sumr = 0.0; j = 0;
			for (i = filt; i < n - filt; i++)
			{
				if (!is_peak[i])
				{
					sumr = sumr + res_array[i];
					j++;
				}
			}
			mean = sumr / (double)j;

			// find standard deviation
			sumr = 0.0; j = 0;
			for (i = filt; i < n - filt; i++)
			{
				if (!is_peak[i])
				{
					sumr = sumr + (mean - res_array[i]) * (mean - res_array[i]);
					j++;
				}
			}
			sd = Math.sqrt(sumr / (double)j);

			thresh = thresh / 2.0;

		} while (sd > bthresh * mean);      // end of loop, add other tests to 


		is_peak[0] = is_peak[1] = false;
		is_peak[n - 1] = is_peak[n - 2] = false;

		boolean[] temp = new boolean[n];

		//widen the peak windows a bit
		for (j = 0; j < 5; j++)     //  depends on size???
		{
			for (i = 0; i < n; i++)
				temp[i] = is_peak[i];

			for (i = 1; i < n - 2; i++)
			{
				if (!temp[i] && temp[i + 1])
					is_peak[i] = true;

				if (temp[i] && !temp[i + 1])
					is_peak[i + 1] = true;
			}
		}

		//erosion filter
		for (i = 0; i < n; i++)
			temp[i] = is_peak[i];
		int count = 0;
		for (i = 2; i < n - 3; i++)
		{
			if (!temp[i] && temp[i - 1] && temp[i + 1])
			{
				is_peak[i] = true;
				count++;
			}
			if (!temp[i] && !temp[i + 1] && temp[i - 1] && temp[i + 2])
			{
				is_peak[i] = is_peak[i + 1] = true;
				count++;
			}
			if (!temp[i] && !temp[i + 1] && !temp[i + 2]
					&& temp[i - 1] && temp[i + 3])
			{
				is_peak[i] = is_peak[i + 1] = is_peak[i + 2] = true;
				count++;
			}
		}

		if(SW_main.gp.log_on)
			Util.log_error(String.format("Whittiker BL: erosion count = %d", count));

		double[] w = new double[n];
		double[] z = new double[n];

		for (i = 0; i < n; i++)
		{               
			if (is_peak[i])
				w[i] = 0.0;
			else
				w[i] = 1.0;
		}

		FiniteDiff.fdiff_1a(w, array, z, lambda, n);

		for (i = 0; i < n; i++)
		{
			res_array[i] = z[i];
			array[i] = array[i] - z[i];
		} 
	}
	
//-------------------------------------------------------------------------------------------------
//		The following functions are (very) loosly adapted from "Numerical Recipes in C"
//-------------------------------------------------------------------------------------------------

	private static double lfit(double[] x, double[] y, double[] sig, int ndata, double[] a, 
			int ma, int[] lista, int mfit, double[][] covar)
	{
		int k, kk, j, ihit, i;
		double ym, wt, sum, sig2i, chisq;	

		double[][] beta = new double[ma][1];
		double[] afunc = new double[ma];

		kk = mfit+1;	

		for (j=0; j < ma; j++) 
		{
			ihit = 0;

			for (k=0; k < mfit; k++)
			{
				if (lista[k] == j) ihit++;
			} 

			if (ihit == 0)
				lista[kk++] = j;
			else if (ihit > 1) 
			{				
				Util.show_and_log("Bad LISTA permutation in LFIT: 1");
			}
		}

		if (kk != (ma + 1))
		{
			Util.show_and_log("Bad LISTA permutation in LFIT: 2");
		}

		for (j=0; j < mfit; j++) 
		{
			for (k=0; k < mfit; k++)
			{
				covar[j][k] = 0.0;
			}
			beta[j][0] = 0.0;
		}

		for (i=0; i < ndata; i++)
		{
			fpoly(x[i], afunc, ma);

			ym = y[i];

			if (mfit < ma)
				for (j=mfit; j < ma; j++)
					ym -= a[lista[j]] * afunc[lista[j]];

			sig2i = 1.0/(sig[i] * sig[i]);

			for (j = 0; j < mfit;  j++)
			{
				wt = afunc[lista[j]] * sig2i;
				for (k = 0; k <= j; k++)			
					covar[j][k] += wt * afunc[lista[k]];
				beta[j][0] += ym * wt;
			}
		}

		if (mfit > 1)	
			for (j=1; j < mfit; j++)
				for (k = 0; k <= j - 1; k++)
				{
					covar[k][j] = covar[j][k];
				}

		gaussj(covar, mfit, beta, 1);

		for (j = 0; j < mfit; j++)
		{
			a[lista[j]] = beta[j][0];
		}

		chisq = 0.0;

		for (i=0; i < ndata; i++) 
		{
			fpoly(x[i], afunc, ma);

			for (sum=0.0, j=0; j < ma; j++) 
			{
				sum += a[j]*afunc[j];
			}
			chisq += (y[i] - sum) / sig[i] * (y[i] - sum) / sig[i];
		}

		return chisq;
	}
	
//-------------------------------------------------------------------------------------------------

	private static void fpoly(double x, double[] p, int np)
	{
		int j;

		p[0] = 1.0;

		for (j = 1; j < np; j++)
			p[j] = p[j - 1] * x;
	}

//-------------------------------------------------------------------------------------------------
	private static void gaussj(double[][] a, int n, double[][] b, int m)
	{
		int i, icol = 0, irow = 0, j, k, l, ll;
		double big, dum, pivinv, temp;

		int[] indxc = new int[n];
		int[] indxr = new int[n];
		int[] ipiv  = new int[n];

		for (j=0; j < n; j++) 
			ipiv[j] = 0;

		for (i=0; i < n; i++) 
		{
			big = 0.0;
			for (j=0; j < n; j++)
				if (ipiv[j] != 1)
					for (k=0; k < n; k++)
					{
						if (ipiv[k] == 0) 
						{
							if (Math.abs(a[j][k]) >= big) 
							{
								big = Math.abs(a[j][k]);
								irow = j;
								icol = k;
							}
						} 
						else if (ipiv[k] > 1)
						{ 
							Util.show_and_log("GAUSSJ: Singular Matrix: 1");
						}
					}

			++(ipiv[icol]);

			if (irow != icol) 
			{
				for (l=0; l < n; l++)
				{
					temp = a[irow][l];
					a[irow][l] = a[icol][l];
					a[icol][l] = temp;
				}
				for (l = 0; l < m; l++)	
				{
					temp = b[irow][l];
					b[irow][l] = b[icol][l];
					b[icol][l] = temp;
				}
			}

			indxr[i] = irow;
			indxc[i] = icol;

			if (a[icol][icol] == 0.0)
			{
				Util.show_and_log("GAUSSJ: Singular Matrix: 2");
			}

			pivinv = 1.0/a[icol][icol];
			a[icol][icol] = 1.0;

			for (l=0; l < n; l++) 
				a[icol][l] *= pivinv;

			for (l=0; l < m; l++) 
				b[icol][l] *= pivinv;

			for (ll=0; ll < n; ll++)
				if (ll != icol) 
				{
					dum = a[ll][icol];
					a[ll][icol] = 0.0;
					for (l = 0; l < n; l++)
					{
						a[ll][l] -= a[icol][l] * dum;
					}
					for (l = 0; l < m; l++)
					{
						b[ll][l] -= b[icol][l] * dum;
					}
				}
		}	// for i

		for (l=n-1; l>=0; l--) 
		{

			if (indxr[l] != indxc[l])
				for (k = 0; k < n; k++)	
				{
					temp = a[k][indxr[l]];
					a[k][indxr[l]] = a[k][indxc[l]];
					a[k][indxc[l]] = temp;
				}
		}
	}
//-------------------------------------------------------------------------------------------------
}
