package spinworks;

/**
 * Cubic spline routines for SpinWorks J, adapted from
 * "Numerical Recipes in C", used for baseline correction
 * and interpolation of 2D matrices
 * @author marat,
 * University of Manitoba
 */
public class Cspline 
{
	/**
	 * Calculate a cubic spline, adapted from "Numerical recipes in C"
	 * @param x
	 * @param y
	 * @param n
	 * @param yp1
	 * @param ypn
	 * @param y2
	 * @author marat
	 */
	public static void spline(double[] x, double[] y, int n, double yp1, double ypn, 
			double[] y2)
	{
		int		i, k;
		double p, qn, sig, un;

		double[] u = new double[n];

		if(yp1 > 0.99e30)			// lower boundary condition "natural"
		{
			y2[1] = 0.0;
			u[1] = 0.0;
		}
		else
		{
			y2[1] = -0.5;
			u[1] = (3.0/(x[2]-x[1])) * ((y[2]-y[1]) / (x[2]-x[1])-yp1);
		}

		for(i = 2; i <= n-1; i++)
		{
			sig = (x[i]-x[i-1]) / (x[i+1] - x[i-1]);
			p = sig * y2[i-1] + 2.0;
			y2[i] = (sig-1.0)/p;
			u[i] = (y[i+1] - y[i]) / (x[i+1] - x[i]) 
					- (y[i] - y[i-1]) / (x[i] - x[i-1]);
			u[i] = (6.0 * u[i] / (x[i+1] - x[i-1])
					- sig * u[i-1]) / p;
		}

		if(ypn > 0.99e30)			// upper boundary "natural"
		{
			qn = 0.0;
			un = 0.0;
		}
		else
		{
			qn = 0.5;
			un = (3.0/(x[n] - x[n-1])) * (ypn - (y[n] - y[n-1]) / (x[n] - x[n-1]));
		}


		y2[n] = (un - qn * u[n-1]) / (qn * y2[n-1] + 1.0);

		for(k = n-1; k >= 1; k--)
			y2[k] = y2[k] * y2[k+1] + u[k];
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Use the calculated spline to interpolate points, adapted from 
	 * "Numerical recipes in C"
	 * @param xa
	 * @param ya
	 * @param y2a
	 * @param n
	 * @param x
	 * @return the interpolated y value
	 * @author marat
	 */
	public static double splint(double[] xa, double[] ya, double[] y2a, int n, double x)
	{
		double y;

		int klo, khi, k;
		double h, b, a;

		klo = 1;
		khi = n;

		while(khi-klo > 1)
		{
			k = (khi + klo) >> 1;
			if(xa[k] > x) 
				khi = k;
			else
				klo = k;
		}

		h = xa[khi] - xa[klo];

		if(h == 0.0)
			Util.log_error("splint(): Warning: identical x values");

		a = (xa[khi] - x)/h;
		b = (x - xa[klo])/h;

		y = a * ya[klo]
				+ b * ya[khi]
				+ ((a*a*a - a) * y2a[klo] + (b*b*b - b) * y2a[khi]) * (h*h)/6.0;

		return y;
	}
//-------------------------------------------------------------------------------------------------	
}
