package spinworks;

import simulation.Simpson;

public class Fft 
{

	// sine table for fft.  Variable names correspond to names in fft algorithm
	public static double[] _wtemp;
	public static double[] _wpi; 
  
/**
 * Cooley-Tukey fast Fourier transform, zero index (C style) based
 * Standard Cooley-Tukey FFT algorithm, based on routine of N. Brenpairser, as described
 * in "Numerical Recipes".   
 * Note that the data file for this routine is standard 0 based C type routine.  NOT the
 * FORTRAN style 1 based indexing used in "Numerical recipes in C"
 * @param data the data are alternating real/imaginary pairs
 * @param npairs the number of real/imaginary pairs (size of data[]/2)
 * @param sign 1 for forward transform, -1 for inverse transform
 */
    public static void fft(double[] data, int npairs, int sign)
    {
        int n, mmax, m, j, istep, i;
        double wtemp, wr, wpr, wpi, wi, theta;
        double temp, tempr, tempi;

        double two_Pi = 2.0 * Math.PI;

        // put data in bit-reversed order 
        n = 2 * npairs;
        j = 1;

        for (i = 1; i < n; i += 2)
        {
            if (j > i)
            {
                temp = data[j-1];         data[j-1] = data[i-1];          data[i-1] = temp;
                temp = data[j];           data[j] = data[i];              data[i] = temp;
            }
            m = n / 2;

            while (m >= 2 && j > m)
            {
                j = j - m;
                m = m / 2;
            }
            j = j + m;
        }
        mmax = 2;

        while (n > mmax)
        {
            istep = 2 * mmax;
            theta = two_Pi / (double)(sign * mmax);
            wtemp = Math.sin(0.5 * theta);
            wpr = -2.0 * wtemp * wtemp;
            wpi = Math.sin(theta);
            wr = 1.0;
            wi = 0.0;
            for (m = 1; m < mmax; m += 2)
            {
                for (i = m; i <= n; i += istep)
                {
                    j = i + mmax;
                    tempr = wr * data[j-1] - wi * data[j];
                    tempi = wr * data[j] + wi * data[j-1];
                    data[j-1] = data[i-1] - tempr;
                    data[j] = data[i] - tempi;
                    data[i-1] += tempr;
                    data[i] += tempi;
                }
                wtemp = wr;
                wr = wtemp * wpr - wi * wpi + wr;
                wi = wi * wpr + wtemp * wpi + wi;
            }
            mmax = istep;
        }
    }
    
//------------------------------------------------------------------------------------------------   
/**
 * phase correction (1D)
 * @param reals the real data points (64 bit double)
 * @param imags the imaginary data points (64 bit double)
 * @param size the number of complex points
 * @param ph_0 zero order phase constant
 * @param ph_1 first order phase constant
 */
    public static void do_phase(double[] reals, double[] imags, int size, 
    		double ph_0, double ph_1)
    {
    	double a, t_r, t_i;

    	try
    	{
    		for (int i = 0; i < size; i++)      // now do the phasing 
    		{
    			//  The value 0.017453293 converts from degrees to radians
    			a = 0.017453293 * (ph_0 + ((double)i / (double)size) * ph_1);
    			t_r = reals[i];
    			t_i = imags[i];
    			reals[i] = t_r * Math.cos(a) - t_i * Math.sin(a);
    			imags[i] = t_i * Math.cos(a) + t_r * Math.sin(a);
    		}
    	}
    	catch (Exception ex)
    	{
    		Util.show_and_log("do_phase: " + ex.getMessage());
    	}
    }
	    
//-------------------------------------------------------------------------------------------------
  /**
   * do magnitude calculation on processed data
   * @param reals the real data points
   * @param imags the imaginary data points
   * @param size the number of real/imaginary pairs
   */
    public static void do_mc(double[] reals, double[] imags, int size)
    {
    	try
    	{
    		for (int i = 0; i < size; i++)
    			reals[i] = Math.sqrt(reals[i] * reals[i] + imags[i] * imags[i]);
    	}
    	catch (Exception ex)
    	{
    		Util.show_and_log("do_mc: " + ex.getMessage());
    		return;
    	}

    	imags = null; // imaginary data no longer valid
    }

//-------------------------------------------------------------------------------------------------
/**
 * do power spectrum on processed data
 * @param reals the real data points
 * @param imags the imaginary data points
 * @param size the number of real/imaginary pairs
 */
    public static void do_ps(double[] reals, double[] imags, int size)
    {
    	try
    	{
    		for (int i = 0; i < size; i++)
    			reals[i] = reals[i] * reals[i] + imags[i] * imags[i];
    	}
    	catch (Exception ex)
    	{
    		Util.show_and_log("do_ps: " + ex.getMessage());
    		return;
    	}

    	imags = null;        // imaginary data no longer valid
    }

//-------------------------------------------------------------------------------------------------
/**
 * do appropriate FFT transform for the data type, 
 * generic overload, free of SpinWorks objects
 * @param proc_buffer the array containing the time domain data
 * @param td_eff the number of time domain points (real + imaginary)
 * @param SI the size of the final real and imaginary arrays (number of complex pairs)
 * @param aq_mode the acquisition mode: 0, 2, 3 are complex, 1 is real 
 * (simultaneous sampling)
 */
    public static void do_fft(double[] proc_buffer, int td_eff, int SI, SW_main.det_modes aq_mode)
    {
    	int i, td;

    	//  silly case where somebody sets the size less than the number of acquired pata pairs
    	if (td_eff > 2 * SI)
    		td = 2 * SI;
    	else
    		td = td_eff;

    	// for sequential (real) data expand into real locations only
    	if (aq_mode == SW_main.det_modes.SEQ)  
    		for (i = td - 1; i > 0; i--)
    		{
    			proc_buffer[2 * i] = proc_buffer[i];
    			proc_buffer[i] = 0.0;
    		}

    	// process data.  
    	if (aq_mode == SW_main.det_modes.SIM   
    			|| aq_mode == SW_main.det_modes.DSP
    			|| aq_mode == SW_main.det_modes.CUSTOM_DSP)
    	{
    		// Fft.fft(proc_buffer, SI, 1);           // normal complex data
    		Fft.fftSineTable(proc_buffer, SI);			  // newer faster (???) fft
    	}

    	else if (aq_mode == SW_main.det_modes.SEQ)
    	{
    		Fft.fft(proc_buffer, 2 * SI, 1);       // sequential sampling (old Bruker)
    	}
    }
    
//GRPDLY
//-------------------------------------------------------------------------------------------------
/**
 * post fft processing for Bruker AVANCE DSP data, 1D, 2D and 3D
 * @param row_id row_id specifies which row is being processed, -1 for 1D data
 * @param gp is ths SpinWorks global parameter/variable object
 * @param q is the "status" data parameter object
 */
    public static void dsp_phase(int row_id, global_par gp, data_par q )
    {
    	double phase = 0.0, a, t_r, t_i;
    	int i;

    	// Early Bruker DSP.
    	// Now find the appropiate first-order phase constant.  No nice neat formula, 
    	// just a big UGLY empirical table based on the DSPFVS and DECIM parameters. 

    	int decim = (int) Math.round(q.DECIM);
    	if(q.DSPFVS == 10 && q.aq_mode != SW_main.det_modes.CUSTOM_DSP)
    	{
    		switch(decim)
    		{
    		case 2: phase = 16110.0; break;
    		case 3: phase = 12060.0; break;
    		case 4: phase = 23985.0; break;
    		case 6: phase = 21270.0; break;
    		case 8: phase = 24682.5; break;
    		case 12: phase = 21735.0; break;
    		case 16: phase = 25031.25; break;
    		case 24: phase = 21967.5; break;
    		case 32: phase = 25205.62; break;
    		case 48: phase = 22083.75; break;
    		case 64: phase = 25292.8; break;
    		case 96: phase = 22141.87; break;
    		case 128: phase = 25336.4; break;
    		case 192: phase = 22170.93; break;
    		case 256: phase = 25358.2; break;
    		case 384: phase = 22185.46; break;
    		case 512: phase = 25369.1; break;
    		case 768: phase = 22192.73; break;
    		case 1024: phase = 25374.55; break;
    		case 1536: phase = 22196.36; break;
    		case 2048: phase = 25377.27; break;
    		}
    	}
    	else if (q.DSPFVS == 11 && q.aq_mode != SW_main.det_modes.CUSTOM_DSP)
    	{
    		switch(decim)
    		{
    		case 2: phase = 16560.0; break;
    		case 3: phase = 13140.0; break;
    		case 4: phase = 17280.0; break;
    		case 6: phase = 18060.0; break;
    		case 8: phase = 19170.0; break;
    		case 12: phase = 25020.0; break;
    		case 16: phase = 26010.0; break;
    		case 24: phase = 25260.0; break;
    		case 32: phase = 26190.0; break;
    		case 48: phase = 25380.0; break;
    		case 64: phase = 26280.0; break;
    		case 96: phase = 25440.0; break;
    		case 128: phase = 26100.0; break;
    		case 192: phase = 25680.0; break;
    		case 256: phase = 26010.0; break;
    		case 384: phase = 25800.0; break;
    		case 512: phase = 25965.0; break;
    		case 768: phase = 25860.0; break;
    		case 1024: phase = 25942.5; break;
    		case 1536: phase = 25890.0; break;
    		case 2048: phase = 25931.25; break;
    		}
    	}
    	else if (q.DSPFVS == 12 && q.aq_mode != SW_main.det_modes.CUSTOM_DSP)
    	{
    		switch(decim)
    		{
    		case 2: phase = 16560.0; break;
    		case 3: phase = 13140.0; break;
    		case 4: phase = 17280.0; break;
    		case 6: phase = 18060.0; break;
    		case 8: phase = 19170.0; break;
    		case 12: phase = 25020.0; break;
    		case 16: phase = 25785.0; break;
    		case 24: phase = 25260.0; break;
    		case 32: phase = 25965.0; break;
    		case 48: phase = 25380.0; break;
    		case 64: phase = 26055.0; break;
    		case 96: phase = 25440.0; break;
    		case 128: phase = 26100.0; break;
    		case 192: phase = 25680.0; break;
    		case 256: phase = 26010.0; break;
    		case 384: phase = 25800.0; break;
    		case 512: phase = 25965.0; break;
    		case 768: phase = 25860.0; break;
    		case 1024: phase = 25942.5; break;
    		case 1536: phase = 25890.0; break;
    		case 2048: phase = 25931.25; break;
    		}
    	}   

    	//  Later Bruker DSP
    	//  For newer (Avance II and up) data, there actually is a nice neat formula based on the filter
    	//  group delay value!
    	else if((q.DSPFVS >=20 && q.DSPFVS <= 23) 
    			|| q.aq_mode == SW_main.det_modes.CUSTOM_DSP)
    	{                                           // new for Avance DRU data
    		phase = q.GRPDLY * 360.0;               // as per Mike Englehardt (Bruker)
    	}

    	else if (q.DSPFVS == -1)                    // data has been converted to analog
    	{
    		return;                                 // no DSP correction required
    	}

    	else                                        // Unknown DSP firmware
    	{
    		Util.show_and_log(String.format
    				("Data appears to be DSP, but with unknown DSP firmware version: %d    DECIM: %d",
    						q.DSPFVS, q.DECIM));
    		return;
    	}   

    	// now do the phasing  
    	// Note that since we must apply the negative of the phase value, the linear combination is slightly
    	// different than the do_phase(...  method
    	
    	if(row_id < 0 && phase != 0.0)              // 1D case
    	{
    		if (gp.log_on) Util.log_error(String.format(
    				"DSP phasing, phase = %7.2f", phase));
    		
    		for(i=0; i < q.SI; i++)   
    		{
    			a = 0.017453293 * (((double)i/(double)q.SI) * phase);
    			t_r = gp.data_array[i];	t_i = gp.i_data_array[i];
    			gp.data_array[i]   = t_r * Math.cos(a) + t_i * Math.sin(a);
    			gp.i_data_array[i] = t_r * Math.sin(a) - t_i * Math.cos(a);     			
    		}  
    	}
    	else if(row_id >= 0 && phase != 0.0)		// 2D case
    	{
    		for(i=0; i < q.SI; i++)   
    		{
    			a = 0.017453293 * (((double)i/(double)q.SI) * phase);
    			t_r = (double)gp.rows_real[row_id][i];	
    			t_i = (double)gp.rows_imag[row_id][i];
    			gp.rows_real[row_id][i]   = (float)(t_r * Math.cos(a) + t_i * Math.sin(a));
    			gp.rows_imag[row_id][i] = (float)(t_r * Math.sin(a) - t_i * Math.cos(a)); 
    		}
    	}
    }
          
//-------------------------------------------------------------------------------------------------
/**
 * reverse data order, F2 or 1D.  Usually done by inverting the sign of the 
 * imaginaries before transform.
 * @param row_id  specifies which row is being processed,  row_id = -1 for 1D 
 * @param gp the SpinWorks global parameter/variable structure
 * @param q the "status" data parameter structure, direct dimension
 * @param q_ind the "status" data parameter structure in the indirect dimension
 */
    public static void rev_data(int row_id, global_par gp, data_par q, data_par q_ind)
    {
    	//  gp.data_array[] and gp.i_data_array[] are the real and imaginary parts of
    	//  the processed data (spectrum), as arrays of doubles
    	int  i;


    	if (row_id < 0)		// 1D
    	{
    		double temp;
    		for (i = 0; i < (q.SI / 2); i++)
    		{
    			temp = gp.data_array[i];
    			gp.data_array[i] = gp.data_array[q.SI - 1 - i];
    			gp.data_array[q.SI - 1 - i] = temp;

    			temp = gp.i_data_array[i];
    			gp.i_data_array[i] = gp.i_data_array[q.SI - 1 - i];
    			gp.i_data_array[q.SI - 1 - i] = temp;
    		}
    	}
    	else				// 2D
    	{
    		float temp;
    		for (i = 0; i < (q.SI / 2); i++)
    		{
    			temp = gp.rows_real[row_id][i];
    			gp.rows_real[row_id][i] = gp.rows_real[row_id][q.SI - 1 - i];
    			gp.rows_real[row_id][q.SI - 1 - i] = temp;

    			temp = gp.rows_imag[row_id][i];
    			gp.rows_imag[row_id][i] = gp.rows_imag[row_id][q.SI - 1 - i];
    			gp.rows_imag[row_id][q.SI - 1 - i] = temp;
    		}
    	}
    }
        
//-------------------------------------------------------------------------------------------------
/**
 * Extract data from processing buffer into real and imaginary data arrays after the fft
 * @param proc_buffer The processing buffer (after fft)
 * @param reals The real part of the data
 * @param imag The imaginary part of the data
 * @param SI The number of complex pairs (size of reals and imaginaries)
 * @param aq_mode The acquisition mode.  0 = complex, 1 = sequential, 2 = DSP, 3 = custom DSP
 */
    public static void extract_data(double[] proc_buffer, double[] reals, double[] imag,
    		int SI, SW_main.det_modes aq_mode)    
    {
    	int i;

    	if (aq_mode == SW_main.det_modes.SIM 	    		// simultaneous 
    			|| aq_mode == SW_main.det_modes.DSP   		// sampling
    			|| aq_mode == SW_main.det_modes.CUSTOM_DSP )
    	{
    		for (i = SI / 2; i < SI; i++)                  	// extract real part of data 
    			reals[i] = proc_buffer[2 * (i - SI / 2)];
    		for (i = 0; i < SI / 2; i++)
    			reals[i] = proc_buffer[2 * (i + SI / 2)];

    		for (i = SI / 2; i < SI; i++)               	// extract imaginary part of the data  
    			imag[i] = proc_buffer[2 * (i - SI / 2) + 1];
    		for (i = 0; i < SI / 2; i++)
    			imag[i] = proc_buffer[2 * (i + SI / 2) + 1];
    	}

    	else if (aq_mode == SW_main.det_modes.SEQ  ) 		// sequential              
    		for (i = 0; i < SI; i++)                 		// sampling
    		{
    			reals[i]
    					= proc_buffer[(2 * SI - 1) - (2 * i + 1)];
    			imag[i]
    					= proc_buffer[(2 * SI - 1) - (2 * i)];
    		}
    }
      
//-------------------------------------------------------------------------------------------------
/**
 *  first point and bias correction
 * @param gp the SpinWorks global parameter/variable structure
 * @param p the "set" data parameter structure  (see README file!)
 * @param q the "status" data parameter structure  (see README file!)
 */
    public static void do_fp_and_bias(global_par gp, data_par p, data_par q)
    {
    	// TODO  pass only array
    	int i;
    	double sumr, sumi, fid_bias_r, fid_bias_i;

    	gp.proc_buffer[0] = gp.proc_buffer[0] * p.fcor;     // first point correction
    	if (q.aq_mode == SW_main.det_modes.SIM 
    			|| q.aq_mode == SW_main.det_modes.DSP
    			|| q.aq_mode == SW_main.det_modes.CUSTOM_DSP)
    		gp.proc_buffer[1] = gp.proc_buffer[1] * p.fcor;

    	q.fcor = p.fcor;                                    // update status parameter

    	if (p.fid_bias)
    	{
    		sumr = 0.0; sumi = 0.0;	                        // FID bias correction 
    		int j = 0;
    		int start = q.td_eff - q.td_eff / 8;
    		if (start % 2 != 0) start++;                    // make sure we start on even 
    		// (real) point

    		for (i = start; i < q.td_eff; i = i + 2)        // last 1/8 of FID 
    		{
    			sumr = sumr + gp.proc_buffer[i];
    			sumi = sumi + gp.proc_buffer[i + 1];
    			j++;
    		}

    		fid_bias_r = sumr / ((double)j);                // real channel 
    		fid_bias_i = sumi / ((double)j);                // imaginary channel 

    		for (i = 0; i < q.td_eff; i = i + 2)            // apply correction
    		{
    			gp.proc_buffer[i] = gp.proc_buffer[i] - fid_bias_r;
    			gp.proc_buffer[i + 1] = gp.proc_buffer[i + 1] - fid_bias_i;
    		} 

    		q.fid_bias = p.fid_bias;                        // update status parameter
    	}
    }
   
//--------------------------------------------------------------------------------------------------
/**
 * load an fid into the processing buffer, 1D
 * @param gp the SpinWorks global parameter/variable structure
 * @param p the "set" data parameter structure  (see README file!)
 * @param q the "status" data parameter structure  (see README file!)
 * @return true if O.K.  false on error
 */
    public static boolean load_fid(global_par gp, data_par p, data_par q)
    {
    	int rc = -1;
    	int i, j = 0;

    	if (q.data_type == SW_main.data_types.BRUKER)
    		if(q.DTYPA == 2)
    			rc = FileUtilBruker.get_double_fid(q.data_path, q, gp);
    		else
    			rc = FileUtilBruker.get_fid(q.data_path, q, gp);

    	else if (q.data_type == SW_main.data_types.VNMR)
    		rc = FileUtilVarian.getVarianFid(q.data_path, gp, q);        

    	else if (q.data_type == SW_main.data_types.JCAMP_DX)
    	{
    		if (gp.fid != null)
    			rc = gp.fid.length;
    	}

    	/*else if (q.data_type == SW_main.data_types.CDFF)
                 rc = file_util.get_cdff_type_2_fid(q.data_path, gp.vblock, gp, q);*/ // TODO

    	else if (q.data_type == SW_main.data_types.SIMPSON)
    		rc = Simpson.readSimpFID(q.data_path, gp, p, q);

    	else if (q.data_type == SW_main.data_types.JEOL)
    		rc = JDF_Util.get_jdf_fid_1D(q.data_path, gp, q); 

    	else if (q.data_type == SW_main.data_types.MAGRITEK)
                 rc = fileUtilMagritek.get_fid(q.data_path, q, gp);

    	/*else if (q.data_type == SW_main.data_types.TECMAG)
                 rc = tecmag_util.get_tecmag_file_1D(q.data_path, gp, p, q); */// TODO

    	if (rc <= 0)
    	{			
    		Util.show_and_log("cannot load FID");
    		return false;
    	}

    	Util.free_proc_buffer(gp);

    	if (q.aq_mode == SW_main.det_modes.SIM	                // simultaneous  
    			|| q.aq_mode == SW_main.det_modes.DSP
    			|| q.aq_mode == SW_main.det_modes.CUSTOM_DSP)	
    		Util.alloc_proc_buffer(2*p.SI, gp);       
    	else if(q.aq_mode == SW_main.det_modes.SEQ)            // sequential 
    		Util.alloc_proc_buffer(4*p.SI, gp);

    	int rem = 0;
    	// Math.DivRem(p.left_shift,  2, out rem);	TODO

    	if ((q.aq_mode == SW_main.det_modes.SIM	            // simultaneous 
    			|| q.aq_mode == SW_main.det_modes.DSP
    			|| q.aq_mode == SW_main.det_modes.CUSTOM_DSP)
    			&& rem != 0)
    	{              
    		Util.show_and_log("For complex data, left or right shift value must be even");
    	}

    	j = p.left_shift;

    	Util.zero_array(gp.proc_buffer);      

    	if(q.TD <= (2 * p.SI))
    		q.td_eff = q.TD;
    	else
    		q.td_eff = 2 * p.SI;

    	// extract actual fid data data into processing buffer		
    	if(j >= 0)                             // left shift or no shift case
    		for(i=0; i < (q.td_eff - j) ; i++)	                                       
    			gp.proc_buffer[i] = gp.fid[i+j];            
    	else if(j < 0)
    		for(i=0; (i-j) < q.td_eff; i++)    // right shift case
    			gp.proc_buffer[i-j] = gp.fid[i];

    	// invert imaginaries to reverse data, if required.
    	if (q.data_type == SW_main.data_types.VNMR 
    			|| q.data_type == SW_main.data_types.TECMAG
    			|| q.data_type == SW_main.data_types.JEOL 
    			|| p.REVERSE)
    		for (i = 1; i < q.td_eff; i = i + 2)
    			gp.proc_buffer[i] = -gp.proc_buffer[i];


    	if(q.aq_mode == SW_main.det_modes.SEQ)           	   // special treatment
    		for(i=0; i < q.td_eff; i = i + 4)	          	   // for sequential sampling
    		{
    			gp.proc_buffer[i+1] = -(gp.proc_buffer[i+1]); // this is the Redfield trick: 
    			gp.proc_buffer[i+2] = -(gp.proc_buffer[i+2]); // mult data by 1 -1 -1  1 
    		}


    	Util.free_proc(gp);				// free any previously allocated space 
    	Util.alloc_proc(gp, p, q);
    	q.left_shift = p.left_shift;    	// update status parameter

    	return true;
    }
   
//-------------------------------------------------------------------------------------------------
/**
* apply a circular left shift (rotation) to the fid.  This is an alternative way of correcting
* filter group delay
* @param gp global parameter object
* @param p set parameters
* @param q status parameters
*/
   public static void circular_ls(global_par gp, data_par p, data_par q)
   {
       int i, j;
       double[] array = new double[q.td_eff];

       for (i = 0; i < q.td_eff; i++)
           array[i] = gp.proc_buffer[i];

       for (i = 0; i < q.td_eff - p.cls; i++)
           gp.proc_buffer[i] = array[i + p.cls];

       j = 0;
       for (i = q.td_eff - p.cls; i < q.td_eff; i++)
       {
           gp.proc_buffer[i] = array[j];
           j++;
       }

       q.cls = p.cls;  // update status parameter
   }
       
//-------------------------------------------------------------------------------------------------
/**
* Do a 1D Hilbert Transform on the data (re-generates imaginaries)
* @param gp the "global parameters" object
* @param p the "set" data parameters
* @param q the "status" data parameters.
*/
	public static void do_HT(global_par gp, data_par p, data_par q)
	{
		if (gp.data_array == null)
        {
            Util.show_and_log("Hilbert transform: no real data???");
            return;
        }

        int i;

        //  if there are no or inconsistent imaginary data (probably a JCAMP-DX file?) 
        //  allocate space for imag points
        
        if (gp.i_data_array == null || gp.i_data_array.length != gp.data_array.length)
        {
            gp.i_data_array = new double[p.SI];

            for (i = 0; i < q.SI; i++)	                // zero the arrays for good measure
                gp.i_data_array[i] = 0.0;               // create zeroed imaginary array
        }

        Util.alloc_proc_buffer(2 * q.SI, gp);           // allocate space

        for (i = q.SI / 2; i < q.SI; i++)               // load real part of data 
            gp.proc_buffer[2 * (i - q.SI / 2)] = gp.data_array[i];
        for (i = 0; i < q.SI / 2; i++)
            gp.proc_buffer[2 * (i + q.SI / 2)] = gp.data_array[i];


        for (i = q.SI / 2; i < q.SI; i++)                // zero imaginary part of data              
            gp.proc_buffer[2 * (i - q.SI / 2) + 1] = 0.0;
        for (i = 0; i < q.SI / 2; i++)
            gp.proc_buffer[2 * (i + q.SI / 2) + 1] = 0.0;

        Fft.fft(gp.proc_buffer, q.SI, -1);               // inverse FFT

        for(i = q.SI; i < 2 * q.SI; i++)
            gp.proc_buffer[i] = 0.0;

        gp.proc_buffer[0] = 0.5 * gp.proc_buffer[0];     // first point correction
        gp.proc_buffer[1] = 0.5 * gp.proc_buffer[1];

        Fft.fft(gp.proc_buffer, q.SI, 1);                 // forward FFT

        for (i = q.SI / 2; i < q.SI; i++)                 // extract real part of data 
            gp.data_array[i] = gp.proc_buffer[2 * (i - q.SI / 2)] / ((double)q.SI  /2.0);
        for (i = 0; i < q.SI / 2; i++)
            gp.data_array[i] = gp.proc_buffer[2 * (i + q.SI / 2)] / ((double)q.SI / 2.0); 

        for (i = q.SI / 2; i < q.SI; i++)                 // extract imaginary part of data 
            gp.i_data_array[i]
                = gp.proc_buffer[2 * (i - q.SI / 2) + 1] / ((double)q.SI / 2.0); 
        for (i = 0; i < q.SI / 2; i++)
            gp.i_data_array[i]
                = gp.proc_buffer[2 * (i + q.SI / 2) + 1] / ((double)q.SI / 2.0); 
	}
	
//-------------------------------------------------------------------------------------------------	
/**
 * Alternate fft routine using sine table.  May be faster for 2D/3D
 * Forward fft only!  Do not use for Hilbert transform, etc.
 * @param data  the data are alternating real/imaginary pairs
 * @param npairs the number of complex pairs
 */
	public static void fftSineTable(double[] data, int npairs)
    {
        int n, mmax, m, j, istep, i;
        double wtemp, wr, wpr, wpi, wi;
        double temp, tempr, tempi;

        int k = 0;
        // put data in bit-reversed order 
        n = 2 * npairs;
        j = 1;

        for (i = 1; i < n; i += 2)
        {
            if (j > i)
            {
                temp = data[j-1];         data[j-1] = data[i-1];          data[i-1] = temp;
                temp = data[j];           data[j] = data[i];              data[i] = temp;
            }
            m = n / 2;

            while (m >= 2 && j > m)
            {
                j = j - m;
                m = m / 2;
            }
            j = j + m;
        }
        mmax = 2;

        while (n > mmax)
        {
            istep = 2 * mmax;            
            wtemp = _wtemp[k];  		// from sine table:  Math.sin(0.5 * theta);        
            wpr = -2.0 * wtemp * wtemp;            
            wpi = _wpi[k];				// from sine table: Math.sin(theta);
            k++;            
            wr = 1.0;
            wi = 0.0;           
            for (m = 1; m < mmax; m += 2)
            {
                for (i = m; i <= n; i += istep)
                {
                    j = i + mmax;
                    tempr = wr * data[j-1] - wi * data[j];
                    tempi = wr * data[j] + wi * data[j-1];
                    data[j-1] = data[i-1] - tempr;
                    data[j] = data[i] - tempi;
                    data[i-1] += tempr;
                    data[i] += tempi;
                }
                wtemp = wr;
                wr = wtemp * wpr - wi * wpi + wr;
                wi = wi * wpr + wtemp * wpi + wi;
            }
            mmax = istep;
        }
    }
	
//-------------------------------------------------------------------------------------------------
/**
 * generate sine table for fft
 * must be called before first fft (i.e at program start-up)
 * @param Length length of sine table.  Max FT size 2**length
 */
	public static void doSineTable(int length)
	{
		double theta = Math.PI;
		
		 _wtemp = new double[length];
		 _wpi = new double[length];
		
		for(int i = 0; i < 20; i++)
		{
			_wtemp[i] = Math.sin(theta / 2.0);
			_wpi[i] = Math.sin(theta);
			theta = theta / 2.0;
		}
	}
	
//-------------------------------------------------------------------------------------------------	      
}// end class Fft
