
/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

package spinworks;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.MessageBox;

/**
 * Utilities for opening an NMR data set
 * @author Kirk Marat, 
 * University of Manitoba
 */
public class FileOpen extends SW_main
{
//-----------------------------------------------------------------------------------------------
/**
 * Auto detect data type
 * @param fileName  Full path to file
 */
	private static void getDataType(String fileName)
	{		
		// split name and get extension
		String[] name = fileName.split("[.]+");
		String extension = name[name.length - 1];	// gets last component separated by "."
		if(gp.log_on && name.length == 2)
			Util.log_error("getDataType(): name and extension = " + name[0] + "  " + name[1]);

		//  Check extensions for non-Bruker non-Varian files with extensions first

		//		Look for .dx and .jdx extensions (JCAMP DX)
		if(extension.equalsIgnoreCase("dx") || extension.equalsIgnoreCase("jdx"))
		{
			q.data_type = data_types.JCAMP_DX;			// set type to JCAMP-DX
			q.data_mode = modes.MODE_1D;				// 1D only for JDX
			if (gp.log_on) Util.log_error("data type from extension: JCAMP-DX");
			return;										// no need to go any further
		}

		//  Look for .ft or .ft2  .ft3 extensions (NMRPIPE)
		if (extension.equalsIgnoreCase("ft"))
		{
			q.data_type = data_types.NMRPIPE;			// set type to NMRPIPE
			q.data_mode = modes.MODE_1D;				// 1D
			if (gp.log_on) Util.log_error("data type from extension: NMRPIPE 1D");
			return;										// no need to go any further
		}

		if (extension.equalsIgnoreCase("ft2"))
		{
			q.data_type = data_types.NMRPIPE;			// set type to NMPPIPE 2D
			q.data_mode = modes.MODE_2D;				// 2D
			if (gp.log_on) Util.log_error("data type from extension: NMRPIPE 2D");
			return;										// no need to go any further
		}

		if (extension.equalsIgnoreCase("ft3"))
		{
			q.data_type = data_types.NMRPIPE;			// set type to NMRPIPE 3D
			q.data_mode = modes.MODE_3D;				// 3D!  
			if (gp.log_on) Util.log_error("data type from extension: NMRPIPE 3D");
			return;										// no need to go any further
		}           

		//  Look for jdf extension (JEOL) and check 1D vs 2D
		if (extension.equalsIgnoreCase("jdf"))
		{
			q.data_type = data_types.JEOL;			     // set type to JEOL
			JDF_Util.get_jdf_pars(fileName, gp, p, p_f1, q, q_f1); // check for 2D
			if (gp.log_on) Util.log_error("data type from extension: jdf");
			return;										 // no need to go any further
		}
		//  Look for .1d or .2d extension (Magritek)
		if (extension.equalsIgnoreCase("1d") || extension.equals("2d"))
		{
			q.data_type = data_types.MAGRITEK;			       // set type to Magritek              
			if (gp.log_on) Util.log_error("data type from extension: Magritek");
			return;											   // no need to go any further
		}

		// Test read to see if it a CDFF (Anasazi) file
		/*if (file_util.get_cdff_type_2_pars(fileName, gp, p, q, p_f1, q_f1) > 0)
		{
			q.data_type = data_types.CDFF;             // set type
			if (gp.log_on) util.log_error("data type: CDF");
			return;                                    // no need to go any further
		}*/

		//  Must be Bruker or Varian
		//		create some needed paths and strings
						
		int j;
		for (j = fileName.length() - 1; j > 0; j--)
			if (fileName.charAt(j) == '/' || fileName.charAt(j) == '\\') break;
		
		String basename = fileName.substring(0, j + 1);		
		
		String bruker_par_path = basename + "acqus";
		String bruker_par_path_3D = basename + "acqu3s";
		String varian_par_path = basename + "procpar";
		
		File fi = null, bfi = null, bfi_3D = null, vfi = null;

		try
		{
			fi = new File(fileName);
			bfi = new File(bruker_par_path);
			bfi_3D = new File(bruker_par_path_3D);
			vfi = new File(varian_par_path);
		}
		catch (Exception ex)
		{
			Util.show_and_log("get_data_type(): Error getting FileInfo " + fileName,
					ex.getMessage());
		}		

		if(bfi.exists())					// an acqus file exists, must be Bruker
			q.data_type = data_types.BRUKER;
		else if(vfi.exists())			    // a procpar file exists, must be Varian
			q.data_type = data_types.VNMR;

		//		auto detect 1D vs 2D: Bruker is handled by determining whether the file name
		//		is "fid" (1D) or "ser" (2D/3D)				 		
		if(q.data_type == data_types.BRUKER)
		{			
			if (fi.getName().equals("fid") || fi.getName().equals("fid1")
					|| fi.getName().equals("fid2") || fi.getName().equals("fid3")
					|| fi.getName().equals("1D_real" ) || fi.getName().equals("1D_imag" ))  // 1D
				InterfaceUtil.set1D();

			else if (fi.getName().equals("ser") || fi.getName().equals("ser1")
					|| fi.getName().equals("ser2") || fi.getName().equals("ser3")
					|| fi.getName().equals("2D_rr"))  // 2D or 3D
			{
				if (fi.getName().equals("ser") && bfi_3D.exists())  // 3D
					InterfaceUtil.set3D();
				else
					InterfaceUtil.set2D();                          // 2D				
			}

			else
			{
				Util.show_and_log("Selected name doesn't appear to correspond to a known data type");
				// TODO add data type selection dialog here??
			}
		}

//		Varian 1D/2D/3D selection is handled by a trial read of the "procpar" file, looking for
//		a non-zero value for "SW1 and or SW2".  Always confirm a change in mode  This should be 
//      improved
		else if(q.data_type == data_types.VNMR)
		{			
			if(FileUtilVarian.checkModeVarian(varian_par_path) == 1
					&& (q.data_mode == modes.MODE_2D || q.data_mode == modes.MODE_3D))
			{			
				 MessageBox messageBox = new MessageBox(shell, SWT.ICON_QUESTION
            			 | SWT.YES | SWT.NO);
            	 messageBox.setMessage("This appears to 1D data.  Switch to 1D mode?");
            	 messageBox.setText("SpinWorks Java: Confirm Data Format");
            	 int response = messageBox.open();            	 
            	 if (response == SWT.YES)
				{
					InterfaceUtil.set1D();
				}

			}

			else if(FileUtilVarian.checkModeVarian(varian_par_path) == 2
					&& (q.data_mode == modes.MODE_1D || q.data_mode == modes.MODE_3D))
			{
				MessageBox messageBox = new MessageBox(shell, SWT.ICON_QUESTION
           			 | SWT.YES | SWT.NO);
           	    messageBox.setMessage("This appears to 2D data.  Switch to 2D mode?");
           	    messageBox.setText("SpinWorks Java: Confirm Data Format");
           	    int response = messageBox.open();	
           	    if(response == SWT.YES)
				{
					InterfaceUtil.set2D();
				}
			}

			else if (FileUtilVarian.checkModeVarian(varian_par_path) == 3
					&& (q.data_mode == modes.MODE_1D || q.data_mode == modes.MODE_2D))
			{
				MessageBox messageBox = new MessageBox(shell, SWT.ICON_QUESTION
           			 | SWT.YES | SWT.NO);
           	    messageBox.setMessage("This appears to 3D data.  Switch to 3D mode?");
           	    messageBox.setText("SpinWorks Java: Confirm Data Format");
           	    int response = messageBox.open();	
           	    if(response == SWT.YES)
				{
					InterfaceUtil.set3D();
				}
			}
		}
	}
//-------------------------------------------------------------------------------------------------
/**
 * Open the NMR data set selectd by either the File Open dialog, the Recent Files Dialog,
 * or as a command line argument
 * @param fileName  Full path to file
 */
    public static void openSelectedFile(String fileName)
    {
        //	this may take a while, so...
        shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_WAIT));
    	
        statusStripLabel1.setText("Opening Data Set");

        gp.new_data();                               // reset global parameter block
        if(!gp.retain_integral_regions)     		 // clear integral regions if desired
            gp.n_integrals = 0;

        data_types current_data_type = q.data_type;	 // save current data type  TODO  find cleaner way to do this!!!
        modes current_data_mode = q.data_mode;		 // save current data mode
        p = new data_par(); q = new data_par();		 // new data set, so...
        p_f1 = new data_par(); q_f1 = new data_par();// fresh parameters  "Z"
        p_f2 = new data_par(); q_f2 = new data_par();// fresh parameters  "Y"
        q.data_type = current_data_type;			 // restore data type 
        q.data_mode = current_data_mode;			 // restore data type 
                                                     //   (new data par re-sets it)

        p.lp = LP_modes.LP_OFF;		// turn off lp as default
        p_f1.lp = LP_modes.LP_OFF;	// both dimensions
        p.fcor = 0.5;               // default value for first point multiplier in F2

        // determine the data type, but only if user hasn't selected 
        if(auto_data_format);
		     getDataType(fileName);	// determine type of data

				
        if (q.data_type == data_types.BRUKER && q.data_mode == modes.MODE_1D)
            fileOpenBruker1D(fileName);

        else if (q.data_type == data_types.BRUKER && q.data_mode == modes.MODE_2D)        
            fileOpenBruker2D(fileName);

        else if (q.data_type == data_types.BRUKER && q.data_mode == modes.MODE_3D)
            fileOpenBruker3D(fileName);
        
        else if (q.data_type == data_types.VNMR && q.data_mode == modes.MODE_1D)
		fileOpenVarian1D(fileName);
        
		else if (q.data_type == data_types.VNMR && q.data_mode == modes.MODE_2D)
			fileOpenVarian2D(fileName);

        else if (q.data_type == data_types.VNMR && q.data_mode == modes.MODE_3D)
            fileOpenVarian3D(fileName);
       
        else if (q.data_type == data_types.JEOL && q.data_mode == modes.MODE_1D)
            fileOpenJEOL_1D(fileName);

        else if (q.data_type == data_types.JEOL && q.data_mode == modes.MODE_2D)
            fileOpenJEOL_2D(fileName);

        /*else if (q.data_type == data_types.TECMAG && q.data_mode == modes.MODE_1D)
            FileOpenTECMAG_1D(fileName);*/

        else if (q.data_type == data_types.JCAMP_DX)
            fileOpenJDX(fileName);

        /*else if (q.data_type == data_types.CDFF)
            FileOpenCDFF(fileName);*/

        else if (q.data_type == data_types.NMRPIPE && (q.data_mode == modes.MODE_2D
            ||  q.data_mode == modes.MODE_3D))
            fileOpenNMRpipe(fileName);

        /*else if (q.data_type == data_types.NMRPIPE && q.data_mode == modes.MODE_1D)
            FileOpenNMRPipe_1D(fileName);*/

        else if (q.data_type == data_types.MAGRITEK &&
            (q.data_mode == modes.MODE_1D || q.data_mode == modes.MODE_2D))
            FileOpenMagritek(fileName); // common file open for 1D and 2D

        else
        {
            Util.show_and_log("Unknown File Format", fileName);
            shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));
            return;
        }

        if(q.data_mode == modes.MODE_2D || q.data_mode == modes.MODE_3D)
            gp.raw_max = Util_2D.max_ser(gp, q, q_f1);				// calculate scaling factor
        else
            gp.raw_max = Util.dscan(gp.fid, q.TD, 4, q.TD - 1);
       
		gp.v_rescale = true;				// new data set, so rescale on next transform

        if (q.data_type == data_types.BRUKER || q.data_type == data_types.VNMR)
		    shell.setText(version + q.data_folder);
        else
        	shell.setText(version + q.data_path);

//set initial scaling based on display mode
        if (gp.display_mode == modes.MODE_FID)
        {
            gp.fid_scale = (double)gp.dysize * 0.80 / gp.raw_max;
            gp.t1 = 0.0;
            gp.t2 = (double)(q.TD /2 ) * q.dwell;
        }
        else
            gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;

        q.SW_Hz = q.SW * q.SFO1;    // used ???
         
		InterfaceUtil.updateToolbar();	// update menu and tool bar     
		gp.expt_read = true;
		SW_main.file_obj.add_file(fileName, q.pulse_sequence, gp.plot_title);	// add this file to recent files list
        statusStripLabel1.setText("Ready");
		statusStripLabel2.setText(String.format("%s  %d scans", q.pulse_sequence, q.NS));
		shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));	// restore cursor
    }
    
//-------------------------------------------------------------------------------------------------   
/**
 * Open a Bruker (UXNMR/XwinNMR/TopSPin) data set 1D version
 * @param fileName  Full path to file
 */
    static void fileOpenBruker1D(String fileName)
	{
		q.data_path = fileName;
        boolean local_autoload = false;

		File fi = null;
		try
		{
			fi = new File(fileName);
			q.data_folder = fi.getParent() + '/';
		}
		catch (Exception ex)
		{
            Util.show_and_log("FileOpenBruker(): " + fileName, ex.getMessage());
            ex.printStackTrace();
			return;
		}
		
        gp.display_mode = modes.MODE_FID;

//user has selected processed data, so override autoload setting if necessary
        if (fi.getName().equals("1D_real")
            || fi.getName().equals("1D_imag"))
        {
            local_autoload = true;
            q.data_path = q.data_folder + "fid";
        }        

//status acquisition parameters
		q.par_path = q.data_folder + "acqus";		// data_folder had trailing slash

		if(!FileUtilBruker.get_par(q.par_path, p, q, gp))
		{
            Util.show_and_log("FileOpenBruker(): cannot open Bruker acqus file: ",
                q.par_path + "\nPossible wrong data format selected?");
            gp.display_mode = modes.MODE_1D;
			return;
		}

//status processing parameters, TO DO try other proc numbers!
		q.par_path = q.data_folder + "pdata/1/procs";

		if(!FileUtilBruker.get_par(q.par_path, p, q, gp))
		{
            Util.show_and_log("FileOpenBruker(): cannot open Bruker procs file: ",
                q.par_path);
            gp.display_mode = modes.MODE_1D;
			return;
		}

		gp.fid = null;					// free space if fid already exists 
		
        if (q.DTYPA == 2)				// data is 64 bit floating point
        {      	
            if (FileUtilBruker.get_double_fid(q.data_path, q, gp) <= 0)
            {
                Util.show_and_log("FileOpenBruker(): cannot open Bruker DP fid file: ",
                    q.data_path);
                gp.display_mode = modes.MODE_1D;
                return;
            }
        }
        else if (q.DTYPA == 1)			// data is 32 bit floating point ???
        {      							// UNTESTED!  Actually used??
            if (FileUtilBruker.get_float_fid(q.data_path, q, gp) <= 0)
            {
                Util.show_and_log("FileOpenBruker(): cannot open Bruker 32 bit float fid file: ",
                    q.data_path);
                gp.display_mode = modes.MODE_1D;
                return;
            }
        }
        else							// data must be standard 32 bit signed integer
        {
            if (FileUtilBruker.get_fid(q.data_path, q, gp) <= 0)
            {
                Util.show_and_log("FileOpenBruker(): cannot open Bruker integer fid file: ",
                    q.data_path);
                gp.display_mode = modes.MODE_1D;
                return;
            }
        }

//now get title	
        q.titlepath = q.data_folder + "pdata/1/title";                   
        try 
        {       	
            String line;
            BufferedReader br = new BufferedReader(new FileReader(q.titlepath)); 
        	
            gp.plot_title = "";
        	while ((line=br.readLine())!=null)
        	{
                gp.plot_title += line + "\n";
            }
            br.close(); 
        }
        catch (Exception ex) //no title, provide really annoying message
        {               
            gp.plot_title = "no title";		// well, not so annoying now
            Util.log_error("FileOpenBruker()", ex.getMessage());	// do we need this??
        }
        
		Util.alloc_proc(gp, p, q);

        if (gp.auto_load || local_autoload)	          // if requested,try to read
            if (FileUtil.readProc1D(gp, p, q) > 0)    // any previously stored data  
            {
                gp.display_mode = modes.MODE_1D;
                gp.cursor_flag = 0;	                  // flag no cursors set                               					
                FileUtil.readPeaks1D(gp, q);      	  // read 1D peaks if available
                if(!gp.retain_integral_regions)
                	FileUtil.readInteg1D();
            }

//calculate internal offset in Hz., and re-set display limits			
		q.hzperpt = q.SW * q.SFO1 / (double)(q.SI);
        q.offset = (q.SFO1 - p.SF) * 1.0e06 + (q.SW * q.SFO1) / 2.0;
		gp.freq1 = q.offset; 
	    gp.freq2 = q.offset - (q.SFO1 * q.SW);

//set reasonable size if too small, can override with parameter editor
		if (p.SI < q.TD / 2) p.SI = q.SI / 2;
	
		Util.hscale(gp);
		Util.rescale(gp, q);

		gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;       
        p.lp = LP_modes.LP_OFF;
		InterfaceUtil.set1D();				// update interface
		ws[aws].has_data = true;
		SW_main.statusStripLabel2.setText("File Opened");
        if (gp.log_on) Util.log_error(fileName + " Opened");
	}
    
//-------------------------------------------------------------------------------------------------
/**
 * Open a Bruker (XwinNMR/UXNMR/TopSpin) fid file, 2D version
 * @param fileName Full path to the data file
 * @return True if OK, false on error
 */
    static boolean fileOpenBruker2D(String fileName)	// TODO  Clean Up !!!
    {
    	boolean par_ok = false, data_ok = false, proc_par_ok = false;
    	boolean proc_par_ok_f1 = false;
    	boolean local_autoload = false;
    	int i;

    	Util_2D.free_ser(gp);
    	Util_2D.free_rows(true, true, true, true, gp);

    	q.data_path = fileName;

		File fi = null;
		try
		{
			fi = new File(fileName);
			q.data_folder = fi.getParent() + '/';
		}
		catch (Exception ex)
		{
            Util.show_and_log("FileOpenBruker_2D(): " + fileName, ex.getMessage());
            ex.printStackTrace();
			return false;
		}   	

    	if (fi.getName().equals("2D_rr"))
    	{
    		local_autoload = true;
    		q.data_path = q.data_folder + "ser";
    	}

    	gp.display_mode = modes.MODE_FID;			// display the FID at file open

    	// get the acquisition parameters
    	q.par_path = q.data_folder + "acqus";
    	p_f1.REVERSE = false;						// switch to default;
    	p.REVERSE = false;

    	par_ok = FileUtilBruker.get_par(q.par_path, p, q, gp);	// TO DO
    	if (par_ok == false)
    	{
    		Util.show_and_log("FileOpenBruker_2D(): cannot read 'acqus': ", q.par_path
    				+ "\nPossible wrong data format selected?");
    	}

    	// now the F1 acquisition parameters
    	q.par_path = q.data_folder + "acqu2s";

    	par_ok = FileUtilBruker.get_par(q.par_path, p_f1, q_f1, gp);	// TO DO
    	if (par_ok == false)
    	{
    		Util.show_and_log("FileOpenBruker_2D(): cannot read 'acqu2s': ", q.par_path);
    	}

    	// try for the processing parameters and title, look at the first three proc numbers
    	i=1;
    	while(proc_par_ok == false && i <= 3)	
    	{
    		q.par_path = String.format("%spdata/%d/proc",  q.data_folder, i);
    		q.titlepath = String.format("%spdata/%d/title", q.data_folder, i);
    		String procs_path = String.format("%spdata/%d/procs", q.data_folder, i);

    		proc_par_ok = FileUtilBruker.get_par(q.par_path, p, q, gp);	// TO DO
    		FileUtilBruker.get_SF(procs_path, p, q, gp);

    		if(proc_par_ok == true)			// found at least a proc file
    		{
    			q.par_path = q.data_folder + "pdata/1/proc2";
    			String proc2s_path = String.format("%spdata/%d/proc2s", 
    					q.data_folder, i);

    			proc_par_ok_f1 = FileUtilBruker.get_par(q.par_path, p_f1, q_f1, gp);
    			FileUtilBruker.get_SF(proc2s_path, p_f1, q_f1, gp);

    			if (proc_par_ok_f1 == false)
    			{
    				Util.show_and_log("FileOpenBruker_2D(): cannot read 'acqus2': ", q.par_path);
    			}

    			// now get plot title
    			 q.titlepath = q.data_folder + String.format("pdata/%d/title", i);   		                     
    			 try 
    			 {       	
    				 String line;
    				 BufferedReader br = new BufferedReader(new FileReader(q.titlepath)); 

    				 gp.plot_title = "";
    				 while ((line=br.readLine())!=null)
    				 {
    					 gp.plot_title += line + "\n";
    				 }
    				 br.close(); 
    			 }
    			 catch (Exception ex) //no title, provide really annoying message
    			 {               
    				 gp.plot_title = "no title";		// well, not so annoying now
    				 Util.log_error("FileOpenBruker_2D()", ex.getMessage());	// do we need this??
    			 }
    		}
    		i++;
    	}	// end while

    	// no processing parameters, make some up.
    	if(proc_par_ok == false)		// F2
    	{
    		Util.show_and_log("FileOpenBruker_2D(): creating default F2 processing parameters");
    		q.SI = q.TD;				p.SI = q.SI;
    		q.SF = q.SFO1;			p.SF = q.SF;
    		q.hzperpt = q.SW * p.SF / (double)(p.SI);
    		q.offset = (q.SW * q.SF)/2.0;
    		q.OFFSET = (float)(q.SW/2.0);		p.OFFSET = q.OFFSET;			
    	}

    	if(proc_par_ok_f1 == false)		// F1
    	{
    		Util.show_and_log("FileOpenBruker_2D(): creating default F1 processing parameters");
    		q.SI = q.TD;						p.SI = q.SI;
    		q_f1.SI = q_f1.TD;					p_f1.SI = q_f1.SI;
    		q_f1.SF = q_f1.SFO1;				p_f1.SF = q_f1.SF;
    		q_f1.offset = (q_f1.SW * q_f1.SF)/2.0;
    		q_f1.OFFSET = (float)(q_f1.SW/2.0);
    		p_f1.OFFSET = q_f1.OFFSET;
    		q_f1.hzperpt = q_f1.SW*p_f1.SF/(double)(p_f1.SI);
    	}

    	if(q.SI != 0)
    		q.hzperpt = q.SW*q.SFO1/(double)(p.SI);	// digital resolution 
    	if(q_f1.SI != 0)
    		q_f1.hzperpt = q_f1.SW * q_f1.SFO1 / (double)(p_f1.SI);		

    	// freq of first data point in Hz 
    	q.offset = (q.SFO1 - p.SF) * 1.0e06 + (q.SW*q.SFO1)/2.0;
    	q_f1.offset = (q_f1.SFO1 - p_f1.SF) * 1.0e06 + (q_f1.SW*q_f1.SFO1)/2.0;

    	if(Math.abs(q.SF - q.SFO1) / q.SFO1 > 0.1)			// F2
    	{
    		Util.show_and_log("FileOpenBruker_2D(): reference frequency out of range (F2)!\nmanual referencing required");
    		q.SI = q.TD;			p.SI = q.SI;
    		q.SF = q.SFO1;			p.SF = p.SFO1;
    	}

    	if(Math.abs(q_f1.SF - q_f1.SFO1) / q_f1.SFO1 > 0.1)	// F1
    	{
    		Util.show_and_log("FileOpenBruker_2D(): reference frequency out of range (F1)!\nmanual referencing required");
    		q_f1.SF = q_f1.SFO1;	p_f1.SF = p_f1.SFO1;
    	}


    	// now read ser file, force Intel byte order on SpinWorks 4 FID Math generated FIDs  TODO
    	/*if (!Equals(fi.Name, "ser"))
    		q.BYTORDA = SW_main.byte_orders.LITTLE;*/

    	if (FileUtilBruker.get_ser(q.data_path, gp, q, q_f1)> 0)
    		data_ok = true;
    	if( data_ok == false)
    	{
    		Util.show_and_log("FileOpenBruker_2D(): cannot read 'ser': ", q.data_path);
    		gp.display_mode = modes.MODE_FID;
    		return false;
    	}

    	//  TopSpin uses a new parameter called FnMODE for F1 detection.  If this is defined, use it, 
    	//  otherwise use MC2 parameter    	
    	if(q_f1.FnMODE != FnMODEs.UNDEFINED)
    		p_f1.MC2 = SW_main.det_modes_IND.values()[p_f1.FnMODE.ordinal() - 1]; // slight difference in Fn MODE and MC2 enumeration
    	
    	//	We have valid data and parameters
    	if(data_ok && par_ok)
    	{
    		Util.free_proc(gp);									// free and alloc allocate some 
    		Util.alloc_proc(gp, p, q);							// space for the data
    		//Util.free_calc(gp);

    		Util_2D.load_display_2D_ser(0, gp, q);

    		gp.cursor_flag = 0;									// flag new freq1 set 
    		gp.freq1 = q.offset; 
    		gp.freq2 = q.offset - (q.SW * q.SFO1);
    		gp.freq1_indirect = q_f1.offset;
    		gp.freq2_indirect = q_f1.offset - (q_f1.SW * q_f1.SFO1);
    		Util_2D.calc_scale_factors_2D(gp, q, q_f1);
    	}

    	p_f1.PHC0 = -p_f1.PHC0;									// Spinworks and XwinNMR have
    	p_f1.PHC1 = -p_f1.PHC1;									// different sign convention 
    	// F1 phase
    	q.PHC0 = 0.0;
    	q.PHC1 = 0.0;

    	p_f1.lp_pred = q_f1.TD;									// default LP parameters
    	p_f1.lp_input = q_f1.TD/2;
    	p.lp = LP_modes.LP_OFF;
    	p_f1.lp = LP_modes.LP_OFF;
    	if (q_f1.TD < 128)
    		p_f1.lp_coef = 8;

    	if (p.SI < q.TD/2) p.SI = q.TD/2;                       // set default minumum sizes
    	if (p_f1.SI < q_f1.TD/2) p_f1.SI = q_f1.TD/2;

    	if (gp.auto_load || local_autoload)						// TODO
    	{
    		statusStripLabel2.setText("Loading Processed Data");
    		if (FileUtil.readProc2D(gp, p, p_f1, q, q_f1) > 0)	// if successful, calculate 
    		{														// scale factors, etc.
    			gp.cursor_flag = 0;									// flag no cursors set 
    			gp.data_max_2D = Util_2D.max_2D(gp.rows_real, q.SI, q_f1.SI);		// find global maximum
    			gp.data_noise = Util_2D.noise_2D(gp, q, q_f1);		// noise level
    			Util_2D.calc_scale_factors_2D(gp, q, q_f1);			// scaling factors
    			Util_2D.calc_default_levels_2D(gp);					// default contour levels		
    			gp.display_mode = modes.MODE_2D;				    // switch to 2D display
    			gp.new_2D_view = true;								// re-contour
    		}
    	}

    	if (gp.rows_real == null || !gp.auto_load)	// in no processed data, scale to fid  
    	{											// intensity for display
    		Util.hscale(gp);
    		Util.rescale(gp, q);
    		gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;
    	}

    	if (gp.a4_paper)
    	{
    		if (gp.plot_xsize > 260.0) gp.plot_xsize = 260.0;
    	}
    	else
    	{
    		if (gp.plot_xsize > 240.0) gp.plot_xsize = 240.0;
    	}

    	statusStripLabel2.setText("File Opened");
    	if (gp.log_on) Util.log_error(fileName + " Opened");

    	return (par_ok && data_ok);	
    }

//-------------------------------------------------------------------------------------------------
/**
 * Open a Bruker 3D data set
 * @param FileName  Full path to file
 */
    static void fileOpenBruker3D(String fileName)
    {
        boolean par_ok = false, data_ok = false, proc_par_ok = false;
        boolean proc_par_ok_f1 = false;
        boolean proc_par_ok_f2 = false;
        int i;

        Util_2D.free_ser(gp);
        Util_2D.free_rows(true, true, true, true, gp);

        q.data_path = fileName;

        File fi = null;
		try
		{
			fi = new File(fileName);
			q.data_folder = fi.getParent() + '/';
		}
		catch (Exception ex)
		{
            Util.show_and_log("FileOpenBruker_2D(): " + fileName, ex.getMessage());
            ex.printStackTrace();
			return;
		}   	
        gp.display_mode = modes.MODE_FID;					// display the FID at file open
             
        //switch to defaults
        p.REVERSE = false;      // X
        p_f2.REVERSE = false;   // Y						
        p_f1.REVERSE = false;   // Z						

        // get acquisition parameters

        // get the acquisition parameters in the "X" dimension
        q.par_path = q.data_folder + "acqus";
        par_ok = FileUtilBruker.get_par(q.par_path, p, q, gp);
        if (par_ok == false)
        {
            Util.show_and_log("FileOpenBruker_3D(): cannot read 'acqus': " + q.par_path);
        }
        if(gp.log_on && par_ok)Util.log_error( q.par_path + " has been read");


        // now the F2 or "Y" dimension acquisition parameters
        q.par_path = q.data_folder + "acqu2s";
        par_ok = FileUtilBruker.get_par(q.par_path, p_f2, q_f2, gp);	// TO DO
        if (par_ok == false)
        {
            Util.show_and_log("FileOpenBruker_3D(): cannot read 'acqu2s': " + q.par_path);
        }
        if (gp.log_on && par_ok) Util.log_error(q.par_path + " has been read");


        // now the F1 or "Z" dimension acquisition parameters
        q.par_path = q.data_folder + "acqu3s";
        par_ok = FileUtilBruker.get_par(q.par_path, p_f1, q_f1, gp);
        if (par_ok == false)
        {
            Util.show_and_log("FileOpenBruker_3D(): cannot read 'acqu3s': " + q.par_path);
        }
        if (gp.log_on && par_ok) Util.log_error(q.par_path + " has been read");

//try for the processing parameters and title, look at the first three proc numbers
        i = 1;
        while (proc_par_ok == false && i <= 3)
        {
            q.par_path = String.format("%spdata/%d/proc", q.data_folder, i);
            q.titlepath = String.format("%spdata/%d/title", q.data_folder, i);
            String procs_path = String.format("%spdata/%d/procs", q.data_folder, i);

            proc_par_ok = FileUtilBruker.get_par(q.par_path, p, q, gp);  // processing parameters inthe "X" dimension
            if (gp.log_on && proc_par_ok) Util.log_error(q.par_path + " has been read");

            FileUtilBruker.get_SF(procs_path, p, q, gp); // get SF from "status" parameters
                                                    	 // to reflect any referencing on spectrometer

            if (proc_par_ok == true)			   		 // found at least a proc file
            {

                //   "Y" dimension
                //q.par_path = String.Concat(q.data_folder, String.format("pdata{0}1{1}proc2", sep, sep));
                //q.par_path = q.data_folder + "pdata/" + "1" + "/proc2";  
            	
            	q.par_path = String.format("%spdata/%d/proc2", q.data_folder, i);
                String proc2s_path = String.format("%spdata/%d/proc2s",
                    q.data_folder, i);                  
                		
                proc_par_ok_f2 = FileUtilBruker.get_par(q.par_path, p_f2, q_f2, gp);
                if (gp.log_on && proc_par_ok) Util.log_error(q.par_path + " has been read");

                FileUtilBruker.get_SF(proc2s_path, p_f2, q_f2, gp);  // get SF from "status" parameters
                                                                // to reflect any referencing on spectrometer  
                if (proc_par_ok_f2 == false)
                {
                    Util.show_and_log("FileOpenBruker_3D(): cannot read 'proc2': " + proc2s_path);
                }

                //   "Z" dimension
                //q.par_path = q.data_folder + "pdata/" + "1" + "/proc3";   
                q.par_path = String.format("%spdata/%d/proc3", q.data_folder, i); 
                String proc3s_path = String.format("%spdata/%d/proc3s",
                    q.data_folder, i);    

                proc_par_ok_f1 = FileUtilBruker.get_par(q.par_path, p_f1, q_f1, gp);
                if (gp.log_on && proc_par_ok) Util.log_error(q.par_path + " has been read");

                FileUtilBruker.get_SF(proc3s_path, p_f1, q_f1, gp);  // get SF from "status" parameters
                                                                // to reflect any referencing on spectrometer
                if (proc_par_ok_f1 == false)
                {
                    Util.log_error("FileOpenBruker_3D(): cannot read 'proc3s': " + proc3s_path);
                }

             // now get plot title
                q.titlepath = q.data_folder + String.format("pdata/%d/title", i);   		                     
                try 
                {       	
                	String line;
                	BufferedReader br = new BufferedReader(new FileReader(q.titlepath)); 

                	gp.plot_title = "";
                	while ((line=br.readLine())!=null)
                	{
                		gp.plot_title += line + "\n";
                	}
                	br.close(); 
                }
                catch (Exception ex) //no title, provide really annoying message
                {               
                	gp.plot_title = "no title";		// well, not so annoying now
                	Util.log_error("FileOpenBruker_2D()", ex.getMessage());	// do we need this??
                }
            }
            i++;
        }	// end while

        // no processing parameters, make some up.
        // if you are using this code, your data is probably toast, anyway
        if (proc_par_ok == false)		// F3  or "X" dimension
        {
            Util.show_and_log("FileOpenBruker_3D(): creating default X (F3) processing parameters");
            q.SI = q.TD; p.SI = q.SI;
            q.SF = q.SFO1; p.SF = q.SF;
            q.hzperpt = q.SW * p.SF / (double)(p.SI);
            q.offset = (q.SW * q.SF) / 2.0;
            q.OFFSET = (float)(q.SW / 2.0); p.OFFSET = q.OFFSET;
        }
        
        if (proc_par_ok_f2 == false)		// F2  or "Y" dimension
        {
            Util.show_and_log("FileOpenBruker_3D(): creating default Z (F2) processing parameters");
            //q.SI = q.TD; p.SI = q.SI;   // check
            q_f2.SI = q_f2.TD; p_f2.SI = q_f2.SI;
            q_f2.SF = q_f2.SFO1; p_f2.SF = q_f1.SF;
            q_f2.offset = (q_f2.SW * q_f2.SF) / 2.0;
            q_f2.OFFSET = (float)(q_f2.SW / 2.0);
            p_f2.OFFSET = q_f2.OFFSET;
            q_f2.hzperpt = q_f2.SW * p_f2.SF / (double)(p_f2.SI);
        }

        if (proc_par_ok_f1 == false)		// F1  or "Z" dimension
        {
            Util.show_and_log("FileOpenBruker_3D(): creating default Y (F1) processing parameters");
            q_f1.SI = q_f1.TD; p_f1.SI = q_f1.SI;
            q_f1.SF = q_f1.SFO1; p_f1.SF = q_f1.SF;
            q_f1.offset = (q_f1.SW * q_f1.SF) / 2.0;
            q_f1.OFFSET = (float)(q_f1.SW / 2.0);
            p_f1.OFFSET = q_f1.OFFSET;
            q_f1.hzperpt = q_f1.SW * p_f1.SF / (double)(p_f1.SI);
        }

        if (q.SI != 0)
            q.hzperpt = q.SW * q.SFO1 / (double)(p.SI);	              // digital resolution "X"

        if (q_f2.SI != 0)
            q_f2.hzperpt = q_f2.SW * q_f2.SFO1 / (double)(p_f2.SI);   // digital resolution "y"

        if (q_f1.SI != 0)
            q_f1.hzperpt = q_f1.SW * q_f1.SFO1 / (double)(p_f1.SI);   // digital resolution "Z"

        // freq of first data point in Hz 
        q.offset = (q.SFO1 - p.SF) * 1.0e06 + (q.SW * q.SFO1) / 2.0;                 // "X"
        q_f2.offset = (q_f2.SFO1 - p_f2.SF) * 1.0e06 + (q_f2.SW * q_f2.SFO1) / 2.0;  // "Y"
        q_f1.offset = (q_f1.SFO1 - p_f1.SF) * 1.0e06 + (q_f1.SW * q_f1.SFO1) / 2.0;  // "Z"

        if (Math.abs(q.SF - q.SFO1) / q.SFO1 > 0.1)			    // F3  or "X"
        {
            Util.show_and_log("FileOpenBruker_3D(): reference frequency out of range (X dimension)!\nmanual referencing required");
            q.SI = q.TD; p.SI = q.SI;
            q.SF = q.SFO1; p.SF = p.SFO1;
        }

        if (Math.abs(q_f2.SF - q_f2.SFO1) / q_f2.SFO1 > 0.1)	// F1 or "Y"
        {
            Util.show_and_log("FileOpenBruker_3D(): reference frequency out of range (Y dimension)!\nmanual referencing required");
            q_f2.SF = q_f2.SFO1; p_f2.SF = p_f2.SFO1;
        }

        if (Math.abs(q_f1.SF - q_f1.SFO1) / q_f1.SFO1 > 0.1)	// F1 or "Z"
        {
            Util.show_and_log("FileOpenBruker_3D(): reference frequency out of range (Z dimension)!\nmanual referencing required");
            q_f1.SF = q_f1.SFO1; p_f1.SF = p_f1.SFO1;
        }
        
        // read the entire ser file into a buffer
        if(FileUtilBruker3D.getSerBuffer(q.data_path, gp) > 0)
        {
            data_ok = true;
        }           

        if (data_ok == false)
        {
            Util.show_and_log("FileOpenBruker_3D(): cannot read 'ser': ", q.data_path);
            gp.display_mode = modes.MODE_FID;
            return;
        }

        try { gp.ser = new double[q_f1.TD][q.TD]; }
        catch (Exception ex)
        {
            Util.show_and_log("FileOpenBruker_3D(): allocation error " + ex.getMessage());
            return;
        }

        // get one plane, mostly for display purposes (very wasteful)
        FileUtilBruker3D.getSerXZ(0, gp, q, q_f1, q_f2);
        //                          X    Z     Y
        
        //  TopSpin uses a new parameter called FnMODE for F1 detection.  If this is defined, use it, 
        //  otherwise use MC2 parameter

        // slight difference in Fn MODE and MC2 enumeration
        if (q_f2.FnMODE != FnMODEs.UNDEFINED)  // "Y"
        	p_f2.MC2 = SW_main.det_modes_IND.values()[p_f2.FnMODE.ordinal() - 1]; 

        if (q_f1.FnMODE != FnMODEs.UNDEFINED)  // "Z"
            p_f1.MC2 = SW_main.det_modes_IND.values()[p_f1.FnMODE.ordinal() - 1]; 
        	
        //	We have valid data and parameters
        if (data_ok && par_ok)
        {
            Util.free_proc(gp);									// free and alloc allocate some 
            Util.alloc_proc(gp, p, q);							// space for the data

            Util_2D.load_display_2D_ser(0, gp, q);

            gp.cursor_flag = 0;									// flag new freq1 set 
            gp.freq1 = q.offset;
            gp.freq2 = q.offset - (q.SW * q.SFO1);              // use "XZ" planes as display mode
            gp.freq1_indirect = q_f1.offset;
            gp.freq2_indirect = q_f1.offset - (q_f1.SW * q_f1.SFO1);
            Util_2D.calc_scale_factors_2D(gp, q, q_f1);
        }

        p_f2.PHC0 = -p_f2.PHC0;									// Spinworks and TopSpin have    "Y"
        p_f2.PHC1 = -p_f2.PHC1;									// different sign convention   ??? why not acq dim ???

        p_f1.PHC0 = -p_f1.PHC0;									// Spinworks and TopSpin have    "Z"
        p_f1.PHC1 = -p_f1.PHC1;									// different sign convention 

        // re-set status phase constants
        //  X              Y                  Z
        q.PHC0 = 0.0;   q_f2.PHC0 = 0.0;    q_f1.PHC0 = 0.0;
        q.PHC1 = 0.0;   q_f2.PHC1 = 0.0;    q_f1.PHC1 = 0.0;

        // default LP parameters
        p.lp = LP_modes.LP_OFF;        		// X
        p_f2.lp = LP_modes.LP_OFF;     		// Y
        p_f1.lp = LP_modes.LP_OFF;     		// Z

        p_f2.lp_pred = q_f2.TD;             // Y
        p_f2.lp_input = q_f2.TD / 2;

        p_f1.lp_pred = q_f1.TD;             // Z
        p_f1.lp_input = q_f1.TD / 2;

        if (q_f2.TD < 128)                  // y
            p_f2.lp_coef = 8;

        if (q_f1.TD < 128)                  // Z
            p_f1.lp_coef = 8;

        // set default minumum sizes
        if (p.SI < q.TD / 2) p.SI = q.TD / 2;               // X
        if (p_f2.SI < q_f2.TD / 2) p_f2.SI = q_f2.TD / 2;   // Y
        if (p_f1.SI < q_f1.TD / 2) p_f1.SI = q_f1.TD / 2;   // Z

       
        if (gp.rows_real == null || !gp.auto_load)	// in no processed data, scale to fid  
        {											// intensity for display
            Util.hscale(gp);
            Util.rescale(gp, q);
            gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;
        }

        if (gp.a4_paper)
        {
            if (gp.plot_xsize > 260.0) gp.plot_xsize = 260.0;
        }
        else
        {
            if (gp.plot_xsize > 240.0) gp.plot_xsize = 240.0;
        }

        statusStripLabel2.setText("File Opened");

        if (gp.log_on) Util.log_error(fileName + " Opened", gp.scratch_folder);

        //	return (par_ok && data_ok);							// TO DO!
    }
    
//-------------------------------------------------------------------------------------------------
/**
 * Open a Varian 1D fid file
 * @param fileName  Full path to file
 */
    private static void fileOpenVarian1D(String fileName)
    {
    	q.data_path = fileName;
    	boolean local_autoload = false;
    	
    	File fi = null;
		try
		{
			fi = new File(fileName);
			q.data_folder = fi.getParent() + '/';
			q.par_path = q.data_folder + "procpar";
		}
		catch (Exception ex)
		{
            Util.show_and_log("FileOpenVarian_1D(): " + fileName, ex.getMessage());
            ex.printStackTrace();
			return;
		}

//user has selected processed data, so override autoload setting if necessary
        if (fi.getName().equals("1D_real")
            || fi.getName().equals("1D_imag"))
        {
            local_autoload = true;
            q.data_path = q.data_folder + "fid";
        }        
        
    	gp.display_mode = modes.MODE_FID;

    	if(!FileUtilVarian.getVarianPars(q.par_path, p, p_f1, p_f2, q, q_f1, q_f2))
    	{
    		Util.show_and_log("FileOpenVarian(): cannot open Varian parameters, File Open aborted\nPossible wrong data format selected?", 
    				q.par_path);
    		gp.display_mode = modes.MODE_1D;
    		return;
    	}

    	if (FileUtilVarian.getVarianFid(q.data_path, gp, q) <= 0)
    	{
    		Util.show_and_log("FileOpenVarian(): cannot open Varian fid file, File Open aborted", 
    				q.data_path);
    		gp.display_mode = modes.MODE_1D;
    		return;
    	}

    	//	now get the title (legacy code)
    	q.titlepath = q.data_folder + "text";

    	try 
        {       	
            String line;
            BufferedReader br = new BufferedReader(new FileReader(q.titlepath)); 
        	
            gp.plot_title = "";
        	while ((line=br.readLine())!=null)
        	{
                gp.plot_title += line + "\n";
            }
            br.close(); 
        }
        catch (Exception ex) 
        {               
            gp.plot_title = "no title";	
        }

    	if (gp.plot_title == null)
    		gp.plot_title = "-";

    	Util.free_proc(gp);									// necessary???? 
    	Util.alloc_proc(gp, p, q);

    	if (gp.vblock == 1)
    		if (gp.auto_load || local_autoload)				// if requested
    			if(FileUtil.readProc1D(gp, p, q) > 0)     // read processed data
    			{
    				gp.display_mode = modes.MODE_1D;   		// set 1D display mode
    				gp.cursor_flag = 0;						// flag no cursors set    				
    				FileUtil.readPeaks1D(gp, q);       		// read 1D peaks if available
    				if(!gp.retain_integral_regions)
    					FileUtil.readInteg1D();             // load integrals if available
    			}

    	gp.freq1 = q.offset;                      
    	gp.freq2 = q.offset - (q.SFO1 * q.SW);

    	Util.hscale(gp);
    	Util.rescale(gp, q);
    	p.lp = LP_modes.LP_OFF;
    	gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;
    	InterfaceUtil.set1D();				// update interface
    	ws[aws].has_data = true;
    	statusStripLabel2.setText("File Opened");
    	if (gp.log_on) Util.log_error(fileName + " Opened", gp.scratch_folder);
    }

//-------------------------------------------------------------------------------------------------
/**
 * Open a Varian 2D fid file
 * @param fileName Full path to file
 */
    private static void fileOpenVarian2D(String fileName)
    {
    	boolean par_ok = false, data_ok = false;
    	boolean local_autoload = false;

    	Util_2D.free_ser(gp);
    	Util_2D.free_rows(true, true, true, true, gp);			

    	q.data_path = fileName;

    	File fi = null;
    	try
    	{
    		fi = new File(fileName);
    		q.data_folder = fi.getParent() + '/';
    	}
    	catch (Exception ex)
    	{
    		Util.show_and_log("FileOpenVarian_2D(): " + fileName, ex.getMessage());
    		ex.printStackTrace();
    		return;
    	}   	

    	// if user has selected processed data temporarily switch to autoload mode TODO
    	if (fi.getName().equals("2D_rr"))
    	{
    		local_autoload = true;
    		q.data_path = q.data_folder + "fid";
    	}

    	q.par_path = q.data_folder + "procpar";
    	gp.display_mode = modes.MODE_FID;			// display FID at file open

    	par_ok = FileUtilVarian.getVarianPars(q.par_path, p, p_f1, p_f2, q, q_f1, q_f2);

    	if (par_ok == false)
    	{
    		Util.show_and_log("FileOpenVarian_2D(): , cannot read: ", q.par_path + 
    				"  - Possibly wrong data format selected?");
    		return;
    	}

    	Util.free_proc(gp);							// free and alloc allocate some 
    	Util.alloc_proc(gp, p, q);					// space for the data

    	//	get the title 
    	q.titlepath = q.data_folder + "text";
    	try 
    	{       	
    		String line;
    		BufferedReader br = new BufferedReader(new FileReader(q.titlepath)); 

    		gp.plot_title = "";
    		while ((line=br.readLine())!=null)
    		{
    			gp.plot_title += line + "\n";
    		}
    		br.close(); 
    	}
    	catch (Exception ex) 
    	{               
    		gp.plot_title = "no title";	
    	}

    	if(FileUtilVarian.getVarianSer(q.data_path, gp, q, q_f1) > 0)
    		data_ok = true;

    	if (data_ok == false)
    	{
    		Util.show_and_log("FileOpenVarian_2D(): cannot oepen: ", q.data_path);
    		gp.display_mode = modes.MODE_1D;			// return to 1D display
    		return;
    	}

    	if (data_ok && par_ok)
    	{	
    		Util_2D.load_display_2D_ser(0, gp, q);
    		gp.cursor_flag = 0;								// flag new freq1 set 
    		gp.freq1 = q.offset;
    		gp.freq2 = q.offset - (q.SW * q.SFO1);
    		gp.freq1_indirect = q_f1.offset;
    		gp.freq2_indirect = q_f1.offset - (q_f1.SW * q_f1.SFO1);
    		Util_2D.calc_scale_factors_2D(gp, q, q_f1);
    	}

    	if (gp.auto_load || local_autoload)	// TODO
    	{
    		statusStripLabel2.setText("Loading Processed Data");
    		if (FileUtil.readProc2D(gp, p, p_f1, q, q_f1) > 0)	// if successful, calculate 
    		{														// scale factors, etc.
    			gp.cursor_flag = 0;									// flag no cursors set 
    			//Util.check_order(gp);								// check order of freq.
    			gp.data_max_2D = Util_2D.max_2D(gp.rows_real, q.SI, q_f1.SI);		// find global maximum
    			gp.data_noise = Util_2D.noise_2D(gp, q, q_f1);		// noise level
    			Util_2D.calc_scale_factors_2D(gp, q, q_f1);			// scaling factors
    			Util_2D.calc_default_levels_2D(gp);					// default contour levels		
    			gp.display_mode =  modes.MODE_2D;					// switch to 2D display
    			gp.new_2D_view = true;								// re-contour
    		}
    	} 

    	p_f1.lp_pred = q_f1.TD;
    	p_f1.lp_input = q_f1.TD / 2;
    	p.lp = LP_modes.LP_OFF;
    	p_f1.lp = LP_modes.LP_OFF;
    	if (q_f1.TD < 128)
    		p_f1.lp_coef = 8;

    	if (gp.rows_real == null || !gp.auto_load)	// if no processed data, scale to fid intensity for
    	{											// for display
    		Util.hscale(gp);
    		Util.rescale(gp, q);
    		gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;
    	}

    	if (gp.a4_paper)
    	{
    		if (gp.plot_xsize > 260.0) gp.plot_xsize = 260.0;
    	}
    	else
    	{
    		if (gp.plot_xsize > 240.0) gp.plot_xsize = 240.0;
    	}

    	statusStripLabel2.setText("File Opened");
    	statusStripLabel1.setText("Ready");

    	if (gp.log_on) Util.log_error(fileName + " Opened");
    }

//-------------------------------------------------------------------------------------------------	
/**
 * Open Varian/Agilent 3D data set
 * @param FileName full path to the fid file
 */
    private static void fileOpenVarian3D(String fileName)
    {
    	boolean par_ok = false, data_ok = false;

    	Util_2D.free_ser(gp);
    	Util_2D.free_rows(true, true, true, true, gp);

    	q.data_path = fileName;

    	File fi = null;
    	try
    	{
    		fi = new File(fileName);
    		q.data_folder = fi.getParent() + '/';
    	}
    	catch (Exception ex)
    	{
    		Util.show_and_log("FileOpenVarian_2D(): " + fileName, ex.getMessage());
    		ex.printStackTrace();
    		return;
    	}   	           

    	q.par_path = q.data_folder + "procpar";

    	gp.display_mode = modes.MODE_FID;			// display FID at file open

    	par_ok = FileUtilVarian.getVarianPars(q.par_path, p, p_f1, p_f2, q, q_f1, q_f2);

    	if (par_ok == false)
    	{
    		Util.show_and_log("FileOpenVarian3D(): cannot read procpar",
    				"Possibly wrong data format selected?");
    		return;
    	}

    	Util.free_proc(gp);							// free and alloc allocate some 
    	Util.alloc_proc(gp, p, q);					// space for the data

//    	get the title 
    	q.titlepath = q.data_folder + "text";
    	try 
    	{       	
    		String line;
    		BufferedReader br = new BufferedReader(new FileReader(q.titlepath)); 

    		gp.plot_title = "";
    		while ((line = br.readLine())!=null)
    		{
    			gp.plot_title += line + "\n";
    		}
    		br.close(); 
    	}
    	catch (Exception ex) 
    	{               
    		gp.plot_title = "no title";	
    	}
    	
    	if (gp.plot_title == null)	// necessary????
    		gp.plot_title = "-";

    	if (FileUtilVarian3D.getSerBuffer(q.data_path, gp) > 0)
    		data_ok = true;

    	// allocate spave for one plane
    	try { gp.ser = new double[q_f1.TD][q.TD]; }
    	catch (Exception ex)
    	{
    		Util.show_and_log("fileOpenVarian3D: allocation error " + ex.getMessage());
    		return;
    	}

    	// load one plane, mostly for display
    	if(q.varian_array ==1)
    		FileUtilVarian3D.get_varian_ser_3D_XZ_order_1(0, gp, q, q_f1, q_f2);
    	else if (q.varian_array == 2)
    		FileUtilVarian3D.get_varian_ser_3D_XZ_order_2(0, gp, q, q_f1, q_f2);
    	else
    	{
    		Util.show_and_log("FileOpenVarian_3D(): unknown Varian/Agilent 3D array ordering", 
    				fileName);
    	}

    	if (data_ok == false)
    	{
    		Util.show_and_log("FileOpenVarian_3D(): cannot read Varian 'fid'", fileName);
    		gp.display_mode = modes.MODE_1D;			// return to 1D display
    		return;
    	}

    	if (data_ok && par_ok)
    	{
    		gp.display_mode = modes.MODE_FID;
    		Util.free_proc(gp);
    		Util.alloc_proc(gp, p, q);

    		Util_2D.load_display_2D_ser(0, gp, q);
    		gp.cursor_flag = 0;								// flag new freq1 set 
    		gp.freq1 = q.offset;
    		gp.freq2 = q.offset - (q.SW * q.SFO1);
    		gp.freq1_indirect = q_f1.offset;
    		gp.freq2_indirect = q_f1.offset - (q_f1.SW * q_f1.SFO1);
    		Util_2D.calc_scale_factors_2D(gp, q, q_f1);
    	}

    	// Default LP values
    	p.lp = LP_modes.LP_OFF;			// X
    	
    	p_f2.lp_pred = q_f2.TD;			// Y
    	p_f2.lp_input = q_f2.TD / 2;
    	p_f2.lp = LP_modes.LP_OFF;
    	if (q_f1.TD < 128)
    		p_f1.lp_coef = 8;
    	
    	p_f1.lp_pred = q_f1.TD;			// Z
    	p_f1.lp_input = q_f1.TD / 2;
    	p_f1.lp = LP_modes.LP_OFF;
    	if (q_f1.TD < 128)
    		p_f1.lp_coef = 8;

    	if (gp.rows_real == null || !gp.auto_load)	// if no processed data, scale to fid intensity for
    	{											// for display
    		Util.hscale(gp);
    		Util.rescale(gp, q);
    		gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;
    	}

    	if (gp.a4_paper)
    	{
    		if (gp.plot_xsize > 260.0) gp.plot_xsize = 260.0;
    	}
    	else
    	{
    		if (gp.plot_xsize > 240.0) gp.plot_xsize = 240.0;
    	}

    	statusStripLabel2.setText("File Opened");
    	statusStripLabel1.setText("Ready");
    	if (gp.log_on) Util.log_error(fileName + " Opened", gp.scratch_folder);
    }
           
//-------------------------------------------------------------------------------------------------
/**
 * Open a JCAMP_DX 1D spectrum file
 * @param fileName Full path to file
 */
		private static void fileOpenJDX(String fileName)
		{
            int ret, npoints = 0;

            ret = DX_Util.read_jdx_pars(fileName, gp, p, q);      

            if(ret == 5)
            	npoints = DX_Util.read_jdx_fid(fileName, gp, p, q);
            else if(ret == 3)
            	npoints = DX_Util.read_jdx_spect(fileName, gp, p, q);

            
			if (ret <= 0)
			{
                Util.show_and_log("Cannot read: " + fileName,  " Possible corrupt file?"
                + "\nPossible wrong data format selected?");
				return;
			}

			q.data_path = fileName;
			statusStripLabel2.setText(String.format("JDX Read: %d points", npoints));
		
			gp.cursor_flag = 0;							// flag new freq1 set 
            
			gp.freq1 = q.offset; 
			gp.freq2 = q.offset - (q.SW * q.SFO1);
		
			if (gp.auto_load)	          				  // if requested,try to read
	            if (FileUtil.readProc1D(gp, p, q) > 0)    // any previously stored data  
	            {
	                gp.display_mode = modes.MODE_1D;
	                gp.cursor_flag = 0;	                  // flag no cursors set                               					
	                FileUtil.readPeaks1D(gp, q);      	  // read 1D peaks if available
	                if(!gp.retain_integral_regions)
	                	FileUtil.readInteg1D();
	            }	
			
			Util.hscale(gp);
			Util.rescale(gp, q);
			gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;

            if (gp.data_max < 10.0)
                gp.integ_scale = gp.integ_scale * 10000.0;
            else if(gp.data_max < 100.0)
                gp.integ_scale = gp.integ_scale * 1000.0;
            else if (gp.data_max < 1000.0)
                gp.integ_scale = gp.integ_scale * 100.0;

			InterfaceUtil.set1D();
            statusStripLabel2.setText("File Opened");
            if (gp.log_on) Util.log_error(fileName + " Opened", gp.scratch_folder);
		}

//-------------------------------------------------------------------------------------------------
/**
 * Open a JEOL (.jdf) 1D data set
 * @param fileName Full path to file
 */
		private static void fileOpenJEOL_1D(String fileName)
		{		
			// get the parent folder
			try
			{
				File file = new File(fileName);
				q.data_folder = file.getParent() + "/";
				System.out.println(q.data_folder);
			}
			catch(Exception ex)
			{
				Util.show_and_log("cannot get file information on: ", q.data_path );
				ex.printStackTrace();
				return;
			}

			q.data_path = fileName;

			int rc = JDF_Util.get_jdf_pars(q.data_path, gp, p, p_f1, q, q_f1);

			if (rc <= 0)
			{
				Util.show_and_log("FileOpenJEOL_1D(): cannot open JEOL parameter file: ", q.data_path
						+ "\nPossible wrong data format selected");
				gp.display_mode = modes.MODE_1D;
				return;
			}


			if (rc == 2)        // FID
			{
				gp.display_mode = modes.MODE_FID;

				if (JDF_Util.get_jdf_fid_1D(q.data_path, gp, q) <= 0)
				{
					Util.show_and_log("FileOpenJEOL_1D(): cannot open JEOL fid file: ", q.data_path);
					gp.display_mode = modes.MODE_1D;
					return;
				}
			}

			if (rc == 1)        // Spectrum
			{
				gp.display_mode = modes.MODE_1D;
				if (JDF_Util.get_jdf_spect_1D(q.data_path, gp, q) <= 0)
				{
					Util.show_and_log("FileOpenJEOL_1D(): cannot open JEOL spectrum` file: ", q.data_path);
					gp.display_mode = modes.MODE_1D;
					return;
				}
			}

			if(gp.display_mode == modes.MODE_FID)
				gp.freq2 = q.offset - q.TD * q.hzperpt;
			else
				gp.freq1 = q.offset;
			gp.freq2 = q.offset - (q.SFO1 * q.SW);

			if (gp.auto_load)	          				  // if requested,try to read
	            if (FileUtil.readProc1D(gp, p, q) > 0)    // any previously stored data  
	            {
	                gp.display_mode = modes.MODE_1D;
	                gp.cursor_flag = 0;	                  // flag no cursors set                               					
	                FileUtil.readPeaks1D(gp, q);      	  // read 1D peaks if available
	                if(!gp.retain_integral_regions)
	                	FileUtil.readInteg1D();
	            }
			
			Util.hscale(gp);
			Util.rescale(gp, q);

			p.lp = LP_modes.LP_OFF;
			InterfaceUtil.set1D();		

			ws[aws].has_data = true;

			statusStripLabel2.setText("JEOL File Opened");
			if (gp.log_on) Util.log_error(fileName + " Opened", gp.scratch_folder);
		}

//-------------------------------------------------------------------------------------------------
/**
 * Open a JEOL (.jdf) 2D NMR data set.
 * @param fileName Full path to file
 */
		private static void fileOpenJEOL_2D(String fileName)
		{
			// get the parent folder
			try
			{
				File file = new File(fileName);
				q.data_folder = file.getParent() + "/";
				System.out.println(q.data_folder);
			}
			catch(Exception ex)
			{
				Util.show_and_log("cannot get file information on: ", q.data_path );
				ex.printStackTrace();
				return;
			}
						
			q.data_path = fileName;
			
			int rc = JDF_Util.get_jdf_pars(q.data_path, gp, p, p_f1, q, q_f1);

			if (rc <= 0)
			{
				Util.show_and_log("FileOpenJEOL_1D(): cannot open JEOL parameter file: ", q.data_path
						+ "\nPossible wrong data format selected?");
				gp.display_mode = modes.MODE_1D;
				return;
			}


			if (rc == 2)        // FID
			{
				gp.display_mode = modes.MODE_FID;

				if (JDF_Util.get_jdf_fid_2D(q.data_path, gp, q, q_f1) <= 0)
				{
					Util.show_and_log("FileOpenJEOL_2D(): cannot open JEOL 2D fid file: ", q.data_path);
					gp.display_mode = modes.MODE_1D;
					return;
				}
			}

			Util.alloc_proc(gp, p, q);

			Util_2D.load_display_2D_ser(0, gp, q);
			gp.cursor_flag = 0;								// flag new freq1 set 

			Util_2D.calc_scale_factors_2D(gp, q, q_f1);

			if (gp.display_mode == modes.MODE_FID)
				gp.freq2 = q.offset - q.TD * q.hzperpt;
			else
				gp.freq1 = q.offset;

			gp.freq2 = q.offset - (q.SFO1 * q.SW);

			gp.freq1_indirect = q_f1.offset;
			gp.freq2_indirect = q_f1.offset - (q_f1.SW * q_f1.SFO1);

			Util.hscale(gp);
			Util.rescale(gp, q);

			p.lp = LP_modes.LP_OFF;

			InterfaceUtil.set2D();

			if (gp.auto_load)						// TODO
	    	{
	    		statusStripLabel2.setText("Loading Processed Data");
	    		if (FileUtil.readProc2D(gp, p, p_f1, q, q_f1) > 0)	// if successful, calculate 
	    		{														// scale factors, etc.
	    			gp.cursor_flag = 0;									// flag no cursors set 
	    			gp.data_max_2D = Util_2D.max_2D(gp.rows_real, q.SI, q_f1.SI);		// find global maximum
	    			gp.data_noise = Util_2D.noise_2D(gp, q, q_f1);		// noise level
	    			Util_2D.calc_scale_factors_2D(gp, q, q_f1);			// scaling factors
	    			Util_2D.calc_default_levels_2D(gp);					// default contour levels		
	    			gp.display_mode = modes.MODE_2D;				    // switch to 2D display
	    			gp.new_2D_view = true;								// re-contour
	    		}
	    	}

	    	if (gp.rows_real == null || !gp.auto_load)	// in no processed data, scale to fid  
	    	{											// intensity for display
	    		Util.hscale(gp);
	    		Util.rescale(gp, q);
	    		gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;
	    	}
	    	
			ws[aws].has_data = true;
			statusStripLabel2.setText("File Opened");
			if (gp.log_on) Util.log_error(fileName + " Opened", gp.scratch_folder);
		}


		
//-------------------------------------------------------------------------------------------------
//	Magritek
		
//-------------------------------------------------------------------------------------------------
        private static void FileOpenMagritek(String fileName)
        {
        	File file;
        	try
			{
				file = new File(fileName);				
			}
			catch(Exception ex)
			{
				Util.show_and_log("cannot get file information on: ", q.data_path );
				ex.printStackTrace();
				return;
			}
        	
        	q.data_path = fileName;
        	q.data_folder = file.getParent() + "/";
			//System.out.println(q.data_folder);			
            String name = file.getName();
            
            if (name.equalsIgnoreCase("spectrum.1d"))	// do the same for 2D???
            {
                Util.show_and_log("processed Magritek data not supported, using fid file: data.1d\n Click OK, then click Process button");
                q.data_path = q.data_folder + "data.1d";
            }

            String par_path = q.data_folder + "acqu.par";
            if(fileUtilMagritek.get_par(par_path, gp, p, q, p_f1, q_f1) == false)
            {
                Util.show_and_log("FileMagritek(): cannot open Magritek parameter file: ",
                           par_path + "\nPossible wrong data format selected?");
                //Cursor = System.Windows.Forms.Cursors.Default;
                gp.display_mode = modes.MODE_1D;
                return;
            }
         
            Util.free_proc(gp);
            //Util.free_calc(gp);
            Util.alloc_proc(gp, p, q);

            if (q.data_mode == modes.MODE_1D)
            {
                if (fileUtilMagritek.get_fid(q.data_path, q, gp) <= 0)
                {
                    Util.show_and_log("FileMagritek(): cannot open Magritek fid file: ",
                        q.data_path + "\nPossible wrong data format selected?");
                    //Cursor = System.Windows.Forms.Cursors.Default;
                    gp.display_mode = modes.MODE_1D;
                    return;
                }
            }
           
            else
            {
                if (fileUtilMagritek.get_ser(q.data_path, q, q_f1, gp) <= 0)
                {
                    Util.show_and_log("FileMagritek(): cannot open Magritek ser file: ",
                        q.data_path + "\nPossible wrong data format selected?");
                    //Cursor = System.Windows.Forms.Cursors.Default;
                    gp.display_mode = modes.MODE_1D;
                    return;
                }
            }          

            gp.freq1 = q.offset;
            gp.freq2 = q.offset - (q.SW * q.SFO1);
            p.SI = q.SI;                                    // set size = status size
            gp.cursor_flag = 0;	                            // flag no cursors set 

            if (q.data_mode == modes.MODE_2D)
            {
                p_f1.SI = q_f1.SI;
                gp.freq1_indirect = q_f1.offset;
                gp.freq2_indirect = q_f1.offset - (q_f1.SW * q_f1.SFO1);
                //gp.data_max_2D = Util_2D.max_2D(gp, q, q_f1);		// find global maximum
                gp.data_noise = Util_2D.noise_2D(gp, q, q_f1);		// noise level
                Util_2D.calc_scale_factors_2D(gp, q, q_f1);			// scaling factors
                Util_2D.calc_default_levels_2D(gp);					// default contour levels		
                gp.display_mode = modes.MODE_2D;				    // switch to 2D display
                InterfaceUtil.set2D();
                gp.new_2D_view = true;								// re-contour
            }
            else													// must be 1D
            {
                Util.hscale(gp);
                Util.rescale(gp, q);
                gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max;
                InterfaceUtil.set1D();
            }

            gp.display_mode = modes.MODE_FID;	// only raw data supported for Magritek
            ws[aws].has_data = true;
            //DrawingPanel.Invalidate();
            statusStripLabel2.setText("File Opened");
            if (gp.log_on) Util.log_error(fileName + " Opened");
        }

        
		
//-------------------------------------------------------------------------------------------------
/**
 * Open an NMRPipe processed 2D/3D data file.
 * @param fileName  Full path to file
 */
		private  static void fileOpenNMRpipe(String fileName)
		{
			try
			{
				File fi = new File(fileName);
				q.data_folder = fi.getParent() + '/';
			}
			catch (Exception ex)
			{
				Util.show_and_log("fileOpenNMRPipe(): " + q.data_folder, ex.getMessage());
				return;
			}

			int ret_code;

			ret_code = NMRpipe.readPipeData(fileName, gp, q, q_f1);

			if (ret_code  <= 0)
			{
				Util.show_and_log("cannot open: " + fileName, "possible corrupt file?"
						+ "\nPossible wrong data format selected?");
				return;
			}

			p.SI = q.SI;
			p_f1.SI = q_f1.SI;

			Util.free_proc(gp);
			Util.alloc_proc(gp, p, q);

			gp.freq1 = q.offset;
			gp.freq2 = q.offset - (q.SW * q.SFO1);
			gp.freq1_indirect = q_f1.offset;
			gp.freq2_indirect = q_f1.offset - (q_f1.SW * q_f1.SFO1);

			gp.cursor_flag = 0;									// flag no cursors set 
			gp.data_max_2D = Util_2D.max_2D(gp.rows_real, q.SI, q_f1.SI); // find global maximum
			gp.data_noise = Util_2D.noise_2D(gp, q, q_f1);		// noise level
			Util_2D.calc_scale_factors_2D(gp, q, q_f1);			// scaling factors
			Util_2D.calc_default_levels_2D(gp);					// default contour levels		
			gp.display_mode = modes.MODE_2D;					// switch to 2D display

			if (q.data_mode == modes.MODE_2D)
				InterfaceUtil.set2D();
			else if (q.data_mode == modes.MODE_3D)
				InterfaceUtil.set3D();

			gp.new_2D_view = true;								// re-contour
			ws[aws].has_data = true;
			update_spect = true;					         	// force screen re-draw
			canvas.redraw();
			statusStripLabel2.setText("File Opened");
			if (gp.log_on) Util.log_error(fileName + " Opened", gp.scratch_folder);
		}

//-------------------------------------------------------------------------------------------------
/**
 * Opens an NMR data set when the filename is dropped into the canvas from a file exploree, etc.
 * @param fileName the name of the file or, for Bruker and Vagilent data, an experiment folder
 */
		public static void openFromDragDrop(String fileName)
        {
            if (gp.log_on) Util.log_error("Opening " + fileName + "from drag and drop",
                gp.scratch_folder);

            /*p.lp = LP_modes.LP_OFF;		// turn off lp as defaul.t  why??
            p_f1.lp = LP_modes.LP_OFF;	// both dimensions
            p.fcor = 0.5;               // default value for first point multiplier in F2
*/
            File file = new File(fileName);			// for evaluating the file name
            
            if(file.isDirectory())					// is it a folder (directory)?
    		{
            	String directory = fileName;
            	fileName = directory + "/fid";		// add "fid" for Vagilent and Bruker 1D
            	file = new File(fileName);			// does this exist?
            	if(!file.exists())
            		fileName = directory + "/ser";	// no,try adding "ser"  Bruker 2D/3D
    		}
            
            file = new File(fileName);				// see if result exists
            if(!file.exists()) return;				// file doesn't exist, so bail out
            
            openSelectedFile(fileName);      		// open it     
            update_spect = true;					// update display
            canvas.redraw();
        }
		
//-------------------------------------------------------------------------------------------------
/**
 * Read any processed data in SpinWorks format.
 */
		public static void readProcessed()	// Perhaps this should be moved somewhere else???
		{
			shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_WAIT));
			if (q.data_mode == modes.MODE_1D)
			{
				statusStripLabel2.setText("Loading Processed Data");
				statusStripLabel1.setText("");
				if (FileUtil.readProc1D(gp, p, q) > 0)		// if successful, calculate scale
				{												// factors, etc.
					gp.cursor_flag = 0;							// flag no cursors set                 
					FileUtil.readPeaks1D(gp, q);            	// read 1D peaks if available
					if(!gp.retain_integral_regions)
						FileUtil.readInteg1D();                 // also any integrals
				}
				else
				{
					Util.show_and_log("No processed 1D data to load, or corrupt file (must re-process)");
					statusStripLabel2.setText("");
					statusStripLabel1.setText("Ready");
					gp.cursor_flag = 0;
					shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));
					return;				
				}
			}
			else if (q.data_mode == modes.MODE_2D)
			{
				statusStripLabel2.setText("Loading Processed Data");
				statusStripLabel1.setText("");

				if (FileUtil.readProc2D(gp, p, p_f1, q, q_f1) > 0)	// if successful, calculate 
				{														// scale factors, etc.
					gp.cursor_flag = 0;									// flag no cursors set 
					gp.data_max_2D = Util_2D.max_2D(gp.rows_real, q.SI, q_f1.SI);		// find global maximum
					gp.data_noise = Util_2D.noise_2D(gp, q, q_f1);		// noise level
					Util_2D.calc_scale_factors_2D(gp, q, q_f1);			// scaling factors
					Util_2D.calc_default_levels_2D(gp);					// default contour levels		
					gp.display_mode = modes.MODE_2D;					// switch to 2D display
					slider.setVisible(false);							// no scrollbar in 2D
					gp.new_2D_view = true;								// re-contour
				}
				else
				{
					Util.show_and_log("No processed 2D data to load or corrupt file (re-process)");
					statusStripLabel2.setText("");
					statusStripLabel1.setText("Ready");
					gp.cursor_flag = 0;
					shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));
					return;
				}
			}
			else if (q.data_mode == modes.MODE_3D)	// Processed 3D
			{
				statusStripLabel2.setText("Loading Processed 3D Data");
				statusStripLabel1.setText("");

				String par_file = q.data_folder + "3D_par";

				if (FileUtil3D.getPar3D(par_file, gp, p, p_f2, p_f1, q, q_f2, q_f1)
						== false)
				{
					Util.show_and_log("No 3D parameters");
					statusStripLabel2.setText("");
					statusStripLabel1.setText("Ready");
					gp.cursor_flag = 0;
					shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));
					return;
				}

				try
				{
					Proc_3D.XYZ = new float[q.SI][q_f2.SI][p_f1.SI];                   
				}
				catch (Exception ex)
				{
					Util.show_and_log("XYZ cube allocation error, probable insufficient memory",
							ex.getMessage());
					shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));
					return;
				}

				if(FileUtil3D.readCube(q.data_folder, q.SI, q_f2.SI, q_f1.SI) 
						> 0)	                                        // if successful, calculate 
				{														// scale factors, etc.  
					Util_2D.alloc_rows(q_f1.SI, q.SI, gp, q, q_f1);     // alloc. space but
					Util_2D.free_rows(false, true, true, true, gp);     // release imaginaries

					Util_3D.load_XZ_plane(gp.current_plane, Proc_3D.XYZ, gp.rows_real, 
							q.SI, q_f1.SI);
					gp.display_mode = modes.MODE_2D;		       		// switch to 2D display
					gp.mode_3D = modes_3D.XZ;       					// default 3D is XZ mode            
					Util_2D.calc_scale_factors_2D(gp, q, q_f1);			// scaling factors
					// find global maximum
					gp.data_max_3D = Util_3D.max_3D(Proc_3D.XYZ, q.SI, q_f2.SI, q_f1.SI);	
					gp.expt_scale = (double)gp.dysize * 0.80 / gp.data_max_3D;    
					Util_2D.calc_default_levels_2D(gp);					// default contour levels                  
					InterfaceUtil.set_tool_tips();					    // due to change in display mode
					gp.new_2D_view = true;								// re-contour
					gp.dimensions_processed = 3;
					gp.current_plane = 0;
					gp.cursor_flag = 0;									// flag no cursors set  
					slider.setVisible(false);							// no scrollbar in 2D
					DisplayButtons.fullButtonClick();					// triggers screen re-draw
				}
				else
				{
					Util.show_and_log("No processed 3D data to load or corrupt file (re-process)");
					statusStripLabel2.setText("");
					statusStripLabel1.setText("Ready");
					gp.cursor_flag = 0;
					shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));
					return;
				}
			}

			statusStripLabel2.setText("Processed Data Loaded");
			statusStripLabel1.setText("Ready");
			gp.cursor_flag = 0;;
			update_spect = true;
			canvas.redraw();
			shell.setCursor(new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));
		}

//-------------------------------------------------------------------------------------------------
/**
 * load an fid into the processing buffer
 * @param gp
 * @param p
 * @param q
 * @return true if OK, false on error
 */
    public static boolean get_data(global_par gp, data_par p, data_par q)
    {
        if (q.data_mode == modes.MODE_1D)
        {
            if (Fft.load_fid(gp, p, q) == false)	            // get data and bail out if 
				return false;									// load failed
        }
        else if (q.data_mode == modes.MODE_2D)			
        {
            if (gp.proc_buffer == null || gp.proc_buffer.length == 0)
                Util.alloc_proc_buffer(4 * p.SI, gp);

            if(Util_2D.load_fid_2D_dir(gp.block_2D, gp, p, q) == false)     
           		return false;
        }

        q.PHC0 = 0.0;       q.PHC1 = 0.0;				         // re-set phase constants

        return true;
    }
 //------------------------------------------------------------------------------------------------
}	// end class
