/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

package spinworks;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * This class handles SpinWorks specific (not vendor specific)
 * file IO for SpinWorks Java
 * @author marat 
 * University of Manitoba
 */
public class FileUtil3D 
{
//-------------------------------------------------------------------------------------------------
/**
 * Writes the 3D parameters to a text file
 * @param gp global (now workspace) parameter object
 * @param q status parameters in X (F3)
 * @param q_f2 status parameters in Y (F2)
 * @param q_f1 status parameters in Z (F1)
 * @return -1 if wrong data type, -4 if file could not be opened, -3 if exception thrown, 1 if OK
 */
    public static int save_par_3D(global_par gp, 
    		data_par q, data_par q_f2, data_par q_f1)
    {
        double scale_factor = 1.0;

        if (gp.data_max_2D > 2.0e9)
            scale_factor = gp.data_max_2D / 2.0e9;

        if (!(q.data_type == SW_main.data_types.BRUKER
            || q.data_type == SW_main.data_types.VNMR))
            return -1;

        String fileNamePar = q.data_folder + "3D_par";
        PrintWriter pw;
        
        DateFormat df = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
    	Date dateobj = new Date();
    	
        //	write the parameter file
        try { pw = new PrintWriter(fileNamePar); }
        catch (SecurityException ex)
        {
            Util.show_and_log("SecurityException: save_proc_3D(): " + fileNamePar, ex.getMessage());
            return -4;
        }
        catch (Exception ex)
        {
            Util.show_and_log("save_proc_3D(): " + fileNamePar, ex.getMessage());
            return -4;
        }

        try
        {
        	pw.println("## SpinWorksJ 0.0 3D parameters for: " + q.data_folder);
        	pw.println("## saved: " + df.format(dateobj) + SW_main.lineSep );
			if (gp.plot_title == null)
				gp.plot_title = " ";

			String temp = gp.plot_title.replace('\n', ';'); // replace CR LF pair
			temp = temp.replace('\r', ';');
			pw.println("title= " + temp);

            pw.println(String.format("size= %d", q.SI));
            pw.println(String.format("size_f1= %d", q_f1.SI));
            pw.println(String.format("size_f2= %d", q_f2.SI));
            

            pw.println(String.format("sf= %.8f", q.SF));
            pw.println(String.format("sf_f1= %.8f", q_f1.SF));
            pw.println(String.format("sf_f2= %.8f", q_f2.SF));

            pw.println(String.format("sf_trans= %.8f", q.SFO1));
            pw.println(String.format("sf_trans_f1= %.8f", q_f1.SFO1));
            pw.println(String.format("sf_trans_f2= %.8f", q_f2.SFO1));

            pw.println(String.format("sw= %.3f", q.SFO1 * q.SW));
            pw.println(String.format("sw_f1= %.3f", q_f1.SFO1 * q_f1.SW));
            pw.println(String.format("sw_f2= %.3f", q_f2.SFO1 * q_f2.SW));

            pw.println(String.format("scale_factor= %.8f", scale_factor));

            pw.println(String.format("roi1_freq1= %.4f", gp.roi_1_f2_left));
            pw.println(String.format("roi1_freq2= %.4f", gp.roi_1_f2_right));
            pw.println(String.format("roi1_freq1_indirect= %.4f", gp.roi_1_f1_left));
            pw.println(String.format("roi1_freq2_indirect= %.4f", gp.roi_1_f1_right));

            pw.println(String.format("roi2_freq1= %.4f", gp.roi_2_f2_left));
            pw.println(String.format("roi2_freq2= %.4f", gp.roi_2_f2_right));
            pw.println(String.format("roi2_freq1_indirect= %.4f", gp.roi_2_f1_left));
            pw.println(String.format("roi2_freq2_indirect= %.4f", gp.roi_2_f1_right));

            pw.println(String.format("roi3_freq1= %.4f", gp.roi_3_f2_left));
            pw.println(String.format("roi3_freq2= %.4f", gp.roi_3_f2_right));
            pw.println(String.format("roi3_freq1_indirect= %.4f", gp.roi_3_f1_left));
            pw.println(String.format("roi3_freq2_indirect= %.4f", gp.roi_3_f1_right));

            pw.println(String.format("phase0= %.4f", q.PHC0));
            pw.println(String.format("phase1= %.4f", q.PHC1));
            pw.println(String.format("phase0_f1= %.4f", q_f1.PHC0));
            pw.println(String.format("phase1_f1= %.4f", q_f1.PHC1));
            pw.println(String.format("phase0_f2= %.4f", q_f2.PHC0));
            pw.println(String.format("phase1_f2= %.4f", q_f2.PHC1));

            pw.println(String.format("wdw= %d", q.WDW.ordinal()));
            pw.println(String.format("wdw_f1= %d", q_f1.WDW.ordinal()));
            pw.println(String.format("wdw_f2= %d", q_f2.WDW.ordinal()));
            pw.println(String.format("lb= %.4f", q.LB));
            pw.println(String.format("lb_f1= %.4f", q_f1.LB));
            pw.println(String.format("lb_f2= %.4f", q_f2.LB));
            pw.println(String.format("gf= %.4f", q.GB));
            pw.println(String.format("gf_f1= %.4f", q_f1.GB));
            pw.println(String.format("gf_f2= %.4f", q_f2.GB));
            pw.println(String.format("ssb= %.4f", q.ssb));
            pw.println(String.format("ssb_f1= %.4f", q_f1.ssb));
            pw.println(String.format("ssb_f2= %.4f", q_f2.ssb));

            pw.println(String.format("f2_path= %s", gp.f2_trace_path));
            pw.println(String.format("f1_path= %s", gp.f1_trace_path));

            pw.println(String.format("x_size= %.1f", gp.plot_xsize));
            pw.println(String.format("y_size= %.1f", gp.plot_ysize));
            pw.println(String.format("plot_axis= %b", gp.plot_axis));
            pw.println(String.format("plot_frame= %b", gp.plot_frame));
            pw.println(String.format("plot_integrals= %b", gp.plot_integrals));
            pw.println(String.format("plot_parameters= %b", gp.plot_parameters));
            pw.println(String.format("plot_peak_labels= %b", gp.plot_peak_labels));
            pw.println(String.format("plot_simulated= %b", gp.plot_simulated));
            pw.println(String.format("plot_spectrum= %b", gp.plot_spectrum));
            pw.println(String.format("plot_grid= %b", gp.plot_grid));
            pw.println(String.format("parameters_right= %b", gp.parameters_right));

            pw.println(String.format("axis_font_height= %.1f",
                gp.axis_font_height));
            pw.println(String.format("peak_font_height= %.1f",
                gp.peak_font_height));
            pw.println(String.format("integ_font_height= %.1f",
                gp.integ_font_height));
            pw.println(String.format("title_font_height= %.1f",
                gp.title_font_height));
            pw.println(String.format("par_font_height= %.1f",
                gp.par_font_height));
            pw.println(String.format("spect_pen_thickness= %.2f",
                gp.spect_pen_thickness));
        }
        catch (Exception ex)
        {
            Util.show_and_log("savePar3D() raised the following exception: ", ex.getMessage());
            pw.close();
            return -3;
        }        
        pw.close();
        return 1;
    }
    
//------------------------------------------------------------------------------------------------- 
/**
 * Gets parameters from a SpinWorks 3D parameter file
 * @param FileName full path name of the parameter set to read
 * @param gp a SpinWorks "global parameters" object for currrent workspace
 * @param p set parameters in X (F3)
 * @param p_f2 set parameters in Y (F2)
 * @param p_f1 set parameters in Z (F1)
 * @param q status parameters in X (F3)
 * @param q_f2 status parameters in Y (F2)
 * @param q_f1 status parameters in Z (F1)
 * @return true if OK, false on error
 */
    public static boolean getPar3D(String fileName, global_par gp,
        data_par p, data_par p_f2, data_par p_f1, 
        data_par q, data_par q_f2, data_par q_f1 )
    //           X           Y              Z
    {
        String instring = "";
        String delim = "[ =]+";
        String[] token;

        int size = 0, size_f1 = 0, size_f2 = 0;
        double sf = 500.0, sf_f1 = 500.0, sf_f2 = 500.0;
        double sw = 0.0, sw_f1 = 0.0, sw_f2 = 0.0;

//    	read the parameter file
        BufferedReader br;
        try { br = new BufferedReader(new FileReader(fileName)); }
        catch (Exception ex)
        {
        	if(!gp.auto_load)
        		Util.show_and_log("read_proc_2D():" + ex.getMessage());
        	return false;
        }

        //	loop through all lines in parameter file looking for desired parameters
        //  add others as needed
        try
        {
        	for(;;)
        	{
        		try {instring = br.readLine();} 
        		catch (IOException ex) 
        		{
        			ex.printStackTrace();
        			Util.show_and_log("get_par(), read error: " + fileName, ex.getMessage());
        			try { br.close(); } 
        			catch (IOException e) { e.printStackTrace(); }
        			return false;
        		}

        		if(instring == null) break;
        		if (instring.length() > 3)
        		{
        			token = instring.split(delim);

        			if (token[0].equals("title"))				// title
        			{
        				String temp =  instring.substring(7);
        				gp.plot_title = temp.replace(";;", "\r\n");
        			}

        			if (token[0].equals( "size"))				// get the size
        				size = Integer.parseInt(token[1]);

        			if (token[0].equals( "size_f1"))			// get the size
        				size_f1 = Integer.parseInt(token[1]);

        			if (token[0].equals( "size_f2"))			// get the size
        				size_f2 = Integer.parseInt(token[1]);

        			/*if (token[0].equals( "scale_factor"))		// get scaling factor 
        				scale_factor = Double.parseDouble(token[1]);*/

        			if (token[0].equals( "sf"))					// get SF (reference frequency) 
        				sf = Double.parseDouble(token[1]);

        			if (token[0].equals( "sf_f1"))				// get SF (reference frequency) 
        				sf_f1 = Double.parseDouble(token[1]);               

        			if (token[0].equals( "sf_f2"))				// get SF (reference frequency) 
        				sf_f2 = Double.parseDouble(token[1]);

        			if (token[0].equals( "sf_trans"))			// get SFO1 (trans. frequency) 
        			{
        				q.SFO1 = Double.parseDouble(token[1]);
        				p.SFO1 = q.SFO1;
        			}

        			if (token[0].equals( "sf_trans_f1"))		// get SFO1 (f1) (trans. frequency) 
        			{
        				q_f1.SFO1 = Double.parseDouble(token[1]);
        				p_f1.SFO1 = q_f1.SFO1;
        			}

        			if (token[0].equals( "sf_trans_f2"))		// get SFO1 (f1) (trans. frequency) 
        			{
        				q_f2.SFO1 = Double.parseDouble(token[1]);
        				p_f2.SFO1 = q_f2.SFO1;
        			}

        			if (token[0].equals( "sw"))				    // get sw in F2 in Hz 
        			{
        				sw = Double.parseDouble(token[1]);
        			}

        			if (token[0].equals( "sw_f1"))				 // get sw in F2 in Hz 
        			{
        				sw_f1 = Double.parseDouble(token[1]);
        			}

        			if (token[0].equals( "sw_f2"))				 // get sw in F2 in Hz 
        			{
        				sw_f2 = Double.parseDouble(token[1]);
        			}

        			if (token[0].equals( "phase0"))
        				q.PHC0 = Double.parseDouble(token[1]);

        			if (token[0].equals( "phase1"))
        				q.PHC1 = Double.parseDouble(token[1]);             

        			if (token[0].equals( "phase0_f1"))
        				q_f1.PHC0 = Double.parseDouble(token[1]);

        			if (token[0].equals( "phase1_f1"))
        				q_f1.PHC1 = Double.parseDouble(token[1]);
        			
        			if (token[0].equals( "phase0_f2"))
        				q_f2.PHC0 = Double.parseDouble(token[1]);

        			if (token[0].equals( "phase1_f2"))
        				q_f2.PHC1 = Double.parseDouble(token[1]);

        			if (token[0].equals("wdw"))				// get the window function used
        				q.WDW = SW_main.windows.values()[Integer.parseInt(token[1])];

        			if (token[0].equals("wdw_f1"))			// get the window function used
        				q_f1.WDW = SW_main.windows.values()[Integer.parseInt(token[1])]; 

        			if (token[0].equals("wdw_f2"))			// get the window function used
        				q_f2.WDW = SW_main.windows.values()[Integer.parseInt(token[1])];

        			if (token[0].equals( "lb"))
        				q.LB = Double.parseDouble(token[1]);

        			if (token[0].equals( "lb_f1"))
        				q_f1.LB = Double.parseDouble(token[1]);
        			
        			if (token[0].equals( "lb_f2"))
        				q_f2.LB = Double.parseDouble(token[1]);

        			if (token[0].equals( "gf"))
        				q.GB = Double.parseDouble(token[1]);

        			if (token[0].equals( "gf_f1"))
        				q_f1.GB = Double.parseDouble(token[1]);
        			
        			if (token[0].equals( "gf_f2"))
        				q_f2.GB = Double.parseDouble(token[1]);

        			if (token[0].equals( "ssb"))
        				q.ssb = Double.parseDouble(token[1]);               

        			if (token[0].equals( "ssb_f1"))
        				q_f1.ssb = Double.parseDouble(token[1]);
        			
        			if (token[0].equals( "ssb_f2"))
        				q_f2.ssb = Double.parseDouble(token[1]);  
        		}	// if(instring.length() > 3)
        	}	// for(;;)
        }
        catch(Exception ex)	// Oh Oh...
        {
        	Util.show_and_log("Error parsing 3D parameter file for: " + instring,
        			ex.getMessage());
        	ex.printStackTrace();
        	return false;	// bail out
        }
        
        try {	br.close();	} 
        catch (IOException e) {	e.printStackTrace();	}

        if (size != q.SI)
        {
            p.SI = size; q.SI = p.SI;
            q.hzperpt = q.SW * q.SFO1 / (double)(q.SI);
        }

        if (size_f1 != q_f1.SI)
        {
            p_f1.SI = size_f1; q_f1.SI = p_f1.SI;
            q_f1.hzperpt = q_f1.SW * q_f1.SFO1 / (double)(q_f1.SI);
        }

        if (size_f2 != q_f2.SI)
        {
            p_f2.SI = size_f2; q_f2.SI = p_f2.SI;
            q_f2.hzperpt = q_f2.SW * q_f2.SFO1 / (double)(q_f2.SI);
        }

        if (sf != q.SF)
        {
            q.SF = sf; p.SF = q.SF;
            q.offset = (q.SFO1 - p.SF) * 1.0e06 + (q.SW * q.SFO1) / 2.0;
        }          

        if (sf_f1 != q_f1.SF)
        {
            q_f1.SF = sf_f1; p_f1.SF = q_f1.SF;
            q_f1.offset = (q_f1.SFO1 - p_f1.SF) * 1.0e06 + (q_f1.SW * q_f1.SFO1) / 2.0;
        }

        if (sf_f2 != q_f2.SF)
        {
            q_f2.SF = sf_f2; p_f2.SF = q_f2.SF;
            q_f2.offset = (q_f2.SFO1 - p_f2.SF) * 1.0e06 + (q_f2.SW * q_f2.SFO1) / 2.0;
        }

        if (sw != 0.0)
        {
            q.SW = sw / q.SFO1;
            q.hzperpt = sw / (double)(q.SI);
            q.offset = (q.SFO1 - p.SF) * 1.0e06 + (q.SW * q.SFO1) / 2.0;
        }

        if (sw_f1 != 0.0)
        {
            q_f1.SW = sw_f1 / q_f1.SFO1;
            q_f1.hzperpt = sw_f1 / (double)(q_f1.SI);
            q_f1.offset = (q_f1.SFO1 - p_f1.SF) * 1.0e06 + (q_f1.SW * q_f1.SFO1) / 2.0;
        }

        if (sw_f2 != 0.0)
        {
            q_f2.SW = sw_f2 / q_f2.SFO1;
            q_f2.hzperpt = sw_f2 / (double)(q_f2.SI);
            q_f2.offset = (q_f2.SFO1 - p_f2.SF) * 1.0e06 + (q_f2.SW * q_f2.SFO1) / 2.0;
        }
     
        return true; // all OK
    }

//-------------------------------------------------------------------------------------------------
/**
 * saves a precessed 3D plane either XZt or XYt
 * @param mode 0 for XZt and 1 for XYt
 * @param plane the particular time domain Y or Z plane number
 * @param gp a SpinWorks "global parameters" object for currrent workspace
 * @param q status parameters in X (F3)
 * @param q_ind status parameters in current indirect (Y or Z) dimension
 * @return -1 if wrong data type, -3 if exception thrown, bytes written if OK
 */
    public static int savePlane3D(SW_main.modes_3D mode, int plane, global_par gp, 
        data_par q, data_par q_ind)
    //           X           Y or Z
    {
        int i, j;
        int bytes_written = 0;
        /*double scale_factor = 1.0;

        if (gp.data_max_2D > 2.0e9)
            scale_factor = gp.data_max_2D / 2.0e9;*/

        if (!(q.data_type == SW_main.data_types.BRUKER
            || q.data_type == SW_main.data_types.VNMR))
            return -1;

        String fileNameRR = "", fileNameRI = "", fileNameIR = "", fileNameII = "";
        
        try
        {
        	if(!Files.exists(Paths.get(q.data_folder + "SW_3D")))	// scratch folder doesn't exist
        		Files.createDirectory(Paths.get(q.data_folder + "SW_3D"));
        }
        catch(Exception ex)
    	{
    		Util.show_and_log("Error creating 3D data folder: " + q.data_folder + "SW_3D",
    				ex.getMessage());
    		ex.printStackTrace();
    		return -3;
    	}

        if (mode == SW_main.modes_3D.XZ)
        {
            fileNameRR = String.format("%sSW_3D/3D_XZt_rr.%d", q.data_folder, plane);
            fileNameRI= String.format("%sSW_3D/3D_XZt_ri.%d", q.data_folder, plane);
            fileNameIR = String.format("%sSW_3D/3D_XZt_ir.%d", q.data_folder, plane);
            fileNameII = String.format("%sSW_3D/3D_XZt_ii.%d", q.data_folder, plane);
        }
        else if (mode == SW_main.modes_3D.XY)
        {
        	fileNameRR = String.format("%sSW_3D/3D_XYt_rr.%d", q.data_folder, plane);
        	fileNameRI = String.format("%sSW_3D/3D_XYt_ri.%d", q.data_folder, plane);
        	fileNameIR = String.format("%sSW_3D/3D_XYt_ir.%d", q.data_folder, plane);
        	fileNameII = String.format("%sSW_3D/3D_XYt_ii.%d", q.data_folder, plane);
        }
        
        FileOutputStream bw;
        
        // the rr (real,real) data
        if (gp.rows_real != null)
        {       	
			try
			{
				bw = new FileOutputStream(fileNameRR);
			}
			catch (Exception ex)
			{
				Util.show_and_log("savePlane3D() cannot open rr: ", ex.getMessage());
				return -3;
			}
			
			byte[] buffer = new byte[4 * q.SI * q_ind.SI];

            try
            {

            	int index = 0;
                for (i = 0; i < q_ind.SI; i++)
                    for (j = 0; j < q.SI; j++)
                    {
                    	// JVM is BigENDIAN while Intel processors are LittleENDIAN
                    	FileUtil.revFloatToByteArray(gp.rows_real[i][j], buffer, index);
						index += 4;
                    }
                bytes_written += index;
				bw.write(buffer);
            }
            catch (Exception ex)
            {
                Util.show_and_log("save_plane_3D(): ", ex.getMessage());
                try { bw.close(); } 
    			catch (IOException e) { e.printStackTrace(); }
                return -3;
            }
            try { bw.close(); } 
			catch (IOException e) { e.printStackTrace(); }
        }

        //	and the ri data
        if (gp.rows_imag != null)
        {
			try
			{
				bw = new FileOutputStream(fileNameRI);
			}
			catch (Exception ex)
			{
				Util.show_and_log("savePlane3D() cannot open ri: ", ex.getMessage());
				return -3;
			}
			
			byte[] buffer = new byte[4 * q.SI * q_ind.SI];

            try
            {

            	int index = 0;
                for (i = 0; i < q_ind.SI; i++)
                    for (j = 0; j < q.SI; j++)
                    {
                    	// JVM is BigENDIAN while Intel processors are LittleENDIAN
                    	FileUtil.revFloatToByteArray(gp.rows_imag[i][j], buffer, index);
						index += 4;
                    }
                bytes_written += index;
				bw.write(buffer);
            }
            catch (Exception ex)
            {
                Util.show_and_log("save_plane_3D(): ", ex.getMessage());
                try { bw.close(); } 
    			catch (IOException e) { e.printStackTrace(); }
                return -3;
            }
            try { bw.close(); } 
			catch (IOException e) { e.printStackTrace(); }
        }

        //	and the ir data
        if (gp.irows_real != null)
        {
			try
			{
				bw = new FileOutputStream(fileNameIR);
			}
			catch (Exception ex)
			{
				Util.show_and_log("savePlane3D() cannot open ir: ", ex.getMessage());
				return -3;
			}
			
			byte[] buffer = new byte[4 * q.SI * q_ind.SI];

            try
            {

            	int index = 0;
                for (i = 0; i < q_ind.SI; i++)
                    for (j = 0; j < q.SI; j++)
                    {
                    	// JVM is BigENDIAN while Intel processors are LittleENDIAN
                    	FileUtil.revFloatToByteArray(gp.irows_real[i][j], buffer, index);
						index += 4;
                    }
                bytes_written += index;
				bw.write(buffer);
            }
            catch (Exception ex)
            {
                Util.show_and_log("save_plane_3D(): ", ex.getMessage());
                try { bw.close(); } 
    			catch (IOException e) { e.printStackTrace(); }
                return -3;
            }
            try { bw.close(); } 
			catch (IOException e) { e.printStackTrace(); }
        }

        //	and the ii data
        if (gp.irows_imag != null)
        {
			try
			{
				bw = new FileOutputStream(fileNameII);
			}
			catch (Exception ex)
			{
				Util.show_and_log("savePlane3D() cannot open ii: ", ex.getMessage());
				return -3;
			}
			
			byte[] buffer = new byte[4 * q.SI * q_ind.SI];

            try
            {

            	int index = 0;
                for (i = 0; i < q_ind.SI; i++)
                    for (j = 0; j < q.SI; j++)
                    {
                    	// JVM is BigENDIAN while Intel processors are LittleENDIAN
                    	FileUtil.revFloatToByteArray(gp.irows_imag[i][j], buffer, index);
						index += 4;
                    }
                bytes_written += index;
				bw.write(buffer);
            }
            catch (Exception ex)
            {
                Util.show_and_log("save_plane_3D(): ", ex.getMessage());
                try { bw.close(); } 
    			catch (IOException e) { e.printStackTrace(); }
                return -3;
            }
            try { bw.close(); } 
			catch (IOException e) { e.printStackTrace(); }
        }

        if (gp.log_on)
            Util.log_error(String.format("save_plane_3D(): bytes written = %d", bytes_written));

        return bytes_written;
    }

//-------------------------------------------------------------------------------------------------
/**
 * loads a processed XYt plane into memory, all 4 quadrants
 * @param plane the Z plane number to load
 * @param gp a SpinWorks "global parameters" object for currrent workspace
 * @param q status parameters in X (F3)
 * @param q_ind status parameters indirect dimension (F2 or Y)
 * @return -1 if wrong data type, -3 if exception thrown, bytes read if OK
 */
     public static int loadPlane_XYt(int plane, global_par gp,
        data_par q, data_par q_ind)
     {

    	 FileInputStream fs;
    	 File file;
    	 byte[] b;
    	 ByteBuffer bb;
         int bytes_read = 0;
         int i, j;

         if (!(q.data_type == SW_main.data_types.BRUKER
            || q.data_type == SW_main.data_types.VNMR))
             return -1;

         String fileNameRR = "", fileNameRI = "", fileNameIR = "", fileNameII = "";

         if(!Files.exists(Paths.get(q.data_folder + "SW_3D")))	// scratch folder doesn't exist
         {
        	 Util.show_and_log(q.data_folder +  "SW_3D doesn't exist");
        	 return -2;
         }
     
         fileNameRR = String.format("%sSW_3D/3D_XYt_rr.%d", q.data_folder, plane);
         fileNameRI = String.format("%sSW_3D/3D_XYt_ri.%d", q.data_folder, plane);
         fileNameIR = String.format("%sSW_3D/3D_XYt_ir.%d", q.data_folder, plane);
         fileNameII = String.format("%sSW_3D/3D_XYt_ii.%d", q.data_folder, plane);
                               
   // rr data
         try 
         { 
        	 file = new File(fileNameRR);
        	 fs = new FileInputStream(file);  	 
         }
    	 catch (IOException ex)
    	 {
    		 Util.show_and_log("load_XY_t(): cannot open: " + fileNameRR, ex.getMessage());
    		 return -1;
    	 }

    	 b = new byte[(int)file.length()];
    	 try { fs.read(b); } 
    	 catch (IOException ex) 
    	 {
    		 ex.printStackTrace();
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -3;
    	 }

    	 bb = ByteBuffer.wrap(b);		
    	 bb.order(ByteOrder.LITTLE_ENDIAN); //SpinWorks processed data always LITTLE ENDIAN
      
    	 try
    	 {
    		 for (i = 0; i < q_ind.SI; i++)
                 for (j = 0; j < q.SI; j++)
                 {
                     gp.rows_real[i][j] = bb.getFloat();
                     bytes_read = bytes_read + 4;
                 }
    	 }
    	 catch (Exception ex)
    	 {
    		 Util.show_and_log(String.format("loadPlane_XYt():\n  {0}:  " + ex.getMessage()));
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -2;  // file opened but read error
    	 }

    	 try { fs.close(); } 
		 catch (IOException e) { e.printStackTrace(); }
                          
// ri data        
    	 try 
         { 
        	 file = new File(fileNameRI);
        	 fs = new FileInputStream(file); 
         }
    	 catch (IOException ex)
    	 {
    		 Util.show_and_log("load_XY_t(): cannot open: " + fileNameRI, ex.getMessage());
    		 return -1;
    	 }

    	 b = new byte[(int)file.length()];
    	 try { fs.read(b); } 
    	 catch (IOException ex) 
    	 {
    		 ex.printStackTrace();
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -3;
    	 }

    	 bb = ByteBuffer.wrap(b);		
    	 bb.order(ByteOrder.LITTLE_ENDIAN); //SpinWorks processed data always LITTLE ENDIAN
      
    	 try
    	 {
    		 for (i = 0; i < q_ind.SI; i++)
                 for (j = 0; j < q.SI; j++)
                 {
                     gp.rows_imag[i][j] = bb.getFloat();
                     bytes_read = bytes_read + 4;
                 }
    	 }
    	 catch (Exception ex)
    	 {
    		 Util.show_and_log(String.format("loadPlane_XYt():\n  {0}:  " + ex.getMessage()));
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -2;  // file opened but read error
    	 }

    	 try { fs.close(); } 
		 catch (IOException e) { e.printStackTrace(); }        
         
  // ir data       
    	 try 
         { 
        	 file = new File(fileNameIR);
        	 fs = new FileInputStream(file); 
         }
    	 catch (IOException ex)
    	 {
    		 Util.show_and_log("load_XY_t(): cannot open: " + fileNameIR, ex.getMessage());
    		 return -1;
    	 }

    	 b = new byte[(int)file.length()];
    	 try { fs.read(b); } 
    	 catch (IOException ex) 
    	 {
    		 ex.printStackTrace();
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -3;
    	 }

    	 bb = ByteBuffer.wrap(b);		
    	 bb.order(ByteOrder.LITTLE_ENDIAN); //SpinWorks processed data always LITTLE ENDIAN
      
    	 try
    	 {
    		 for (i = 0; i < q_ind.SI; i++)
                 for (j = 0; j < q.SI; j++)
                 {
                     gp.irows_real[i][j] = bb.getFloat();
                     bytes_read = bytes_read + 4;
                 }
    	 }
    	 catch (Exception ex)
    	 {
    		 Util.show_and_log(String.format("loadPlane_XYt():\n  {0}:  " + ex.getMessage()));
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -2;  // file opened but read error
    	 }

    	 try { fs.close(); } 
		 catch (IOException e) { e.printStackTrace(); }
       
 // ii data     
    	 try 
         { 
    		 file = new File(fileNameII);
        	 fs = new FileInputStream(file);         	
         }
    	 catch (IOException ex)
    	 {
    		 Util.show_and_log("load_XY_t(): cannot open: " + fileNameII, ex.getMessage());
    		 return -1;
    	 }

    	 b = new byte[(int)file.length()];
    	 try { fs.read(b); } 
    	 catch (IOException ex) 
    	 {
    		 ex.printStackTrace();
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -3;
    	 }

    	 bb = ByteBuffer.wrap(b);		
    	 bb.order(ByteOrder.LITTLE_ENDIAN); //SpinWorks processed data always LITTLE ENDIAN
      
    	 try
    	 {
    		 for (i = 0; i < q_ind.SI; i++)
                 for (j = 0; j < q.SI; j++)
                 {
                     gp.irows_imag[i][j] = bb.getFloat();
                     bytes_read = bytes_read + 4;
                 }
    	 }
    	 catch (Exception ex)
    	 {
    		 Util.show_and_log(String.format("loadPlane_XYt():\n  {0}:  " + ex.getMessage()));
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -2;  // file opened but read error
    	 }

    	 try { fs.close(); } 
		 catch (IOException e) { e.printStackTrace(); }
                        
         if (gp.log_on)
             Util.log_error(String.format("load_plane_XYt(): bytes read = {0}", bytes_read));

         return bytes_read;  
     }

//-------------------------------------------------------------------------------------------------
// do we really need a separate function here?  Mostly duplicates XYt
 /**
  * loads a processed XZt plane into memory, all 4 quadrants
  * @param plane the y plane number to load
  * @param gp a SpinWorks "global parameters" object for currrent workspace
  * @param q status parameters in X (F3)
  * @param q_ind status parameters indirect dimension (F1 or Z)
  * @return -1 if wrong data type, -3 if exception thrown, bytes read if OK
  */
     public static int loadPlane_XZt(int plane, global_par gp,
    		 data_par q, data_par q_ind)
     {

    	 FileInputStream fs;
    	 File file;
    	 byte[] b;
    	 ByteBuffer bb;
    	 int bytes_read = 0;
    	 int i, j;

    	 if (!(q.data_type == SW_main.data_types.BRUKER
    			 || q.data_type == SW_main.data_types.VNMR))
    		 return -1;

    	 String fileNameRR = "", fileNameRI = "", fileNameIR = "", fileNameII = "";

    	 if(!Files.exists(Paths.get(q.data_folder + "SW_3D")))	// scratch folder doesn't exist
    	 {
    		 Util.show_and_log(q.data_folder +  "SW_3D doesn't exist");
    		 return -2;
    	 }

    	 fileNameRR = String.format("%sSW_3D/3D_XZt_rr.%d", q.data_folder, plane);
    	 fileNameRI = String.format("%sSW_3D/3D_XZt_ri.%d", q.data_folder, plane);
    	 fileNameIR = String.format("%sSW_3D/3D_XZt_ir.%d", q.data_folder, plane);
    	 fileNameII = String.format("%sSW_3D/3D_XZt_ii.%d", q.data_folder, plane);

    // rr data
    	 try 
    	 { 
    		 file = new File(fileNameRR);
    		 fs = new FileInputStream(file);  	 
    	 }
    	 catch (IOException ex)
    	 {
    		 Util.show_and_log("load_XZ_t(): cannot open: " + fileNameRR, ex.getMessage());
    		 return -1;
    	 }

    	 b = new byte[(int)file.length()];
    	 try { fs.read(b); } 
    	 catch (IOException ex) 
    	 {
    		 ex.printStackTrace();
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -3;
    	 }

    	 bb = ByteBuffer.wrap(b);		
    	 bb.order(ByteOrder.LITTLE_ENDIAN); //SpinWorks processed data always LITTLE ENDIAN

    	 try
    	 {
    		 for (i = 0; i < q_ind.SI; i++)
    			 for (j = 0; j < q.SI; j++)
    			 {
    				 gp.rows_real[i][j] = bb.getFloat();
    				 bytes_read = bytes_read + 4;
    			 }
    	 }
    	 catch (Exception ex)
    	 {
    		 Util.show_and_log(String.format("loadPlane_XZt():\n  {0}:  " + ex.getMessage()));
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -2;  // file opened but read error
    	 }

    	 try { fs.close(); } 
    	 catch (IOException e) { e.printStackTrace(); }

    // ri data        
    	 try 
    	 { 
    		 file = new File(fileNameRI);
    		 fs = new FileInputStream(file); 
    	 }
    	 catch (IOException ex)
    	 {
    		 Util.show_and_log("load_XZ_t(): cannot open: " + fileNameRI, ex.getMessage());
    		 return -1;
    	 }

    	 b = new byte[(int)file.length()];
    	 try { fs.read(b); } 
    	 catch (IOException ex) 
    	 {
    		 ex.printStackTrace();
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -3;
    	 }

    	 bb = ByteBuffer.wrap(b);		
    	 bb.order(ByteOrder.LITTLE_ENDIAN); //SpinWorks processed data always LITTLE ENDIAN

    	 try
    	 {
    		 for (i = 0; i < q_ind.SI; i++)
    			 for (j = 0; j < q.SI; j++)
    			 {
    				 gp.rows_imag[i][j] = bb.getFloat();
    				 bytes_read = bytes_read + 4;
    			 }
    	 }
    	 catch (Exception ex)
    	 {
    		 Util.show_and_log(String.format("loadPlane_XZt():\n  {0}:  " + ex.getMessage()));
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -2;  // file opened but read error
    	 }

    	 try { fs.close(); } 
    	 catch (IOException e) { e.printStackTrace(); }        

    // ir data       
    	 try 
    	 { 
    		 file = new File(fileNameIR);
    		 fs = new FileInputStream(file); 
    	 }
    	 catch (IOException ex)
    	 {
    		 Util.show_and_log("load_XZ_t(): cannot open: " + fileNameIR, ex.getMessage());
    		 return -1;
    	 }

    	 b = new byte[(int)file.length()];
    	 try { fs.read(b); } 
    	 catch (IOException ex) 
    	 {
    		 ex.printStackTrace();
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -3;
    	 }

    	 bb = ByteBuffer.wrap(b);		
    	 bb.order(ByteOrder.LITTLE_ENDIAN); //SpinWorks processed data always LITTLE ENDIAN

    	 try
    	 {
    		 for (i = 0; i < q_ind.SI; i++)
    			 for (j = 0; j < q.SI; j++)
    			 {
    				 gp.irows_real[i][j] = bb.getFloat();
    				 bytes_read = bytes_read + 4;
    			 }
    	 }
    	 catch (Exception ex)
    	 {
    		 Util.show_and_log(String.format("loadPlane_XZt():\n  {0}:  " + ex.getMessage()));
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -2;  // file opened but read error
    	 }

    	 try { fs.close(); } 
    	 catch (IOException e) { e.printStackTrace(); }

    // ii data     
    	 try 
    	 { 
    		 file = new File(fileNameII);
    		 fs = new FileInputStream(file);         	
    	 }
    	 catch (IOException ex)
    	 {
    		 Util.show_and_log("load_XZ_t(): cannot open: " + fileNameII, ex.getMessage());
    		 return -1;
    	 }

    	 b = new byte[(int)file.length()];
    	 try { fs.read(b); } 
    	 catch (IOException ex) 
    	 {
    		 ex.printStackTrace();
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -3;
    	 }

    	 bb = ByteBuffer.wrap(b);		
    	 bb.order(ByteOrder.LITTLE_ENDIAN); //SpinWorks processed data always LITTLE ENDIAN

    	 try
    	 {
    		 for (i = 0; i < q_ind.SI; i++)
    			 for (j = 0; j < q.SI; j++)
    			 {
    				 gp.irows_imag[i][j] = bb.getFloat();
    				 bytes_read = bytes_read + 4;
    			 }
    	 }
    	 catch (Exception ex)
    	 {
    		 Util.show_and_log(String.format("loadPlane_XZt():\n  {0}:  " + ex.getMessage()));
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    		 return -2;  // file opened but read error
    	 }

    	 try { fs.close(); } 
    	 catch (IOException e) { e.printStackTrace(); }

    	 if (gp.log_on)
    		 Util.log_error(String.format("load_plane_XZt(): bytes read = {0}", bytes_read));

    	 return bytes_read;  
     }
     
//-------------------------------------------------------------------------------------------------   
 /**
  * Save a processed 3D cube to disk, each Z plane stored individually !
  * @param folder folder for the processed data - must already have trailing \ or /
  * @param size_f3 size of data in F3 (X)
  * @param size_f2 size of data in F2 (Y)
  * @param size_f1 size of data in F1 (Z)
  * @return bytes written, -1 for file open error, -2 for write error
  */
     public static int saveCube(String folder, int size_f3, int size_f2, int size_f1)
     //                                            X            Y            Z
     {
         int bytes_written = 0;
         int i = 0, j = 0;
         String fileName = "";
         FileOutputStream bw;

         for (int k = 0; k < size_f1; k++)	// loop over Z
         {
             fileName 
                 = String.format("%sSW_3D/3D_XYZ_rrr.%d", folder, k);
             try
 			{
 				bw = new FileOutputStream(fileName);
 			}
 			catch (Exception ex)
 			{
 				Util.show_and_log("saveCube() cannot open: " + fileName, ex.getMessage());
 				return -3;
 			}
 			
 			byte[] buffer = new byte[4 * size_f3 * size_f2];

 			try
 			{
 				int index = 0;
 				for (j = 0; j < size_f2; j++)	// loop over Y
 				{
 					for (i = 0; i < size_f3; i++)	// loop over Y
 					{
 						// JVM is BigENDIAN while Intel processors are LittleENDIAN
 						FileUtil.revFloatToByteArray(Proc_3D.XYZ[i][j][k], buffer, index);
 						index += 4;					
 					}
 				}
 				bytes_written += index;
				bw.write(buffer);
 			}
 			catch (Exception ex)
			{
				Util.show_and_log(String.format("saveCube():  %d  %d  %d",
						i, j, k), ex.getMessage());
				try { bw.close(); } 
	            catch (IOException e) {	e.printStackTrace();	}
				return -2;
			}
             
            try {	bw.close(); } 
            catch (IOException e) {	e.printStackTrace();	}
         }
         return bytes_written;
     }

 //-------------------------------------------------------------------------------------------------
 /**
  * Read a processed 3D cube from disk, each Z plane read individually
  * @param folder folder for the processed data - must already have trailing \ or /
  * @param size_f3 size of data in F3 (X)
  * @param size_f2 size of data in F2 (Y)
  * @param size_f1 size of data in F1 (Z)
  * @return bytes read, -1 for file open error, -2 for read error
  */
     public static int readCube(String folder, int size_f3, int size_f2, int size_f1)
     //                                            X            Y            Z
     {
    	 int bytes_read = 0;
    	 int i = 0, j = 0;
    	 String fileName = "";
    	 byte[] b;
    	 ByteBuffer bb;
    	 File file;
    	 FileInputStream fs;

    	 for (int k = 0; k < size_f1; k++)
    	 {
    		 fileName 
    		 	= String.format("%sSW_3D/3D_XYZ_rrr.%d", folder, k);
    		 
    		 try 
             { 
            	 file = new File(fileName);
            	 fs = new FileInputStream(file);  	 
             }
        	 catch (IOException ex)
        	 {
        		 Util.show_and_log("load_XY_t(): cannot open: " + fileName, ex.getMessage());
        		 return -1;
        	 }

    		 b = new byte[(int)file.length()];
        	 try { fs.read(b); } 
        	 catch (IOException ex) 
        	 {
        		 ex.printStackTrace();
        		 try { fs.close(); } 
        		 catch (IOException e) { e.printStackTrace(); }
        		 return -3;
        	 }

        	 bb = ByteBuffer.wrap(b);		
        	 bb.order(ByteOrder.LITTLE_ENDIAN); //SpinWorks processed data always LITTLE ENDIAN
        	 
    		 try
    		 {
	    		 for (j = 0; j < size_f2; j++)	// loop over X
	    		 {
	    			 for (i = 0; i < size_f3; i++) 	// loop over Y
	    			 {	                  
	    				 Proc_3D.XYZ[i][j][k] = bb.getFloat();
	    				 bytes_read = bytes_read + 4;  				 
	    			 }
	    		 }
    		 }
    		 catch (Exception ex)
			 {
				 Util.show_and_log(String.format("read_cube_3D():  %d  %d  %d",
						 i, j, k), ex.getMessage());
				 try { fs.close(); } 
	    		 catch (IOException e) { e.printStackTrace(); }
				 return -2;
			 }
   		 
    		 try { fs.close(); } 
    		 catch (IOException e) { e.printStackTrace(); }
    	 }
    	 return bytes_read;
     }
     
//-------------------------------------------------------------------------------------------------
}
