
/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

package spinworks;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.widgets.Display;

import spinworks.SW_main.byte_orders;
import spinworks.SW_main.det_modes_IND;
import spinworks.SW_main.modes;
import spinworks.SW_main.phasing;
import spinworks.SW_main.windows;

/**
 * Utilities for reading Bruker NMR data
 * @author marat
 */
public class FileUtilBruker 
{
	/**
	 * Load parameters from a Bruker XwinNMR/TopSpin parameter file (acqu, acqus, proc procs, etc.)
	 * Add other parameters as needed.  Can read F1, F2 or F3 parameters depending on the file 
	 * name passed and the  particular data_par object.
	 * @param fileName  fully qualified file name
	 * @param p set parameters (a data_par object)
	 * @param q status parameters (a data_par object)
	 * @param gp a SpinWorks global parameter object
	 * @return true if OK, false on error
	 */
    public static boolean get_par(String fileName, data_par p, data_par q, global_par gp)
    {
        String instring;
        String delim = "[ =]+";
        String[] token;

        if(gp.log_on)
            Util.log_error("Opening parameter file:  " + fileName);

        // try to open the parameter file 
        BufferedReader br;
        try { br = new BufferedReader(new FileReader(fileName)); }
        catch (Exception ex)
        {
            Util.show_and_log("get_par(), cannot open: " + fileName, ex.getMessage());
            return false;
        }

//		loop through all input lines looking for parameters.  Each parameter conversion has
//		its own try/catch block in order to catch all possible errors and so that parameter conversion
//		will continue despite a bad input line.
              
        for(;;)
        {
        	 try {instring = br.readLine();} 
             catch (IOException ex) 
             {
     			ex.printStackTrace();
     			Util.show_and_log("get_par(), read error: " + fileName, ex.getMessage());
     			try { br.close(); } 
     			catch (IOException e) { e.printStackTrace(); }
     			return false;
     		}
        	 
        	if(instring == null) break;
        	
            if (instring.length() > 3)
            {
            	token = instring.split(delim);       
                if (token[0].equals("##END")) break;

                try
                {
                    if (token[0].equals("##$SFO1"))				// transmitter frequency
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("SFO1 = %s", 
                                token[1]));
                        q.SFO1 = Double.parseDouble(token[1]);
                        p.SFO1 = q.SFO1;                      
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }               

                try
                {
                    if (token[0].equals("##$SFO2"))			    // decoupler frequency
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("SFO2 = %s", 
                                token[1]));
                        q.SFO2 = Double.parseDouble(token[1]); 
                        p.SFO2 = q.SFO2;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$SFO3"))				// second decoupler freq.
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("SFO3 = %s", 
                                token[1]));
                        q.SFO3 = Double.parseDouble(token[1]); 
                        p.SFO3 = q.SFO3;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$SW"))					// spectral width (PPM!)
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("SW = %s", 
                                token[1]));
                        q.SW = Double.parseDouble(token[1]); 
                        p.SW = q.SW;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$SF"))					// freq. of 0 ppm.
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("SF = %s", 
                                token[1]));
                        q.SF = Double.parseDouble(token[1]); 
                        p.SF = q.SF;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$GB"))					// GM factor
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("GB = %s", 
                                token[1]));
                        p.GB = Double.parseDouble(token[1]); 
                        q.GB = 0.0;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$LB"))					// line broadening (Hz.)
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("LB = %s", 
                                token[1]));
                        p.LB = Double.parseDouble(token[1]);
                        q.LB = 0.0;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$TD"))					// acquired points (r+i)
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("TD = %s", 
                                token[1]));
                        q.TD = Integer.parseInt(token[1]);
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$DECIM"))				// DSP decimation factor
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("DECIM = %s", 
                                token[1]));
                        q.DECIM = Integer.parseInt(token[1]);
                    }
                }
                catch (Exception ex)
                {
                    Util.show_and_log("get_par(): " + instring, ex.getMessage());
                    q.DECIM = 2;		// kludge!  TO DO change to something more rational
                }

                try
                {
                    if (token[0].equals("##$DSPFVS"))				// DSP firmware version
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("DSPFVS = %s", 
                                token[1]));
                        q.DSPFVS = Integer.parseInt(token[1]);
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$GRPDLY"))				// DSP group delay
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("GRPDLY = %s", 
                                token[1]));
                        q.GRPDLY = Double.parseDouble(token[1]); 
                        p.GRPDLY = q.GRPDLY;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$SI"))					// transform size (complex)
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("SI = %s", 
                                token[1]));
                        p.SI = Integer.parseInt(token[1]); q.SI = p.SI;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$BYTORDA"))				// byte order
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("BYTORDA = %s", 
                                token[1]));
                        int temp =  Integer.parseInt(token[1]);
                        //q.BYTORDA = Integer.parseInt(token[1]);
                        if (temp == 0)
                        	q.BYTORDA = byte_orders.LITTLE;
                        else
                        	q.BYTORDA = byte_orders.BIG;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$AQ_mod"))				// acquisition mode
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("AQ_mod = %s", 
                                token[1]));
                        q.AQ_mod = Integer.parseInt(token[1]); p.AQ_mod = q.AQ_mod;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$AQSEQ"))				// acquisition order 3D
                    {                                               // changed from AQORDER
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format(
                                "  file = %s   AQSEQ = %s", fileName, token[1]));
                        q.AQORDER = Integer.parseInt(token[1]); 
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$DIGMOD"))				// filter type
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("DIGMOD = %s", 
                                token[1]));
                        q.DIGMOD = Integer.parseInt(token[1]);
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$NS"))					// number of scans
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("NS = %s",
                                token[1]));
                        q.NS = Integer.parseInt(token[1]);
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$PULPROG"))				// pulse program name
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("PULPROG = %s", 
                                token[1]));
                        q.pulse_sequence = token[1];
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$NUC1"))				// observe nucleus
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("NUC1 = %s", 
                                token[1]));
                        q.NUC1 = token[1];
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$INSTRUM"))				// instrument name
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("INSTRUM = %s", 
                                token[1]));
                        q.INSTRUM = token[1];
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$WDW"))					// window function type
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("WDW = %s",
                                token[1]));
                        int temp = Integer.parseInt(token[1]);
                        //{ NONE, EM, GM, SINE, QSINE, TRAF, TRAFS, GAUSSIAN };
                        
                        if(temp == 0) p.WDW = windows.NONE;
                        else if(temp == 0) p.WDW = windows.NONE;
                        else if(temp == 1) p.WDW = windows.EM;
                        else if(temp == 2) p.WDW = windows.GM;
                        else if(temp == 3) p.WDW = windows.SINE;
                        else if(temp == 4) p.WDW = windows.QSINE;
                        else if(temp == 6) p.WDW = windows.TRAF;
                        else if(temp == 7) p.WDW = windows.TRAFS;
                        else if(temp == 8) p.WDW = windows.GAUSSIAN;
                        
                        q.WDW = SW_main.windows.NONE;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$PH_mod"))				// phasing type
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("PH_mod = %s", token[1]));
                        int temp = Integer.parseInt(token[1]); 
                        //{ NONE, CONSTANTS, MAGNITUDE, POWER, AUTO };
                        
                        if(temp== 0) p.PH_mod  = phasing.NONE;
                        else if(temp== 1) p.PH_mod  = phasing.CONSTANTS;
                        else if(temp== 2) p.PH_mod  = phasing.MAGNITUDE;
                        else if(temp== 3) p.PH_mod  = phasing.POWER;
                        else if(temp== 4) p.PH_mod  = phasing.AUTO;
                                  
                        q.PH_mod = SW_main.phasing.NONE;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$PHC0"))				// zero order phase
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("PHC0 = %s", 
                                token[1]));
                        p.PHC0 = Double.parseDouble(token[1]); 
                        q.PHC0 = 0.0;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$PHC1"))				// first order phase
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("PHC1 = %s", 
                                token[1]));
                        p.PHC1 = Double.parseDouble(token[1]); 
                        q.PHC1 = 0.0;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$SSB"))					// sine bell shift
                    {
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("SSB = %s", 
                                token[1]));
                        p.SSB = Integer.parseInt(token[1]); q.SSB = 0;
                        if (p.SSB >= 1)								// convert XwinNMR to
                            p.ssb = (1.0 / (double)p.SSB) * 180.0;	// SpinWorks definition
                        else p.ssb = 0.0;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$MC2"))					// F1 detection mode
                    {
                    	int temp = Integer.parseInt(token[1]);
                        
                        //{ QF, SEQUENTIAL, TPPI, STATES, STATES_TPPI, ECHO_ANTIECHO }; 
                        
                    	if(temp == 0)p.MC2 = det_modes_IND.QF;
                    	else if(temp == 1)p.MC2 = det_modes_IND.SEQUENTIAL;
                    	else if(temp == 2)p.MC2 = det_modes_IND.TPPI;
                    	else if(temp == 3)p.MC2 = det_modes_IND.STATES;
                    	else if(temp == 4)p.MC2 = det_modes_IND.STATES_TPPI;
                    	else if(temp == 5)p.MC2 = det_modes_IND.ECHO_ANTIECHO;
                    				
                        q.MC2 = SW_main.det_modes_IND.QF;	// ???
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format(" MC2 = %s ", p.MC2));
                        
                        //System.out.println(fileName + String.format(":  MC2 = %s ", p.MC2));
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$FnMODE"))				// F1 detection mode (new!)
                    {
                        q.FnMODE = SW_main.FnMODEs.values()[Integer.parseInt(token[1])]; 
                        p.FnMODE = q.FnMODE;
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("FnMODE = %s ", token[1]));
                        
                        //System.out.println(fileName + String.format(":  FnMODE = %s ", token[1]));
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$DTYPA"))				// data type  0 = int
                    {                                               // 1 = 64 bit doubles
                        q.DTYPA = Integer.parseInt(token[1]);
                        if (gp.log_on)
                            Util.log_error("getpar()  " + String.format("DTYPA = %s ", token[1]));
                    }
                }
                catch (Exception ex)
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }

                try
                {
                    if (token[0].equals("##$REVERSE"))			    // reverse, 1 = TRUE
                    {                                                           
                        if (token[1].equals("yes")) p.REVERSE = true;
                    }
                }
                catch (Exception ex) 
                { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }
            }
        } // end for(;;)

// Horrible kluge held over from Xsim, SpinWorks defines acquisition mode differently 
// than Bruker.
	    if (q.AQ_mod == 2)
	        q.aq_mode = SW_main.det_modes.SEQ;				// sequential
	    else if (q.AQ_mod == 1 || q.AQ_mod == 3)			// simultaneous or DQD
	        q.aq_mode = SW_main.det_modes.SIM;
	
	    if (q.DECIM > 1) q.aq_mode = SW_main.det_modes.DSP;	// simultaneous DSP
	
	    if (gp.log_on) Util.log_error(String.format(
	        "get_par(%s): aq_mode = %d", fileName, q.aq_mode.ordinal()));
	
	    q.dwell = 1.0 / (q.SW * q.SFO1);
	    
	    try {br.close();} 
	    catch (IOException ex) { ex.printStackTrace(); }
	
	    if (gp.log_on) Util.log_error(String.format(
	        "get_par(%s): SW in Hz = %f", fileName, q.SW * q.SFO1));
	
	    return (true);
	}	// end get_par
    
//-------------------------------------------------------------------------------------------------
    /**
     * this function loads a Bruker XwinNMR/TopSpin 32 bit integer fid (if available)
     * @param fileName fully qualified file name
     * @param q  status data parameters
     * @param gp a SpinWorks global parameter object
     * @return bytes read, negative on error
     */
	  public static int get_fid(String fileName, data_par q, global_par gp)
	  {
	      FileInputStream fs;
	      int bytes_read = 0;
	
	      File file = new File(fileName);

	      try { fs = new FileInputStream(file); }
	      catch (IOException ex)
	      {
	          Util.show_and_log("get_fid(): cannot open: " + fileName, ex.getMessage());
	          return -1;
	      }
	
	      try { gp.fid = new double[q.TD]; }
	      catch (Exception ex)
	      {
	    	  try { fs.close(); } 
		      catch (IOException e) { e.printStackTrace(); }
	          Util.show_and_log("get_fid(): cannot allocate space for fid "
	              + fileName, ex.getMessage());
	          return -2;
	      }
	      
	      byte[] b = new byte[(int)file.length()];
	      
	      try { fs.read(b); } 
	      catch (IOException ex) 
	      {
			  ex.printStackTrace();
			  try { fs.close(); } 
		      catch (IOException e) { e.printStackTrace(); }
			  return -3;
		  }
		      
	      ByteBuffer bb = ByteBuffer.wrap(b);
	      
	      // handle byte order issues here!
	      if (q.BYTORDA == byte_orders.LITTLE)
	    	  bb.order(ByteOrder.LITTLE_ENDIAN);
	      
	      try
	      {
	    	  for(int i = 0; i < q.TD; i++)
	    	  {
	    		  gp.fid[i] = (double)bb.getInt();
	    		  bytes_read += 4;
	    	  }
	      }
	      catch(Exception ex) { ex.printStackTrace();}
	
	      try { fs.close(); } 
	      catch (IOException ex) { ex.printStackTrace(); }
	      
	      return bytes_read;
	  }

	  
//-------------------------------------------------------------------------------------------------
	  /**
	   * this function loads a Bruker XwinNMR/TopSpin 64 bit floating point fid (if available)
	   * @param fileName fully qualified file name
	   * @param q  status data parameters
	   * @param gp a SpinWorks global parameter object
	   * @return bytes read, negative on error
	   */
	  public static int get_double_fid(String fileName, data_par q, global_par gp)
	  {
		  FileInputStream fs;
		  int bytes_read = 0;

		  File file = new File(fileName);

		  try { fs = new FileInputStream(file); }
		  catch (IOException ex)
		  {
			  Util.show_and_log("get_fid(): cannot open: " + fileName, ex.getMessage());
			  return -1;
		  }

		  try { gp.fid = new double[q.TD]; }
		  catch (Exception ex)
		  {
			  try { fs.close(); } 
			  catch (IOException e) { e.printStackTrace(); }
			  Util.show_and_log("get_fid(): cannot allocate space for fid "
					  + fileName, ex.getMessage());
			  return -2;
		  }

		  byte[] b = new byte[(int)file.length()];

		  try { fs.read(b); } 
		  catch (IOException ex) 
		  {
			  ex.printStackTrace();
			  try { fs.close(); } 
			  catch (IOException e) { e.printStackTrace(); }
			  return -3;
		  }

		  ByteBuffer bb = ByteBuffer.wrap(b);

		  // handle byte order issues here!
		  // does this work for doubles??? seems to 
		  if (q.BYTORDA == byte_orders.LITTLE)
			  bb.order(ByteOrder.LITTLE_ENDIAN);

		  try
		  {
			  for(int i = 0; i < q.TD; i++)
			  {
				  gp.fid[i] = bb.getDouble();
				  bytes_read += 8;
			  }
		  }
		  catch(Exception ex) { ex.printStackTrace();}

		  try { fs.close(); } 
		  catch (IOException ex) { ex.printStackTrace(); }

		  return bytes_read;
	  }
		
	  //-------------------------------------------------------------------------------------------------
	  /**
	   * this function loads a Bruker XwinNMR/TopSpin 32 bit floating point fid (if available)  UNTESTED!
	   * @param fileName fully qualified file name
	   * @param q  status data parameters
	   * @param gp a SpinWorks global parameter object
	   * @return bytes read, negative on error
	   */
	  public static int get_float_fid(String fileName, data_par q, global_par gp)
	  {
		  FileInputStream fs;
		  int bytes_read = 0;

		  File file = new File(fileName);

		  try { fs = new FileInputStream(file); }
		  catch (IOException ex)
		  {
			  Util.show_and_log("get_fid(): cannot open: " + fileName, ex.getMessage());
			  return -1;
		  }

		  try { gp.fid = new double[q.TD]; }
		  catch (Exception ex)
		  {
			  try { fs.close(); } 
			  catch (IOException e) { e.printStackTrace(); }
			  Util.show_and_log("get_fid(): cannot allocate space for fid "
					  + fileName, ex.getMessage());
			  return -2;
		  }

		  byte[] b = new byte[(int)file.length()];

		  try { fs.read(b); } 
		  catch (IOException ex) 
		  {
			  ex.printStackTrace();
			  try { fs.close(); } 
			  catch (IOException e) { e.printStackTrace(); }
			  return -3;
		  }

		  ByteBuffer bb = ByteBuffer.wrap(b);

		  // handle byte order issues here!
		  if (q.BYTORDA == byte_orders.LITTLE)
			  bb.order(ByteOrder.LITTLE_ENDIAN);

		  try
		  {
			  for(int i = 0; i < q.TD; i++)
			  {
				  gp.fid[i] = bb.getFloat();
				  bytes_read += 4;
			  }
		  }
		  catch(Exception ex) { ex.printStackTrace();}

		  try { fs.close(); } 
		  catch (IOException ex) { ex.printStackTrace(); }

		  return bytes_read;
	  }
		  
//-------------------------------------------------------------------------------------------------
	  /**
	   * this function loads a Bruker BRUKER/XwinNMR ser file (if available) 
	   * into the ser array
	   * @param fileName fully qualified path to the file
	   * @param gp a SpinWorks global parameter object
	   * @param q status parameters, acquisition dimension
	   * @param q_f1 status parameters, indirect dimension
	   * @return bytes read, negative if there is an error
	   */
	  public static int get_ser(String fileName, global_par gp, data_par q, data_par q_f1)
	  {
		  int block_size;
		  int i, j;
		  boolean error = false;         // to break nested loop

		  FileInputStream fs;
	      int bytes_read = 0;
	
	      File file = new File(fileName);

	      try { fs = new FileInputStream(file); }
	      catch (IOException ex)
	      {
	          Util.show_and_log("get_ser(): cannot open: " + fileName, ex.getMessage());
	          return -1;
	      }
	
	      try { gp.fid = new double[q.TD]; }
	      catch (Exception ex)
	      {
	    	  try { fs.close(); } 
		      catch (IOException e) { e.printStackTrace(); }
	          Util.show_and_log("get_ser(): cannot allocate space for fid "
	              + fileName, ex.getMessage());
	          return -2;
	      }
	      
	      byte[] b = new byte[(int)file.length()];
	      
	      try { fs.read(b); } 
	      catch (IOException ex) 
	      {
			  ex.printStackTrace();
			  try { fs.close(); } 
		      catch (IOException e) { e.printStackTrace(); }
			  return -3;
		  }
		      
	      ByteBuffer bb = ByteBuffer.wrap(b);
	      
	      // handle byte order issues here!
	      if (q.BYTORDA == byte_orders.LITTLE)
	    	  bb.order(ByteOrder.LITTLE_ENDIAN);

		  try { gp.ser = new double[q_f1.TD][q.TD]; }
		  catch (Exception ex)
		  {
			  try { fs.close(); } 
		      catch (IOException ioe) {ioe.printStackTrace(); }
			  
			  Util.show_and_log("get_ser(): allocation error " + fileName, ex.getMessage());
			  return -1;
		  }

		  // is this block size stuff being handled correctly?  TODO
		  bytes_read = 0;									// re-set bytes read;
          double n_k = (q.TD * 4)/ 1024.0;
		  block_size = (int) Math.round(n_k + 0.49999) * 1024;
		  
		  int[] buffer = new int[block_size / 4];

		  for (i = 0; i < q_f1.TD; i++)						// now read the data 
		  {
			  for (j = 0; j < block_size / 4; j++)
			  {
				  try
				  { 
					  buffer[j] = bb.getInt();
					  bytes_read += 4;
				  }
				  catch (Exception ex)
				  {
					  error = true;
					  Util.show_and_log("get_ser(): size error " + fileName, ex.getMessage());
					  break;
				  }
			  }

			  try
			  {
				  for (j = 0; j < q.TD; j++)
					  gp.ser[i][j] = (double)buffer[j];
			  }
			  catch (Exception ex)
			  {
				  error = true;
				  Util.show_and_log("get_ser(): file corrupt?? " + fileName, ex.getMessage());
				  break;
			  }

			  if (error) break;
		  }

		  try { fs.close(); } 
	      catch (IOException ex) { ex.printStackTrace(); }
		  
		  return bytes_read;
	  }

 //-------------------------------------------------------------------------------------------------
	  /**
	   * gets only the spectrum reference frequency (SF) from the procs or proc2s file
	   * @param fileName
	   * @param p set parameters of the dimension
	   * @param q status parameters of the dimension
	   * @param gp a SpinWorks global parameter object
	   * @return true if OK, false on error
	   */
	  public static boolean get_SF(String fileName, data_par p, data_par q, global_par gp)
	  {
		  String instring;
		  String delim = "[ =]+";
		  String[] token;

		  // try to open the parameter file 
		  BufferedReader br;
		  try { br = new BufferedReader(new FileReader(fileName)); }
		  catch (Exception ex)
		  {
			  Util.show_and_log("get_par(), cannot open: " + fileName, ex.getMessage());
			  return false;
		  }

		  //		loop through all input lines looking for parameters.  
		  for(;;)
		  {
			  try {instring = br.readLine();} 
			  catch (IOException ex) 
			  {
				  ex.printStackTrace();
				  Util.show_and_log("get_par(), read error: " + fileName, ex.getMessage());
				  try { br.close(); } 
				  catch (IOException e) { e.printStackTrace(); }
				  return false;
			  }

			  if(instring == null) break;

			  if (instring.length() > 3)
			  {
				  token = instring.split(delim);       
				  if (token[0].equals("##END")) break;

				  try
				  {
					  if (token[0].equals("##$SF"))					// freq. of 0 ppm.
					  {
						  if (gp.log_on)
							  Util.log_error("getpar()  " + String.format("SF = %s", 
									  token[1]));
						  q.SF = Double.parseDouble(token[1]); 
						  p.SF = q.SF;
					  }
				  }
				  catch (Exception ex) 
				  { Util.show_and_log("get_par(): " + instring, ex.getMessage()); }
			  }
		  } // end for

		  try { br.close(); } 
		  catch (IOException ex) { ex.printStackTrace(); }
		  return (true);
	  }
	  
//-------------------------------------------------------------------------------------------------
	  /**
	   * Get the experiment number from a Bruker path.
	   * @param path fully qualified path ti the folder containing the data
	   * @returnreturns the experiment number if successful, 0 if unable to 
	   * give an expt. number and -1 if an exception was thrown
	   */
	  public static int get_ex_no(String path)
	  {
		  int i = 1, j;
		  int n = 0;

		  try
		  {
			  String temp = path.substring(0, path.length() - 1);

			  for (j = temp.length() - 1; j > 0; j--)
			  {
				  if (temp.charAt(j) == '/' || temp.charAt(j) == '\\') break;
				  n = n + Integer.parseInt(String.format("%c", temp.charAt(j))) * i;
				  i = i * 10;
			  }
		  }
		  catch (Exception ex)
		  {
			  Util.show_and_log("get_ex_no():  " + path, ex.getMessage());
			  return -1;
		  }

		  //System.out.println(String.format(" expno = %d", n));
		  return n;
	  }
	        
	        
//-------------------------------------------------------------------------------------------------
	  /**
	   * Utility for Bruker data only, to load in data with a new experiment number
	   * @param expno the new experiment number
	   * @return true if OK, false on error
	   */
	  public static boolean load_new_exno(int expno)
	  {
		  SW_main.shell.setCursor( new Cursor(Display.getDefault(), SWT.CURSOR_WAIT));	
		  boolean switch_ok = false;
		  int j = 0, k = 0;
		  String new_path = ""; ;
		  String new_folder = "";
		  String temp = "";            
	  
		  temp = SW_main.q.data_folder.substring(0, SW_main.q.data_folder.length() - 1);

		  /*//keep track of current limits and scale 1D
		  double temp_freq1 = SW_main.gp.freq1;
		  double temp_freq2 = SW_main.gp.freq2;
		  double temp_expt_scale = SW_main.gp.expt_scale;*/

		  for (j = temp.length() - 1; j > 0; j--)
		  {
			  if (temp.charAt(j) == '/' || temp.charAt(j) == '\\') break;
		  }
		  k = temp.length() - j - 1;			// length of expno # field

		  temp = temp.substring(0, temp.length() - k);
		  //System.out.println("temp (2) = " + temp);

		  new_folder = String.format("%s%d/", temp, expno);
		  new_path =   String.format("%s%d/fid", temp, expno);	// try 1D first
		  
		  //  check to see if the new folder exists.  If not, bail out.
		  File di = new File(new_folder);
		  
		  if (!di.exists() || !di.isDirectory())
		  {
			  SW_main.shell.setCursor( new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));	
			  return false;
		  }

		  try
		  {
			  File fi = new File(new_path);

			  if (fi.exists())					    	// fid file exists, so is 1D
			  {
				  switch_ok = true;
				  SW_main.q.data_path = new_path;		// update data path
				  SW_main.q.data_folder = new_folder;	// and data folder
				  if (SW_main.q.data_mode == SW_main.modes.MODE_2D)
					  InterfaceUtil.set1D();
			  }

			  if (switch_ok == false)					// couldn't find fid file, so let's try for 2D
			  {
				  new_path = String.format("%s%d/ser", temp, expno);
				  fi = new File(new_path);

				  if (fi.exists())
				  {
					  switch_ok = true;
					  SW_main.q.data_path = new_path;		// update data path
					  SW_main.q.data_folder = new_folder;	// and data folder
					  if (SW_main.q.data_mode == SW_main.modes.MODE_1D)
						  InterfaceUtil.set2D();
				  }
			  }
		  }	// try
		  catch (Exception ex)
		  {
			  Util.show_and_log(String.format(
					  "Trying to switch to data set %s\n caused the following exception: ",
					  SW_main.q.data_path), ex.getMessage());
		  }

		  if (!switch_ok) return false;				// couldn't find a valid data set, so bail out

		  SW_main.gp.new_data();              		// reset some of the global parameter block

		  if (SW_main.q.data_mode == SW_main.modes.MODE_1D)
			  FileOpen.fileOpenBruker1D(new_path);
		  else if (SW_main.q.data_mode == SW_main.modes.MODE_2D)
			  FileOpen.fileOpenBruker2D(new_path);

		  SW_main.p.fcor = 0.5;                   	// reset default fp multiplier, F2 only
		  SW_main.gp.v_rescale = true;				// new data set, so rescale on next transform

		  if (SW_main.q.data_mode == modes.MODE_2D)
			  SW_main.gp.raw_max = Util_2D.max_ser(SW_main.gp, SW_main.q, SW_main.q_f1); // calculate scaling factor
		  else
			  SW_main.gp.raw_max = Util.dscan(SW_main.gp.fid, SW_main.q.TD, 4, SW_main.q.TD - 1);

		  //update_toolbar();					             // update menu and tool bar

		  SW_main.file_obj.add_file(SW_main.q.data_path,			 // add this file to recent files list
				  SW_main.q.pulse_sequence, SW_main.gp.plot_title);

		  if (SW_main.gp.display_mode == modes.MODE_FID)     // set default vertical scaling
		  {
			  SW_main.gp.fid_scale = (double)SW_main.gp.dysize * 0.80 / SW_main.gp.raw_max;
			  SW_main.gp.t1 = 0.0;
			  SW_main.gp.t2 = (double)(SW_main.q.TD / 2) * SW_main.q.dwell;
		  }
		  else if (SW_main.gp.display_mode == modes.MODE_1D)  // set default vertical scaling
		  {                                                   // and limits to 
			  /*SW_main.gp.expt_scale = temp_expt_scale;        // first spectrum if 1D
			  SW_main.gp.freq1 = temp_freq1;
			  SW_main.gp.freq2 = temp_freq2;*/
		  }
		  else                                                // 2D reset to default
			  SW_main.gp.expt_scale = (double)SW_main.gp.dysize * 0.80 / SW_main.gp.data_max;
		  

		  SW_main.shell.setCursor( new Cursor(Display.getDefault(), SWT.CURSOR_ARROW));				// re-set cursor
		  SW_main.statusStripLabel1.setText("Ready");		  // and update embelishments
		  SW_main.statusStripLabel2.setText(
				  String.format("%s  %d scans", SW_main.q.pulse_sequence, SW_main.q.NS));

		  SW_main.shell.setText(SW_main.version + "  " + SW_main.q.data_folder);

		  if (SW_main.gp.log_on)
			  Util.log_error(String.format("Bruker experiment #%d loaded", expno));

		  DisplayButtons.fullButtonClick();	// also redraws canvas
		  
		  return true;
	  }

//-------------------------------------------------------------------------------------------------
}	// end class
