/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

/**
 * 
 */
package spinworks;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import spinworks.SW_main.byte_orders;

/**
 * @author marat
 *
 */
public class FileUtilBruker3D 
{
/**
 * this function loads a Bruker BRUKER/XwinNMR ser file (if available) 
 * into the ser array, 3D version, 312 acquisition order
 * @param plane The Y plane to load
 * @param gp global parameter object
 * @param q  status parameters in acquisition dimension
 * @param q_f1 
 * @param q_f2
 * @return
 */
	public static int getSerXZ(int plane, global_par gp,
			data_par q, data_par q_f1, data_par q_f2)
//					 X              Z              Y
	{
		int ser_size = gp.ser_buffer.length;            // total length of ser file in bytes
		int block_size = ser_size / (q_f1.TD * q_f2.TD);// length of each block in bytes
		int offset = 0;                                 // byte offset to start of current block
		int bytes_read = 0;
		int i, j;
		boolean error = false;                          // to break nested loop

		if (gp.log_on) Util.log_error(String.format(
				"get_ser_XZ(): ser_buffer length = %d   block_size = %d", ser_size, block_size));

		int[] buffer = new int[block_size / 4];

		offset = block_size * q_f1.TD * plane;          // offset to current plane in bytes         
		
		ByteBuffer bb = ByteBuffer.wrap(gp.ser_buffer);

		// handle byte order issues here!
		if (q.BYTORDA == byte_orders.LITTLE)
			bb.order(ByteOrder.LITTLE_ENDIAN);

		for (i = 0; i < q_f1.TD; i++)					// now read the data 
		{
			for (j = 0; j < block_size / 4; j++)
			{
				try
				{
					buffer[j] = bb.getInt(offset);
					offset = offset + 4;
					bytes_read = bytes_read + 4;
				}
				catch (Exception ex)
				{                    
					error = true;
					Util.show_and_log("get_ser_XZ(): size error ",  ex.getMessage());
					break;
				}
			}

			try
			{			
				for (j = 0; j < q.TD; j++)
					  gp.ser[i][j] = (double)buffer[j];
			}
			catch (Exception ex)
			{
				error = true;
				Util.show_and_log("get_ser_XZ(): file corrupt?? ",  ex.getMessage());
				break;
			}

			if (error) break;
		}

		if(gp.log_on)Util.log_error(String.format("bytes read = %d", bytes_read));

		return bytes_read;
	}
	
//-------------------------------------------------------------------------------------------------
/**
 * this function loads a Bruker BRUKER/XwinNMR ser file (if available) 
 * into the ser array, 3D version, 321 acquisition order
 * @param plane
 * @param gp
 * @param q
 * @param q_f1
 * @param q_f2
 * @return
 */
	public static int getSerXZ_321(int plane, global_par gp,
            data_par q, data_par q_f1, data_par q_f2)
//                   X              Z              Y
 {
     int ser_size = gp.ser_buffer.length;             // total length of ser file in bytes
     int block_size = ser_size / (q_f1.TD * q_f2.TD); // length of each block in bytes
     int offset = 0;                                  // byte offset to start of current block
     int bytes_read = 0;
     int i, j;
     boolean error = false;                           // to break nested loop

     if (gp.log_on) Util.log_error(String.format
         ("get_ser_XZ_321(): ser_buffer length = %d   block_size = %d", ser_size, block_size));

     int[] buffer = new int[block_size / 4];
     ByteBuffer bb = ByteBuffer.wrap(gp.ser_buffer);

     // handle byte order issues here!
     if (q.BYTORDA == byte_orders.LITTLE)
    	 bb.order(ByteOrder.LITTLE_ENDIAN);

     // now read the data
     for (i = 0; i < q_f1.TD; i++)	                    // loop over all Z rows				 
     {
         offset = plane * block_size + i * (q_f2.TD * block_size);
         for (j = 0; j < block_size / 4; j++)           // loop over all X columns
         {
        	 try
        	 {
        		 buffer[j] = bb.getInt(offset);
        		 offset = offset + 4;
        		 bytes_read = bytes_read + 4;
             }
             catch (Exception ex)
             {
                 error = true;
                 Util.show_and_log("get_ser_XZ_321: size error ", ex.getMessage());
                 break;
             }
         }   // for j

         try
         {
             for (j = 0; j < q.TD; j++)
                 gp.ser[i][j] = (double)buffer[j];
         }
         catch (Exception ex)
         {
             error = true;
             Util.show_and_log("get_ser_XZ_312(): file corrupt?? ", ex.getMessage());
             break;
         }

         if (error) break;
     }   // for i

     if (gp.log_on)	//TODO  remove after testing!!
     {
         for (i = 0; i < 4; i++)  
              Util.log_error(String.format("  plane = %d   i (row) = %d     data = %f   ,   %f",
                   plane, i, gp.ser[i][0], gp.ser[i][1]));
     }

     return bytes_read;
 }
	
//-------------------------------------------------------------------------------------------------
/**
 * 
 * @param plane
 * @param gp
 * @param q
 * @param q_f1
 * @param q_f2
 * @return
 */
    public static int getSerXY(int plane, global_par gp,
               data_par q, data_par q_f1, data_par q_f2)
//                      X              Z              Y
    {
        int ser_size = gp.ser_buffer.length;          	// total length of ser file in bytes
        int block_size = ser_size / (q_f1.TD * q_f2.TD);// length of each block in bytes;
        int offset = 0;                                	// byte offset to start of current block
        int bytes_read = 0;
        int i, j;
        boolean error = false;                         	// to break nested loop

        if (gp.log_on) Util.log_error(String.format
            ("get_ser_XY(): ser_buffer length = %d   block_size = %d", ser_size, block_size));

        int[] buffer = new int[block_size / 4];
        ByteBuffer bb = ByteBuffer.wrap(gp.ser_buffer);

        // handle byte order issues here!
        if (q.BYTORDA == byte_orders.LITTLE)
       	 bb.order(ByteOrder.LITTLE_ENDIAN);
        
        // now read the data
        for (i = 0; i < q_f2.TD; i++)	             	// loop over all Y rows				 
        {
            offset = plane * block_size + i * (q_f1.TD * block_size);
            for (j = 0; j < block_size / 4; j++)     	// loop over all X columns
            {
            	try
            	{
            		buffer[j] = bb.getInt(offset);
            		offset = offset + 4;
            		bytes_read = bytes_read + 4;
            	}
                catch (Exception ex)
                {
                    error = true;
                    Util.show_and_log("get_ser_XY(): size error ", ex.getMessage());
                    break;
                }
            }   // for j

            try
            {
                for (j = 0; j < q.TD; j++)
                    gp.ser[i][j] = (double)buffer[j];
            }
            catch (Exception ex)
            {
                error = true;
                Util.show_and_log("get_ser_XY_312(): file corrupt?? " + ex.getMessage());
                break;
            }
            if (error) break;
        }   // for i

        if (gp.log_on) Util.log_error(String.format("get_ser_XY (312) bytes read = %d", bytes_read));
       
        return bytes_read;
    }

//-------------------------------------------------------------------------------------------------
/**
 * 
 * @param plane
 * @param gp
 * @param q
 * @param q_f1
 * @param q_f2
 * @return
 */
    public static int getSerXY_321(int plane, global_par gp,
               data_par q, data_par q_f1, data_par q_f2)
//                      X              Z              Y
    {
        int ser_size = gp.ser_buffer.length;               // total length of ser file in bytes
        int block_size = ser_size / (q_f1.TD * q_f2.TD);   // length of each F3 (X) block in bytes
        int offset = plane * block_size * q_f2.TD;         // byte offset to start of current block
        int bytes_read = 0;
        int i, j, k;
        boolean error = false;         					   // to break nested loop

        if (gp.log_on) Util.log_error(String.format
            ("get_ser_XY_321(): ser_buffer length = %d   block_size = %d", ser_size, block_size));

        int[] buffer = new int[block_size / 4];
        ByteBuffer bb = ByteBuffer.wrap(gp.ser_buffer);

        // handle byte order issues here!
        if (q.BYTORDA == byte_orders.LITTLE)
        	bb.order(ByteOrder.LITTLE_ENDIAN);

        if(gp.log_on)
            Util.log_error(String.format("get_XY_321(): plane = {0}  offset = {1}",
                plane, offset));

        // now read the data
        for (i = 0; i < q_f2.TD; i++)	                    // loop over all Y rows				 
        {             
            for (j = 0; j < block_size / 4; j++)            // loop over all X columns
            {
                try
                {               	
                	buffer[j] = bb.getInt(offset);
            		offset = offset + 4;
            		bytes_read = bytes_read + 4;
                }
                catch (Exception ex)
                {
                    error = true;
                    Util.show_and_log("get_ser_XY_321(): size error ",  ex.getMessage());
                    break;
                }
            }   // for j

            try
            {
                for (k = 0; k < q.TD; k++)
                    gp.ser[i][k] = (double)buffer[k];
            }
            catch (Exception ex)
            {
                error = true;
                Util.show_and_log("get_ser_XY_321(): file corrupt?? ", ex.getMessage());
                break;
            }
            if (error) break;
        }   // for i

        if (gp.log_on) Util.log_error(String.format("get_ser_XY (321) bytes read = %d}", bytes_read));

        if (gp.log_on)	//TODO remove after testing!!
        {
            for (i = 0; i < 4; i++)
                for (j = 0; j < 4; j = j + 2)
                    Util.log_error(String.format("  plane = %d   i = %d  j = %f   data = {3}   ,   %f",
                        plane, i, j, gp.ser[i][j], gp.ser[i][j + 1]));
        }

        return bytes_read;
    }
    
//-------------------------------------------------------------------------------------------------
/**
 * Loads a 3D ser file into a buffer 
 * @param fileName full path to the file
 * @param serBuffer A buffer for the ser file
 * @return bytes read if successful, -1 if file can't be opened, -3 on read error
 */
	 public static int getSerBuffer(String fileName, global_par gp)  
     {       
		 FileInputStream fs;

		 File file = new File(fileName);

		 try { fs = new FileInputStream(file); }
		 catch (IOException ex)
		 {
			 Util.show_and_log("get_ser(): cannot open: " + fileName, ex.getMessage());
			 return -1;
		 }

		 gp.ser_buffer = new byte[(int)file.length()];

		 try { fs.read(gp.ser_buffer); } 
		 catch (IOException ex) 
		 {
			 ex.printStackTrace();
			 try { fs.close(); } 
			 catch (IOException e) { e.printStackTrace(); }
			 return -3;
		 }

		 try {	fs.close();	} 
		 catch (IOException e) {	e.printStackTrace();	}

         return gp.ser_buffer.length;
     }
//-------------------------------------------------------------------------------------------------	
}
