
/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

package spinworks;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;

/**
 * 3D File handling for Varian/Agilent data
 * @author Kirk Marat, 
 * University of Manitoba
 */
public class FileUtilVarian3D 
{
//-------------------------------------------------------------------------------------------------
/**
 * read Varian 3D fid file, XZ (F3-F1) planes  for phase2, phase1 ordering
 * @param plane
 * @param gp
 * @param q
 * @param q_f1
 * @param q_f2
 * @return bytes read, -1 on error
 */
	public static int get_varian_ser_3D_XZ_order_2(int plane, global_par gp,
			data_par q, data_par q_f1, data_par q_f2)
	{
		DataFileHeader fh = new DataFileHeader();
		DataBlockHeader bh = new DataBlockHeader();

		byte[] fh_byte = new byte[32];             // file header is 32 bytes
		byte[] bh_byte = new byte[28];             // block header is 28 bytes
		int bytes_read = 0;
		int offset = 0;                            // offset to current block in bytes
		int block_size = q.TD * 4 + 28;            // size of data block in bytes
		boolean error = false;                     // for breaking nested loops

		// read file header into byte array
		try
		{
			for (int i = 0; i < 32; i++)
				fh_byte[i] = gp.ser_buffer[i];
		}
		catch (Exception ex)
		{
			Util.show_and_log("get_varian_ser_3D_XZ 2():  " , ex.getMessage());
			return -1;
		}
		bytes_read += 32;	

		ByteBuffer bb_fh = ByteBuffer.wrap(fh_byte);
		fh.nblocks = bb_fh.getInt(0);
		fh.np = bb_fh.getInt(8);
		fh.status = bb_fh.getShort(26);
		fh.nbheaders = bb_fh.getInt(28);
		q.nblocks = fh.nblocks;

		// calculate offset for bizzare Varian file ordering
		offset = 32 + (2 * (plane / 2) * block_size * q_f1.TD) 
				+ (2 * (plane % 2) * block_size);

		//  loop through all increments  (Z dimension)
		for (int i = 0; i < q_f1.TD; i = i+2)
		{
		//  read block header, and get status field
			for (int k = 0; k < 28; k++)
				bh_byte[k] = gp.ser_buffer[k + 32 + (i) * (4 * fh.np + 28)];
			ByteBuffer bb_bh = ByteBuffer.wrap(bh_byte);
			
			bh.status = bb_bh.getShort(2);				
			bh.index = bb_bh.getShort(4);

			offset = offset + 28;                       // have read the block header
			bytes_read = bytes_read + 28;

			//  now read the data, even rows

			byte[] b = new byte[4];
			for (int j = 0; j < q.TD; j++)
			{
				try
				{
					for (int ii = 0; ii < 4; ii++)
						b[ii] = gp.ser_buffer[ii + offset];
					
					ByteBuffer bb_data = ByteBuffer.wrap(b);
										
					if ((bh.status & 0x4) != 0)			//32 bit int
						gp.ser[i][j] = (double)bb_data.getInt();
					
					if ((bh.status & 0x8) != 0)			//32 bit float
						gp.ser[i][j] = (double)(8192.0 * bb_data.getFloat());
					
					bytes_read = bytes_read + 4;
					offset = offset + 4;
				}
				catch (Exception ex)
				{
					Util.show_and_log("get_varian_ser_XZ 2():  " , ex.getMessage());
					error = true;
					break;
				}               
			}  // for j  (X dimension)

			//  read block header, and get status field
			for (int k = 0; k < 28; k++)
				bh_byte[k] = gp.ser_buffer[k + offset];	
			ByteBuffer bb_bh2 = ByteBuffer.wrap(bh_byte);
			
			bh.status = bb_bh2.getShort(2);	
			bh.index = bb_bh.getShort(4);

			offset = offset + 28;                       // have read the block header
			bytes_read = bytes_read + 28;

			// read the data, odd rows
			for (int j = 0; j < q.TD; j++)
			{
				try
				{
					for (int ii = 0; ii < 4; ii++)
						b[ii] = gp.ser_buffer[ii + offset];
					ByteBuffer bb_data = ByteBuffer.wrap(b);

					if ((bh.status & 0x4) != 0)			//32 bit int
					gp.ser[i + 1][j] = (double)bb_data.getInt();
					if ((bh.status & 0x8) != 0)			//32 bit float
						gp.ser[i + 1][j] = (double)(8192.0 * bb_data.getFloat());
					
					bytes_read = bytes_read + 4;
					offset = offset + 4;
				}
				catch (Exception ex)
				{
					Util.show_and_log("get_varian_ser_XZ 2():  ", ex.getMessage());
					error = true;
					break;
				}
			}  // for j  (X dimension)

			if (error) break;   // break outer loop    
			offset = offset + 2 * block_size;   // skip two blocks!

		}   // for i (Z dimension)

		//  clean-up    

		if (gp.log_on)

			Util.log_error(String.format("get_varian_ser_3D_XZ 2(): bytes read = %d",
					bytes_read));

		return (bytes_read);
	}

//-------------------------------------------------------------------------------------------------
/**
 * read Varian 3D fid file, XZ (F3-F1) planes  for phase, phase2 ordering
 * @param plane
 * @param gp
 * @param q
 * @param q_f1
 * @param q_f2
 * @return bytes read, -1 on error
 */
	public static int get_varian_ser_3D_XZ_order_1(int plane, global_par gp,
			data_par q, data_par q_f1, data_par q_f2)
	{
		DataFileHeader fh = new DataFileHeader();
		DataBlockHeader bh = new DataBlockHeader();

		byte[] fh_byte = new byte[32];             // file header is 32 bytes
		byte[] bh_byte = new byte[28];             // block header is 28 bytes
		int bytes_read = 0;
		int offset = 0;                            // offset to current block in bytes
		int block_size = q.TD * 4 + 28;            // size of data block in bytes
		boolean error = false;                     // for breaking nested loops

		// read file header into byte array
		try
		{
			for (int i = 0; i < 32; i++)
				fh_byte[i] = gp.ser_buffer[i];
		}
		catch (Exception ex)
		{
			Util.show_and_log("get_varian_ser_3D_XZ 1():  " , ex.getMessage());
			return -1;
		}
		bytes_read = bytes_read + 32;

		ByteBuffer bb_fh = ByteBuffer.wrap(fh_byte);
		fh.nblocks = bb_fh.getInt(0);
		fh.np = bb_fh.getInt(8);
		fh.status = bb_fh.getShort(26);
		fh.nbheaders = bb_fh.getInt(28);
		q.nblocks = fh.nblocks;

		// calculate offset for bizzare Varian file ordering

		offset = 32 + (2 * (plane / 2) * block_size * q_f1.TD)
				+ (plane % 2 * block_size);

		//  loop through all increments  (Z dimension)
		for (int i = 0; i < q_f1.TD; i++)
		{			
		//  read block header, and get status field
			for (int k = 0; k < 28; k++)
				bh_byte[k] = gp.ser_buffer[k + offset];

			ByteBuffer bb_bh = ByteBuffer.wrap(bh_byte);
			bh.status = bb_bh.getShort(2);	
			bh.index = bb_bh.getShort(4);

			offset = offset + 28;                       // have read the block header
			bytes_read = bytes_read + 28;

			byte[] b = new byte[4];
			for (int j = 0; j < q.TD; j++)
			{
				try
				{
					for (int ii = 0; ii < 4; ii++)
						b[ii] = gp.ser_buffer[ii + offset];				
					ByteBuffer bb_data = ByteBuffer.wrap(b);
					
					if ((bh.status & 0x4) != 0)			//32 bit int
						gp.ser[i][j] = (double)bb_data.getInt();
					if ((bh.status & 0x8) != 0)			//32 bit float
						gp.ser[i][j] = (double)(8192.0 * bb_data.getFloat());
					
					bytes_read = bytes_read + 4;
					offset = offset + 4;
				}
				catch (Exception ex)
				{
					Util.show_and_log("get_varian_ser()_XZ 1:  " , ex.getMessage());
					error = true;
					break;
				}
			}  // for j  (X dimension)

			if (error) break;   // break outer loop    
			offset = offset +  block_size;   // skip a block!

		}   // for i (Z dimension)

		//  clean-up    

		if (gp.log_on)

			Util.log_error(String.format("get_varian_ser_3D_XZ 1(): bytes read = %d",
					bytes_read));

		return (bytes_read);
	}
        
//-------------------------------------------------------------------------------------------------
/**
 * read Varian 3D fid file, XY (F3-F2) planes for phase2, phase ordering (eg, HNCO)
 * @param plane
 * @param gp
 * @param q
 * @param q_f1
 * @param q_f2
 * @return bytes read, -1 on error
 */
	public static int get_varian_ser_3D_XY_order_2(int plane, global_par gp,
			data_par q, data_par q_f1, data_par q_f2)
	{
		DataFileHeader fh = new DataFileHeader();
		DataBlockHeader bh = new DataBlockHeader();

		byte[] fh_byte = new byte[32];             // file header is 32 bytes
		byte[] bh_byte = new byte[28];             // block header is 28 bytes
		int bytes_read = 0;
		int offset = 0;                            // offset to current block in bytes
		int block_size = q.TD * 4 + 28;            // size of data block in bytes
		boolean error = false;                     // for breaking nested loops

		// read file header into byte array
		try
		{
			for (int i = 0; i < 32; i++)
				fh_byte[i] = gp.ser_buffer[i];
		}
		catch (Exception ex)
		{
			Util.show_and_log("get_varian_ser_3D_XY 2():  ", ex.getMessage());
			return -1;
		}
		bytes_read = bytes_read + 32;

		ByteBuffer bb_fh = ByteBuffer.wrap(fh_byte);
		fh.nblocks = bb_fh.getInt(0);
		fh.np = bb_fh.getInt(8);
		fh.status = bb_fh.getShort(26);
		fh.nbheaders = bb_fh.getInt(28);
		q.nblocks = fh.nblocks;

		// calculate offset for bizzare Varian file ordering
		offset = 32 
				+ 4 * (plane / 2) * block_size
				+ (plane % 2) * block_size;

		byte[] b = new byte[4];

		//  loop through all increments  (Y dimension)
		for (int i = 0; i < q_f2.TD; i = i + 2)
		{
			//  read first row of pair
			//  read block header, and get status field
			
			for (int k = 0; k < 28; k++)
				bh_byte[k] = gp.ser_buffer[k + offset];
			ByteBuffer bb_bh = ByteBuffer.wrap(bh_byte);
			
			bh.status = bb_bh.getShort(2);	
			bh.index = bb_bh.getShort(4);

			offset = offset + 28;                       // have read the block header
			bytes_read = bytes_read + 28;

			for (int j = 0; j < q.TD; j++)
			{
				try
				{
					for (int ii = 0; ii < 4; ii++)
						b[ii] = gp.ser_buffer[ii + offset];
					ByteBuffer bb_data = ByteBuffer.wrap(b);
					
					if ((bh.status & 0x4) != 0)			//32 bit int
						gp.ser[i][j] = (double)bb_data.getInt();
					if ((bh.status & 0x8) != 0)			//32 bit float
						gp.ser[i][j] = (double)(8192.0 * bb_data.getFloat());
					
					bytes_read = bytes_read + 4;
					offset = offset + 4;
				}
				catch (Exception ex)
				{
					Util.show_and_log("get_varian_ser_XY 2():  ", ex.getMessage());
					error = true;
					break;
				}
			}  // for j  (X dimension)

			offset = offset + block_size + 28;

			for (int j = 0; j < q.TD; j++)
			{
				try
				{
					for (int ii = 0; ii < 4; ii++)
						b[ii] = gp.ser_buffer[ii + offset];
					ByteBuffer bb_data = ByteBuffer.wrap(b);
					
					if ((bh.status & 0x4) != 0)			//32 bit int
						gp.ser[i + 1][j] = (double)bb_data.getInt();
					if ((bh.status & 0x8) != 0)			//32 bit float
						gp.ser[i + 1][j] = (double)(8192.0 * bb_data.getFloat());
					
					bytes_read = bytes_read + 4;
					offset = offset + 4;
				}
				catch (Exception ex)
				{
					Util.show_and_log("get_varian_ser_XY 2():  ", ex.getMessage());
					error = true;
					break;
				}
			}  // for j  (X dimension)


			if (error) break;   // break outer loop    

			offset = offset + 2 * q_f1.TD * block_size
					- 3 * block_size;   // check!!  TO DO  f1 or f2!!!!

		}   // for i (Y dimension)

		//  clean-up    

		if (gp.log_on)

			Util.log_error(String.format("get_varian_ser_3D_XY 2(): bytes read = %d",
					bytes_read));

		return (bytes_read);
	}
		
//-------------------------------------------------------------------------------------------------
/**
 * read Varian 3D fid file, XY (F3-F2) planes for phase, phase2 ordering
 * @param plane
 * @param gp
 * @param q
 * @param q_f1
 * @param q_f2
 * @return
 */
	public static int get_varian_ser_3D_XY_order_1(int plane, global_par gp,
			data_par q, data_par q_f1, data_par q_f2)
	{
		DataFileHeader fh = new DataFileHeader();
		DataBlockHeader bh = new DataBlockHeader();

		byte[] fh_byte = new byte[32];             // file header is 32 bytes
		byte[] bh_byte = new byte[28];             // block header is 28 bytes
		int bytes_read = 0;
		int offset = 0;                            // offset to current block in bytes
		int block_size = q.TD * 4 + 28;            // size of data block in bytes
		boolean error = false;                     // for breaking nested loops

		// read file header into byte array
		try
		{
			for (int i = 0; i < 32; i++)
				fh_byte[i] = gp.ser_buffer[i];
		}
		catch (Exception ex)
		{
			Util.show_and_log("get_varian_ser_3D_XZ 1():  ", ex.getMessage());
			return -1;
		}
		bytes_read = bytes_read + 32;

		//  extract important fields from file header
		ByteBuffer bb_fh = ByteBuffer.wrap(fh_byte);
		fh.nblocks = bb_fh.getInt(0);
		fh.np = bb_fh.getInt(8);
		fh.status = bb_fh.getShort(26);
		fh.nbheaders = bb_fh.getInt(28);
		q.nblocks = fh.nblocks;

		// calculate offset for bizzare Varian file ordering

		offset = 32 + 2 * plane * block_size;
		byte[] b = new byte[4];

		//  loop through all increments  (Y dimension)
		for (int i = 0; i < q_f2.TD; i = i+2)
		{
			//  read first row of pair
			//  read block header, and get status field
			for (int k = 0; k < 28; k++)
				bh_byte[k] = gp.ser_buffer[k + offset];
			ByteBuffer bb_bh = ByteBuffer.wrap(bh_byte);
			
			bh.status = bb_bh.getShort(2);	
			bh.index = bb_bh.getShort(4);
			
			offset = offset + 28;                       // have read the block header
			bytes_read = bytes_read + 28;           

			for (int j = 0; j < q.TD; j++)
			{
				try
				{
					for (int ii = 0; ii < 4; ii++)
						b[ii] = gp.ser_buffer[ii + offset];	
					ByteBuffer bb_data = ByteBuffer.wrap(b);
					
					if ((bh.status & 0x4) != 0)			//32 bit int
						gp.ser[i][j] = (double)bb_data.getInt();
					if ((bh.status & 0x8) != 0)			//32 bit float
						gp.ser[i][j] = (double)(8192.0 * bb_data.getFloat());
					
					bytes_read = bytes_read + 4;
					offset = offset + 4;
				}
				catch (Exception ex)
				{
					Util.show_and_log("get_varian_ser_XY 1(): ",  ex.getMessage());
					error = true;
					break;
				}
			}  // for j  (X dimension)

			//  read second row of pair
			//  read block header, and get status field
			for (int k = 0; k < 28; k++)
				bh_byte[k] = gp.ser_buffer[k + offset];
			ByteBuffer bb_bh2 = ByteBuffer.wrap(bh_byte);
			
			bh.status = bb_bh2.getShort(2);	
			bh.index = bb_bh2.getShort(4);

			offset = offset + 28;                       // have read (skip??) the block header
			bytes_read = bytes_read + 28;

			for (int j = 0; j < q.TD; j++)
			{
				try
				{
					for (int ii = 0; ii < 4; ii++)
						b[ii] = gp.ser_buffer[ii + offset];					
					ByteBuffer bb_data = ByteBuffer.wrap(b);
					
					if ((bh.status & 0x4) != 0)			//32 bit int
						gp.ser[i + 1][j] = (double)bb_data.getInt();
					if ((bh.status & 0x8) != 0)			//32 bit float
						gp.ser[i + 1][j] = (double)(8192.0 * bb_data.getFloat());
					
					bytes_read = bytes_read + 4;
					offset = offset + 4;
				}
				catch (Exception ex)
				{
					Util.show_and_log("get_varian_ser_XY 1(): ",  ex.getMessage());
					error = true;
					break;
				}
			}  // for j  (X dimension)

			if (error) break;   // break outer loop    
			offset = offset + 2 * q_f1.TD * block_size
					- 2 * block_size;   // check!!  TO DO  f1 or f2!!!!
			
		}   // for i (Y dimension)

		//  clean-up    

		if (gp.log_on)

			Util.log_error(String.format("get_varian_ser_3D_XY 1(): bytes read = %d",
					bytes_read));

		return (bytes_read);
	}

//-------------------------------------------------------------------------------------------------
/**
 * Loads a 3D ser file into a buffer 
 * @param fileName full path to the file
 * @param serBuffer A buffer for the ser file
 * @return bytes read if successful, -1 if file can't be opened, -3 on read error
 */
	public static int getSerBuffer(String fileName, global_par gp)  
	{       
		FileInputStream fs;

		File file = new File(fileName);

		try { fs = new FileInputStream(file); }
		catch (IOException ex)
		{
			Util.show_and_log("get_ser(): cannot open: " + fileName, ex.getMessage());
			return -1;
		}

		gp.ser_buffer = new byte[(int)file.length()];

		try { fs.read(gp.ser_buffer); } 
		catch (IOException ex) 
		{
			ex.printStackTrace();
			try { fs.close(); } 
			catch (IOException e) { e.printStackTrace(); }
			return -3;
		}

		try {	fs.close();	} 
		catch (IOException e) {	e.printStackTrace();	}

		return gp.ser_buffer.length;
	}
		 
//-------------------------------------------------------------------------------------------------
}
