package spinworks;

import spinworks.SW_main.filters;

/**
 * Time domain filter functions for SpinWorks Java
 * @author marat,
 * University of Manitoba
 */
public class Filter 
{

//-------------------------------------------------------------------------------------------------
	/**
	 * Apply a filter function to the fid.  Apply to reals and imaginaries separately.
	 * @param npts
	 * @param data
	 * @param convolve
	 * @param width
	 * @param filter
	 * @param sum
	 */
	public static void apply_filter(int npts, double[] data, double[] convolve,
			int width, double[] filter, double[] sum)
	{
		int i, j, k, w2, first, last;
		double c, d;

		w2 = width / 2;

		//  first width/2 points  
		for (i = 0; i < w2; i ++)
		{
			d = sum[(w2+i)];

			first = 0;
			last = i + w2;

			k = (w2 - i);

			c = 0; 

			for(j = first; j <= last; j++) 
			{
				c += data[j] * filter[k];
				k++;
			}

			convolve[i] = c / d;
		}

		//  bulk of points  
		for (i = w2; i < npts-w2; i++)
		{
			d = sum[width-1];

			first = i - w2;
			last = i + w2;

			k = 0;

			c = 0; 

			for (j = first; j <= last; j++)
			{
				c += data[j] * filter[k];
				k++;
			}

			convolve[i] = c / d;
		}

		//  last width/2 points  
		for (i = npts-w2; i < npts; i++)
		{
			d = sum[width-2-(i+w2-npts)];

			first = i - w2;
			last = npts - 1;

			k = 0;

			c = 0; 

			for (j = first; j <= last; j++)
			{
				c += data[j] * filter[k];
				k++;
			}

			convolve[i] = c / d;
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * calculate a sine filter function
	 * @param width
	 * @param filter
	 * @param sum
	 */
	private static void calculate_sine_filter(int width, double[] filter, double[] sum)
	{
		int i;
		double s = 0;

		for (i = 0; i < width; i++)
		{
			filter[i] = Math.sin(Math.PI * ((double)(i + 1)) / ((double)(width + 1)));
			s += filter[i];
			sum[i] = s;
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * calculate a square or "box" filter function
	 * @param width
	 * @param filter
	 * @param sum
	 */
	private static void calculate_box_filter(int width, double[] filter, double[] sum)
	{
		int i;

		for (i = 0; i < width; i++)
		{
			filter[i] = 1;
			sum[i] = i + 1;
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * calculate triangle filter function, assumes that width is odd, and width > 0 
	 * @param width
	 * @param filter
	 * @param sum
	 */
	private static void calculate_triangle_filter(int width, double[] filter, double[] sum)
	{
		int i, w2 = width / 2;
		double s = 0;

		filter[w2] = 1;

		for (i = 1; i <= w2; i++)
		{
			filter[w2 + i] = ((double)(w2 + 1 - i)) / ((double)(w2 + 1));
			filter[w2 - i] = filter[w2 + i];
		}

		for (i = 0; i < width; i++)
		{
			s += filter[i];
			sum[i] = s;
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * gaussian filter function, assumes that width is odd, width > 0, and 0 < end < 1 
	 * @param width
	 * @param end
	 * @param filter
	 * @param sum
	 */
	private static void calculate_gaussian_filter(int width, double end, double[] filter, 
			double[] sum)
	{
		int i, w2 = width / 2;
		double s = 0.9;
		double a;

		a = -Math.log((double)end) / (w2 * w2);

		filter[w2] = 1;

		for (i = 1; i <= w2; i++)
		{
			filter[w2 + i] = Math.exp(-a * i * i);
			filter[w2 - i] = filter[w2 + i];
		}

		for (i = 0; i < width; i++)
		{
			s += filter[i];
			sum[i] = s;
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * used for down-sampling of oversampled data
	 * @param width
	 * @param cutoff
	 * @param filter
	 * @param sum
	 */
	@SuppressWarnings("unused")
	private static void calculate_windowed_sinc_filter(int width, double cutoff, 
			double[] filter, double[] sum)
	{
		//	width is the number of points (poles or taps) for the filter
		//	cutoff is the filter cutoff frequency as a fraction of the spectral window.
		int i;
		int w2 = width / 2;
		double s = 0.0;

		for (i = 0; i < width; i++)
		{
			if ((i + 1) == (w2 + 1))
				filter[i] = Math.PI * cutoff;		// avoids divide by zero
			else
				filter[i] = Math.sin(Math.PI * cutoff * (double)(i + 1 - (w2 + 1)))
				/ (double)(i + 1 - (w2 + 1));	    // sinc function

			filter[i] = filter[i]				    // apply Hamming window
					* (0.54 - 0.46 * Math.cos(2 * Math.PI * (double)(i + 1) 
							/ (double)(width + 1)));

			s += filter[i];
			sum[i] = s;
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Solvent filter function.  Applies a heavy low pass filter to the fid and then subtracts 
	 * the result from the fid, essentially creating a high pass filter.
	 * @param type the filter function enumeration
	 * @param filter_points the number of filter points (must be odd)
	 * @param td the number of time domain points (real + imag)
	 * @param data the time domain data (interleaved reals and imags)
	 */
	public static void do_filter(filters type, int filter_points, int td, double[] data)
	{
		if (type == filters.FILT_NONE) return;

		int i, j;

		double[] fid_real = new double[td / 2];		// allocate space 
		double[] fid_imag = new double[td / 2];
		double[] convolute_real = new double[td / 2];
		double[] convolute_imag = new double[td / 2];
		double[] filter = new double[filter_points];
		double[] sum = new double[filter_points];

		j = 0;										// split reals and imaginaries
		for (i = 0; (i < td - 1); i = i + 2)
		{
			fid_real[j] = data[i];
			fid_imag[j] = data[i + 1];
			j++;
		}

		if (type == filters.FILT_SINE)
			calculate_sine_filter(filter_points, filter, sum);
		else if (type == filters.FILT_GAUSS)
			calculate_gaussian_filter(filter_points, 0.1, filter, sum);
		else if (type == filters.FILT_TRIANGLE)
			calculate_triangle_filter(filter_points, filter, sum);
		else
			calculate_box_filter(filter_points, filter, sum);


		apply_filter(td / 2, fid_real, convolute_real,
				filter_points, filter, sum);
		apply_filter(td / 2, fid_imag, convolute_imag,
				filter_points, filter, sum);

		j = 0;
		for (i = 0; (i < td - 1); i = i + 2)
		{
			data[i] = fid_real[j] - convolute_real[j];
			data[i + 1] = fid_imag[j] - convolute_imag[j];
			j++;
		}
	}
//-------------------------------------------------------------------------------------------------
}
