package spinworks;

/**
 * smoothing and interpolation with first differences
 * Used by Whittaker smoother
 * @author marat, 
 * University of Manitoba
 */
public class FiniteDiff 
{

//  smoothing and interpolation with first differences

    public static void fdiff_1(double[] w, double[] y, double[] z, double lambda, int n)
    {
        int i, i1;
        double[] c = new double[n+1];
        double[] d = new double[n+1];

        d[1] = w[1] + lambda;
        c[1] = -lambda / d[1];
        z[1] = w[1] * y[1];

        for (i = 2; i < n; i++)
        {
            i1 = i - 1;
            d[i] = w[i] + 2 * lambda - c[i1] * c[i1] * d[i1];
            c[i] = -lambda / d[i];
            z[i] = w[i] * y[i] - c[i1] * z[i1];
        }

        d[n] = w[n] + lambda - c[n - 1] * c[n - 1] * d[n - 1];
        z[n] = (w[n] * y[n] - c[n - 1] * z[n - 1]) / d[n];

        for (i = n - 1; 1 <= i; i--)
            z[i] = z[i] / d[i] - c[i] * z[i + 1];
    }

//-------------------------------------------------------------------------------------------------
//  smoothing and interpolation with first differences (Whittaker method)

    public static void fdiff_1a(double[] w, double[] y, double[] z, double lambda, int n)
    {
        int i, i1;
        double[] c = new double[n];
        double[] d = new double[n];

        d[0] = w[0] + lambda;
        c[0] = -lambda / d[0];
        z[0] = w[0] * y[0];

        for (i = 1; i < n-1; i++)
        {
            i1 = i - 1;
            d[i] = w[i] + 2 * lambda - c[i1] * c[i1] * d[i1];
            c[i] = -lambda / d[i];
            z[i] = w[i] * y[i] - c[i1] * z[i1];
        }

        d[n - 1] = w[n - 1] + lambda - c[n - 2] * c[n - 2] * d[n - 2];
        z[n - 1] = (w[n - 1] * y[n - 1] - c[n - 2] * z[n - 2]) / d[n - 1];

        for (i = n - 2; 0 <= i; i--)
            z[i] = z[i] / d[i] - c[i] * z[i + 1];
    }

//-------------------------------------------------------------------------------------------------
//  smoothing and interpolation with second differences

    public static void fdiff_2(double[] w, double[] y, double[] z, double lambda, int m)
    {
        int i, i1, i2;
        double[] c = new double[m];
        double[] d = new double[m];
        double[] e = new double[m];

        d[0] = w[0] + lambda;
        c[0] = -2 * lambda / d[0];
        e[0] = lambda / d[0];
        z[0] = w[0] * y[0];
        d[1] = w[1] + 5 * lambda - d[0] * c[0] * c[0];
        c[1] = (-4 * lambda - d[0] * c[0] * e[0]) / d[1];
        e[1] = lambda / d[1];
        z[1] = w[1] * y[1] - c[0] * z[0];

        for (i = 2; i < m - 2; i++)
        {
            i1 = i - 1; 
            i2 = i - 2;
            d[i] = w[i] + 6 * lambda - c[i1] * c[i1] * d[i1] - e[i2] * e[i2] * d[i2];
            c[i] = (-4 * lambda - d[i1] * c[i1] * e[i1]) / d[i];
            e[i] = lambda / d[i];
            z[i] = w[i] * y[i] - c[i1] * z[i1] - e[i2] * z[i2];
        }

        i1 = m - 3; 
        i2 = m - 4;

        d[m - 2] = w[m - 2] + 5 * lambda - c[i1] * c[i1] * d[i1] - e[i2] * e[i2] * d[i2];
        c[m - 2] = (-2 * lambda - d[i1] * c[i1] * e[i1]) / d[m - 2];
        z[m - 2] = w[m - 2] * y[m - 2] - c[i1] * z[i1] - e[i2] * z[i2];

        i1 = m - 2; 
        i2 = m - 3;

        d[m - 1] = w[m - 1] + lambda - c[i1] * c[i1] * d[i1] - e[i2] * e[i2] * d[i2];
        z[m - 1] = (w[m - 1] * y[m - 1] - c[i1] * z[i1] - e[i2] * z[i2]) / d[m - 1];
        z[m - 2] = z[m - 2] / d[m - 2] - c[m - 2] * z[m - 1];

        for (i = m - 3; 0 <= i; i--)
            z[i] = z[i] / d[i] - c[i] * z[i + 1] - e[i] * z[i + 2];
    }

//-------------------------------------------------------------------------------------------------
}
