
/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

package spinworks;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;

import simulation.SimPar;

/**
 * screen graphics routines for SpinWorks Java
 * @author marat
 */
public class GraphUtil
{
	public final static int MAX_DISP_BUFF = 200000;
	public static double[] tempi;
	public static double[] tempr;
	public final static double M_LN2 = 0.693147181;		// ln(2.0)
	
	private static boolean out_of_sw = false;
	
	static int[] max_yvalues = new int[MAX_DISP_BUFF];	
	static int[] max_xvalues = new int[MAX_DISP_BUFF];  // the x value of a maximum y

	/**
	 * Draw a 1D view on the screen
	 * @param gr the current graphics context
	 * @param gp a SpinWorks "global parameter" object
	 * @param q data parameters in the direct (acquisition) dimension
	 * @param q_ind data parameters in the current indirect dimension 
	 * (needed to label freq on extracted slices)
	 */
	public static void draw_1D_view(GC gr, global_par gp, SimPar sp, 
			data_par q, data_par q_ind)
	{
		if (gp.freq1 == gp.freq2) return;               // impossible limits
		Util.check_order(gp);

		//	adjust xsize to accomodate any stacked trace offset
		//int dxsize = gp.dxsize - gp.stack_x_offset * gp.n_traces; // TODO  check
		gp.hzperpix = (gp.freq1 - gp.freq2) / (double)(gp.dxsize - 1);

		gr.setBackground(gp.back_colour);
    	gr.fillRectangle(0, 0, gp.dxsize, gp.dysize);
   	
		if(gp.show_zero_line)
		{
			gr.setForeground(gp.colour_dark_blue);
			gr.drawLine(0, gp.dysize - gp.expt_y_offset, 
					gp.dxsize, gp.dysize - gp.expt_y_offset);	
		}
		
		// draw the title, if desired
		if(gp.show_title)
		{
			gr.setForeground(gp.axis_colour);
	    	gr.setFont(new Font(SW_main.shell.getDisplay(), "Arial", 
	    			SW_main.MEDIUM_FONT, SWT.NORMAL));	// set font   	
			gr.drawText(gp.plot_title, 3, 10);
		}
			
		// draw any stacked traces
		// draw stack first so whiewashing doesn't erase other items
		   draw_stack(gr, gp);
			
		//	draw spectrum
		if (gp.data_array != null && gp.show_spectrum) 
			draw_spect(gr, gp.freq1, gp.freq2, 
					gp.expt_scale, 0, gp.expt_y_offset, 
					gp.dxsize, gp.dysize, gp.data_array, 2, gp.spect_colour, gp, q);
		
		//	draw integrals if desired
		if (gp.data_array != null && (gp.integ_mode || gp.show_integrals)
				&& gp.n_integrals > 0)
			draw_integ(gr, gp.freq1, gp.freq2, gp.integ_scale, 120, gp, q);

		if (gp.calc_array != null && gp.show_simulated  && !gp.phase_mode)	// TODO
		{
			double scale;
			// in deconvolution mode, fix calculated sale to experimental scale
			if (sp.calc_mode == SimPar.DECON)
				scale = gp.expt_scale;
			else
				scale = gp.calc_scale;

			draw_spect(gr, gp.freq1, gp.freq2,
					scale, 0, gp.calc_y_offset,
					gp.dxsize, gp.dysize, gp.calc_array, 2, gp.sim_colour, gp, q);
		}
	
		//	draw axis and labels for scaling buttons 
		draw_axis(gr, gp.freq1, gp.freq2, 0, gp.axis_offset, gp.dxsize, gp, q, q_ind);
		//draw_scale_button_labels_1D(gr, gp);	// TODO ???
		
		double y1 = -gp.axis_offset/ gp.expt_scale / gp.data_max;
		double y2 = (gp.dysize - gp.expt_y_offset) / gp.expt_scale / gp.data_max;
				
		if(gp.data_max > .001 && gp.expt_scale > 1.0e-9 && gp.show_yaxis)
			draw_y_axis(gp, gr, y1, y2, gp.dxsize - 5, gp.dysize);

		//	draw any necessary cursor lines		
		gr.setForeground(gp.colour_red);	// Red
		if(gp.cursor_flag == 1)
			gr.drawLine(SW_main.cursor1_x, 0, SW_main.cursor1_x, gp.dysize);
		else if(gp.cursor_flag == 2)
		{
			gr.drawLine(SW_main.cursor1_x, 0, SW_main.cursor1_x, gp.dysize);
			gr.drawLine(SW_main.cursor2_x, 0, SW_main.cursor2_x, gp.dysize);
		}

		// Draw a labelled line at the PP minimum intensity 
		if (gp.adjust_mi)
		{
			int y = gp.dysize - (int)(gp.expt_scale * gp.data_max * gp.pp_thresh)
				  - gp.expt_y_offset;
					
			if (y > 10 && gp.data_array != null) // show the mi for peak picking
			{
				gr.setFont(new Font(SW_main.shell.getDisplay(), "Arial", 
						SW_main.MEDIUM_FONT, SWT.NORMAL));
				gr.setForeground(gp.colour_magenta);	// whatever
				gr.drawString("PP Minimum", 10, y - 20);
				gr.drawLine(0, y, gp.dxsize, y);
			}
		}
		   	  	
		//	now the block number if drawing 1D slices of 2D data
		if(q.data_mode == SW_main.modes.MODE_2D
				&& gp.display_mode == SW_main.modes.MODE_1D && gp.data_array != null)
		{	
			gr.setForeground(gp.axis_colour);
			gr.setFont(new Font(SW_main.shell.getDisplay(), "Arial", 
					SW_main.SMALL_FONT, SWT.NORMAL)); // set font 
			gr.drawText(String.format("Block # %d   %.3f ppm", 
					gp.block_2D + 1, 
					((q_ind.offset - gp.block_2D * q_ind.hzperpt)/q_ind.SF)),
					3, 50);
		}
		

		if (!gp.integ_mode && !gp.phase_mode && gp.show_simulated && gp.calc_array != null)
			set_cursor(gp.trans_index, gr, gp, sp, q);
		
		if (gp.match_on)  // peaks and simulation match
		{
			showmatch(gr, gp, sp, q);
			draw_peaks(gr, gp, q);
		}
				
		gr.drawText(String.format("vertical scale = %f", 
				gp.expt_scale * 1000.0), 3, gp.dysize - gp.axis_offset - 25);
						
	} 
	
//-------------------------------------------------------------------------------------------------
    /**
     * Draw the screen in FID more (FID view)
     * @param gr the current graphics context
     * @param gp A SpinWorks "global parameter" object
     * @param q "status" data parameters (direct dimension)
     */
	public static void draw_FID_view(GC gr, global_par gp, data_par q)
		{
			gr.setBackground(gp.back_colour);
			gr.fillRectangle(0, 0, gp.dxsize, gp.dysize);

			gr.setForeground(gp.colour_dark_blue);
			gr.drawLine(0, gp.dysize - gp.expt_y_offset,
						gp.dxsize, gp.dysize - gp.expt_y_offset);

            if (gp.t1 > gp.t2)      // check for order
            {
                double temp = gp.t1;
                gp.t1 = gp.t2;
                gp.t2 = temp;
            }

            double[] reals = new double[q.TD/2];
            double[] imags = new double[q.TD/2];

            try
            {
                if (q.data_mode == SW_main.modes.MODE_1D && gp.fid != null) // 1D fid
                {
                    for (int i = 0; i < q.TD / 2; i++)
                    {
                        reals[i] = gp.fid[i * 2];
                        imags[i] = gp.fid[i * 2 + 1];
                    }
                }
                else // 2D ser
                {
                    for (int i = 0; i < q.TD / 2; i++)
                    {
                        reals[i] = gp.ser[gp.block_2D][i * 2];
                        imags[i] = gp.ser[gp.block_2D][i * 2 + 1];
                    }
                }
            }
            catch (Exception ex)
            {
                Util.log_error("draw_fid_view(): ", ex.getMessage());
                return;
            }

			draw_fid(gr, gp.t1, gp.t2,
					gp.fid_scale, 0, gp.expt_y_offset,
					gp.dxsize, gp.dysize, reals, 2, gp.spect_colour, gp, q);

			gr.drawText("real", gp.dxsize - 50, gp.dysize - 80);
	
			draw_fid(gr, gp.t1, gp.t2,
						gp.fid_scale, 0, gp.expt_y_offset + 300,
						gp.dxsize, gp.dysize, imags, 2, gp.sim_colour, gp, q);

			gr.drawText("imag", gp.dxsize - 50, gp.dysize - 100);


            /*draw_window(gr, gp.t1, gp.t2, gp.fid_scale, 0, gp.expt_y_offset,	// TODO
                gp.dxsize, gp.dysize, 1, Color.Coral,
                gp, p, q);

            draw_window(gr, gp.t1, gp.t2, gp.fid_scale, 0, gp.expt_y_offset + 300,
                gp.dxsize, gp.dysize, 1, Color.Coral,
                gp, p, q);*/

			draw_fid_axis(gr, gp.t1, gp.t2, gp.axis_offset, gp.dxsize, gp, q);
			//draw_scale_button_labels_1D(gr, gp);

//	draw any necessary cursor lines
			gr.setForeground(gp.colour_red);	// Red
			if(gp.cursor_flag == 1)
				gr.drawLine(SW_main.cursor1_x, 0, SW_main.cursor1_x, gp.dysize);
			else if(gp.cursor_flag == 2)
			{
				gr.drawLine(SW_main.cursor1_x, 0, SW_main.cursor1_x, gp.dysize);
				gr.drawLine(SW_main.cursor2_x, 0, SW_main.cursor2_x, gp.dysize);
			}

			gr.setForeground(gp.axis_colour);
			Font font = new Font(SW_main.shell.getDisplay(), "Arial", SW_main.MEDIUM_FONT, SWT.NORMAL);
	    	gr.setFont(font);					// set font
	    	
//	now draw the title
			gr.drawText(gp.plot_title, 3, 10);
            gr.drawText(String.format("vertical scale = %f", gp.fid_scale * 1000.0),
				3, gp.dysize - gp.axis_offset - 20);

			font.dispose();
		}

    
//-------------------------------------------------------------------------------------------------
/// Draws a 1D spectrum in the drawing panel
	
	static void draw_spect(GC gr, double f1, double f2, double yscale,
    		int xoffset, int yoffset, int xsize, int ysize,
    		double[] array, int thresh, Color color,
    		global_par gp, data_par q)
    {
    	if (array == null) return;  // nothing to draw
    	if (gr == null)
    	{
    		Util.show_and_log("draw_spect(), null graphics?? ");
    		return;
    	}

    	int point1, point2;		// start and stop points in data array 
    	int x,y;				// display x and y coord. 
    	int oldx, oldy;			// old x and y values 
    	int npoints;            // number of points to be plotted
    	int i;                  // handy counter
    	double a;				// for phasing   

    	int[] xvalues = new int[MAX_DISP_BUFF];		// for speed, perhaps 
    	int[] yvalues = new int[MAX_DISP_BUFF];     // do this one time at start-up???

    	gr.setForeground(color);
    	
    	double hzperpix = (f1-f2)/(double)(xsize-1);

    	// if(gp.phase_mode) thresh = thresh * 2;		// less resolution while phasing
   	
		point1 = (int) Math.round((q.offset - f1) / q.hzperpt);
		point2 = (int) Math.round((q.offset - f2) / q.hzperpt);

    	// make sure array bounds are not exceeded
    	if (point1 < 0) point1 = 0;
    	if(point2 >= q.SI) point2 = q.SI-1;

    	if (point2 >= array.length)
    	{
    		Util.show_and_log("draw_spect(), space allocation ");
    		return;
    	}
    	if (point1 >= array.length)
    	{
    		return;
    	}
    	if (hzperpix <= 1.0e-10)
    	{
    		Util.show_and_log("draw_spect() Hz per Pix too small!");
    		return;
    	}

    	// initialize point counter 
    	npoints = 0;

    	y = ysize - (((int)(array[point1] * yscale)) - yoffset);

    	if (y > (ysize - yoffset - gp.axis_offset-2))
    		y = ysize - gp.axis_offset-2;

    	if (y < 0) y = 0;

    	oldx = xoffset;
    	oldy = y;

    	try
    	{
    		for (i = point1; i <= point2; i++)
    		{
    			x = (int) (Math.round((f1 - q.offset + i * q.hzperpt) / hzperpix) + xoffset);

    			if (!gp.phase_mode)
    				y = ysize - ((int)(array[i] * yscale)) - yoffset;
    			else if (gp.phase_mode)
    			{
    				a = 0.0174533 * (gp.ph0 + ((double)i / (double)q.SI) * gp.ph1);
    				y = ysize - (int)(yscale * (tempr[i] * Math.cos(a)
    						- tempi[i] * Math.sin(a)))
    						- yoffset;
    			}
    			    		
    			if (y > (ysize - gp.axis_offset - 2))
    				y = ysize - gp.axis_offset - 2;

    			if (y < 0) y = 0;

    			if (((x != oldx) || (Math.abs(y - oldy) > thresh)) &&
    					npoints < MAX_DISP_BUFF)
    			{
    				xvalues[npoints] = x;
    				yvalues[npoints] = y;
    				npoints++;
    				oldx = x;
    				oldy = y;
    			}
    		}
    	}
    	catch (Exception ex)
    	{            
    		Util.show_and_log("draw_spect() " + ex.getMessage());
    		ex.printStackTrace();
    		return;
    	}

//  	draw Spectrum 
    	int[] pts = new int[2 * npoints];

    	for(i = 0; i < npoints; i++) 
    	{
    		pts[2 * i] = xvalues[i];
    		pts[2 * i + 1] = yvalues[i];
    	}
    	
    	if(npoints > 2)
    		gr.drawPolyline(pts);   // polyLine requires single array with alternating x, y values
    	
    	
    	// for Autophase development  remove for release version
    	if(gp.phase_mode)
    	{
    		double[] reals = new double[q.SI];
    		double[] imags = new double[q.SI];
    		
    		for(int k = 0; k < q.SI; k++)
    		{
    			reals[k] = gp.data_array[k];
    			imags[k] = gp.i_data_array[k];	   
    		}
    		Fft.do_phase(reals, imags, q.SI, gp.ph0, gp.ph1);
 			//double integ = AutoPhase.totalInteg(reals, 0, q.SI);
 			
 			//gr.drawText(String.format("integ = %f", integ / 1.0e8), 200, 200);
 			
 			for (int k = 0; k < gp.n_peaks; k++)	// loop over all peaks
 			{
 				if(gp.peaks[k].freq < gp.freq1 && gp.peaks[k].freq > gp.freq2) // only for peaks in display region
 				{
 					int start = Util.freq_to_index(gp.peaks[k].freq, q) - 10;
 					if(start < 0) start = 0;
 					int end = Util.freq_to_index(gp.peaks[k].freq, q) + 10;
 					if(end >= q.SI) end = q.SI - 1;
 					
 					double sum = 0.0;
 					for(int l = start; l <= end; l++)
 						sum = sum + reals[l];
 						
 					gr.drawText(String.format("peak = %d at %f ppm   integ = %f", k, 
 							gp.peaks[k].freq / q.SFO1,
 							sum / 1.0e8), 100, 200 + k * 20);
 				}
 			}
    	}

    	//  	clean up
    }		//end draw_spect
	
//-------------------------------------------------------------------------------------------------
// Draws a 1D spectrum on the canvas
// overloaded version for stack trace
		
	private static void draw_spect(GC gr, double f1, double f2,
    		int xoffset, int yoffset, int xsize, int ysize,
    		Trace stack_trace, int thresh, Color colour,
    		global_par gp)
    {
    	if (stack_trace.data == null) return;  // nothing to draw
    	if (gr == null)
    	{
    		Util.show_and_log("draw_spect(), null graphics?? ");
    		return;
    	}

    	int point1, point2;		// start and stop points in data array 
    	int x,y;				// display x and y coord. 
    	int oldx, oldy;			// old x and y values 
    	int npoints;            // number of points to be plotted
    	int i;                  // handy counter

    	int[] xvalues = new int[MAX_DISP_BUFF];			// for speed, perhaps 
    	int[] yvalues = new int[MAX_DISP_BUFF];    		// do this one time at start-up???   	

    	gr.setForeground(colour);
    	
    	double hzperpix = (f1-f2)/(double)(xsize-1);
   	
		point1 = (int) Math.round((stack_trace.offset - f1) / stack_trace.hzperpt);
		point2 = (int) Math.round((stack_trace.offset - f2) / stack_trace.hzperpt);

    	// make sure array bounds are not exceeded
    	if (point1 < 0) point1 = 0;
    	if(point2 >= stack_trace.size) point2 = stack_trace.size-1;

    	if (point2 >= stack_trace.data.length)
    	{
    		Util.show_and_log("draw_spect(), space allocation ");
    		return;
    	}
    	if (point1 >= stack_trace.data.length)
    	{
    		return;
    	}
    	if (hzperpix <= 1.0e-10)
    	{
    		Util.show_and_log("draw_spect() Hz per Pix too small!");
    		return;
    	}

    	// initialize point counter 
    	npoints = 0;

    	y = ysize - (((int)(stack_trace.data[point1] * stack_trace.yscale)) - yoffset);

    	if (y > (ysize - yoffset - gp.axis_offset-2))
    		y = ysize - gp.axis_offset-2;

    	if (y < 0) y = 0;

    	oldx = xoffset;
    	oldy = y;

    	try
    	{
    		for (i = point1; i <= point2; i++)
    		{
    			x = (int) (Math.round((f1 - stack_trace.offset + i * stack_trace.hzperpt) / hzperpix) + xoffset);
    			y = ysize - ((int)(stack_trace.data[i] * stack_trace.yscale)) - yoffset;
    			
    			if (y > (ysize - gp.axis_offset - 2))
    				y = ysize - gp.axis_offset - 2;

    			if (y < 0) y = 0;

    			if (((x != oldx) || (Math.abs(y - oldy) > thresh)) &&
    					npoints < MAX_DISP_BUFF)
    			{
    				xvalues[npoints] = x;
    				yvalues[npoints] = y;
    				npoints++;
    				oldx = x;
    				oldy = y;
    			}
    		}
    	}
    	catch (Exception ex)
    	{            
    		Util.show_and_log("draw_spect() " + ex.getMessage());
    		ex.printStackTrace();
    		return;
    	}

//	  	draw Spectrum 
    	int[] pts = new int[2 * npoints];
    	  	
    	for(i = 0; i < npoints; i++) 
    	{ 		
    		pts[2 * i] = xvalues[i];   		
    		pts[2 * i + 1] = yvalues[i];   		
    	}
    		  	 
    	if(gp.stack_whitewash)
    	{
    		int[] vertices = new int[2 * npoints + 4];
    		int min = 0;	// actually a maximum since origin is upper left and + y is down
    		for(i = 0; i < npoints; i++)
    		{
    			vertices[2 * i] = xvalues[i];
    			vertices[2 * i + 1] = yvalues[i];
    			if(min < yvalues[i]) min = yvalues[i];
    		}    		
    		vertices[2 * npoints] = vertices[2 * npoints - 2];
        	vertices[2 * npoints + 1] = min;
        	vertices[2 * npoints + 2] = vertices[0];
        	vertices[2 * npoints + 3] = min; 
	    	gr.setBackground(gp.back_colour);
	    	gr.fillPolygon(vertices);
    	}
    	
    	if(npoints > 2)
    		gr.drawPolyline(pts);   // polyLine requires single array with alternating x, y values
    		   	
    }	//end draw_spect
    
//-------------------------------------------------------------------------------------------------
// draw the x axis and the phase pivot point, and label shifts
// and couplings if doing a simulation. Used for both 1D and 2D display modes
	
    public static void draw_axis(GC gr, double f1, double f2,
    		int x_offset, int y_offset, int xsize, global_par gp, data_par q, data_par q_ind)
    {
    	
     	double range, tic_space = 0.1, first_tic, tic, pow10;
    	int x, y, old_x;
    	int i, iexp;
    	String number;
    	double unitperpix;

    	if (Math.abs(f1 - f2) < 0.000001)       // kludge
    	{
    		Util.log_error("draw axis(): freq range too small");
    		return;
    	}

    	// pass font as agrument???
    	gr.setFont(new Font(SW_main.shell.getDisplay(), "Arial", SW_main.MEDIUM_FONT, SWT.NORMAL));	// set font
    	gr.setForeground(gp.axis_colour);	// set axis colour
    	
    	if(gp.axis_unit == SW_main.units.PPM)
    	{
    		if (gp.rotate_2D && (gp.display_mode == SW_main.modes.MODE_2D))
    		{
    			f1 = f1 / q_ind.SF;
    			f2 = f2 / q_ind.SF;
    		}
    		else
    		{
    			f1 = f1 / q.SF;
    			f2 = f2 / q.SF;
    		}
    	}

    	iexp = 0;
    	range = f1 - f2;
    	unitperpix = (f1 - f2)/(double)(xsize-1);

    	// scale range to between 1 to 10, keep track of power of 10 
    	if(range >= 1.0)
    		while(range >= 10.0)
    		{
    			range = range/10.0;
    			iexp++;
    		}

    	else		//(range < 1.0)
    		while(range < 1.0)
    		{
    			range = range*10.0;
    			iexp--;
    		}

    	if(Math.abs(f1) >= 100000.0 || xsize < 300) 
    	{
    		if(range < 1.0 && range >= 0.5)	tic_space = 0.1;			
    		else if(range < 2.0 && range >= 1.0)	tic_space = 0.2;	
    		else if(range < 5.0 && range >= 2.0)	tic_space = 0.5;	
    		else if(range < 10.0 && range >= 5.0)	tic_space = 1.0;
    	}
    	else if(Math.abs(f1) >= 10000.0 || xsize < 600)  
    	{
    		if(range < 1.0 && range >= 0.5)	tic_space = 0.05;			
    		else if(range < 2.0 && range >= 1.0)	tic_space = 0.1;	
    		else if(range < 5.0 && range >= 2.0)	tic_space = 0.2;	
    		else if(range < 10.0 && range >= 5.0)	tic_space = 0.5;
    	}
    	else
    	{
    		if(range < 1.0 && range >= 0.5)	tic_space = 0.02;			
			else if(range < 2.0 && range >= 1.0)	tic_space = 0.05;	
			else if(range < 5.0 && range >= 2.0)	tic_space = 0.1;	
			else if(range < 10.0 && range >= 5.0)	tic_space = 0.2;	
    	}

    	pow10 = Math.pow(10.0, (double)iexp);

    	// calculate real tic spacing and rational first tic position 
    	double ts = tic_space;
    	tic_space = tic_space * pow10;
    	first_tic = tic_space * ((int) Math.round((f1 - tic_space) / tic_space)); 
    	
    	double ft = first_tic/pow10;

    	// set up graphics 
    	if(gp.display_mode == SW_main.modes.MODE_1D)
    		x = x_offset;
    	else
    		x = gp.left_2D + x_offset;

    	y = gp.dysize - y_offset;

    	if(gp.display_mode == SW_main.modes.MODE_1D)
    	{
    		if(gp.axis_unit == SW_main.units.HZ) 
    			gr.drawText("Hz", 3, y + 10);
    		else if (gp.axis_unit == SW_main.units.PPM)
    			gr.drawText("PPM", 3, y + 10);
    	}
    	else
    	{
    		String str = "";

    		if (gp.rotate_2D)
    		{
    			if (q.data_mode == SW_main.modes.MODE_3D &&
    					gp.mode_3D == SW_main.modes_3D.XY)
    			{
    				if (gp.axis_unit == SW_main.units.HZ)
    					str = "Hz Y (F2)";
    				else if (gp.axis_unit == SW_main.units.PPM)
    					str = "PPM Y (F2)";
    			}
    			else if (q.data_mode == SW_main.modes.MODE_3D &&
    					gp.mode_3D == SW_main.modes_3D.XZ)
    			{
    				if (gp.axis_unit == SW_main.units.HZ)
    					str = "Hz Z (F1)";
    				else if (gp.axis_unit == SW_main.units.PPM)
    					str = "PPM Z (F1)";
    			}
    			else  // 2D
    			{
    				if (gp.axis_unit == SW_main.units.HZ)
    					str = "Hz (F1)";
    				else if (gp.axis_unit == SW_main.units.PPM)
    					str = "PPM (F1)";
    			}
    		}
    		else // non-rotated
    		{
    			if (q.data_mode == SW_main.modes.MODE_3D)
    			{
    				if (gp.axis_unit == SW_main.units.HZ)
    					str = "Hz X (F3)";
    				else if (gp.axis_unit == SW_main.units.PPM)
    					str = "PPM X (F3)";
    			}
    			else  // 2D
    			{
    				if (gp.axis_unit == SW_main.units.HZ)
    					str = "Hz (F2)";
    				else if (gp.axis_unit == SW_main.units.PPM)
    					str = "PPM (F2)";
    			}
    		}
    		gr.drawText(str, gp.left_2D + x_offset, y + 10);
    	}

    	tic = first_tic;

    	// now draw the axis, every second tic mark labelled 

    	if (tic_space < 0.001 || tic_space > 10000.0)
    	{
    		if ((ft / (4.0 * ts) - (int) Math.round(ft / (4.0 * ts))) > 0.001)
    			i = 1;
    		else
    			i = 0;
    	}
    	else
    	{
    		if ((ft / (2.0 * ts) - (int) Math.round(ft / (2.0 * ts))) > 0.001)
    			i = 1;
    		else
    			i = 0;
    	}
    	
    	int xmax;
    	if(gp.display_mode == SW_main.modes.MODE_1D) xmax = xsize;
    	else xmax = xsize + gp.left_2D + x_offset;

    	while(x < xmax)
    	{
    		old_x = x;

    		try
    		{
    			if (gp.display_mode == SW_main.modes.MODE_1D)
    				x = (int) Math.round((f1 - tic) / unitperpix);
    			else
    				x = (int) (Math.round((f1 - tic) / unitperpix) + gp.left_2D + x_offset);
    		}
    		catch (Exception ex)
    		{
    			Util.show_and_log("draw_axis(), bad x " + ex.getMessage());
    			return;
    		}

    		if(x >= xmax) x = xmax;

    		gr.drawLine(old_x, y, x, y);

    		if(i == 0)
    		{
    			gr.drawLine(x, y, x, y+10);

    			int right, left;
    			if(gp.display_mode == SW_main.modes.MODE_1D) 
    			{
    				right = xmax - 20;
    				left = 70;
    			}
    			else
    			{
    				right = xmax;
    				left = gp.left_2D + 80 + x_offset;
    			}

    			// decide on amount of right padding to centre number under tic
    			if(x > left  &&  x < right)
    			{
    				if(tic_space < 0.001)
    					number = String.format("%.4f" , tic);
    				else if(tic_space < 0.01)
    					number = String.format("%.3f" , tic);
    				else if(tic_space <0.1)
    					number = String.format("%.2f" , tic);
    				else if(tic_space < 1.0 )
    					number = String.format("%.1f" , tic);
    				else if(tic_space <10.0 )
    					number = String.format("%.0f" , tic);
    				else if(tic_space < 100.0 )
    					number = String.format("%.0f" , tic);
    				else if(tic_space < 1000.0 )
    					number = String.format("%.0f" , tic);
    				else if(tic_space < 10000.0 )
    					number = String.format("%.0f" , tic);
    				else if(tic_space < 100000.0 )
    					number = String.format("%.0f" , tic);
    				else
    					number = String.format("%.0f" , tic);

    				number = number.trim();							// remove extra whitespace
    				Point size = gr.textExtent(number);				// and centre number
    				gr.drawText(number, x - size.x / 2, y+10);
    			}
    			i=1;
    		}
    		else
    		{ 
    			gr.drawLine(x, y, x, y+5);
    			i = 0;
    		}
    		tic = tic-tic_space;
    	}		// end while

    	// now clean up 
    	gr.drawLine(x, y, xmax, y);

    	// show phase pivot point if in phase mode 
    	if(gp.phase_mode)
    	{
    		gr.setForeground(SW_main.shell.getDisplay().getSystemColor(SWT.COLOR_DARK_YELLOW));

    		try
    		{
    			if (gp.axis_unit == SW_main.units.HZ)
    				x = (int) Math.round((f1
    						- (q.offset - (double)gp.pivot * q.hzperpt)) / unitperpix);

    			else if (gp.axis_unit == SW_main.units.PPM)
    				x = (int) Math.round((f1
    						- (q.offset - (double)gp.pivot * q.hzperpt) / q.SF) / unitperpix);
    		}
    		catch (Exception ex)
    		{
    			Util.show_and_log("draw_axis(), bad pivot x " + ex.getMessage());
    			return;
    		}

    		if(gp.display_mode == SW_main.modes.MODE_2D) 
    			x = x + gp.left_2D + x_offset;

    		gr.setForeground(new Color(Display.getCurrent(), 255, 100, 100));
			gr.setLineWidth(3);			
    		gr.drawLine(x, y + gp.axis_offset, x, gp.dysize - gp.axis_offset);
    	}
    	
    	if(gp.bpt_mode)	// mark baseline points if in baseline point mode
    	{
    		gr.setForeground(gp.colour_magenta);
    		for(i = 0; i < gp.n_bl_pts; i++)
    		{	  			
    			x = (int) Math.round((gp.freq1 - (q.offset - (gp.baseline_x[i] * q.hzperpt))) 
    					/ gp.hzperpix);
    			gr.drawLine(x, gp.dysize - gp.axis_offset,
    					x, gp.dysize - gp.axis_offset - 20);
    		}
    	}
    }	// end draw_axis
    
//------------------------------------------------------------------------------------------------
// TODO needs work, draw the Y axis for 1D

 	private static void draw_y_axis(global_par gp,
 			GC gr, double y1, double y2, int x_offset, int ysize)
 	{
 		double range, tic_space = 0.1, first_tic, tic, pow10;
 		int x, y, old_y;
 		int i, iexp;
 		String number;
 		double unitperpix;
 			   
 		// pass font as agrument???
    	gr.setFont(new Font(SW_main.shell.getDisplay(), "Arial", SW_main.MEDIUM_FONT, SWT.NORMAL));					// set font
    	gr.setForeground(gp.axis_colour);	// sex axis colour

 		iexp = 0;
 		range = y2 - y1;
 		unitperpix = range / (double)(ysize  - gp.axis_offset - 1);

 		// scale range to between 1 to 10, keep track of power of 10 

 		if(range >= 1.0)
 			while(range >= 10.0)
 			{
 				range = range/10.0;
 				iexp++;
 			}
 		else		//(range < 1.0)
 			while(range < 1.0)
 			{
 				range = range*10.0;
 				iexp--;
 			}

		if(range < 1.0 && range >= 0.5)	tic_space = 0.02;			
		else if(range < 2.0 && range >= 1.0)	tic_space = 0.05;	
		else if(range < 5.0 && range >= 2.0)	tic_space = 0.1;	
		else if(range < 10.0 && range >= 5.0)	tic_space = 0.2;	
		
 		pow10 = Math.pow(10.0, (double)iexp);

 		// calculate real tic spacing and rational first tic position 
 		double ts = tic_space;
 		tic_space = tic_space * pow10;

 		first_tic = tic_space * (Math.round((y1 + tic_space) / tic_space));// ??????  why round??

 		double ft = first_tic/pow10;

 		// set up graphics, initial location
 		x = x_offset;
 		y = gp.dysize - gp.axis_offset;

 		tic = first_tic;
 		//tic = 0.0;

 		if (tic_space < 0.001 || tic_space > 10000.0)
 		{
 			if ((ft / (4.0 * ts) - Math.round(ft / (4.0 * ts))) > 0.001)
 				i = 1;
 			else
 				i = 0;
 		}
 		else
 		{
 			if ((ft / (2.0 * ts) - Math.round(ft / (2.0 * ts))) > 0.001)
 				i = 1;
 			else
 				i = 0;
 		}
 		
 		while(y1 + tic < y2)
 		{
 			old_y = y;
 			y = ysize - (int)Math.round((y1 + tic) / unitperpix) - gp.axis_offset - gp.expt_y_offset; 

 			gr.drawLine(x, old_y, x, y);

 			if(i == 0)
 			{
 				gr.drawLine(x, y, x-10, y);
 				if(y >  30 &&  y < ysize - 20)
 				{
 					// decide on amount of right padding to centre number under tic
 					if(tic_space < 0.0001)
 						number = String.format("%.4f", tic);
 					else if(tic_space < 0.001)
 						number = String.format("%.4f", tic);
 					else if(tic_space < 0.01)
 						number = String.format("%.3f", tic);
 					else if(tic_space <0.1)
 						number = String.format("%.2f", tic);
 					else if(tic_space < 1.0 )
 						number = String.format("%.1f", tic);
 					else if(tic_space <10.0 )
 						number = String.format("%.0f", tic);
 					else if(tic_space < 100.0 )
 						number = String.format("%.0f", tic);
 					else if(tic_space < 1000.0 )
 						number = String.format("%.0f", tic);
 					else if(tic_space < 10000.0 )
 						number = String.format("%.0f", tic);
 					else if(tic_space < 100000.0 )
 						number = String.format("%.0f", tic);
 					else
 						number = String.format("%.0f", tic);

 					number = number.trim();							// remove white space
 					Point size = gr.textExtent(number);				// and centre number   				
 					gr.drawText(number, x - 15 - size.x, y - size.y / 2);
 				}
 				i=1;
 			}
 			else
 			{ 
 				gr.drawLine(x, y, x-5, y);
 				i=0;
 			}
 			tic = tic + tic_space;
 		}		// end while

 		gr.drawLine(x, y, x, gp.top_2D);	// now clean up 
 		gr.drawLine(x_offset, gp.top_2D, gp.xsize_2D + x_offset, gp.top_2D);
 		gr.drawLine(gp.xsize_2D + x_offset, gp.top_2D, 
 				gp.xsize_2D + x_offset, gp.dysize - gp.bottom_2D);

 	} // End of draw_y_axis()
//-------------------------------------------------------------------------------------------------
//  Draw the FID

    private static void draw_fid(GC gr, double t1, double t2, double yscale,
    		int xoffset, int yoffset, int xsize, int ysize,
    		double[] array, int thresh, Color color,
    		global_par gp, data_par q)
    {
    	if (array == null)
    	{
    		Util.show_and_log("   draw_fid(): FID array is null");
    		return;
    	}

    	if (gr == null)
    	{
    		Util.show_and_log("draw_fid(), null graphics?? ");
    		return;
    	}

    	int point1, point2;		// start and stop points in data array 
    	int x, y;				// display x and y coord. 
    	int oldx, oldy;			// old x and y values 
    	int npoints;            // number of points to be plotted
    	int i;                  // handy counter

    	int[] xvalues = new int[MAX_DISP_BUFF];		// for speed, perhaps 
    	int[] yvalues = new int[MAX_DISP_BUFF];     // do this one time at start-up???

    	gr.setForeground(color);
    	double secperpix = (t2 - t1) / (double)(xsize - 1);


    	point1 = (int) Math.round(t1 / q.dwell);
    	point2 = (int) Math.round(t2 / q.dwell);

    	// make sure array bounds are not exceeded
    	if (point1 < 0) point1 = 0;
    	if (point2 >= q.TD/2) point2 = q.TD/2 - 1;

    	if (point2 >= array.length || point1 >= array.length)
    	{
    		Util.show_and_log("draw_fid(), Space allocation error, SI > array length ");
    		return;
    	}

    	if (secperpix <= 1.0e-10)
    	{
    		Util.show_and_log("draw_fid() sec per Pix too small!");
    		return;
    	}

    	// initialize point counter 
    	npoints = 0;

    	y = ysize - (((int)(array[point1] * yscale)) - yoffset);

    	if (y > (ysize - yoffset - gp.axis_offset - 2))
    		y = ysize - gp.axis_offset - 2;

    	if (y < 0) y = 0;

    	oldx = xoffset;
    	oldy = y;

    	try
    	{
    		for (i = point1; i <= point2; i++)
    		{
    			x = (int) (Math.round((i * q.dwell - t1) / secperpix) + xoffset);                                     
    			y = ysize - ((int)(array[i] * yscale)) - yoffset;

    			if (y > (ysize - gp.axis_offset - 2))
    				y = ysize - gp.axis_offset - 2;

    			if (y < 0) y = 0;

    			if (((x != oldx) || (Math.abs(y - oldy) > thresh)) &&
    					npoints < MAX_DISP_BUFF)
    			{
    				xvalues[npoints] = x;
    				yvalues[npoints] = y;
    				npoints++;
    				oldx = x;
    				oldy = y;
    			}
    		}
    	}
    	catch (Exception ex)
    	{
    		Util.show_and_log("draw_fid() " + ex.getMessage());
    		return;
    	}

    	//	draw FID 
    	int[] pts = new int[2 * npoints];

    	for(i = 0; i < npoints; i++) 
    	{
    		pts[2 * i] = xvalues[i];
    		pts[2 * i + 1] = yvalues[i];
    	}

    	if(npoints > 2)
    		gr.drawPolyline(pts);   // polyLine requires single array with alternating x, y values

    	//	clean up
    }
       
       
//-------------------------------------------------------------------------------------------------        
    private static void draw_fid_axis(GC gr, double t1, double t2, int axis_offset,
    		int xsize, global_par gp, data_par q)
    {
    	double range, tic_space = 0.1, first_tic, tic, pow10;
    	int x, y, old_x;
    	int i, iexp;
    	String number = "";

    	if (q.TD == 0) return;  // avoid possible race condition

    	double unitperpix;;

    	gr.setFont(new Font(SW_main.shell.getDisplay(), "Arial", SW_main.MEDIUM_FONT, SWT.NORMAL));	// set font
    	gr.setForeground(gp.axis_colour);	// sex axis colour

    	iexp = 0;
    	range = t2 - t1;
    	unitperpix = range / (double)(xsize - 1);

    	//scale range to between 1 to 10, keep track of power of 10 
    	if (range >= 1.0)
    		while (range >= 10.0)
    		{
    			range = range / 10.0;
    			iexp++;
    		}

    	else		//(range < 1.0)
    		while (range < 1.0)
    		{
    			range = range * 10.0;
    			iexp--;
    		}

    	if (range < 1.0 && range >= 0.5) tic_space = 0.02;
    	else if (range < 2.0 && range >= 1.0) tic_space = 0.05;
    	else if (range < 5.0 && range >= 2.0) tic_space = 0.1;
    	else if (range < 10.0 && range >= 5.0) tic_space = 0.2;

    	pow10 = Math.pow(10.0, (double)iexp);

    	//calculate real tic spacing and rational first tic position 
    	double ts = tic_space;
    	tic_space = tic_space * pow10;

    	first_tic = tic_space * (Math.round((t1 + tic_space) / tic_space));

    	double ft = first_tic / pow10;

    	//set up graphics 
    	x = 0;
    	y = gp.dysize - axis_offset; 

    	gr.drawText("Sec", 3, y + 10);
    	tic = first_tic;

    	//now draw the axis, every second tic mark labelled 

    	if (Math.abs(ft / (2.0 * ts) - Math.round(ft / (2.0 * ts))) > 0.001)
    		i = 1;
    	else
    		i = 0;

    	int xmax = gp.dxsize;

    	while (x < xmax)
    	{
    		old_x = x;
    		x = (int) Math.round((tic - t1) / unitperpix);
    		if (x >= xmax) x = xmax;

    		gr.drawLine(old_x, y, x, y);

    		if (i == 0)
    		{
    			gr.drawLine(x, y, x, y + 10);

    			int temp, temp2;
    			temp = xmax - 20;
    			temp2 = 70;

    			if (x > temp2 && x < temp)
    			{
    				//decide on amount of right padding to centre number under tic
    				if (tic_space < 0.0001)
    					number = String.format("%.5f", tic);
    				else if (tic_space < 0.001)
    					number = String.format("%.4f", tic);
    				else if (tic_space < 0.01)
    					number = String.format("%.3f", tic);
    				else if (tic_space < 0.1)
    					number = String.format("%.2f", tic);
    				else if (tic_space < 1.0)
    					number = String.format("%.1f", tic);
    				else if (tic_space < 10.0)
    					number = String.format("%.4f", tic);
    				else if (tic_space < 100.0)
    					number = String.format("%f", tic);
    				else if (tic_space < 1000.0)
    					number = String.format("%f", tic);
    				else
    					number = String.format("%f", tic);

    				String label = number.trim();
    				gr.drawText(label, x, y + 10);
    			}
    			i = 1;
    		}
    		else
    		{
    			gr.drawLine(x, y, x, y + 5);
    			i = 0;
    		}
    		tic = tic + tic_space;
    	}		// end while
    	gr.drawLine(x, y, xmax, y);
    } 
        
//-------------------------------------------------------------------------------------------------
//  Draw the 1D Peaks
    
    private static void draw_peaks(GC gr, global_par gp, data_par q)
    {
    	int[] datax = new int[PeakPick.MAX_PEAKS];   // x position for data mark
    	int[] labelx = new int[PeakPick.MAX_PEAKS];  // x position for label
    	int datay;                                   // y position for data mark
    	int data_index;                              // array index
    	int i;
    	String str;

    	gr.setForeground(gp.axis_colour);
    	Font peaks_font = new Font(SW_main.shell.getDisplay(), "Arial", SW_main.SMALL_FONT, SWT.NORMAL);
    	gr.setFont(peaks_font);	

    	//  set format string for number of digits
    	String pp_digits_str = "%." + String.format("%d", gp.pp_digits) + "f";
   	
    	//  make a mark for any "picked peaks" 
    	for (i = 0; i < gp.n_peaks; i++)
    	{
    		datax[i] = (int) Math.round((gp.freq1 - gp.peaks[i].freq) / gp.hzperpix);
    		data_index = (int) Math.round((q.offset - gp.peaks[i].freq) / q.hzperpt);
    		
    		labelx[i] = datax[i];

    		if (data_index >= 0 && data_index < q.SI && datax[i] > 0 && datax[i] < gp.dxsize)
    		{
    			datay = gp.dysize
    					- (int)(gp.data_array[data_index] * gp.expt_scale)
    					- gp.expt_y_offset;

    			if (datay <= 100) datay = 101;
    			gr.drawLine(datax[i], datay - 15, datax[i], datay - 30);  
    			
    			// identify any peaks with labels
    			if(gp.peaks[i].label.length() > 0 &&  gp.peaks[i].label != null)
    			{
    				try
    				{
    					Image textImage = RotateText.createRotatedText(gp.peaks[i].label, 
    							peaks_font, gp.axis_colour, gp.back_colour, SWT.NORMAL);
    					gr.drawImage(textImage, datax[i] - SW_main.SMALL_FONT / 2 - 2, 
    							datay - textImage.getBounds().height - 40);
    					
    					if(SW_main.currentOS == SW_main.osTypes.MACOS)	// fix for small MacOS graphics bug
    	    			{												// (spurious whie lines at edge of label)
    	    				gr.setForeground(gp.back_colour);
    	    				gr.drawLine(
    	    						datax[i] - SW_main.SMALL_FONT / 2 - 2, 
    	    						datay - textImage.getBounds().height - 40, 
    	    						datax[i] - SW_main.SMALL_FONT / 2 - 2, 
    	    						datay  - 40 );
    	    				
    	    				gr.drawLine(
    	    						datax[i] - SW_main.SMALL_FONT / 2 - 2, 
    	    						datay  - 41, 
    	    						datax[i] - SW_main.SMALL_FONT / 2 - 2 + textImage.getBounds().width, 
    	    						datay - 41);
    	    				
    	    				gr.setForeground(gp.axis_colour);	// restore colour
    	    			}
    					
    					textImage.dispose();	
    				}
    				catch(Exception ex) {	ex.printStackTrace();	}
    			}
    		}
    	}

    	boolean changed = true;
    	int count = 0;

    	//  Adjust spacing to prevent overlap of labels
    	while (changed == true && count < 1200)
    	{
    		changed = false;
    		for (i = 0; i < gp.n_peaks - 1; i++)
    		{
    			if (Math.abs(labelx[i + 1] - labelx[i]) < SW_main.SMALL_FONT + 5)  // overlap
    			{
    				labelx[i] = labelx[i] - 1;  // move label left 1 pixel
    				changed = true;
    				break;
    			}
    		}

    		for (i = gp.n_peaks - 1; i > 0; i--)
    		{
    			if (Math.abs(labelx[i - 1] - labelx[i]) < SW_main.SMALL_FONT + 5)  // overlap
    			{
    				labelx[i] = labelx[i] + 1;  // move label right 1 pixel
    				changed = true;
    				break;
    			}
    		}
    		count++;
    	}
    		      	      	    
    	// now label the peaks 
    	// TODO need some adjustments for MacOS here, expecially for larger fonts
    	for (i = 0; i < gp.n_peaks; i++)
    	{
    		if (labelx[i] > 0 && labelx[i] < gp.dxsize)
    		{
    			if (gp.peak_unit == SW_main.units.PPM)
    				str = String.format(pp_digits_str, gp.peaks[i].freq / q.SF);
    			else
    				str = String.format(pp_digits_str, gp.peaks[i].freq);
    			
    			Image textImage = RotateText.createRotatedText(str, peaks_font, 
    					          gp.axis_colour, gp.back_colour, SWT.NORMAL);
    			gr.drawImage(textImage, labelx[i] - SW_main.SMALL_FONT / 2 - 2, 35);
    			
    			if(SW_main.currentOS == SW_main.osTypes.MACOS)	// fix for small MacOS graphics bug
    			{												// (spurious white likes on rotated labels)
    				gr.setForeground(gp.back_colour);
    				gr.drawLine(
    						labelx[i] - SW_main.SMALL_FONT / 2 -2, 
    						35, 
    						labelx[i] - SW_main.SMALL_FONT / 2 - 2, 
    						35 + textImage.getBounds().height);
    				
    				gr.drawLine(
    						labelx[i] - SW_main.SMALL_FONT / 2 -2, 
    						35 + textImage.getBounds().height -1, 
    						labelx[i] - SW_main.SMALL_FONT / 2 - 2 + textImage.getBounds().width, 
    						35 + textImage.getBounds().height -1);
    				
    				gr.setForeground(gp.axis_colour);	// restore colour
    			}
    			
    			textImage.dispose();

    			int offset = 45 + str.length() * SW_main.SMALL_FONT;

    			gr.drawLine(datax[i], offset, datax[i], offset - 10);
    			gr.drawLine(datax[i], offset - 10, labelx[i], offset - 20);
    			gr.drawLine(labelx[i], offset - 20, labelx[i], offset - 30);
    		}
    	}
    }
            
//-------------------------------------------------------------------------------------------------
// Draw 1D integrals

    static void draw_integ(GC gr, double f1, double f2, double yscale, 
    		int yoffset, global_par gp, data_par q)
    {
    	if (gp.data_array == null) return;

    	double ratio;       // ratio of actual to display digital resolution 
    	double hzperpix;
    	int point1, point2; // start and stop points in data array
    	int point1a;        // first point in array, uncorrected
    	int start, stop;    // start and stop points in integral 
    	int x,y;            // display x and y coord. 
    	int npoints;
    	int i, j;
    	double integral;
    	String number;
    	  	
    	Font integ_font = new Font(SW_main.shell.getDisplay(), 
    			"Arial", SW_main.SMALL_FONT, SWT.NORMAL);
    	gr.setFont(integ_font);

    	int[] xvalues = new int[MAX_DISP_BUFF];
    	int[] yvalues = new int[MAX_DISP_BUFF];

    	hzperpix = (f1-f2)/(double)(gp.dxsize-1);
    	ratio = q.hzperpt/hzperpix; // pix per pt 

		point1 = (int) Math.round((q.offset - f1) / q.hzperpt);
		point2 = (int) Math.round((q.offset - f2) / q.hzperpt);
    	
    	// make sure array bounds are not exceeded 
    	point1a = point1;   // if left end of spectrum outside array
    	if(point1 < 0) point1 = 0;
    	if(point2 >= q.SI) point2 = q.SI-1;

    	if (point2 >= gp.data_array.length)
    	{
    		Util.show_and_log("draw_integ(), Space allocation error, SI > array length  ");
    		return;
    	}

    	if (point1 >= gp.data_array.length)
    		return;

    	for(i=0; i < gp.n_integrals; i++)
    	{
    		if (gp.integrals[i] == null) break;

    		if (gp.integrals[i].start > -1.0e30)
    		{
    			
				start = (int) Math.round((q.offset - gp.integrals[i].start) / q.hzperpt);
				if (start < 0) start = 0;

				stop = (int) Math.round((q.offset - gp.integrals[i].end) / q.hzperpt);
				if (stop > q.SI - 1) stop = q.SI - 1;
    			
    			integral = 0.0; npoints = 0;

    			for(j=start; j < stop; j++)
    			{
    				x = (int) Math.round((double)(j - point1a) * ratio);

    				integral = integral + gp.data_array[j] 
    						+ gp.integrals[i].bias
    						+ gp.integrals[i].slope * ((double)j - (double)start);

    				y = gp.dysize-((int) (integral * yscale))-yoffset;


    				if (y > (gp.dysize - 2))	y = gp.dysize - 2;
    				if (y < 2)	y = 2;

    				if((npoints < MAX_DISP_BUFF) && (x > 0) && (x < gp.dxsize))
    				{  
    					xvalues[npoints] = x;
    					yvalues[npoints] = y;
    					npoints++;
    				}
    			}

    			gp.integrals[i].value = integral;

    			if(gp.n_integrals == 1 && !gp.integ_calibrated)
    			{
    				gp.integ_cal = integral;
    			}

    			// paranoid divide by zero check 
    			if(gp.integ_cal > 1.0 || gp.integ_cal < -1.0) 
    				gp.integrals[i].scaled = integral/gp.integ_cal;      

    			int oldx = xvalues[0];		int oldy = yvalues[0];
    			for(j = 1; j < npoints; j++)
    			{
    				if(i == gp.current_integral) // red
    				{
    					gr.setForeground(gp.active_integ_colour);
    					gr.drawLine(oldx, oldy, xvalues[j], yvalues[j]);	
    				}
    				else
    				{
    					gr.setForeground(gp.integ_colour);
    					gr.drawLine(oldx, oldy, xvalues[j], yvalues[j]);
    				}

    				oldx = xvalues[j]; oldy = yvalues[j];
    			}

    			String int_digits_str = "%." + String.format("%d", gp.int_digits) + "f";
    			number = String.format(int_digits_str, gp.integrals[i].scaled);

    			String label = number.trim();

    			gr.setForeground(gp.integ_colour);
    			
    			Image textImage = RotateText.createRotatedText(label, gr.getFont(), 
				          gp.integ_colour, gp.back_colour, SWT.NORMAL);
    			gr.drawImage(textImage, (int)((((start-point1a) + (stop-point1a))/2) * ratio) 
    					- SW_main.SMALL_FONT / 2, gp.dysize - 97);
    			textImage.dispose();
		
    			if(i == gp.current_integral) // red
    			{
    				gr.setForeground(gp.active_integ_colour);
    				gr.drawLine((int)((start - point1a) * ratio), gp.dysize - gp.axis_offset - 1,
    						(int)((stop - point1a) * ratio), gp.dysize - gp.axis_offset - 1);
    			}
    			else // normal colour
    			{
    				gr.setForeground(gp.integ_colour);
    				gr.drawLine((int)((start - point1a) * ratio), gp.dysize - gp.axis_offset - 1,
    						(int)((stop - point1a) * ratio), gp.dysize - gp.axis_offset - 1);
    			}
    		}
    	}
    	integ_font.dispose();   
    }
    
//--------------------------------------------------------------------------------------------------
//	Draw the set of stack traces

    private static void draw_stack(GC gr, global_par gp)
    {
    	int j;
    	int dxsize = gp.stack_x_size;

    	if (gp.stack_x_size_auto)
    		dxsize = gp.dxsize - gp.stack_x_offset * gp.n_traces;
    	else
    		dxsize = gp.stack_x_size;
    	
    	for(int i = 0; i < MAX_DISP_BUFF; i++)
    		max_yvalues[i] = 0;

    	/*for (int i = 0; i < gp.n_traces; i++)*/
    	for(int i = gp.n_traces - 1; i >= 0; i--)
    	{
    		String label;
    		if (gp.stack_label_type == 0)
    			label = "";
    		else if (gp.stack_label_type == 1)
    			label = String.format("%d", i + 1);
    		else if (gp.stack_label_type == 2)
    			label = gp.stack_trace[i].title;
    		else if (gp.stack_label_type == 3)
    			label = gp.stack_trace[i].folder;
    		else
    			label = "";

    		if (gp.show_spectrum == false || gp.stack_y_offset == 0)
    			j = i;
    		else
    			j = i + 1;

    		int off;
    		if (gp.stack_y_offset == 0)
    			off = 50;
    		else
    			off = gp.stack_y_offset * (j);
    		
    		gr.setForeground(gp.axis_colour);
    		gr.drawText(label, gp.stack_x_offset * (j), 
                    gp.dysize - gp.expt_y_offset - off - 20);

    		if (gp.stack_trace[i] != null)
    			draw_spect(gr, gp.freq1, gp.freq2,
						gp.stack_x_offset * (j),
						gp.expt_y_offset + off,
						dxsize, gp.dysize, gp.stack_trace[i], 2,
						gp.stack_trace[i].colour,
						gp);
    	}
    }
    
//--------------------------------------------------------------------------------------------------
//  This function sets the transition cursor on the calculated spectrum at the 
//  peak of index n. 
//  For Simpson and other situations where no transitions are
//  actually used then no cursor is drawn.
		
    public static void set_cursor(int n, GC gr, global_par gp, SimPar sp, data_par q)
    {
    	int x, y, curx, cury, i;

    	//	fail-safe bail-out
    	if(sp.calc_mode != SimPar.HIRES || gp.calc_array == null || sp.n_trans == 0) return;

    	//  set up font, etc.		
    	gr.setForeground(gp.axis_colour);
    	Font font = new Font(SW_main.shell.getDisplay(), "Arial", SW_main.SMALL_FONT, SWT.NORMAL);
    	gr.setFont(font);	

    	// check that n is a valid transition 
    	if(n < 0) n = 0;
    	if(n > sp.n_trans) n = sp.n_trans - 1;

    	// calculate cursor x and y coordinates 
    	try
    	{
    		x = (int) Math.round((gp.freq1 - gp.calc_peaks[n].calc_freq)
    				/ gp.hzperpix);     // x screen coordinate

    		i = (int) Math.round((q.offset - gp.calc_peaks[n].calc_freq)
    				/ q.hzperpt);	    // index in array
    	}
    	catch (Exception ex)
    	{
    		Util.show_and_log("set_cursor() bad x or i " + ex.getMessage());
    		return;
    	}

    	if(i < 0) i = 0;
    	if(i >= q.SI) i = q.SI-1;

    	y = gp.dysize - (int)((double)gp.calc_array[i] * gp.calc_scale) 
    			- gp.calc_y_offset - 40;

    	// position the cursor 
    	curx = x; cury = y;					// add the cursor offsets 
    	if(cury < 20 ) cury = 20;

    	if(curx > gp.dxsize && (gp.calc_peaks[n].calc_freq < q.offset) && (gp.calc_peaks[n].calc_freq
    			> (q.offset-(q.SW*q.SF))))	// cursor outside display (right)
    	{
    		gr.drawText(">>>", gp.dxsize-15, gp.dysize - gp.calc_y_offset - 20);
    		
    		if(out_of_sw)
    		{
    			out_of_sw = false;
    		}
    	}

    	else if(curx < 0 && (gp.calc_peaks[n].calc_freq < q.offset) &&
    			(gp.calc_peaks[n].calc_freq > (q.offset-(q.SW*q.SF))))  //outside display (left)
    	{
    		gr.drawText("<<<", 0, gp.dysize - gp.calc_y_offset - 20);

    		if(out_of_sw)
    		{
    			out_of_sw = false;
    		}
    	}

    	else // O.K. to display cursor
    	{
    		if(gp.calc_peaks[n].calc_freq < q.offset && gp.calc_peaks[n].calc_freq > 
    		(q.offset-(q.SFO1*q.SW)))
    		{					
    			// now display the cursor 
    			gr.drawLine(curx, 60, curx, cury);

    			gr.drawText(String.format("%.3f", gp.calc_peaks[n].calc_freq),
    					curx, 50);

    			//old_x = curx; old_y = cury;

    			if(out_of_sw)
    			{
    				out_of_sw = false;
    			}
    		} 
    		else
    		{
    			gr.drawText("Transition outside spectral window", 
    					(int)(0.7 * (double)gp.dxsize), 10);
    			out_of_sw = true;
    		}
    	}

    	String str;

    	if (!out_of_sw)
    	{
    		if (gp.calc_peaks[n].obs_freq != 0.0)
    			str = String.format("trans. # %d    as. freq = %.3f",
    					gp.calc_peaks[n].trans_no, gp.calc_peaks[n].obs_freq);
    		else
    			str = String.format("trans. # %d    - not assigned -",
    					gp.calc_peaks[n].trans_no);

    		gr.drawText(str, (int)(0.7 * (double)gp.dxsize), 10);
    	}
    	SW_main.statusStripLabel4.setText("  Mouse:  (set cursor,  zoom,  assign line)");
    	font.dispose();	// clean-up
    }

//-------------------------------------------------------------------------------------------------
//	This function draws a series of lines which show the match between 
//  the experimental and calculated spectra. 
    public static void showmatch(GC gr, global_par gp, SimPar sp, data_par q)
    {
    	int calcx, datax;            // x position in of calculated and expt. points 
    	int calcy, datay;            // y position of calculated and expt points 
    	int calc_index, data_index;  // array index in calcuated and experimental										 // arrays 
    	int i;

    	//	bail out if called at an inappropriate time 
    	if (gp.data_array == null ||  gp.calc_array == null 
    			|| sp.calc_mode != SimPar.HIRES) return;

    	//  set up colour, etc.		
    	gr.setForeground(gp.axis_colour);    

    	//	now run through and show the match of all assigned peaks 
    	for(i=0; i < sp.n_trans; i++)
    	{
    		if(gp.calc_peaks[i].obs_freq > 0.000001 
    				|| gp.calc_peaks[i].obs_freq < (-0.000001))
    		{ // a zero frequency means that the peak is unassigned 

    			calcx = (int) Math.round
    					((gp.freq1 - gp.calc_peaks[i].calc_freq) / gp.hzperpix);

    			datax = (int) Math.round
    					((gp.freq1 - gp.calc_peaks[i].obs_freq) / gp.hzperpix);

    			calc_index = (int) Math.round
    					((q.offset - gp.calc_peaks[i].calc_freq) / q.hzperpt);

    			data_index = (int) Math.round
    					((q.offset - gp.calc_peaks[i].obs_freq) / q.hzperpt);

    			if((data_index >=0 && data_index < q.SI) &&
    					(calc_index >=0 && calc_index < q.SI))
    			{
    				calcy = gp.dysize 
    						- (int)(gp.calc_array[calc_index] * gp.calc_scale)
    						- gp.calc_y_offset;

    				datay = gp.dysize 
    						- (int)(gp.data_array[data_index] * gp.expt_scale)
    						- gp.expt_y_offset;

    				if((calcx > 0 && calcx < gp.dxsize) && 
    						(datax > 0 && datax < gp.dxsize))
    				{
    					if (calcy <=  30) calcy = 31;
    					if (datay <=  30) datay = 31;

    					gr.drawLine(calcx, calcy + 25, calcx, calcy + 50);
    					gr.drawLine(calcx, calcy + 50, datax, datay - 50);
    					gr.drawLine(datax, datay - 50, datax, datay - 25);					
    				}
    			}
    		}      // if(peaks[i]... 
    	}      // for i 
    }		
//-------------------------------------------------------------------------------------------------
}	// end class

