package spinworks;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * Utilities for reading processed 2D and 3D NMR pipe data
 * @author marat, 
 * University od Manitoba
 */
public class NMRpipe 
{

//  Modified from original NMRPipe header file, as supplied with NMRPipe

/***/
/* fdatap.h: defines the NMRPipe data header array FDATA, and
/*           outlines some data format details.
/***/

/***/
/* The NMRPipe parameter array FDATA currently consists of 512 4-byte 
/* floating-point values which describe the spectral data.  While all numerical
/* values in this array are floating point, many represent parameters
/* (such as size in points) which are integers.  Some parts of the
/* header contain packed ascii text.
/*
/* There are currently three variations of spectral data in the NMRPipe
/* format:
/*
/*   1. Single-File (1D and 2D): the data are stored in a single
/*      binary file consiting of the header followed by the
/*      spectral intensities, stored in sequential order as
/*      4-byte floats.
/*
/*   2. Multi-File (3D and 4D): the data are stored as a series
/*      of 2D file planes, each with its own complete header
/*      followed by the spectral intensities in sequential order.
/*
/*   3. Data Stream (1D-4D): the data are in the form of a pipeline
/*      stream, with a single header at the beginning followed by
/*      all of the spectral intensities in sequential order.
/* 
/* The header values can be manipulated directly, but this is not
/* recommended.  Instead, the functions getParm() and setParm() can
/* be used to extract or set header values according to parameter
/* codes and the dimension of interest (if any).  See the source
/* code distribution for examples of these functions.
/*
/* The NMRPipe format was created to be compatible with an older format
/* which pre-dates phase-sensitive NMR and multidimensional NMR. 
/* So, for historical reasons, there are some potentially confusing
/* aspects regarding definition of dimension sizes, data types,
/* and interleaving of real and imaginary data.
/*
/* In the NMRPipe nomenclature, the dimensions are called the X-Axis,
/* Y-Axis, Z-Axis, and A-Axis.  Some rules of thumb about the data format
/* follow:
/*
/*  1. Complex data in the X-Axis is stored as separated 1D vectors
/*     of real and imaginary points (see below). 
/*
/*  2. Complex data in the Y-Axis, Z-Axis, and A-Axis is stored as
/*     interleaved real and imaginary points.
/*
/*  3. The X-Axis size is recorded as complex points.
/*
/*  4. The Z-Axis and A-Axis sizes are recorded as total points real+imag.
/*
/*  5. If both the X-Axis and Y-Axis are complex, the Y-Axis size
/*     is reported as total points real+imag.
/*
/*  6. If the X-Axis is not complex but the Y-Axis is complex,
/*     the Y-axis size is reported as complex points.
/*
/*  7. TPPI data, and Bruker QSEQ mode data are treated as real data.
/***/

/***/
/* 1D Real Format:
/*  (512-point FDATA)
/*  (N real points...)
/*
/* 1D Complex Format:
/*  (512-point FDATA)
/*  (N real points...)
/*  (N imag points...)
/*
/* 2D Hypercomplex Format;
/* (direct dimension = t2, indirect dimension = t1):
/*
/*  (512-point FDATA)
/*  (N t2=real points... for t1=1 Real)
/*  (N t2=imag points... for t1=1 Real)
/*  (N t2=real points... for t1=1 Imag)
/*  (N t2=imag points... for t1=1 Imag)
/*  (N t2=real points... for t1=2 Real)
/*  (N t2=imag points... for t1=2 Real)
/*  (N t2=real points... for t1=2 Imag)
/*  (N t2=imag points... for t1=2 Imag)
/*  ... etc ...
/*  (N t2=real points... for t1=M Real)
/*  (N t2=imag points... for t1=M Real)
/*  (N t2=real points... for t1=M Imag)
/*  (N t2=imag points... for t1=M Imag)
/*
/* 3D Hypercomplex format: consists of a series of 2D hypercomplex
/* planes above, which are alternating real and imaginary in the third
/* dimension.
/*
/* 4D Hypercomplex format: consists of a series of 3D hypercomplex 
/* spectra above, which are alternating real and imaginary in the
/* fourth dimension.
/***/

/***/
/* Some useful constant definitions:
/***/

final static int FDATASIZE =          512;   // Length of header in 4-byte float values. 

//const int FDIEEECONS =   0xeeeeeeee;  /* Indicates IEEE floating point format.    */
//const int FDVAXCONS =    0x11111111;  /* Indicates DEC VAX floating point format. */
// VAX???  Really???
final static float FDORDERCONS =       2.345F; /* Constant used to determine byte-order.   */ 
//const int FDFMTCONS =    FDIEEECONS;  /* Floating point format on this computer.  */
//const int ZERO_EQUIV =       -666.0;  /* Might be used as equivalent for zero.    */

/***/
/* General Parameter locations:
/***/

final static int FDMAGIC =        0;   /* Should be zero in valid NMRPipe data.            */
final static int FDFLTFORMAT =    1;   /* Constant defining floating point format.         */
final static int FDFLTORDER =     2;   /* Constant defining byte order.                    */

final static int FDSIZE =        99;   /* Number of points in current dim R|I.             */
final static int FDREALSIZE =    97;   /* Number of valid time-domain pts (obsolete).      */
final static int FDSPECNUM =    219;   /* Number of complex 1D slices in file.             */
final static int FDQUADFLAG =   106;   /* See Data Type codes below.                       */
final static int FD2DPHASE =    256;   /* See 2D Plane Type codes below.                   */

/***/
/* Parameters defining number of dimensions and their order in the data;
/* a newly-converted FID has dimension order (2 1 3 4). These dimension
/* codes are a hold-over from the oldest 2D NMR definitions, where the
/* directly-acquired dimension was always t2, and the indirect dimension
/* was t1.
/***/

final static int FDTRANSPOSED = 221; /* 1=Transposed, 0=Not Transposed.                  */
final static int FDDIMCOUNT =     9; /* Number of dimensions in complete data.           */
final static int FDDIMORDER =    24; /* Array describing dimension order.                */

final static int FDDIMORDER1 =   24; /* Dimension stored in X-Axis.                      */
final static int FDDIMORDER2 =   25; /* Dimension stored in Y-Axis.                      */
final static int FDDIMORDER3 =   26; /* Dimension stored in Z-Axis.                      */
final static int FDDIMORDER4 =   27; /* Dimension stored in A-Axis.                      */

/***/
/* The following parameters describe the data when it is
/* in a multidimensional data stream format (FDPIPEFLAG != 0):
/***/

final static int FDPIPEFLAG =    57; /* Dimension code of data stream.    Non-standard.  */
final static int FDPIPECOUNT =   75; /* Number of functions in pipeline.  Non-standard.  */
final static int FDSLICECOUNT = 443; /* Number of 1D slices in stream.    Non-standard.  */
final static int FDFILECOUNT =  442; /* Number of files in complete data.                */

/***/
/* The following definitions are used for data streams which are
/* subsets of the complete data, as for parallel processing:
/***/

final static int FDFIRSTPLANE =  77; /* First Z-Plane in subset.            Non-standard. */
final static int FDLASTPLANE =   78; /* Last Z-Plane in subset.             Non-standard. */
final static int FDPARTITION =   65; /* Slice count for client-server mode. Non-standard. */

final static int FDPLANELOC =    14; /* Location of this plane; currently unused.         */

/***/
/* The following define max and min data values, previously used
/* for contour level setting:
/***/

final static int FDMAX =        247; /* Max value in real part of data.                  */
final static int FDMIN =        248; /* Min value in real part of data.                  */
final static int FDSCALEFLAG =  250; /* 1 if FDMAX and FDMIN are valid.                  */
final static int FDDISPMAX =    251; /* Max value, used for display generation.          */
final static int FDDISPMIN =    252; /* Min value, used for display generation.          */

/***/
/* Locations reserved for User customization:
/***/

final static int FDUSER1  =      70;
final static int FDUSER2  =      71;
final static int FDUSER3  =      72;
final static int FDUSER4  =      73;
final static int FDUSER5  =      74;

/***/
/* Defines location of "footer" information appended to spectral
/* data; currently unused for NMRPipe format:
/***/

final static int FDLASTBLOCK =  359;
final static int FDCONTBLOCK =  360;
final static int FDBASEBLOCK =  361;
final static int FDPEAKBLOCK =  362;
final static int FDBMAPBLOCK =  363;
final static int FDHISTBLOCK =  364;
final static int FD1DBLOCK =    365;

/***/
/* Defines data and time data was converted:
/***/

final static int FDMONTH =      294;
final static int FDDAY =        295;
final static int FDYEAR =       296;
final static int FDHOURS =      283;
final static int FDMINS =       284;
final static int FDSECS  =      285;

/***/
/* Miscellaneous Parameters:
/***/

final static int FDMCFLAG =      135; /* Magnitude Calculation performed.               */
final static int FDNOISE =       153; /* Used to contain an RMS noise estimate.         */
final static int FDRANK =        180; /* Estimate of matrix rank; Non-standard.         */
final static int FDTEMPERATURE = 157; /* Temperature, degrees C.                        */
final static int FD2DVIRGIN =    399; /* 0=Data never accessed, header never adjusted.  */
final static int FDTAU =         199; /* A Tau value (for spectral series).             */

final static int FDSRCNAME =    286;  /* char srcFile[16]  286-289 */
final static int FDUSERNAME =   290;  /* char uName[16]    290-293 */
final static int FDOPERNAME =   464;  /* char oName[32]    464-471 */
final static int FDTITLE =      297;  /* char title[60]    297-311 */
final static int FDCOMMENT =    312;  /* char comment[160] 312-351 */

/***/
/* For meanings of these dimension-specific parameters,
/* see the corresponding ND parameters below.
/***/

final static int FDF2LABEL =     16;
final static int FDF2APOD =      95;
final static int FDF2SW =       100;
final static int FDF2OBS =      119;
final static int FDF2ORIG =     101;
final static int FDF2UNITS =    152;
final static int FDF2QUADFLAG =  56; /* Non-standard. */
final static int FDF2FTFLAG =   220;
final static int FDF2AQSIGN =    64; /* Non-standard. */
final static int FDF2LB =       111;
final static int FDF2CAR =       66; /* Non-standard. */
final static int FDF2CENTER =    79; /* Non-standard. */
final static int FDF2OFFPPM =   480; /* Non-standard. */
final static int FDF2P0 =       109;
final static int FDF2P1 =       110;
final static int FDF2APODCODE = 413;
final static int FDF2APODQ1 =   415;
final static int FDF2APODQ2 =   416;
final static int FDF2APODQ3 =   417;
final static int FDF2C1 =       418;
final static int FDF2ZF =       108;
final static int FDF2X1 =       257; /* Non-standard. */
final static int FDF2XN =       258; /* Non-standard. */
final static int FDF2FTSIZE  =   96; /* Non-standard. */
final static int FDF2TDSIZE =   386; /* Non-standard. */

final static int FDF1LABEL =     18;
final static int FDF1APOD =     428;
final static int FDF1SW =       229; 
final static int FDF1OBS =      218;
final static int FDF1ORIG =     249; 
final static int FDF1UNITS =    234; 
final static int FDF1FTFLAG =   222; 
final static int FDF1AQSIGN =   475; /* Non-standard. */
final static int FDF1LB =       243;
final static int FDF1QUADFLAG =  55; /* Non-standard. */
final static int FDF1CAR =       67; /* Non-standard. */
final static int FDF1CENTER =    80; /* Non-standard. */
final static int FDF1OFFPPM =   481; /* Non-standard. */
final static int FDF1P0 =       245;
final static int FDF1P1 =       246;
final static int FDF1APODCODE = 414;
final static int FDF1APODQ1 =   420;
final static int FDF1APODQ2 =   421; 
final static int FDF1APODQ3 =   422;
final static int FDF1C1 =       423;
final static int FDF1ZF =       437;
final static int FDF1X1 =       259; /* Non-standard. */
final static int FDF1XN =       260; /* Non-standard. */
final static int FDF1FTSIZE =   98;  /* Non-standard. */
final static int FDF1TDSIZE =   387; /* Non-standard. */

final static int FDF3LABEL =     20;
final static int FDF3APOD  =     50; /* Non-standard. */
final static int FDF3OBS =       10;
final static int FDF3SW =        11;
final static int FDF3ORIG =      12;
final static int FDF3FTFLAG =    13;
final static int FDF3AQSIGN =   476; /* Non-standard. */
final static int FDF3SIZE =      15;
final static int FDF3QUADFLAG =  51; /* Non-standard. */
final static int FDF3UNITS =     58; /* Non-standard. */
final static int FDF3P0 =        60; /* Non-standard. */
final static int FDF3P1 =        61; /* Non-standard. */
final static int FDF3CAR =       68; /* Non-standard. */
final static int FDF3CENTER =    81; /* Non-standard. */
final static int FDF3OFFPPM =   482; /* Non-standard. */
final static int FDF3APODCODE = 400; /* Non-standard. */
final static int FDF3APODQ1 =   401; /* Non-standard. */
final static int FDF3APODQ2 =   402; /* Non-standard. */
final static int FDF3APODQ3 =   403; /* Non-standard. */
final static int FDF3C1 =       404; /* Non-standard. */
final static int FDF3ZF =       438; /* Non-standard. */
final static int FDF3X1 =       261; /* Non-standard. */
final static int FDF3XN =       262; /* Non-standard. */
final static int FDF3FTSIZE =   200; /* Non-standard. */
final static int FDF3TDSIZE =   388; /* Non-standard. */

final static int FDF4LABEL =     22;
final static int FDF4APOD =      53; /* Non-standard. */
final static int FDF4OBS =       28;
final static int FDF4SW =        29;
final static int FDF4ORIG =      30;
final static int FDF4FTFLAG =    31;
final static int FDF4AQSIGN  =  477; /* Non-standard. */
final static int FDF4SIZE =      32;
final static int FDF4QUADFLAG =  54; /* Non-standard. */
final static int FDF4UNITS =     59; /* Non-standard. */
final static int FDF4P0 =        62; /* Non-standard. */
final static int FDF4P1 =        63; /* Non-standard. */
final static int FDF4CAR =       69; /* Non-standard. */
final static int FDF4CENTER =    82; /* Non-standard. */
final static int FDF4OFFPPM =   483; /* Non-standard. */
final static int FDF4APODCODE = 405; /* Non-standard. */
final static int FDF4APODQ1 =   406; /* Non-standard. */
final static int FDF4APODQ2 =   407; /* Non-standard. */
final static int FDF4APODQ3 =   408; /* Non-standard. */
final static int FDF4C1 =       409; /* Non-standard. */
final static int FDF4ZF =       439; /* Non-standard. */
final static int FDF4X1 =       263; /* Non-standard. */
final static int FDF4XN =       264; /* Non-standard. */
final static int FDF4FTSIZE =   201; /* Non-standard. */
final static int FDF4TDSIZE =   389; /* Non-standard. */

/***/
/* Header locations in use for packed text:
/***/

/* 286 287 288 289                                                     */
/* 290 291 292 293                                                     */
/* 464 465 466 467  468 469 470 471                                    */
/* 297 298 299 300  301 302 303 304  305 306 307 308  309 310 311      */
/* 312 313 314 315  316 317 318 319  320 321 322 323  324 325 326 327  */
/* 328 329 330 331  332 333 334 335  336 337 338 339  340 341 342 343  */
/* 344 345 346 347  348 349 350 351                                    */

final static int SIZE_NDLABEL =    8;
final static int SIZE_F2LABEL =    8;
final static int SIZE_F1LABEL =    8;
final static int SIZE_F3LABEL =    8;
final static int SIZE_F4LABEL =    8;

final static int SIZE_SRCNAME =   16;
final static int SIZE_USERNAME =  16;
final static int SIZE_OPERNAME =  32;
final static int SIZE_COMMENT =  160;
final static int SIZE_TITLE =     60;

/***/
/* The following are definitions for generalized ND parameters:
/***/

final static int NDPARM =      1000;

final static int BAD_DIM =     -666;
final static int NULL_DIM =       0;
final static int CUR_XDIM =       1;
final static int CUR_YDIM =       2;
final static int CUR_ZDIM =       3;
final static int CUR_ADIM =       4;

final static int ABS_XDIM =      -1;
final static int ABS_YDIM =      -2;
final static int ABS_ZDIM =      -3;
final static int ABS_ADIM =      -4;

final static int CUR_HDIM =       9; 
final static int CUR_VDIM =      10;
 
final static int NDSIZE =        (1+NDPARM);  /* Number of points in dimension.          */
final static int NDAPOD =        (2+NDPARM);  /* Current valid time-domain size.         */
final static int NDSW  =         (3+NDPARM);  /* Sweep Width Hz.                         */
final static int NDORIG =        (4+NDPARM);  /* Axis Origin (Last Point), Hz.           */
final static int NDOBS =         (5+NDPARM);  /* Obs Freq MHz.                           */
final static int NDFTFLAG =      (6+NDPARM);  /* 1=Freq Domain 0=Time Domain.            */
final static int NDQUADFLAG =    (7+NDPARM);  /* Data Type Code (See Below).             */
final static int NDUNITS =       (8+NDPARM);  /* Axis Units Code (See Below).            */
final static int NDLABEL =       (9+NDPARM);  /* 8-char Axis Label.                      */
final static int NDLABEL1 =      (9+NDPARM);  /* Subset of 8-char Axis Label.            */
final static int NDLABEL2 =     (10+NDPARM);  /* Subset of 8-char Axis Label.            */
final static int NDP0 =         (11+NDPARM);  /* Zero Order Phase, Degrees.              */
final static int NDP1 =         (12+NDPARM);  /* First Order Phase, Degrees.             */
final static int NDCAR =        (13+NDPARM);  /* Carrier Position, PPM.                  */
final static int NDCENTER =     (14+NDPARM);  /* Point Location of Zero Freq.            */
final static int NDAQSIGN =     (15+NDPARM);  /* Sign adjustment needed for FT.          */
final static int NDAPODCODE =   (16+NDPARM);  /* Window function used.                   */
final static int NDAPODQ1 =     (17+NDPARM);  /* Window parameter 1.                     */
final static int NDAPODQ2 =     (18+NDPARM);  /* Window parameter 2.                     */
final static int NDAPODQ3 =     (19+NDPARM);  /* Window parameter 3.                     */
final static int NDC1 =         (20+NDPARM);  /* Add 1.0 to get First Point Scale.       */
final static int NDZF =         (21+NDPARM);  /* Negative of Zero Fill Size.             */
final static int NDX1 =         (22+NDPARM);  /* Extract region origin, if any, pts.     */
final static int NDXN =         (23+NDPARM);  /* Extract region endpoint, if any, pts.   */
final static int NDOFFPPM =     (24+NDPARM);  /* Additional PPM offset (for alignment).  */
final static int NDFTSIZE =     (25+NDPARM);  /* Size of data when FT performed.         */
final static int NDTDSIZE =     (26+NDPARM);  /* Original valid time-domain size.        */
final static int MAX_NDPARM =   (27);

/***/
/* Axis Units, for NDUNITS:  // use SpinWorks enumeration instead ?!
/***/

final static int FD_SEC =       1;
final static int FD_HZ =        2;
final static int FD_PPM =       3;
final static int FD_PTS  =      4;

/***/
/* 2D Plane Type, for FD2DPHASE:
/***/

final static int FD_MAGNITUDE = 0;
final static int FD_TPPI =      1;
final static int FD_STATES  =   2;
final static int FD_IMAGE =     3;

/***/
/* Data Type (FDQUADFLAG and NDQUADFLAG) // use SpinWorks enumeration instead ?!
/***/

final static int FD_QUAD =       0;
final static int FD_COMPLEX =    0;
final static int FD_SINGLATURE = 1;
final static int FD_REAL =       1;
final static int FD_PSEUDOQUAD = 2;

/***/
/* Sign adjustment needed for FT (NDAQSIGN):
/***/

final static int ALT_NONE  =           0; /* No sign alternation required.                */
final static int ALT_SEQUENTIAL =      1; /* Sequential data needing sign alternation.    */
final static int ALT_STATES =          2; /* Complex data needing sign alternation.       */
final static int ALT_NONE_NEG =       16; /* As above, with negation of imaginaries.      */
final static int ALT_SEQUENTIAL_NEG = 17; /* As above, with negation of imaginaries.      */
final static int ALT_STATES_NEG =     18; /* As above, with negation of imaginaries.      */

final static int FOLD_INVERT =        -1; /* Folding requires sign inversion.             */
final static int FOLD_BAD =            0; /* Folding can't be performed (extracted data). */
final static int FOLD_ORDINARY =       1; /* Ordinary folding, no sign inversion.         */ 

 
//-------------------------------------------------------------------------------------------------
@SuppressWarnings("unused")
/**
 * read a 2D or 3D NMRPipe file, TO DO: needs clean-up!
 * @param fileName full path to data
 * @param gp a global parametersd object
 * @param q status data paramaters, acquisition dimension
 * @param q_f1 status data paramaters, indirect dimension
 * @return bytes read, negative or zero on failure
 */
	public static int readPipeData(String fileName, global_par gp, data_par q, data_par q_f1)
	{
		FileInputStream fs;
	
		if(gp.log_on) Util.log_error("opening NMRPipe file: " + fileName);
	
		int i, j;       // handy counters
		byte[] header = new byte[4 * FDATASIZE];
		int bytes_read = 0;
		boolean rev_bytes = false;
		float magic, fltorder, size;
		float planeloc;
		float dimcount;
		float dimorder_z, dimorder_y, dimorder_x;
		float f3obs, f2obs, f1obs;
		float f3car, f2car, f1car;
		float f3center, f2center, f1center;
		float f3sw, f2sw, f1sw;
		float f3orig, f2orig, f1orig;
		float f3ftsize, f2ftsize, f1ftsize;
		float f3x1, f3xn;
		float f2x1, f2xn;
		float f1x1, f1xn;
		float temperature;
		float transposed;
	
		String title = "", comment = "", f3label = "", f2label = "", f1label = "";
	
		String plane_ID;
		int plane;
		double offset_f3, hzperpt_f3, SFO1_f3, SF_f3, SW_f3;   // local for determining plane offset
		double plane_shift;
	
		q.data_path = fileName;
	
		try
		{
			plane_ID = fileName.substring(fileName.length() - 7, fileName.length() - 4);
			if(gp.log_on)Util.log_error("plane_id = " + plane_ID);
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
			return -1;
		}
	
		try     // "There is no try, there is only do or not do" - Yoda
		{
			plane = Integer.parseInt(plane_ID);
		}
		catch (Exception ex)
		{
			plane = 1;
			Util.show_and_log(plane_ID, ex.getMessage());
		}
	
		if (gp.log_on)
			Util.log_error(String.format("readPipeData(): plane ID = %s; converted to %d", plane_ID, plane));
	
		// try to open the data file, bail out if fails
		try { fs = new FileInputStream(new File(fileName)); }
		catch (IOException ex)
		{
			Util.show_and_log("readPipeData(): cannot open: " + fileName, ex.getMessage());
			return -1;
		}  		 
	
		// read file header into byte array
		try{	fs.read(header);    }
		catch (Exception ex)
		{
			Util.show_and_log("readPipeData(): can't read data" + ex.getMessage());
			try {	fs.close();	} 
			catch (IOException e) {	e.printStackTrace();	}
			return -2;
		}
		bytes_read += 2048;
	
		// wrap to a byte buffer
		ByteBuffer bb = ByteBuffer.wrap(header);
	
		try{	fltorder = bb.getFloat(FDFLTORDER * 4);	}
		catch (Exception ex)
		{
			Util.show_and_log("getPipeData(): can't get byte order ", ex.getMessage());
			try {	fs.close();	} 
			catch (IOException e) {	e.printStackTrace();	}
			return -2;
		}
		if(gp.log_on) Util.log_error(String.format("readPipeData(): fdfltformat= %f", fltorder));
	
		//------------
		//  Check for byte order and reverse if necessary  reverse the bytes.  Looks for the floating point 
		//  constant 2.345 in the location FDORDERCONS.  If not found, bytes must be reversed.
		if (Math.abs(fltorder - FDORDERCONS) > 0.001)
			rev_bytes = true;
	
		//TODO byte order issues need to be checked
		if(rev_bytes)   // reverse bytes in header
			bb.order(ByteOrder.LITTLE_ENDIAN);
	
	
	
	
		try
		{
			//------------
			//  assorted flags
			magic = bb.getFloat(FDMAGIC * 4);      	
			if (gp.log_on) Util.log_error(String.format("readPipeData(): fdmagic = %f", magic));
	
			dimcount = bb.getFloat(FDDIMCOUNT * 4);       	
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): fddimcount = %f", dimcount));
	
			dimorder_x = bb.getFloat(FDDIMORDER1 * 4);  // X
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): fddimorder1 (X) = %f", dimorder_x));
	
			dimorder_y = bb.getFloat(FDDIMORDER2 * 4); // Y       	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): fddimorder2 (Y) = %f", dimorder_y));
	
			dimorder_z = bb.getFloat(FDDIMORDER3 * 4); // Z       	
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): fddimorder3 (Z) = %f", dimorder_z));
	
			size = bb.getFloat(FDSIZE * 4);
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): fdsize= %f", size));
	
			planeloc = bb.getFloat(FDPLANELOC * 4);   	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): fdplaneloc= %f", planeloc));
	
			//--------------
			//  observe frequencies
	
			f3obs = bb.getFloat(FDF3OBS * 4);
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f3obs = %f", f3obs));
	
			f2obs = bb.getFloat(FDF2OBS * 4);       	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f2obs = %f", f2obs));
	
			f1obs = bb.getFloat(FDF1OBS * 4);   	
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): f1obs = %f", f1obs));
	
			//---------------
			//  "carrier" location in ppm
	
			f3car = bb.getFloat(FDF3CAR * 4);
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f3car = %f", f3car));
	
			f2car = bb.getFloat(FDF2CAR * 4);
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f2car = %f", f2car));
	
			f1car = bb.getFloat(FDF1CAR * 4);   
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): f1car = %f", f1car));
	
			//-------------
			//  "center" points
	
			f3center = bb.getFloat(FDF3CENTER * 4);  	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f3center = %f", f3center));
	
			f2center = bb.getFloat(FDF2CENTER * 4); 	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f2center = %f", f2center));
	
			f1center = bb.getFloat(FDF1CENTER * 4);      	
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): f1center = %f", f1center));
	
			//-------------
			//  spectral widths in Hz.
	
			f3sw = bb.getFloat(FDF3SW * 4);       	
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): f3sw = %f", f3sw));
	
			f2sw = bb.getFloat(FDF2SW * 4);       	
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): f2sw = %f", f2sw));
	
			f1sw = bb.getFloat(FDF1SW * 4);      	
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): f1sw = %f", f1sw));
	
			//-------------
			//  "origin" (offset?) frequencies in Hz  right most point in Hz
	
			f3orig = bb.getFloat(FDF3ORIG * 4);
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f3orig = %f", f3orig));
	
			f2orig = bb.getFloat(FDF2ORIG * 4);      	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f2orig = %f", f2orig));
	
			f1orig = bb.getFloat(FDF1ORIG * 4);
			if (gp.log_on)
				Util.log_error(String.format(" f1orig = %f", f1orig));
	
			//-------------
			//  ft sizes, total of real plus imag.
	
			f3ftsize = bb.getFloat(FDF3FTSIZE * 4);
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f3ftsize = %f", f3ftsize));
	
			f2ftsize = bb.getFloat(FDF2FTSIZE * 4);     	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f2ftsize = %f", f2ftsize));
	
			f1ftsize = bb.getFloat(FDF1FTSIZE * 4);      	
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): f1ftsize = %f", f1ftsize));
	
			//--------------
			//  "extract" information
	
			f3x1 = bb.getFloat(FDF3X1 * 4);       	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f3x1= %f   ", f3x1));
	
			f3xn = bb.getFloat(FDF3XN * 4);
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f3xn = %f", f3xn));
	
			f2x1 = bb.getFloat(FDF2X1* 4);     	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f2x1= %f   ", f2x1));
	
			f2xn = bb.getFloat(FDF2XN * 4);
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f2xn = %f", f2xn));
	
			f1x1 = bb.getFloat(FDF1X1 * 4);      	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): f1x1= %f    ", f1x1));
	
			f1xn = bb.getFloat(FDF1XN * 4);        	
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): f1xn = %f", f1xn));
	
			temperature = bb.getFloat(FDTEMPERATURE * 4);     	
			if (gp.log_on) 
				Util.log_error(String.format("readPipeData(): temperature= %f", temperature));
	
			transposed = bb.getFloat(FDTRANSPOSED * 4);    // transposed flag   
			if (gp.log_on)
				Util.log_error(String.format("readPipeData(): transposed = %f", transposed));
	
			// leave individual try/catch on text, but no abort
			try     // title
			{
				char[] ch = new char[60];
	
				for (i = 0; i < 60; i++)
				{
					ch[i] = (char)header[FDTITLE * 4 + i];
				}
				title = new String(ch);
			}
			catch (Exception e)
			{
				Util.show_and_log("readPipeData(): " + e.getMessage());
				e.printStackTrace();
			}
	
			if (gp.log_on) Util.log_error(" title = " + title);
	
			try     // comment
			{
				char[] ch = new char[160];
	
				for (i = 0; i < 160; i++)
				{
					ch[i] = (char)header[FDCOMMENT * 4 + i];
				}
				comment = new String(ch);
			}
			catch (Exception e)
			{
				Util.show_and_log("readPipeData(): " + e.getMessage());
				e.printStackTrace();
			}
	
			try     // F1 Label
			{
				char[] ch = new char[8];
	
				for (i = 0; i < 8; i++)
				{
					if (header[FDF1LABEL * 4 + i] == 0)
						ch[i] = ' ';
					else
						ch[i] = (char)header[FDF1LABEL * 4 + i];	
				}
				f1label = new String(ch);
			}
			catch (Exception e)
			{
				Util.log_error("readPipeData(): " + e.getMessage());
				e.printStackTrace();
			}
	
			if (gp.log_on) Util.log_error(" f1label = " + f1label);
	
			try     // F2 Label
			{
				char[] ch = new char[8];
	
				for (i = 0; i < 8; i++)
				{
					if (header[FDF2LABEL * 4 + i] == 0)
						ch[i] = ' ';
					else
						ch[i] = (char)header[FDF2LABEL * 4 + i];
				}
	
				f2label = new String(ch);
				f2label = f2label.trim();
			}
			catch (Exception e)
			{
				Util.log_error("readPipeData(): " + e.getMessage());
				e.printStackTrace();
			}
	
			if (gp.log_on) Util.log_error(" f2label = " + f2label);
	
			try     // F3 Label
			{
				char[] ch = new char[8];
	
				for (i = 0; i < 8; i++)
				{
					if (header[FDF3LABEL * 4 + i] == 0)
						ch[i] = ' ';
					else
						ch[i] = (char)header[FDF3LABEL * 4 + i];
				}
	
				f3label = new String(ch);
				f3label = f3label.trim();
			}
			catch (Exception e)
			{
				Util.show_and_log("readPipeData(): " + e.getMessage());
				e.printStackTrace();;
			}
	
			if (gp.log_on) Util.log_error(" f3label = " + f3label);
		}
		catch(Exception ex)
		{
			Util.show_and_log("Error reading NMR Pipe parameters: ", ex.getMessage());
			ex.printStackTrace();
			return -2;
		}
	
	
		//--------------
		//  Convert some parameters into SpinWorks versions
	
		if ((int)f2ftsize > 0 && (int)f1ftsize > 0)
			q.data_mode = SW_main.modes.MODE_2D;
	
		q.SI = (int)f2ftsize;
		q_f1.SI = (int)f1ftsize;
	
	
		q.SFO1 = (double)f2obs;
		q_f1.SFO1 = (double)f1obs;
		SFO1_f3 = (double)f3obs;
	
		if (f2x1 != 0.0F && f2xn != 0.0F)
		{
			q.SI = (int)(f2xn - f2x1) + 1;
		}
	
		q.SF = q.SFO1 - ((double)f2car * q.SFO1) / 1.0e6;
		q_f1.SF = (double)f1obs - (double)(f1car * f1obs) / 1.0e6;
		SF_f3 = (double)f3obs - (double)(f3car * f3obs) / 1.0e6;
	
		q.SW = f2sw / q.SFO1;
		q_f1.SW = f1sw / q_f1.SFO1;
		SW_f3 = f3sw / SFO1_f3;
	
		q.hzperpt = (double)f2sw / (double)(q.SI);
		q.offset = f2orig + f2sw - q.hzperpt;
	
		q_f1.hzperpt = (double)f1sw / (double)(q_f1.SI);
		q_f1.offset = f1orig + f1sw - q_f1.hzperpt;
	
		hzperpt_f3 = (double)f3sw / (double) f3ftsize;
		offset_f3 = (double)f3orig + f3sw - hzperpt_f3;
	
	
		if (gp.log_on)
		{
			Util.log_error(String.format("readPipeData():  SF(F2) = %.7f   SF(F1) = %.7f", q.SF, q_f1.SF));
	
			Util.log_error("readPipeData(): F3:");
			Util.log_error(String.format("   size = %f    sw = %f    hzperpt = %f   offset = %f",
					f3ftsize, SW_f3, hzperpt_f3, offset_f3));
	
			Util.log_error("readPipeData(): F2:");
			Util.log_error(String.format("   size = %d    sw = %f    hzperpt %f   offset = %f",
					q.SI, q.SW, q.hzperpt, q.offset));
	
			Util.log_error("readPipeData(): F1:");
			Util.log_error(String.format("   size = %d    sw = %f    hzperpt = %f   offset = %f",
					q_f1.SI, q_f1.SW, q_f1.hzperpt, q_f1.offset));
	
		}
	
		//	allocate space for all quadrants
	
		if (q.SI == 0 || q_f1.SI == 0)
		{
			Util.show_and_log(String.format("NMRPIPE: read error F1 size = %d  F2 size = %d bad file format?",
					q_f1.SI, q.SI));
			try {	fs.close();	} 
			catch (IOException ex) {	ex.printStackTrace();	}
			return -4;
		}
	
		Util_2D.alloc_rows(q_f1.SI, q.SI, gp, q, q_f1);
	
		//--------------
		//  Now read data
	
		//TODO read only once!
	
		try {	fs.close();	} 
		catch (IOException ex) {	ex.printStackTrace();	}
	
		// try to reopen the data file, bail out if fails
		File file = new File(fileName);
		try { fs = new FileInputStream(file); }
		catch (IOException ex)
		{
			Util.show_and_log("readPipeData(): cannot open: " + fileName, ex.getMessage());
			return -1;
		}  		 
	
		byte[] data_bytes = new byte[(int) file.length()];
	
		try { fs.read(data_bytes); } 
		catch (IOException ex) 
		{
			ex.printStackTrace();
			try { fs.close(); } 
			catch (IOException e) { e.printStackTrace(); }
			return -3;
		}
	
		bb = ByteBuffer.wrap(data_bytes);
		if (rev_bytes)
			bb.order(ByteOrder.LITTLE_ENDIAN);
	
		try
		{
			if (transposed < 0.01)  // data are not transposed (row major)
			{
				if(gp.log_on)Util.log_error("not transposed");
				for (i = 0; i < q_f1.SI; i++)
				{
					for (j = 0; j < q.SI; j++)
					{
						gp.rows_real[i][j] = bb.getFloat(2048 + (4 * j) + (4 * i * q.SI));
						bytes_read += 4;
					}    
				}   // for i
			}
			else        // data ARE transposed (column major)    
			{
				if(gp.log_on)Util.log_error("transposed");
				for (i = 0; i < q.SI; i++)
				{
					for (j = 0; j < q_f1.SI; j++)
					{
						gp.rows_real[j][i] = bb.getFloat(2048 + (4 * j) + (4 * i * q_f1.SI));
						bytes_read += 4;
					}
				}
			}
		}
		catch (Exception ex)
		{
			Util.show_and_log("readPipeData(): data read failed ", ex.getMessage());
			try {	fs.close();	} 
			catch (IOException e) {	e.printStackTrace();	}
			return -3;
		}
	
		if(dimcount == 2.0F)
			gp.plot_title = String.format("X = %s  Y = %s", f2label, f1label);
		else if (dimcount == 3.0F)
		{
			plane_shift = (offset_f3 - ((double)plane - 1) * hzperpt_f3) / SFO1_f3;
	
			gp.plot_title = String.format("X = %s  Y = %s  Z = %s at  %.3f ppm  plane = %d ",
					f2label, f1label, f3label, plane_shift, plane);
		}
		else
		{
			gp.plot_title = "unknown dimension count";
		}
	
		if (gp.log_on) Util.log_error(String.format("readPipeData(): bytes read = %d", bytes_read));
	
		//--------------
		//  clean-up
		try {	fs.close();	} 
		catch (IOException e) {	e.printStackTrace();	}
		return bytes_read;
	}
    
//-------------------------------------------------------------------------------------------------
	/**
	 * Utility for 3D NMRPipe data only, to load a new plane of a 3D experiment
	 * @param slice  the integer slice number
	 */
	public static void load_new_slice(int slice)
	{
		String new_path = "";
		String temp = SW_main.q.data_path;
	
		temp = temp.substring(0, SW_main.q.data_path.length() - 7);
		new_path = String.format("%s%03d.ft3", temp, slice); 
		if(SW_main.gp.log_on)
			Util.log_error("new_path = " + new_path);
	
		File file = new File(new_path);
		if (!file.exists()) return;
	
		SW_main.gp.new_data();                  // reset some of the global parameter block
	
		double old_sf = SW_main.q.SF;           // maintain current referencing
		double old_sf_f1 = SW_main.q_f1.SF;
		double old_offset = SW_main.q.offset;
		double old_offset_f1 = SW_main.q_f1.offset;
	
		NMRpipe.readPipeData(new_path, SW_main.gp, SW_main.q, SW_main.q_f1);
	
		SW_main.q.SF = old_sf;                  // restore current referencing
		SW_main.q_f1.SF = old_sf_f1;
		SW_main.p.SF = SW_main.q.SF;
		SW_main.p_f1.SF = SW_main.q_f1.SF;
		SW_main.q.offset = old_offset;
		SW_main.q_f1.offset = old_offset_f1;
	
		Util.free_proc(SW_main.gp);
		Util.alloc_proc(SW_main.gp, SW_main.p, SW_main.q);
		SW_main.gp.cursor_flag = 0;					// flag no cursors set 
	
		InterfaceUtil.updateToolbar();		    	// update menu and tool bar
	
		SW_main.statusStripLabel1.setText("Ready");	// and update embelishments
	
		if (SW_main.gp.log_on)
			Util.log_error(String.format("NMRPipe slice #%d loaded", slice));
	
		SW_main.update_spect = true;
		SW_main.canvas.redraw();
	}
        
//-------------------------------------------------------------------------------------------------
	public static int getSliceNumber(String dataPath) 
	{
		int plane;
	
		try
		{
			String plane_ID = dataPath.substring(dataPath.length() - 7, dataPath.length() - 4);
			if(SW_main.gp.log_on)
				Util.log_error("plane_ID = " + plane_ID);
	
			try
			{
				plane = Integer.parseInt(plane_ID);
			}
			catch (Exception ex)
			{
				Util.log_error("getSliceNumber(): ", ex.getMessage());
				plane = 1;
			}
		}
		catch (Exception ex)
		{
			Util.show_and_log("getSliceNumber():  " + dataPath, ex.getMessage());
			return -1;
		}
	
		return plane;
	}

 //------------------------------------------------------------------------------------------------       
}
