package spinworks;

import java.awt.Desktop;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import simulation.SimPar;
import spinworks.SW_main.corr;

public class PeakPick 
{
	final public static int MAX_PEAKS = 4000;

/**
 * finds peaks in a specified region
 * @param start start frequency (Hz.)
 * @param end end frequency (Hz.)
 * @param sens peak picking sensitivity (PC)
 * @param thresh peak picking threshold (MI)
 * @param gp a global parameter object
 * @param q status data parameters
 * @return number of peaks
 */
	public static int findPeaks(double start, double end, double sens, double thresh,
			global_par gp, data_par q)
	{
		if (gp.data_array == null)
		{
			Util.show_and_log("No experimental data!");
			return 0;
		}

		int count;				// peak counter 
		int point1, point2;		// start and stop indices 
		int i, j, temp;
		double isens, mi;
		double frequency;

		point1 = (int)((q.offset-start)/q.hzperpt);
		point2 = (int)((q.offset-end)/q.hzperpt);

		//swap order if necessary    necessary?  frequencies already sorted!??
		if(point1 > point2)
		{
			temp = point1;
			point1 = point2;
			point2 = temp;
		}

		//check that points are within array 
		if(point1 < 0) point1 = 0;
		if(point2 >= q.SI) point2 = q.SI;

		isens = 2.0 * sens * gp.data_noise;
		mi = thresh * gp.data_max;

		j = gp.n_peaks;
		count = 0;

		//now find peaks, use parabolic interpolation 
		for(i=point1+1; i < point2-1; i++)
		{
			double intensity;
			if (gp.pp_sign == 0)
				intensity = Math.abs(gp.data_array[i]);
			else if (gp.pp_sign == -1)
				intensity = -gp.data_array[i];
			else
				intensity = gp.data_array[i];

			if ((Math.abs(gp.data_array[i]) - Math.abs(gp.data_array[i-1])) > isens
					&& (Math.abs(gp.data_array[i]) - Math.abs(gp.data_array[i+1])) > isens
					&& intensity > mi)
			{
				frequency = parab(i, gp, q);

				gp.peaks[j] = new PeakInfo();

				if((frequency >= q.offset - (double)(i+1) * q.hzperpt) &&
						(frequency <= q.offset - (double)(i-1) * q.hzperpt))
					gp.peaks[j].freq = frequency;

				gp.peaks[j].intensity =
						gp.data_array[Util.freq_to_index(gp.peaks[j].freq, q)];

				j++; count++;
			}
			if(j >= MAX_PEAKS)  // too many peaks found 
			{
				Util.show_and_log("Too Many Peaks");
				break;
			}
		}
		gp.n_peaks = j;

		return count;
	}

//-------------------------------------------------------------------------------------------------
/**
 * finds the interpolated peak position closest to freq
 * returns the interpolated frequency if a peak is found, else
 * the original cursor frequency (freq) is returned
 * @param freq the input frequency
 * @param range the range in Hz to search for a peak
 * @param gp a global par object
 * @param q the status data parameters
 * @return the peak frequency
 */
	public static double findPeak(double freq, double range, global_par gp, data_par q , SimPar sp)
	{
		double frequency = 0.0;
		int i, point1, point2;
		boolean peak_found;
		double isens, mi;

		isens = 2.0 * gp.pp_sens * gp.data_noise;
		mi = gp.pp_thresh * gp.data_max;
		peak_found = false;
		double factor = 0.5;

		while (!peak_found && factor < range)					
		{			
			point1 = (int)((q.offset - freq - sp.linewidth * factor) / q.hzperpt);
	        point2 = (int)((q.offset - freq + sp.linewidth * factor) / q.hzperpt);

			if (point1 < 0) point1 = 0;                 // check array bounds
			if (point2 >= q.SI) point2 = q.SI;

			for (i = point1; i <= point2; i++)
			{

				double intensity;
				if (gp.pp_sign == 0)
					intensity = Math.abs(gp.data_array[i]);
				else if (gp.pp_sign == -1)
					intensity = -gp.data_array[i];
				else
					intensity = gp.data_array[i];

				if ((Math.abs(gp.data_array[i]) - Math.abs(gp.data_array[i - 1])) > isens
						&& (Math.abs(gp.data_array[i]) - Math.abs(gp.data_array[i + 1]) > isens
								&& intensity > mi))
				{
					frequency = parab(i, gp, q);

					if ((frequency >= q.offset - (double)(i + 1) * q.hzperpt) &&
							(frequency <= q.offset - (double)(i - 1) * q.hzperpt))
						peak_found = true;
				}
				if (peak_found) break;			// stop scanning 
			}

			factor = factor * 2.0;
		}

		if (peak_found) return frequency;
		else return freq;						// if no peak found return cursor frequency 
	}

//-------------------------------------------------------------------------------------------------
// TODO, is this unnecessary duplication of find_peak??  except for "factor" ???
/**
 * finds peak closest to freq.  Used by autoassign
 * @param freq
 * @param gp global parameter object
 * @param q status data parameters
 * @param sp simulation parameters
 * @return frequencty of the found peak
 */
	public static double autoPick(double freq, global_par gp, data_par q, SimPar sp)
	{
		double isens, mi;
		int point1, point2, i;
		boolean peak1_found, peak2_found;
		double frequency1 = 0.0, frequency2 = 0.0;

		isens = 2.0 * gp.pp_sens * gp.data_noise;
		mi = gp.pp_thresh * gp.data_max;

		point1 = (int)((q.offset - (freq + 5.0 * sp.linewidth)) / q.hzperpt);
		point2 = (int)((q.offset - (freq - 5.0 * sp.linewidth)) / q.hzperpt);

		if (point1 < 0) point1 = 0;
		if (point2 >= q.SI) point2 = q.SI;

		peak1_found = false;
		peak2_found = false;

		for (i = ((point1 + point2) / 2) - 1; i <= point2; i++)
		{
			if ((gp.data_array[i] - gp.data_array[i - 1]) > isens
					&& (gp.data_array[i] - gp.data_array[i + 1]) > isens
					&& (gp.data_array[i] > mi))
			{
				frequency1 = parab(i, gp, q);

				if ((frequency1 >= q.offset - (double)(i + 1) * q.hzperpt)
						&& (frequency1 <= q.offset - (double)(i - 1) * q.hzperpt))
				{
					peak1_found = true;
				}
			}
			if (peak1_found) break;				// stop scanning 
		}

		for (i = ((point1 + point2) / 2) + 1; i >= point1; i--)
		{
			if ((gp.data_array[i] - gp.data_array[i - 1]) > isens
					&& (gp.data_array[i] - gp.data_array[i + 1]) > isens
					&& (gp.data_array[i] > mi))
			{
				frequency2 = parab(i, gp, q);

				if ((frequency2 >= q.offset - (double)(i + 1) * q.hzperpt)
						&& (frequency2 <= q.offset - (double)(i - 1) * q.hzperpt))
				{
					peak2_found = true;
				}
			}
			if (peak2_found) break;				// stop scanning 
		}

		if (peak1_found && !peak2_found)
			return frequency1;

		else if (peak2_found && !peak1_found)
			return frequency2;

		else if (peak1_found && peak2_found)
		{
			if ((frequency2 - freq) < (freq - frequency1))
				return frequency2;
			else
				return frequency1;
		}

		else
			return 0.0;

	}

//-------------------------------------------------------------------------------------------------
/**
 * do an inverse parabolic interpolation of a peak at index i
 * @param i the index in the array
 * @param gp a global parameter object
 * @param q a status data parameter object
 * @return the interpolated frequency
 */
	private static double parab(int i, global_par gp, data_par q)
	{
		double a, b, c, fa, fb, fc, numerator, denomonator;

		a = q.offset - (double)(i + 1) * q.hzperpt;
		b = q.offset - (double)(i) * q.hzperpt;
		c = q.offset - (double)(i - 1) * q.hzperpt;

		fa = gp.data_array[i + 1];
		fb = gp.data_array[i];
		fc = gp.data_array[i - 1];

		numerator = fa * (b + c) + fc * (a + b) - 2 * fb * (a + c);
		denomonator = fa - 2 * fb + fc;

		// paranoid check for division by zero 
		if ((denomonator >= 1e-6) || (denomonator <= (-1e-6)))
			return (0.5 * numerator / denomonator);
		else
			return -(1e20); // return impossible frequency if fit failed 
	}

//-------------------------------------------------------------------------------------------------
/**
 * clear the peak picking table
 * @param gp A SpinWorks global parameter object
 */
	public static void clearPeaks(global_par gp)
	{
		int i;

		for (i = 0; i < gp.n_peaks; i++)
			gp.peaks[i].dispose();

		gp.n_peaks = 0;
	}

//-------------------------------------------------------------------------------------------------
/**
 * list the picked peaks into a file and displays a list 
 * @param q status data parameters
 */
	public static void listPeaks(data_par q)
	{	
		if (Desktop.isDesktopSupported()) // new way!  portable??
		{
		    try 
		    {
		        Desktop.getDesktop().open(new File(q.data_folder + "peaks.txt"));
		    } 
		    catch (IOException ex) 
		    {
		        Util.show_and_log("no application for this file type?? " +
		             q.data_folder + "peaks.txt", ex.getMessage());
		    }
		}
		else
			Util.show_and_log("Desktop not supported");
	}

//-------------------------------------------------------------------------------------------------
/**
 * Save picked peaks into a file
 * @param gp a SpinWorks global parameter object
 * @param q status data parameters
 */
	public static void savePeaks(global_par gp, data_par q)
	{
		int i;
		PrintWriter pw;
		String fileName = null;	

		if(q.data_type != SW_main.data_types.JEOL
				|| q.data_type != SW_main.data_types.JCAMP_DX )		//TODO clean this up
			fileName = q.data_folder + "peaks.txt";	
		else if(q.data_type == SW_main.data_types.JEOL)
			fileName = q.data_path + "_SW_proc/peaks.txt";
		
		try
		{
			pw = new PrintWriter(new FileWriter(fileName));
		}
		catch (Exception ex)
		{
			Util.show_and_log(ex.getMessage(), fileName);
			return;
		}

		
		if(q.data_type == SW_main.data_types.JEOL)
			pw.println(String.format("SpinWorks 4 peak list for:    %s", q.data_path));
		else
			pw.println(String.format("SpinWorks 4 peak list for:    %s", q.data_folder));
		
		pw.println("peak #   freq. (Hz)    freq. (ppm)    intensity  intensity(Absolute)  label");
		pw.println();


		for (i = 0; i < gp.n_peaks; i++)
		{
			try
			{
				if (gp.peaks[i].freq != 0.0 && gp.peaks[i].freq > -1.0e09)
				{
					pw.println(String.format("%4d   %10.3f    %10.3f     %10.4f   %14.0f       %s",
							i + 1, gp.peaks[i].freq, gp.peaks[i].freq / q.SF,
							gp.peaks[i].intensity / gp.data_max,
							gp.peaks[i].intensity, gp.peaks[i].label));
				}
			}
			catch (Exception ex)
			{
				Util.show_and_log("1D peak list raised the following exception:", ex.getMessage());
				break;
			}
		}

		pw.close();
	}

//-------------------------------------------------------------------------------------------------
/**
 * this function lists the picked peaks into a file and displays a list 
 * @param gp a SpinWorks global parameter object
 * @param q status parameters in the acquisition dimension
 * @param q_ind status parameters in the indirect dimension
 */
	public static void listPeaks2D(global_par gp, data_par q, data_par q_ind)
	{
		int i;

		String peaks_file_name = q.data_folder + "peaks_2D.txt";
		PrintWriter pw;

		peaks_file_name = q.data_folder + "peaks_2D.txt";

		try
		{
			pw = new PrintWriter(new FileWriter(peaks_file_name));
		}
		catch (Exception ex)
		{
			Util.show_and_log("list_peaks_2D():" + ex.getMessage(), peaks_file_name);
			return;
		}


		pw.println("SpinWorks 2D peak list for:    " + q.data_folder);
		pw.println();
		
		for (i = 0; i < gp.n_peaks; i++)
		{
			if (gp.peaks[i].freq != 0.0 && gp.peaks[i].freq > -1.0e09)
			{
				try
				{
					pw.println();
					pw.println("----------------------------------");
					pw.println(String.format("%d Assignment: %s", i + 1, gp.peaks[i].label));
					
					pw.print(String.format("     F2 (ppm) = %10.3f       F1 (ppm) =  %10.3f     intensity (rel) = %.4f   intensity (abs) = ",
							gp.peaks[i].freq / q.SF, gp.peaks[i].freq_f1 / q_ind.SF,
							gp.peaks[i].intensity / gp.data_max_2D));

					pw.println(String.format("%.0f", gp.peaks[i].intensity));

					pw.print("          COSY:  ");
					for (int j = 0; j < gp.peaks[i].n_cosy; j++)
						pw.print(String.format("  %10.3f", gp.peaks[i].cosy[j]));
					pw.println();
					pw.print(" ASSOCIATED F1:  ");
					for (int j = 0; j < gp.peaks[i].n_cosy; j++)
					{
						if(gp.peaks[i].cosy_f1[j] != 0.0)
							pw.print(String.format("  %10.3f", gp.peaks[i].cosy_f1[j]));
						else
							pw.print("      ------");
					}
					pw.println();	pw.println();
					
					pw.print("         TOCSY: ");
					for (int j = 0; j < gp.peaks[i].n_tocsy; j++)
						pw.print(String.format("  %10.3f", gp.peaks[i].tocsy[j]));
					pw.println();
					pw.print(" ASSOCIATED F1: ");
					for (int j = 0; j < gp.peaks[i].n_tocsy; j++)
					{
						if(gp.peaks[i].tocsy_f1[j] != 0.0)
							pw.print(String.format("  %10.3f", gp.peaks[i].tocsy_f1[j]));
						else
							pw.print("      ------");
					}
					pw.println();	pw.println();

					pw.print("         NOESY: ");
					for (int j = 0; j < gp.peaks[i].n_noesy; j++)
						pw.print(String.format("  %10.3f", gp.peaks[i].noesy[j]));
					pw.println();
					pw.print(" ASSOCIATED F1: ");
					for (int j = 0; j < gp.peaks[i].n_noesy; j++)
					{
						if(gp.peaks[i].noesy_f1[j] != 0.0)
							pw.print(String.format("  %10.3f", gp.peaks[i].noesy_f1[j]));
						else
							pw.print("      ------");
					}
					pw.println();	pw.println();
					
					pw.print("          HMBC:  ");
					for (int j = 0; j < gp.peaks[i].n_hmbc; j++)
						pw.print(String.format("  %10.3f", gp.peaks[i].hmbc[j]));
					pw.println();
					pw.print(" ASSOCIATED F1: ");
					for (int j = 0; j < gp.peaks[i].n_hmbc; j++)
					{
						if(gp.peaks[i].hmbc_f1[j] != 0.0)
							pw.print(String.format("  %10.3f", gp.peaks[i].hmbc_f1[j]));
						else
							pw.print("      ------");
					}
					pw.println();
				}
				catch (Exception ex)
				{
					Util.show_and_log("2D peak list raised the following exception:", 
							ex.getMessage());
					break;
				}
			}
		}

		pw.close();
		
		if (Desktop.isDesktopSupported()) // new way!  portable??
		{
		    try 
		    {
		        Desktop.getDesktop().open(new File(q.data_folder + "peaks_2D.txt"));
		    } 
		    catch (IOException ex) 
		    {
		        Util.show_and_log("no application for this file type?? " +
		             q.data_folder + "peaks_2D.txt", ex.getMessage());
		    }
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * sort picked peaks in decending frequency order
 * @param n number of peaks
 * @param peaks the peaks object
 * @return number of peaks sorted
 */
	public static int sortPeaks(int n, PeakInfo[] peaks)
	{
		int i, j;
		PeakInfo tmp_peak = new PeakInfo();

		for (i = 0; i < n; i++)					// sort peaks
		{										// using simple bubble sort
			for (j = i; j < n; j++)
			{
				if (peaks[i].freq < peaks[j].freq)
				{
					tmp_peak = peaks[i];
					peaks[i] = peaks[j];
					peaks[j] = tmp_peak;
				}
			}
		}

		for (i = 0; i < n; i++)
		{
			if (peaks[i].freq < -1.0e09) break;
		}

		if (i < n) n = i;

		return n;
	}

//-------------------------------------------------------------------------------------------------
/**
 * delete peaks in the dispayed region
 * @param gp A Spinworks global parameter object
 */
	public static void deletePeaks(global_par gp)
	{
		int i;
		for (i = 0; i < gp.n_peaks; i++)
		{
			if (gp.display_mode == SW_main.modes.MODE_1D)
			{
				if (gp.peaks[i].freq < gp.freq1 && gp.peaks[i].freq > gp.freq2)
				{
					gp.peaks[i].freq = -2e09;
					gp.peaks[i].intensity = 0.0;
				}
			}
			else if (gp.display_mode == SW_main.modes.MODE_2D)
			{
				if (gp.peaks[i].freq < gp.freq1 && gp.peaks[i].freq > gp.freq2
						&& gp.peaks[i].freq_f1 < gp.freq1_indirect 
						&& gp.peaks[i].freq_f1 > gp.freq2_indirect)
				{
					gp.peaks[i].freq = -2e09;
					gp.peaks[i].freq_f1 = -2e09;
					gp.peaks[i].intensity = 0.0;
				}
			}
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * Delete any duplicate peaks
 * @param gp A Spinworks global parameter object
 */
	public static void deleteDuplicates(global_par gp)
	{
		int i, j;
		for (i = 0; i < (gp.n_peaks - 1); i++)
			for (j = i + 1; j < gp.n_peaks; j++)
			{
				if (gp.peaks[i].freq == gp.peaks[j].freq)
				{
					gp.peaks[j].freq = -2e09;
					gp.peaks[j].intensity = 0.0;
				}
			}
	}

//-------------------------------------------------------------------------------------------------
/**
 * parse a string from the 2D assignment editor
 * @param str the string to parse
 * @param peak an object containing the peak information
 * @param type tye of correlation:  COSY, TOCSY, NOESY or HMBC
 */
	public static void parseAssignments(String str, PeakInfo peak, SW_main.corr type)
	{
		String[] token;
		String[] subToken;
		String delim = "[\n]+";
		String delim2 = "[,;]+";

		str = str.trim();
		token = str.split(delim);
		
		if (type == corr.COSY) // COSY
		{
			peak.n_cosy = 0;
			for (int i = 0; i < token.length && i < 10; i++)
			{
				subToken = token[i].split(delim2);
				try { peak.cosy[i] = Double.parseDouble(subToken[0]); }
				catch (Exception ex)
				{
					if(!ex.getMessage().equals("empty String"))
							Util.log_error(ex.getMessage(), str);
					break;
				}
				
				if(subToken.length == 1)		// no associated F1 (manually entered)
					peak.cosy_f1[i] = 0.0;
				
				else if(subToken.length > 1)	// associated F1 present
				{
					try { peak.cosy_f1[i] = Double.parseDouble(subToken[1]); }
					catch (Exception ex)
					{
						if(!ex.getMessage().equals("empty String"))
							Util.log_error(ex.getMessage(), str);
						break;
					}
				}
				peak.n_cosy++;
			}	// for(int i = 0...
		}

		else if (type == corr.TOCSY) // TOCSY
		{
			peak.n_tocsy = 0;
			for (int i = 0; i < token.length && i < 10; i++)
			{              
				subToken = token[i].split(delim2);
				try { peak.tocsy[i] = Double.parseDouble(subToken[0]); }
				catch (Exception ex)
				{
					if(!ex.getMessage().equals("empty String"))
							Util.log_error(ex.getMessage(), str);
					break;
				}
				
				if(subToken.length == 1)		// no associated F1 (manually entered)
					peak.tocsy_f1[i] = 0.0;
				
				else if(subToken.length > 1)	// associated F1 present
				{
					try { peak.tocsy_f1[i] = Double.parseDouble(subToken[1]); }
					catch (Exception ex)
					{
						if(!ex.getMessage().equals("empty String"))
							Util.log_error(ex.getMessage(), str);
						break;
					}
				}
				peak.n_tocsy++;
			}	// for(int i = 0...
		}
		
		else if (type == corr.NOESY) // NOESY
		{
			peak.n_noesy = 0;
			for (int i = 0; i < token.length && i < 10; i++)
			{
				subToken = token[i].split(delim2);
				try { peak.noesy[i] = Double.parseDouble(subToken[0]); }
				catch (Exception ex)
				{
					if(!ex.getMessage().equals("empty String"))
							Util.log_error(ex.getMessage(), str);
					break;
				}
				
				if(subToken.length == 1)		// no associated F1 (manually entered)
					peak.noesy_f1[i] = 0.0;
				
				else if(subToken.length > 1)	// associated F1 present
				{
					try { peak.noesy_f1[i] = Double.parseDouble(subToken[1]); }
					catch (Exception ex)
					{
						if(!ex.getMessage().equals("empty String"))
							Util.log_error(ex.getMessage(), str);
						break;
					}
				}
				peak.n_noesy++;
			}
		}
		
		else if (type == corr.HMBC) // HMBC
		{
			peak.n_hmbc = 0;
			for (int i = 0; i < token.length && i < 10; i++)
			{
				subToken = token[i].split(delim2);
							
				if(subToken.length == 1)		// no associated F2 (manually entered)
				{
					peak.hmbc[i] = 0.0;
					try { peak.hmbc_f1[i] = Double.parseDouble(subToken[0]); }
					catch (Exception ex)
					{
						if(!ex.getMessage().equals("empty String"))
								Util.log_error(ex.getMessage(), str);
						break;
					}
				}
				
				else if(subToken.length > 1)	// associated F2 present
				{
					try { peak.hmbc[i] = Double.parseDouble(subToken[0]); }
					catch (Exception ex)
					{
						if(!ex.getMessage().equals("empty String"))
								Util.log_error(ex.getMessage(), str);
						break;
					}
					
					try { peak.hmbc_f1[i] = Double.parseDouble(subToken[1]); }
					catch (Exception ex)
					{
						if(!ex.getMessage().equals("empty String"))
							Util.log_error(ex.getMessage(), str);
						break;
					}
				}
				peak.n_hmbc++;
			}
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * Performs a "binning" of the data and lists the result
 * @param array
 * @param start_ppm
 * @param end_ppm
 * @param width_ppm
 * @param method
 * @param q
 * @param data_max
 */
	public static void binData(double[] array, double start_ppm, double end_ppm, 
			double width_ppm, int method, data_par q, double data_max)
	{
		double start, end, width;

		start = start_ppm * q.SF;
		end = end_ppm * q.SF;
		width = width_ppm * q.SF;


		String bin_file_name = q.data_folder + "bin_file.txt";
		
		PrintWriter pw;
		try
		{
			pw = new PrintWriter(bin_file_name);
		}
		catch (Exception ex)
		{
			Util.show_and_log("list_peaks(): " + ex.getMessage(), bin_file_name);
			return;
		}

		//print a header
		if (method == 0)
			pw.println(" start (ppm)      stop  (ppm)        integral");
		else if (method == 1)
			pw.println(" start (ppm)      stop  (ppm)              max (raw)  max(normalized)");
		else if (method == 2)
			pw.println(" start (ppm)      stop  (ppm)                  mean");


		if (method == 0  || method == 2)        // integrals and mean
		{
			double freq = start;

			double total = 0.0;

			do      // find total of all bins
			{
				for (int i = Util.freq_to_index(freq, q); 
						i < Util.freq_to_index(freq - width, q); i++)
					total = total + array[i];
				freq = freq - width;
			} while (freq > end + width);

			freq = start;

			do      // find integral in each bin
			{
				double sum = 0.0;
				int count = 0;
				for (int i = Util.freq_to_index(freq, q); 
						i < Util.freq_to_index(freq - width, q); i++)
				{
					sum = sum + array[i];
					count++;
				}
				if(method == 0)
					pw.println(String.format(
							" %10.3f       %10.3f       %10.4f", 
							freq / q.SF, (freq - width) / q.SF, sum/total * 100.0));
				else // must be method = 2
					pw.println(String.format(
							" %10.3f       %10.3f       %16.1f",
							freq / q.SF, (freq - width) / q.SF, sum / (double)count));

				freq = freq - width;
			} while (freq > end + width) ;      
		}

		else if (method == 1)   // maximum
		{
			double freq = start;
			do      // find integral in each bin
			{
				double max = -1e50;
				for (int i = Util.freq_to_index(freq, q); 
						i < Util.freq_to_index(freq - width, q); i++)
				{
					if (array[i] > max)
						max = array[i];
				}

				pw.println(String.format(
						" %10.3f       %10.3f       %16.1f       %10.4f ",
						freq / q.SF, (freq - width) / q.SF, max, max/data_max));

				freq = freq - width;
			} while (freq > end + width) ;       
		}               

		else
		{
			Util.show_and_log("unknown binning method???");
		}

		pw.close();

		//List with whatever editor is associated with the ".txt" file type
		//user can the save the file with a different name from the editor
		
		if (Desktop.isDesktopSupported()) // new way!  portable??
		{
		    try 
		    {
		        Desktop.getDesktop().open(new File(bin_file_name));
		    } 
		    catch (IOException ex) 
		    {
		        Util.show_and_log("no application for this file type?? " +
		             bin_file_name, ex.getMessage());
		    }
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * Performs a "binning" of the data and lists the result
 * @param array
 * @param start_ppm
 * @param end_ppm
 * @param width_ppm
 * @param method
 * @param q
 * @param data_max
 * @param pw
 */
	public static void binData(double[] array, double start_ppm, double end_ppm,
			double width_ppm, int method, data_par q, double data_max, PrintWriter pw)
	{
		// TODO update format codes for Java
		double start, end, width;

		start = start_ppm * q.SF;
		end = end_ppm * q.SF;
		width = width_ppm * q.SF;

		// print a header
		if (method == 0)
			pw.println(" start (ppm)      stop  (ppm)        integral");
		else if (method == 1)
			pw.println(" start (ppm)      stop  (ppm)              max (raw)  max(normalized)");
		else if (method == 2)
			pw.println(" start (ppm)      stop  (ppm)                  mean");


		if (method == 0 || method == 2)        // integrals and mean
		{
			double freq = start;

			double total = 0.0;

			do      // find total of all bins
			{
				for (int i = Util.freq_to_index(freq, q);
						i < Util.freq_to_index(freq - width, q); i++)
					total = total + array[i];
				freq = freq - width;
			} while (freq > end + width);

			freq = start;

			do      // find integral in each bin
			{
				double sum = 0.0;
				int count = 0;
				for (int i = Util.freq_to_index(freq, q);
						i < Util.freq_to_index(freq - width, q); i++)
				{
					sum = sum + array[i];
					count++;
				}
				if (method == 0)
					pw.println(String.format(
							" %10.3f       %10.3f       %10.3f",
							freq / q.SF, (freq - width) / q.SF, sum / total * 100.0));
				else // must be method = 2
					pw.println(String.format(
							" %10.3f       %10.3f       %16.1f",
							freq / q.SF, (freq - width) / q.SF, sum / (double)count));

				freq = freq - width;
			} while (freq > end + width);
		}

		else if (method == 1)   // maximum
		{
			double freq = start;
			do      // find integral in each bin
			{
				double max = -1e50;
				for (int i = Util.freq_to_index(freq, q);
						i < Util.freq_to_index(freq - width, q); i++)
				{
					if (array[i] > max)
						max = array[i];
				}

				pw.println(String.format(
						" %10.3f       %10.3f       %16.1f       %10.4f ",
						freq / q.SF, (freq - width) / q.SF, max, max / data_max));

				freq = freq - width;
			} while (freq > end + width);
		}

		else
		{
			Util.show_and_log("unknown binning method");
		}           
	}
	
//-------------------------------------------------------------------------------------------------
}


