package spinworks;

/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.MessageBox;
import lp.Lp;

/**
 * 2D Processing routines
 * @author Kirk Marat, 
 * University of Manitoba
 */
public class Proc_2D 
{	
//-------------------------------------------------------------------------------------------------
/**
 * Process a the rows (direct or acquisition dimension) of 2D file
 * @param gp a SpinWorks global parameter object
 * @param p set acquisition parameters in the acquisition dimension
 * @param q status acquisition parameters in the acquisition dimension
 * @param p_ind set acquisition parameters in the indirect (evolution) dimension
 * @param q_ind status acquisition parameters in the indirect (evolution) dimension
 * @return true if OK, false on error
 */
	public static boolean proc_rows(global_par gp, data_par p, data_par q, 
			data_par p_ind, data_par q_ind)
	{
		int i;

		Util.free_proc(gp);									// free existing space  necessary??
		Util_2D.free_rows(true, true, true, true, gp);
		Util_2D.alloc_rows(p_ind.SI, p.SI, gp, q, q_ind);
		q.td_eff = q.TD;

		if (p_ind.MC2 == SW_main.det_modes_IND.ECHO_ANTIECHO 
				&& !gp.ea_sort_done )							// sort echo-antiecho data into
			sort_echo_antiecho(gp, q, q_ind);   // pure sin-cos data

		// allocate space for proc_buffer
		if (q.aq_mode == SW_main.det_modes.SIM			  // simultaneous  
				|| q.aq_mode == SW_main.det_modes.DSP
				|| q.aq_mode == SW_main.det_modes.CUSTOM_DSP)	
			Util.alloc_proc_buffer(2*q.SI, gp);

		else if (q.aq_mode == SW_main.det_modes.SEQ)	   // sequential 
			Util.alloc_proc_buffer(4*q.SI, gp);

		if (q.data_mode != SW_main.modes.MODE_3D)
			SW_main.progressBar.setMaximum(q_ind.TD);
		
		for(i=0; i < q_ind.TD && i < q_ind.SI; i++)        // loop over all rows
		{
			SW_main.current_row = i + 1;
			
			if (q.data_mode != SW_main.modes.MODE_3D)      // slows down 3D processing ??
			{
				SW_main.statusStripLabel3.setText(String.format("row %d of %d:", 
						SW_main.current_row, q_ind.TD));
				SW_main.progressBar.setSelection(SW_main.current_row);
			}

			if(Util_2D.load_fid_2D_dir(i, gp, p, q) == false)   // load proc. buffer
			{	
				Util.show_and_log(String.format("Cannot read row # %d", i));
				return false;									// bail out on problem
			}

			if (q.lp == SW_main.LP_modes.BACK_REAL
					|| q.lp == SW_main.LP_modes.BACK_COMPLEX)	// do backwards LP if requested
				Lp.do_back_lp_f2(gp, p, q);
			else if (q.lp == SW_main.LP_modes.ZHU_BAX)
				Lp.do_ZhuBax_LP(SW_main.direction.DIRECT, gp, p, q, p_ind, q_ind);

			Fft.do_fp_and_bias(gp, p, q);						// first point and bias correction 

			if (gp.log_on && SW_main.current_row == 1) Util.log_error(String.format(
					"Windowing F2  wdw_fcn = %d", p.WDW.ordinal()));

			window_2D(SW_main.direction.DIRECT, gp, p, q, p_ind, q_ind);	// apply window function

			Filter.do_filter(p.filter_type, p.filter_points, 
						q.td_eff, gp.proc_buffer);					// apply solvent filter
			 
			if (p.cls != 0)											// apply circular left shift 
				Fft.circular_ls(gp, p, q);

			Fft.do_fft(gp.proc_buffer, q.td_eff, q.SI, q.aq_mode);	// transform it
			post_proc_2D_dir(i, gp, q);								// extract into rows arrays
			phase_2D_dir(i, gp, p, q);			    				// apply any requested phasing
		}                                                           // end: loop over rows

		q.WDW = p.WDW;	q.PH_mod = p.PH_mod;

		if (q.PH_mod == SW_main.phasing.CONSTANTS)         // record applied phase 
		{                                                  // constants
			q.PHC0 = p.PHC0;
			q.PHC1 = p.PHC1;
		}

		q.hzperpt = q.SW * q.SFO1 / (double)(q.SI);	       // update digital resolution

		gp.dimensions_processed = 1;
		
		if (q.data_mode != SW_main.modes.MODE_3D)
		{
			SW_main.progressBar.setSelection(0);
			SW_main.statusStripLabel3.setText("");
		}
		return true;									   // for now
	}

//-------------------------------------------------------------------------------------------------
/**
 * Process the columns (evolution dimension) of a 2D file (indirect dimension transforms)
 * @param gp a SpinWorks global parameter object
 * @param p set acquisition parameters in the acquisition dimension
 * @param q status acquisition parameters in the acquisition dimension
 * @param p_ind set acquisition parameters in the indirect (evolution) dimension
 * @param q_ind status acquisition parameters in the indirect (evolution) dimension
 * @return true if OK, false on error
 */
	public static boolean proc_columns(global_par gp, data_par p, data_par q, 
			data_par p_ind, data_par q_ind)
	{
		int i;

		switch(p_ind.MC2)
		{
		case QF:
		case STATES:
		case STATES_TPPI:
		case ECHO_ANTIECHO:
			Util.alloc_proc_buffer(2*q_ind.SI, gp);
			break;

		case TPPI:
		case SEQUENTIAL:
			Util.alloc_proc_buffer(4*q_ind.SI, gp);
			break;

		default:
			Util.show_and_log("proc_columns(): unknown detection mode");
			return false;
		}

		if (gp.log_on) Util.log_error(String.format(
				"Processing columns, size of proc buffer = %d doubles",
				gp.proc_buffer.length));
		
		if (q.data_mode != SW_main.modes.MODE_3D)
			SW_main.progressBar.setMaximum(q.SI);

		if (p_ind.MC2 == SW_main.det_modes_IND.QF)		// two quadrant transform
		{
			if (gp.log_on) Util.log_error("F1: two quadrant transform");

			q_ind.td_eff = 2 * q_ind.TD;        // may be updated by LP routines

			for(i = 0; i < q.SI; i++)		    // loop over columns		
			{
				if(Util_2D.load_fid_2D_ind_QF(i, gp, p, q, p_ind, q_ind) == false)	// load data to proc. buffer	
				{
					Util.show_and_log(String.format("Cannot load column # %d", i));
					return false;
				}

				SW_main.current_column = i + 1;
				
				if (q.data_mode != SW_main.modes.MODE_3D)      // slows down 3D processing ??
				{
					SW_main.statusStripLabel3.setText(String.format("column %d of %d:", 
							SW_main.current_column, q.SI));
					SW_main.progressBar.setSelection(SW_main.current_column);
				}

				if (p_ind.lp == SW_main.LP_modes.FORWARD_COMPLEX
						|| p_ind.lp == SW_main.LP_modes.FORWARD_REAL)	// forward LP
				{

					if (Lp.do_ZhuBax_LP(SW_main.direction.INDIRECT, gp, p, q, p_ind, q_ind) == false)
						break;
				}

				else if (p_ind.lp == SW_main.LP_modes.ZHU_BAX)
				{
					if (Lp.do_ZhuBax_LP(SW_main.direction.INDIRECT, gp, p, q, p_ind, q_ind) == false)
						break;
				}

				gp.proc_buffer[0] = gp.proc_buffer[0] * p_ind.fcor;	// first point correction
				gp.proc_buffer[1] = gp.proc_buffer[1] * p_ind.fcor;

				if (gp.log_on && SW_main.current_column == 1) Util.log_error(String.format(
						"Windowing indirect  2QT wdw_fcn = %d", p_ind.WDW.ordinal()));

				window_2D(SW_main.direction.INDIRECT, gp, p, q, p_ind, q_ind);
				q_ind.WDW = p_ind.WDW;								// update window status parameter

				do_fft_ind(gp, p_ind, q_ind);
				post_proc_2D_ind(i, 0, gp, p_ind, q_ind);		// extract into  arrays
				phase_2D_ind(i, 0, gp, p_ind, q_ind);			// apply any requested phasing
			}	// for i...
		}	// if QF...

		else                                        // four quadrant transform
		{
			q_ind.td_eff = q_ind.TD;                // may be udated by LP routines

			if (gp.log_on) Util.log_error(String.format(
					"Indirect Dimension 4 QT:  TD = %d  SI = %d", q_ind.td_eff, q_ind.SI));

			for(i = 0; i < q.SI; i++)								// load data to proc. buffer
			{
				if(Util_2D.load_fid_2D_ind(i, gp.rows_real, gp, p_ind, q_ind) == false)	// real	
				{
					Util.show_and_log( String.format("Cannot load column # %d", i) + " fqt");
					return false;
				}


				SW_main.current_column = i + 1;
				
				if (q.data_mode != SW_main.modes.MODE_3D)      // slows down 3D processing ??
				{
					SW_main.statusStripLabel3.setText(String.format("column %d of %d:", 
							SW_main.current_column, q.SI));
					SW_main.progressBar.setSelection(SW_main.current_column);
				}

				if (p_ind.lp == SW_main.LP_modes.FORWARD_REAL
						|| p_ind.lp == SW_main.LP_modes.FORWARD_COMPLEX)    // FORWARD LP
				{

					if (Lp.do_ZhuBax_LP(SW_main.direction.INDIRECT, 
							gp, p, q, p_ind, q_ind) == false)
						break;
				}
				else if (p_ind.lp == SW_main.LP_modes.ZHU_BAX)
				{

					if (Lp.do_ZhuBax_LP(SW_main.direction.INDIRECT, 
							gp, p, q, p_ind, q_ind) == false)
						break;
				}

				if (p_ind.MC2 == SW_main.det_modes_IND.TPPI
						|| p_ind.MC2 == SW_main.det_modes_IND.SEQUENTIAL)	// fp correction for
					gp.proc_buffer[0] = gp.proc_buffer[0] * p_ind.fcor;     // real transform
				else
				{
					gp.proc_buffer[0] = gp.proc_buffer[0] * p_ind.fcor;	 // first point correction
					gp.proc_buffer[1] = gp.proc_buffer[1] * p_ind.fcor;  // for complex transform
				}

				if (gp.log_on && SW_main.current_column == 1) Util.log_error(String.format(
						"Windowing indirect  4QT real wdw_fcn = %d", p_ind.WDW.ordinal()));

				window_2D(SW_main.direction.INDIRECT, gp, p, q, p_ind, q_ind);

				do_fft_ind(gp, p_ind, q_ind);
				post_proc_2D_ind(i, 0, gp, p_ind, q_ind);	 	// extract into  arrays
				if(p_ind.PH_mod != SW_main.phasing.MAGNITUDE)
					phase_2D_ind(i, 0, gp, p_ind, q_ind);	// apply any requested phasing

				if(Util_2D.load_fid_2D_ind(i, gp.rows_imag, gp, p_ind, q_ind) == false)	// imag	
				{
					Util.show_and_log(String.format("Cannot load column # %d", i) + " fqt");
					return false;
				}

				if (p_ind.lp == SW_main.LP_modes.FORWARD_REAL
						|| p_ind.lp == SW_main.LP_modes.FORWARD_COMPLEX)	// forward LP
				{

					if (Lp.do_ZhuBax_LP(SW_main.direction.INDIRECT, 
							gp, p, q, p_ind, q_ind) == false)
						break;
				}
				else if (p_ind.lp == SW_main.LP_modes.ZHU_BAX)
				{

					if (Lp.do_ZhuBax_LP(SW_main.direction.INDIRECT, 
							gp, p, q, p_ind, q_ind) == false)
						break;
				}

				if (p_ind.MC2 == SW_main.det_modes_IND.TPPI
						|| p_ind.MC2 == SW_main.det_modes_IND.SEQUENTIAL)// fp correction for
					gp.proc_buffer[0] = gp.proc_buffer[0] * p_ind.fcor;	 // real transform
				else
				{
					gp.proc_buffer[0] = gp.proc_buffer[0] * p_ind.fcor;	// first point correction
					gp.proc_buffer[1] = gp.proc_buffer[1] * p_ind.fcor;	// for complex transform
				}

				if (gp.log_on & SW_main.current_column == 1) Util.log_error(String.format(
						"Windowing indirect  4QT imag wdw_fcn = %d", p_ind.WDW.ordinal()));

				window_2D(SW_main.direction.INDIRECT, gp, p, q, p_ind, q_ind);
				do_fft_ind(gp, p_ind, q_ind);
				post_proc_2D_ind(i, 1, gp, p_ind, q_ind);	    // extract into  arrays

				if (p_ind.PH_mod != SW_main.phasing.MAGNITUDE)                      
					phase_2D_ind(i, 1, gp, p_ind, q_ind);		// apply any requested phasing
				else if (p_ind.PH_mod == SW_main.phasing.MAGNITUDE)
					do_mc_2D_ind(i, gp, q_ind);
			}
		}

		q_ind.WDW = p_ind.WDW;							// update status prameters	
		q_ind.PH_mod = p_ind.PH_mod;
		q_ind.MC2 = p_ind.MC2;
		q_ind.lp = p_ind.lp;
		q_ind.lp_coef = p_ind.lp_coef;
		q_ind.lp_cutoff = p_ind.lp_cutoff;
		q_ind.lp_input = p_ind.lp_input;
		q_ind.lp_pred = p_ind.lp_pred;

		if (q_ind.PH_mod == SW_main.phasing.CONSTANTS)
		{
			q_ind.PHC0 = p_ind.PHC0;        
			q_ind.PHC1 = p_ind.PHC1;
		}

		Util.free_proc_buffer(gp);                      // legacy call

		if (p_ind.MC2 == SW_main.det_modes_IND.QF) 
			Util_2D.free_rows(false, true, false, true, gp);

		gp.data_max_2D = Util_2D.max_2D(gp.rows_real, q.SI, q_ind.SI);	// find global maximum
		gp.data_noise = Util_2D.noise_2D(gp, q, q_ind);

		q_ind.hzperpt = q_ind.SW * q_ind.SFO1 / (double)(q_ind.SI);     // update digital resolution
		
		gp.dimensions_processed = 2;
		
		if (q.data_mode != SW_main.modes.MODE_3D)
		{
			SW_main.progressBar.setSelection(0);
			SW_main.statusStripLabel3.setText("");
		}
		return true;
	}

	//tilt
//-------------------------------------------------------------------------------------------------
/// Phase the rows of a 2D file (separate from transform)
	public static void phase_rows(global_par gp, data_par p, data_par q, data_par q_ind)
	{
		int i;

		if (gp.log_on) Util.log_error(String.format(
				"Phase F2 with: ph0 = %6.2f  ph1 = %6.2f", p.PHC0, p.PHC1));

		for(i=0; i < q_ind.SI; i++)
		{
			SW_main.current_row = i + 1;
			do_pk_2D_f2(i, gp, p, q);
		}
	}

//-------------------------------------------------------------------------------------------------
/// Phase the columns of a 2D file (separate from transform)
	public static void phase_columns(global_par gp, data_par q, data_par p_ind, data_par q_ind)
	{
		int i;

		if (gp.log_on) Util.log_error(String.format(
				"Phase F1 with: ph0 = %6.2f  ph1 = %6.2f",
				p_ind.PHC0, p_ind.PHC1));

		for(i = 0; i < q.SI; i++)						
		{
			SW_main.current_column = i + 1;
			do_pk_2D_f1(i, 0, gp, p_ind, q_ind);		// apply any requested phasing
			do_pk_2D_f1(i, 1, gp, p_ind, q_ind);		// apply any requested phasing
		}
	}

//-------------------------------------------------------------------------------------------------
/// Do a magnitude calculation of the columns of a 2D file (separate from  transform
	public static void mc_columns(global_par gp, data_par q, data_par q_ind)
	{
		int i;
		
		 MessageBox messageBox = new MessageBox(SW_main.shell, SWT.ICON_QUESTION
    			 | SWT.YES | SWT.NO);
    	 messageBox.setMessage("Do F1 magnitide calculation ?");
    	 messageBox.setText("SpinWorks Java ");
    	 int response = messageBox.open();
    	 if (response == SWT.NO) return;

		for(i = 0; i < q.SI; i++)						
		{
			SW_main.current_column = i + 1;
			do_mc_2D_ind(i, gp, q_ind);				// do magnitude calculation
		}

		if (gp.log_on) Util.log_error("F1 magnitude calculation complete");
	}

//-------------------------------------------------------------------------------------------------
/// Window the rows or column of a 2D of 3D file.  Data must be already loaded into the proc_buffer.
	public static void window_2D(SW_main.direction dir, global_par gp, data_par p, data_par q, 
			data_par p_ind, data_par q_ind)
	{
		SW_main.windows wdw_fcn;
		int td;
		double lb, gb, ssb, sw;
		boolean real_data = false;

		if (dir == SW_main.direction.INDIRECT)
		{
			wdw_fcn = p_ind.WDW;
			lb = p_ind.LB;
			gb = p_ind.GB;
			ssb = p_ind.ssb;        // SpinWorks sine shift in degrees
			td = q_ind.td_eff;
			sw = q_ind.SW * q_ind.SF;
			if (p_ind.MC2 == SW_main.det_modes_IND.TPPI)
				real_data = true;

		}
		else        // direct dimension
		{
			wdw_fcn = p.WDW;
			lb = p.LB;
			gb = p.GB;
			ssb = p.ssb;            // SpinWorks sine shift in degrees
			td = q.td_eff;
			sw = q.SW * q.SF;
			if (q.aq_mode == SW_main.det_modes.SEQ)
				real_data = true;
		}

		switch (wdw_fcn)
		{
		case NONE:
			break;

		case EM:
			//wfcn.do_em(dir, gp, p, q, p_ind, q_ind);
			Wfcn.do_em(gp.proc_buffer, td, lb, sw, real_data);
			if (dir == SW_main.direction.INDIRECT)
				q_ind.LB = p_ind.LB;
			else
				q.LB = p.LB;
			break;

		case GM:
			//wfcn.do_gm(dir, gp, p, q, p_ind, q_ind);
			Wfcn.do_gm(gp.proc_buffer, td, lb, gb, sw, real_data);
			if (dir == SW_main.direction.INDIRECT)
			{
				q_ind.LB = p_ind.LB;
				q_ind.GB = p_ind.GB;
			}
			else
			{
				q.LB = p.LB;
				q.GB = p.GB;
			}
			break;

		case SINE:
			//wfcn.do_sine(dir, gp, p, q, p_ind, q_ind);
			Wfcn.do_sine(gp.proc_buffer, td, ssb);
			if (dir == SW_main.direction.INDIRECT)
				q_ind.ssb = p_ind.ssb;
			else
				q.ssb = p.ssb;
			break;

		case QSINE:
			// wfcn.do_qsine(dir, gp, p, q, p_ind, q_ind);
			Wfcn.do_qsine(gp.proc_buffer, td, ssb);
			if (dir == SW_main.direction.INDIRECT)
				q_ind.ssb = p_ind.ssb;
			else
				q.ssb = p.ssb;
			break;

		case TRAF:
			//wfcn.do_traf(dir, gp, p, q, p_ind, q_ind);
			Wfcn.do_traf(gp.proc_buffer, td, lb, sw, real_data);
			if (dir == SW_main.direction.INDIRECT)
				q_ind.LB = p_ind.LB;
			else
				q.LB = p.LB;
			break;

		case TRAFS:
			//wfcn.do_trafs(dir, gp, p, q, p_ind, q_ind);
			Wfcn.do_trafs(gp.proc_buffer, td, lb, sw, real_data);
			if (dir == SW_main.direction.INDIRECT)
				q_ind.LB = p_ind.LB;
			else
				q.LB = p.LB;
			break;

		case GAUSSIAN:
			//wfcn.do_gaussian(dir, gp, p, q, p_ind, q_ind);
			Wfcn.do_gaussian(gp.proc_buffer, td, lb, sw, real_data);
			if (dir == SW_main.direction.INDIRECT)
				q_ind.LB = p_ind.LB;
			else
				q.LB = p.LB;
			break;

		default:
			break;
		}
	}

//-------------------------------------------------------------------------------------------------
/// Phase a row (direct dimension) of a 2D data  set
	public static void phase_2D_dir(int n, global_par gp, data_par p, data_par q)
	{
		switch (p.PH_mod)
		{
		case NONE:
			break;

		case CONSTANTS:
			do_pk_2D_f2(n, gp, p, q);
			break;

		case MAGNITUDE:
			do_mc_2D_f2(n, gp, q);
			break;

		case POWER:
			do_ps_2D_dir(n, gp, q);
			break;

		default:
			break;
		}
	}

//-------------------------------------------------------------------------------------------------
/// Phase a column of a 2D data set
	public static void phase_2D_ind(int n, int quad, global_par gp, data_par p_ind, data_par q_ind)
	{
		switch (p_ind.PH_mod)
		{
		case NONE:
			break;

		case CONSTANTS:
			do_pk_2D_f1(n, quad, gp, p_ind, q_ind);
			break;

		case MAGNITUDE:
			do_mc_2D_ind(n, gp, q_ind);
			break;

		case POWER:
			do_ps_2D_f1(n, gp, q_ind);
			break;

		default:
			break;
		}
	}

//-------------------------------------------------------------------------------------------------
/// do the fft on an F1 row in the processing buffer
	public static void do_fft_ind(global_par gp, data_par p_ind, data_par q_ind)
	{
		int i;

		switch (p_ind.MC2)
		{
		case QF:
		case STATES:
		case STATES_TPPI:
		case ECHO_ANTIECHO:
			//Fft.fft(gp.proc_buffer, p_ind.SI, 1);
			Fft.fftSineTable(gp.proc_buffer, p_ind.SI);
			break;

		case TPPI:
		case SEQUENTIAL:
			int td;
			if (q_ind.td_eff > 2 * q_ind.SI)
				td = 2 * q_ind.SI;
			else
				td = q_ind.td_eff;

			for (i = td - 1; i > 0; i--)	// for sequential (real) data 
			{								// expand into real locations only
				gp.proc_buffer[2 * i] = gp.proc_buffer[i];
				gp.proc_buffer[i] = 0.0;
			}
			//Fft.fft(gp.proc_buffer, 2 * q_ind.SI, 1);
			Fft.fftSineTable(gp.proc_buffer, 2 * q_ind.SI);
			break;

		default:
			break;
		}
		q_ind.hzperpt = q_ind.SW * q_ind.SFO1 / (double)(q_ind.SI);
	}

//------------------------------------------------------------------------------------------------
/// special pre-processing for echo-antiecho data, 
/// turns p-n type data into sin-cos (hypercomplex) data  
/// (This method works for both Bruker and Varian/Agilent data)
	public static void sort_echo_antiecho(global_par gp, data_par q, data_par q_ind)
	{
		double[] fid0,  fid1;
		int i, j;

		fid0 = new double[q.TD]; 
		fid1 = new double[q.TD]; 

		for(i = 0; i < q_ind.TD; i = i+2)
		{
			for(j = 0; j < q.TD; j++)
			{
				fid0[j] = gp.ser[i][j];	
				fid1[j] = gp.ser[i+1][j];
			}
			for(j = 0; j < q.TD; j = j+2)
			{
				gp.ser[i][j]		= -fid1[j+1]	- fid0[j+1];			// re0
				gp.ser[i][j+1]		=  fid1[j]		+ fid0[j];				// im0
				gp.ser[i+1][j]		=  fid1[j]		- fid0[j];				// re1
				gp.ser[i+1][j+1]	=  fid1[j+1]	- fid0[j+1];			// im0
			}
		}

		// in 3D, a fresh "ser" is brought in for every plane
		if(q.data_mode != SW_main.modes.MODE_3D) gp.ea_sort_done = true;
		if(gp.log_on)Util.log_error("Bruker  sort_echo_antiecho done");	
	}

//-------------------------------------------------------------------------------------------------
/// handle data extraction and ordering for of 2D rows (direct dimension)
	public static void post_proc_2D_dir(int n, global_par gp, data_par q)
	{
		int i;

		if (q.aq_mode == SW_main.det_modes.SIM					// simultaneous 
				|| q.aq_mode == SW_main.det_modes.DSP
				|| q.aq_mode == SW_main.det_modes.CUSTOM_DSP)		
		{
			for (i = q.SI / 2; i < q.SI; i++)				// extract real part of data 
				gp.rows_real[n][i] = (float)gp.proc_buffer[2 * (i - q.SI / 2)];
			for (i = 0; i < q.SI / 2; i++)
				gp.rows_real[n][i] = (float)gp.proc_buffer[2 * (i + q.SI / 2)];

			for (i = q.SI / 2; i < q.SI; i++)				// extract imaginary part of data 
				gp.rows_imag[n][i] = (float)gp.proc_buffer[2 * (i - q.SI / 2) + 1];
			for (i = 0; i < q.SI / 2; i++)
				gp.rows_imag[n][i] = (float)gp.proc_buffer[2 * (i + q.SI / 2) + 1];
		}

		else if (q.aq_mode == SW_main.det_modes.SEQ)			// sequential              
			for (i = 0; i < q.SI; i++)
			{
				gp.rows_real[n][i] = (float)gp.proc_buffer[(2 * q.SI - 1) - (2 * i + 1)];
				gp.rows_imag[n][i] = (float)gp.proc_buffer[(2 * q.SI - 1) - (2 * i)];
			}

		// apply special phase correction to Bruker AVANCE DSP processed data 
		if (q.aq_mode == SW_main.det_modes.DSP || q.aq_mode == SW_main.det_modes.CUSTOM_DSP)
			Fft.dsp_phase(n, gp, q);
	}

//-------------------------------------------------------------------------------------------------
/// handle data extraction and ordering for 2D columns
	public static void post_proc_2D_ind(int n, int quad, global_par gp, data_par p_ind, data_par q_ind)
	{
		int i;

		switch (p_ind.MC2)
		{
		case QF:
		case STATES:
		case STATES_TPPI:
		case ECHO_ANTIECHO:
			if (quad == 0)								// rr and ir quadrants
			{
				for (i = q_ind.SI / 2; i < q_ind.SI; i++)		// extract real part of data 
					gp.rows_real[i][n] = (float)(gp.proc_buffer[2 * (i - q_ind.SI / 2)]);
				for (i = 0; i < q_ind.SI / 2; i++)
					gp.rows_real[i][n] = (float)(gp.proc_buffer[2 * (i + q_ind.SI / 2)]);

				for (i = q_ind.SI / 2; i < q_ind.SI; i++)		// extract imag. part of data 
					gp.irows_real[i][n] = (float)(gp.proc_buffer[2 * (i - q_ind.SI / 2) + 1]);
				for (i = 0; i < q_ind.SI / 2; i++)
					gp.irows_real[i][n] = (float)(gp.proc_buffer[2 * (i + q_ind.SI / 2) + 1]);
			}

			else if (quad == 1)							// ri and ii quadrants
			{
				for (i = q_ind.SI / 2; i < q_ind.SI; i++)		// extract real part of data 
					gp.rows_imag[i][n] = (float)(gp.proc_buffer[2 * (i - q_ind.SI / 2)]);
				for (i = 0; i < q_ind.SI / 2; i++)
					gp.rows_imag[i][n] = (float)(gp.proc_buffer[2 * (i + q_ind.SI / 2)]);

				for (i = q_ind.SI / 2; i < q_ind.SI; i++)		// extract imag. part of data 
					gp.irows_imag[i][n] = (float)(gp.proc_buffer[2 * (i - q_ind.SI / 2) + 1]);
				for (i = 0; i < q_ind.SI / 2; i++)
					gp.irows_imag[i][n] = (float)(gp.proc_buffer[2 * (i + q_ind.SI / 2) + 1]);
			}

			break;

		case SEQUENTIAL:
		case TPPI:
			if (quad == 0)
			{
				for (i = 0; i < q_ind.SI; i++)
				{
					gp.rows_real[q_ind.SI - 1 - i][n]
							= (float)(gp.proc_buffer[(2 * q_ind.SI - 1) - (2 * i + 1)]);
					gp.irows_real[q_ind.SI - 1 - i][n]
							= (float)(gp.proc_buffer[(2 * q_ind.SI - 1) - (2 * i)]);
				}
			}
			else if (quad == 1)
			{
				for (i = 0; i < q_ind.SI; i++)
				{
					gp.rows_imag[q_ind.SI - 1 - i][n]
							= (float)(gp.proc_buffer[(2 * q_ind.SI - 1) - (2 * i + 1)]);
					gp.irows_imag[q_ind.SI - 1 - i][n]
							= (float)(gp.proc_buffer[(2 * q_ind.SI - 1) - (2 * i)]);
				}
			}

			break;

		default:
			break;
		}         
	}

//-------------------------------------------------------------------------------------------------
/// phase correction of 2D row n
	public static void do_pk_2D_f2(int n, global_par gp, data_par p, data_par q)
	{
		int i;
		double a, t_r, t_i;
		// The value 0.0174533 converts degrees to radians

		for (i = 0; i < q.SI; i++)   // now do the phasing, real rows
		{
			a = 0.0174533 * (p.PHC0 + ((double)i / (double)q.SI) * p.PHC1);
			t_r = (double)gp.rows_real[n][i];
			t_i = (double)gp.rows_imag[n][i];
			gp.rows_real[n][i] = (float)(t_r * Math.cos(a) - t_i * Math.sin(a));
			gp.rows_imag[n][i] = (float)(t_i * Math.cos(a) + t_r * Math.sin(a));
		}

		if (gp.irows_real == null || gp.irows_imag == null)
			return;

		for (i = 0; i < q.SI; i++)   // now the imaginary rows
		{
			a = 0.0174533 * (p.PHC0 + ((double)i / (double)q.SI) * p.PHC1);
			t_r = (double)gp.irows_real[n][i];
			t_i = (double)gp.irows_imag[n][i];
			gp.irows_real[n][i] = (float)(t_r * Math.cos(a) - t_i * Math.sin(a));
			gp.irows_imag[n][i] = (float)(t_i * Math.cos(a) + t_r * Math.sin(a));
		}
	}

//-------------------------------------------------------------------------------------------------
/// phase correct column n
	public static void do_pk_2D_f1(int n, int quad, global_par gp, data_par p_ind, data_par q_ind)
	{
		int i;
		double a, t_r, t_i;


		if (quad == 0)
			for (i = 0; i < q_ind.SI; i++)   // now do the phasing, reals
			{
				a = 0.0174533 * (p_ind.PHC0 + ((double)i / (double)q_ind.SI) * p_ind.PHC1);
				t_r = (double)gp.rows_real[i][n];
				t_i = (double)gp.irows_real[i][n];
				gp.rows_real[i][n] = (float)(t_r * Math.cos(a) - t_i * Math.sin(a));
				gp.irows_real[i][n] = (float)(t_i * Math.cos(a) + t_r * Math.sin(a));
			}

		else if (quad == 1)
			for (i = 0; i < q_ind.SI; i++)   // now do the phasing, imaginaries 
			{
				a = 0.0174533 * (p_ind.PHC0 + ((double)i / (double)q_ind.SI) * p_ind.PHC1);
				t_r = (double)gp.rows_imag[i][n];
				t_i = (double)gp.irows_imag[i][n];
				gp.rows_imag[i][n] = (float)(t_r * Math.cos(a) - t_i * Math.sin(a));
				gp.irows_imag[i][n] = (float)(t_i * Math.cos(a) + t_r * Math.sin(a));
			}
	}


//-------------------------------------------------------------------------------------------------
/// magnitude calculation of 2D row n (not often used!)
	public static void do_mc_2D_f2(int n, global_par gp, data_par q)
	{
		if (gp.rows_real == null || gp.rows_imag == null) return;

		int i;
		double t_r, t_i;

		for (i = 0; i < q.SI; i++)
		{
			t_r = (double)gp.rows_real[n][i];
			t_i = (double)gp.rows_imag[n][i];
			gp.rows_real[n][i] = (float)(Math.sqrt(t_r * t_r + t_i * t_i));
		}
	}

//-------------------------------------------------------------------------------------------------
/// magnitude calculation of the column n of a 2D file real-real quadrant ONLY
	public static void do_mc_2D_ind(int n, global_par gp, data_par q_ind)
	{
		if (gp.irows_real == null || gp.rows_real == null) return;	

		int i;
		double t_r, t_i;												

		for (i = 0; i < q_ind.SI; i++)
		{
			t_r = (double)gp.rows_real[i][n];
			t_i = (double)gp.irows_real[i][n];
			gp.rows_real[i][n] = (float)(Math.sqrt(t_r * t_r + t_i * t_i));        
		}                                              
	}

//-------------------------------------------------------------------------------------------------
/// power spectrum of 2D row n (not often used!)
	public static void do_ps_2D_dir(int n, global_par gp, data_par q)
	{
		if (gp.rows_real == null || gp.rows_imag == null) return;

		int i;
		double t_r, t_i;

		for (i = 0; i < q.SI; i++)
		{
			t_r = (double)gp.rows_real[n][i];
			t_i = (double)gp.rows_imag[n][i];
			gp.rows_real[n][i] = (float)((t_r * t_r + t_i * t_i) / 100000.0);
		}
	}

//-------------------------------------------------------------------------------------------------
/// power spectrum of 2D column n
	public static void do_ps_2D_f1(int n, global_par gp, data_par q_ind)
	{
		if(gp.rows_real == null || gp.irows_imag == null) return;

		int i;
		double t_r, t_i;

		for (i = 0; i < q_ind.SI; i++)
		{
			t_r = (double)gp.rows_real[i][n];
			t_i = (double)gp.irows_imag[i][n];
			gp.rows_real[i][n] = (float)((t_r * t_r + t_i * t_i) / 100000.0);
		}
	}

//-------------------------------------------------------------------------------------------------
/// Apply a frequency domain smoothing function to 2D real-real data 
/// Each point is replaced by an average of 1/2 the point in question and 
/// 1/16 of each of the adjacent points (binomial smooth) 
/// can improce sensitivity at expense of resolution.  Time domain windowing 
/// generally better.
	public static void smooth_2D(global_par gp,  data_par q, data_par q_ind)
	{
		int i, j;

		float[][] temp_mat = new float[q_ind.SI][q.SI];

		for (i = 0; i < q_ind.SI; i++)
			for (j = 0; j < q.SI; j++)
				temp_mat[i][j] = gp.rows_real[i][j];

		for (i = 1; i < q_ind.SI - 1; i++)
			for (j = 1; j < q.SI - 1; j++)
				gp.rows_real[i][j] = (float)(0.5 * (double)temp_mat[i][j]
						+ 0.0625 * (double)temp_mat[i + 1][j]
						+ 0.0625 * (double)temp_mat[i - 1][j]
						+ 0.0625 * (double)temp_mat[i][j + 1]
						+ 0.0625 * (double)temp_mat[i][j - 1]
						+ 0.0625 * (double)temp_mat[i + 1][j + 1]
						+ 0.0625 * (double)temp_mat[i - 1][j - 1]
						+ 0.0625 * (double)temp_mat[i + 1][j - 1]
						+ 0.0625 * (double)temp_mat[i - 1][j + 1]);
	}

//-------------------------------------------------------------------------------------------------
/// Symmetrize a  2D matrix
	public static void sym_2D(int mode, global_par gp, data_par q_ind)
	{		
		int i, j;

		if(mode == 0)			// magnitude data	
		{
			for(i = 0; i < q_ind.SI; i++)
				for(j = 0; j < i; j++)
					if(gp.rows_real[i][j] < gp.rows_real[j][i])
						gp.rows_real[j][i] = gp.rows_real[i][j];
					else
						gp.rows_real[i][j] = gp.rows_real[j][i];
		}

		else if(mode == 1)		// phase sensitive (necessary???)
		{
			for(i = 0; i < q_ind.SI; i++)
				for(j = 0; j < i; j++)
					if(Math.abs(gp.rows_real[i][j]) < Math.abs(gp.rows_real[j][i]))
						gp.rows_real[j][i] = gp.rows_real[i][j];
					else
						gp.rows_real[i][j] = gp.rows_real[j][i];
		}
	}

//-------------------------------------------------------------------------------------------------
/// Do a 2D MQ MAS shearing transform  UNDER CONSTRUCTION!!!
	public static int proc_rows_shear(global_par gp, data_par p, data_par q,
			data_par p_ind, data_par q_ind)
	{			
		double ph0_orig, ph1_orig;
		int i;

		ph0_orig = (double)p.PHC0;					// save original phase constants
		ph1_orig = (double)p.PHC1;

		Util.free_proc(gp);							// free existing space  necessary??
		Util_2D.free_rows(true, true, true, true, gp);
		Util_2D.alloc_rows(p_ind.SI, p.SI, gp, q, q_ind);

		// allocate space for proc_buffer
		if (q.aq_mode == SW_main.det_modes.SIM			// simultaneous  
				|| q.aq_mode == SW_main.det_modes.DSP
				|| q.aq_mode == SW_main.det_modes.CUSTOM_DSP)
			Util.alloc_proc_buffer(2 * q.SI, gp);

		else if (q.aq_mode == SW_main.det_modes.SEQ)	// sequential 
			Util.alloc_proc_buffer(4 * q.SI, gp);

		for (i = 0; i < q_ind.TD && i < q_ind.SI; i++)
		{
			SW_main.current_row = i + 1;

			if (Util_2D.load_fid_2D_dir(i, gp, p, q) == false)	// load data to proc. buffer
			{
				Util.show_and_log(String.format("proc_rows(): cannot load block # %d", i));				
				return -1;								// bail out on problem
			}

			if (q.lp == SW_main.LP_modes.BACK_REAL
					|| q.lp == SW_main.LP_modes.BACK_COMPLEX)	// do backwards LP if requested
				Lp.do_back_lp_f2(gp, p, q);

			window_2D(SW_main.direction.DIRECT, gp, p, q, p_ind, q_ind);	// apply window function
			q.WDW = p.WDW;											// update window status parameter

			Fft.do_fft(gp.proc_buffer, q.td_eff, q.SI, q.aq_mode);  // do fft

			post_proc_2D_dir(i, gp, q);								// extract into rows arrays

			p.PHC0 = ph0_orig;								        // restore phase angles
			p.PHC1 = ph1_orig;
			phase_2D_dir(i, gp, p, q);                				// apply any requested phasing

			p.PHC0 = ((double)(i / 2) * p.mq_shear_angle / -2.0);	// calculate shearing phase angles
			p.PHC1 = ((double)(i / 2) * p.mq_shear_angle);			// pivot at centre of spectrum

			do_pk_2D_f2(i, gp, p, q);								// apply any requested phasing              
		}


		p.PHC0 = ph0_orig;								    		// restore phase angles
		p.PHC1 = ph1_orig;

		q.WDW = p.WDW; q.PH_mod = p.PH_mod;	
		q.WDW = p.WDW; q.PH_mod = p.PH_mod;

		Util.free_proc_buffer(gp);

		return 1;													// everything O.K.
	}

//-------------------------------------------------------------------------------------------------
/// Hilbert transform of 2D row n (regenerates imaginary data)
	public static void do_ht_2D_dir(int n, global_par gp, data_par q)
	{
		int i;

		for (i = q.SI / 2; i < q.SI; i++)                  // load real part of data 
			gp.proc_buffer[2 * (i - q.SI / 2)] = gp.rows_real[n][ i];
		for (i = 0; i < q.SI / 2; i++)
			gp.proc_buffer[2 * (i + q.SI / 2)] = gp.rows_real[n][ i];


		for (i = q.SI / 2; i < q.SI; i++)                 // zero imaginary part of data              
			gp.proc_buffer[2 * (i - q.SI / 2) + 1] = 0.0;
		for (i = 0; i < q.SI / 2; i++)
			gp.proc_buffer[2 * (i + q.SI / 2) + 1] = 0.0;

		Fft.fft(gp.proc_buffer, q.SI, -1);                // inverse FFT


		for (i = q.SI; i < 2 * q.SI; i++)
			gp.proc_buffer[i] = 0.0;

		gp.proc_buffer[0] = 0.5 * gp.proc_buffer[0];       // first point correction
		gp.proc_buffer[1] = 0.5 * gp.proc_buffer[1];

		Fft.fft(gp.proc_buffer, q.SI, 1);                  // forward FFT

		for (i = q.SI / 2; i < q.SI; i++)                  // extract real part of data 
			gp.rows_real[n][ i] = (float)(gp.proc_buffer[2 * (i - q.SI / 2)] / ((double)q.SI/2.0));
		for (i = 0; i < q.SI / 2; i++)
			gp.rows_real[n][i] = (float)(gp.proc_buffer[2 * (i + q.SI / 2)] / ((double)q.SI / 2.0));

		for (i = q.SI / 2; i < q.SI; i++)                  // extract imaginary part of data 
			gp.rows_imag[n][i]
					= (float)(gp.proc_buffer[2 * (i - q.SI / 2) + 1] / ((double)q.SI / 2.0));
		for (i = 0; i < q.SI / 2; i++)
			gp.rows_imag[n][ i]
					= (float)(gp.proc_buffer[2 * (i + q.SI / 2) + 1] / ((double)q.SI / 2.0));

		//--------


		for (i = q.SI / 2; i < q.SI; i++)                  // load real part of data 
			gp.proc_buffer[2 * (i - q.SI / 2)] = gp.irows_real[n][ i];
		for (i = 0; i < q.SI / 2; i++)
			gp.proc_buffer[2 * (i + q.SI / 2)] = gp.irows_real[n][ i];


		for (i = q.SI / 2; i < q.SI; i++)                 // zero imaginary part of data              
			gp.proc_buffer[2 * (i - q.SI / 2) + 1] = 0.0;
		for (i = 0; i < q.SI / 2; i++)
			gp.proc_buffer[2 * (i + q.SI / 2) + 1] = 0.0;

		Fft.fft(gp.proc_buffer, q.SI, -1);                // inverse FFT


		for (i = q.SI; i < 2 * q.SI; i++)
			gp.proc_buffer[i] = 0.0;

		gp.proc_buffer[0] = 0.5 * gp.proc_buffer[0];       // first point correction
		gp.proc_buffer[1] = 0.5 * gp.proc_buffer[1];

		Fft.fft(gp.proc_buffer, q.SI, 1);                  // forward FFT

		for (i = q.SI / 2; i < q.SI; i++)                  // extract real part of data 
			gp.irows_real[n][ i] = (float)(gp.proc_buffer[2 * (i - q.SI / 2)] / ((double)q.SI / 2.0));
		for (i = 0; i < q.SI / 2; i++)
			gp.irows_real[n][ i] = (float)(gp.proc_buffer[2 * (i + q.SI / 2)] / ((double)q.SI / 2.0));

		for (i = q.SI / 2; i < q.SI; i++)                  // extract imaginary part of data 
			gp.irows_imag[n][ i]
					= (float)(gp.proc_buffer[2 * (i - q.SI / 2) + 1] / ((double)q.SI / 2.0)); 
		for (i = 0; i < q.SI / 2; i++)
			gp.irows_imag[n][ i]
					= (float)(gp.proc_buffer[2 * (i + q.SI / 2) + 1] / ((double)q.SI / 2.0));                
	}

//-------------------------------------------------------------------------------------------------
/// Hilbert transform of 2D column n (regenerates imaginary data)
	public static void do_ht_2D_ind(int n, global_par gp, data_par q_ind)
	{
		int i;

		for (i = q_ind.SI / 2; i < q_ind.SI; i++)         // load real part of data 
			gp.proc_buffer[2 * (i - q_ind.SI / 2)] = gp.rows_real[i][ n];
		for (i = 0; i < q_ind.SI / 2; i++)
			gp.proc_buffer[2 * (i + q_ind.SI / 2)] = gp.rows_real[i][ n];


		for (i = q_ind.SI / 2; i < q_ind.SI; i++)        // zero imaginary part of data              
			gp.proc_buffer[2 * (i - q_ind.SI / 2) + 1] = 0.0;
		for (i = 0; i < q_ind.SI / 2; i++)
			gp.proc_buffer[2 * (i + q_ind.SI / 2) + 1] = 0.0;

		Fft.fft(gp.proc_buffer, q_ind.SI, -1);           // inverse FFT


		for (i = q_ind.SI; i < 2 * q_ind.SI; i++)
			gp.proc_buffer[i] = 0.0;

		gp.proc_buffer[0] = 0.5 * gp.proc_buffer[0];     // first point correction
		gp.proc_buffer[1] = 0.5 * gp.proc_buffer[1];

		Fft.fft(gp.proc_buffer, q_ind.SI, 1);            // forward FFT

		for (i = q_ind.SI / 2; i < q_ind.SI; i++)            // extract real part of data 
			gp.rows_real[i][ n] = (float)(gp.proc_buffer[2 * (i - q_ind.SI / 2)] / ((double)q_ind.SI / 2.0));
		for (i = 0; i < q_ind.SI / 2; i++)
			gp.rows_real[i][ n] = (float)(gp.proc_buffer[2 * (i + q_ind.SI / 2)] / ((double)q_ind.SI / 2.0));

		for (i = q_ind.SI / 2; i < q_ind.SI; i++)            // extract imaginary part of data 
			gp.irows_real[i][ n]
					= (float)(gp.proc_buffer[2 * (i - q_ind.SI / 2) + 1] / ((double)q_ind.SI / 2.0));
		for (i = 0; i < q_ind.SI / 2; i++)
			gp.irows_real[i][ n]
					= (float)(gp.proc_buffer[2 * (i + q_ind.SI / 2) + 1] / ((double)q_ind.SI / 2.0));

		//--------


		for (i = q_ind.SI / 2; i < q_ind.SI; i++)            // load real part of data 
			gp.proc_buffer[2 * (i - q_ind.SI / 2)] = gp.rows_imag[i][ n];
		for (i = 0; i < q_ind.SI / 2; i++)
			gp.proc_buffer[2 * (i + q_ind.SI / 2)] = gp.rows_imag[i][ n];


		for (i = q_ind.SI / 2; i < q_ind.SI; i++)            // zero imaginary part of data              
			gp.proc_buffer[2 * (i - q_ind.SI / 2) + 1] = 0.0;
		for (i = 0; i < q_ind.SI / 2; i++)
			gp.proc_buffer[2 * (i + q_ind.SI / 2) + 1] = 0.0;

		Fft.fft(gp.proc_buffer, q_ind.SI, -1);             // inverse FFT


		for (i = q_ind.SI; i < 2 * q_ind.SI; i++)
			gp.proc_buffer[i] = 0.0;

		gp.proc_buffer[0] = 0.5 * gp.proc_buffer[0];           // first point correction
		gp.proc_buffer[1] = 0.5 * gp.proc_buffer[1];

		Fft.fft(gp.proc_buffer, q_ind.SI, 1);                   // forward FFT

		for (i = q_ind.SI / 2; i < q_ind.SI; i++)               // extract real part of data 
			gp.rows_imag[i][ n] = (float)(gp.proc_buffer[2 * (i - q_ind.SI / 2)] / ((double)q_ind.SI / 2.0));
		for (i = 0; i < q_ind.SI / 2; i++)
			gp.rows_imag[i][ n] = (float)(gp.proc_buffer[2 * (i + q_ind.SI / 2)] / ((double)q_ind.SI / 2.0));

		for (i = q_ind.SI / 2; i < q_ind.SI; i++)               // extract imaginary part of data 
			gp.irows_imag[i][ n]
					= (float)(gp.proc_buffer[2 * (i - q_ind.SI / 2) + 1] / ((double)q_ind.SI / 2.0));
		for (i = 0; i < q_ind.SI / 2; i++)
			gp.irows_imag[i][ n]
					= (float)(gp.proc_buffer[2 * (i + q_ind.SI / 2) + 1] / ((double)q_ind.SI / 2.0));

	}
//--------------------------------------------------------------------------------------------------
}
