
/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

package spinworks;

/**
 * General utility routines for 2D
 * @author marat
 */
public class Util_2D 
{
//-------------------------------------------------------------------------------------------------
/**
 * Extract an fid from a pre-loaded 2D "ser" matrix.
 * @param n the index of the row to extract
 * @param gp global (now workspace) parameter object
 * @param p acquisition dimension set data parameters
 * @param q acquisition dimension status data parameters
 * @return true if OK false otherwise
 */
	public static boolean load_fid_2D_dir(int n, global_par gp,
			data_par p, data_par q)
	{
		if (gp.ser == null) return false;		        // NO DATA!!!
		int i;

		Util.zero_array(gp.proc_buffer);

		if (q.TD <= (2 * p.SI))							// truncate if td longer than
			q.td_eff = q.TD;							// size (rare case!)
		else
			q.td_eff = 2 * p.SI;

		for (i = 0; i < q.td_eff; i++)					// extract actual data 
			gp.proc_buffer[i] = gp.ser[n][i];

		// invert imaginaries to reverse data, if required.
		if (q.data_type == SW_main.data_types.VNMR
				|| q.data_type == SW_main.data_types.TECMAG
				|| q.data_type == SW_main.data_types.JEOL
				|| p.REVERSE)
			for (i = 1; i < q.td_eff; i = i + 2)
				gp.proc_buffer[i] = -gp.proc_buffer[i];

		if (q.aq_mode == SW_main.det_modes.SEQ)
			for (i = 0; i < q.td_eff; i = i + 4)
			{
				gp.proc_buffer[i + 1] = -(gp.proc_buffer[i + 1]);	// this is the Redfield trick: 
				gp.proc_buffer[i + 2] = -(gp.proc_buffer[i + 2]);	// mult data by 1 -1 -1 1 
			}
		return true;									// something more meaningful later
	}

//-------------------------------------------------------------------------------------------------
/**
 * Extract an column (F1 fid) from a processed "rows" matrix.
 * @param n index of the column to extract
 * @param rows the matrix (quadrant) to extract the row from (1 of up to 4
 * @param gp global (now workspace) parameter object
 * @param p_ind indirect dimension set data parameters
 * @param q_ind indirect dimension status data parameters
 * @return true if OK, false otherwise
 */
	public static boolean load_fid_2D_ind(int n, float[][] rows,  global_par gp,
			data_par p_ind, data_par q_ind)
	{
		int i, td;

		if (rows == null || gp.proc_buffer == null)
		{
			Util.show_and_log(
					"load_fid_2D_f1(): Space allocation error: rows matrix or proc_buffer null");
			return false;
		}

		Util.zero_array(gp.proc_buffer);

		//		set effective td and extract data from column
		if (q_ind.TD <= (q_ind.SI))								 // truncate if td longer than
			td = q_ind.TD;										 // size (rare case!)
		else
			td = q_ind.SI;

		for (i = 0; i < td; i = i + 2)							 // load actual data 
		{
			gp.proc_buffer[i] = (double)rows[i][n];
			if(p_ind.REVERSE)                                    // reverse data if
				gp.proc_buffer[i + 1] = -(double)rows[i + 1][n]; // requested by inverting
			else                                                 // alternate data points
				gp.proc_buffer[i + 1] = (double)rows[i + 1][n];
		}

		if (p_ind.MC2 == SW_main.det_modes_IND.STATES_TPPI)
		{
			for (i = 0; i < td; i = i + 4)
			{
				gp.proc_buffer[i + 2] = -(gp.proc_buffer[i + 2]); //  convert to 
				gp.proc_buffer[i + 3] = -(gp.proc_buffer[i + 3]); //  equivalent to STATES
			}
		}

		return true; // something meaningful later (need some try/catch blocks)
	}

//------------------------------------------------------------------------------------------------
/**
 * extract an column (F1 fid) from a processed "rows" matrix.  Special version for 
 * magnitude (COSY/HMBC) type data.
 * @param n the index of the column to extrac
 * @param gp a global (workspace) parameter structure
 * @param p set data parameters, direct dimension
 * @param q status data parameters, direct dimension
 * @param p_ind set data parameters, indirect dimension
 * @param q_ind status data parameters, indirect dimension
 * @return true if OK, false otherwise
 */
	public static boolean load_fid_2D_ind_QF(int n, global_par gp,
			data_par p, data_par q, data_par p_ind, data_par q_ind)
	{
		int i, td;
		boolean reverse;

		if (gp.rows_real == null || gp.rows_imag == null
				|| gp.proc_buffer == null)
		{
			Util.show_and_log(
					"load_fid_2D_f1_QF(): Space allocation error: rows matrix or proc_buffer null");
			return false;
		}

		//  for two quadrant transform since direct dimension has already been reversed in Varian data
		//  must flip sign of F1 reverse
		if (q.data_type == SW_main.data_types.VNMR)
			reverse = !p_ind.REVERSE;
		else
			reverse = p_ind.REVERSE;

		Util.zero_array(gp.proc_buffer);

		//		set effective td and extract data from column
		if (q_ind.TD <= p_ind.SI)			// truncate if td longer than
			td = 2 * q_ind.TD;				// size (rare case!)
		else
			td = 2 * p_ind.SI;

		for (i = 0; i < td; i = i + 2)		// extract actual data 
		{
			gp.proc_buffer[i] = (double)gp.rows_real[i / 2][n];
			if(reverse)
				gp.proc_buffer[i + 1] = -(double)gp.rows_imag[i / 2][n];
			else
				gp.proc_buffer[i + 1] = (double)gp.rows_imag[i / 2][n];
		}

		return true;						// something more meaningful later
	}
	
//-------------------------------------------------------------------------------------------------
/**
 * Load a specific block of the ser array to be displayed, redundant!!! ???
 * @param block
 * @param gp
 * @param q
 */
	public static void load_display_2D_ser(int block, global_par gp, data_par q)
	{
		gp.display_mode = SW_main.modes.MODE_FID;

		if (gp.ser == null) return;
		int n;

		if (q.TD >= q.SI) n = q.SI;
		else n = q.TD;

		int j = 0;
		try
		{
			for (int i = 0; i < n; i = i + 2)	//load fid into data array (for display)
			{
				gp.data_array[j] = gp.ser[block][i];
				gp.i_data_array[j] = gp.ser[block][i + 1];
				j++;
			}
		}
		catch (Exception ex) 
		{   Util.show_and_log("load_display_2D_ser(): ", ex.getMessage());   }
	}

//-------------------------------------------------------------------------------------------------  
/**
 * Load a block (row) of the processed 2D for 1D display.
 * @param block the block to load
 * @param gp a SpinWorks global parameter bloc
 * @param p set data parameters
 * @param q status data parameters
 */
	public static void load_display_2D_direct(int block, global_par gp, data_par p, data_par q)
	{
		Util.free_proc(gp);
		Util.alloc_proc(gp, p, q);

		if (gp.rows_real != null)
			for (int i = 0; i < q.SI; i++)
				gp.data_array[i] = gp.rows_real[block][i];


		if (gp.rows_imag != null)
			for (int i = 0; i < q.SI; i++)
				gp.i_data_array[i] = gp.rows_imag[block][i];
	}

//-------------------------------------------------------------------------------------------------
/**
 * Store a row of the 2D array after processing in a 1D mode.
 * @param row index of the row to stor
 * @param gp a SpinWorks global parameter object
 * @param p set parameters in the direct dimension
 * @param q status parameters in direct dimension
 */
	public static void store_row_direct(int row, global_par gp, data_par p, data_par q)
	{
		if (gp.rows_real != null)
			for (int i = 0; i < q.SI; i++)
				gp.rows_real[row][i] = (float)gp.data_array[i];

		if (gp.rows_imag != null)
			for (int i = 0; i < q.SI; i++)
				gp.rows_imag[row][i] = (float)gp.i_data_array[i];

		Util.free_proc(gp);
		Util.alloc_proc(gp, p, q);
	}
	// TO Do  check efficiency!  go directly from proc buffer to 2D matrix!

//-------------------------------------------------------------------------------------------------
/**
 * "free" space occupied by rows of a 2D data set (flag quadrants as null). 
 * All mermory freeing should be done by JVM garbage collection. 
 * However, this method flags quadrants as being unused. 
 * @param rr the real real quadrant
 * @param ri the real imaginary quadrant
 * @param ir the imaginary real quadrant
 * @param ii the imaginary imaginary quadrant
 * @param gp a SpinWorks global parameter object
 */
	public static void free_rows(boolean rr, boolean ri, boolean ir, boolean ii, global_par gp)		
	{
		if (gp.rows_real != null && rr == true)
			gp.rows_real = null;

		if (gp.rows_imag != null && ri == true)
			gp.rows_imag = null;

		if (gp.irows_real != null && ir == true)
			gp.irows_real = null;

		if (gp.irows_imag != null && ii == true)
			gp.irows_imag =  null;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Allocate space occupied by rows of a 2D data set. 
 * For now, all four quadrants  We could get away with 2 quadrants for COSY type data
 * @param row_size
 * @param col_size
 * @param gp A SpinWorks global parameter object
 * @param q Status acquisition parameters, direct dimension
 * @param q_ind Status acquisition parameters, indirect dimension
 */
	public static void alloc_rows(int row_size, int col_size, global_par gp, 
			data_par q, data_par q_ind)
	{
		gp.rows_real = new float[row_size][col_size];
		gp.rows_imag = new float[row_size][col_size];
		gp.irows_real = new float[row_size][col_size];
		gp.irows_imag = new float[row_size][col_size];

		//		for good measure, lets zero the arrays
		for (int i = 0; i < row_size; i++)
			for (int j = 0; j < col_size; j++)
			{
				gp.rows_real[i][j] = 0.0F;
				gp.rows_imag[i][j] = 0.0F;
				gp.irows_real[i][j] = 0.0F;
				gp.irows_imag[i][j] = 0.0F;
			}
		//		Since the space is now allocated, we can update the status parameter
		//		as well
		q.SI = col_size;
		q_ind.SI = row_size;
	}

//-------------------------------------------------------------------------------------------------
/**
 * flag the "ser" array as null
 * @param gp a SpinWorks global parameter object
 */
	public static void free_ser(global_par gp)
	{   gp.ser = null;  }

//-------------------------------------------------------------------------------------------------
/**
 * zero a 2D floating point array
 * @param row_size
 * @param col_size
 * @param array  The 2D array
 */
	public static void zero_array(int row_size, int col_size, float[][] array)
	{
		for (int i = 0; i < row_size; i++)
			for (int j = 0; j < col_size; j++)
				array[i][j] = 0.0F;               
	}

//-------------------------------------------------------------------------------------------------
/**
 * Find the global absolute maximum of a 2D data set.
 * @param array
 * @param sizeDirect
 * @param sizeIndirect
 * @return The maximum value in the 2D matrix
 */
	public static double max_2D(float array[][], int sizeDirect, int sizeIndirect)
	{
		double max = 0.0;

		try
		{
			for (int i = 0; i < sizeIndirect; i++)
				for (int j = 0; j < sizeDirect; j++)
					if (Math.abs((double)array[i][j]) > max)
						max = Math.abs((double)array[i][j]);
		}
		catch (Exception ex)
		{
			Util.log_error("max_2D(): " + ex.getMessage());
			return 1.0e4;       // return a default max for scaling
		}
		return max;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Find the maximum value in a 2D matrix
 * @param matrix the 2D matrix (usually rr quadrant)
 * @param x_start start index in x (direct) dimension
 * @param y_start start index in y (indirect) dimension
 * @param x_end end index in x (direct) dimension
 * @param y_end end index in y (indirect) dimension
 * @return the absolute maximum value
 */
	public static double max_2D(float[][] matrix, 
			int x_start, int y_start, int x_end, int y_end )
	{
		double max = 0.0;

		try
		{
			for (int i = y_start; i < y_end; i++)
				for (int j = x_start; j < x_end; j++)
					if (Math.abs((double)matrix[i][j]) > max)
						max = Math.abs((double)matrix[i][j]);
		}
		catch (Exception ex)
		{
			Util.log_error("max_2D(): " + ex.getMessage());
			return 1.0e4;       // return a default max for scaling
		}
		return max;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Find the maximum in the ser (2D FID) file.
 * @param gp a SpinWorks global parameter object
 * @param q status parameters in direct dimension
 * @param q_ind status parameters in indirect dimension
 * @return the maximum value in matrix
 */
	public static double max_ser(global_par gp, data_par q, data_par q_ind)
	{
		double max = 0.0;

		try
		{
			for (int i = 0; i < q_ind.TD; i++)					// now read the data 
				for (int j = 0; j < q.TD; j++)
					if (Math.abs(gp.ser[i][j]) > max)
						max = Math.abs(gp.ser[i][j]);
		}
		catch (Exception ex)    // probably null ser or index out if range
		{
			Util.log_error("max_ser(): " + ex.getMessage());
			return 1.0e4;       // return a default max for scaling
		}

		return max;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Calculate the approx. noise level in a 2D matrix from the edges of the matrix.
 * @param gp a SpinWorks global parameter object
 * @param q status parameters in the direct (acquisition) dimension
 * @param q_ind status parameters in the indirect (evolution) dimension
 * @return RMS noise level (approximate)
 */
	public static double noise_2D(global_par gp, data_par q, data_par q_ind)
	{
		double sum, avgl, avgr, rmsl, rmsr;
		int i, j;
		try
		{
			sum = 0.0;
			for (i = 0; i < 10; i++)
				for (j = 0; j < 4; j++)
					sum = sum + gp.rows_real[j][i];
			avgr = sum / 40.0;

			sum = 0.0;
			for (i = q.SI - 10; i < q.SI; i++)
				for (j = q_ind.SI - 4; j < q_ind.SI; j++)
					sum = sum + gp.rows_real[j][i];
			avgl = sum / 40.0;

			sum = 0.0;
			for (i = 0; i < 10; i++)
				for (j = 0; j < 4; j++)
					sum = sum + (avgr - gp.rows_real[j][i]) * (avgr - gp.rows_real[j][i]);
			rmsr = Math.sqrt(sum) / 40.0;

			sum = 0.0;
			for (i = q.SI - 10; i < q.SI; i++)
				for (j = q_ind.SI - 4; j < q_ind.SI; j++)
					sum = sum + (avgl - gp.rows_real[j][i]) * (avgl - gp.rows_real[j][i]);
			rmsl = Math.sqrt(sum) / 40.0;
		}
		catch (Exception ex)
		{
			Util.log_error(ex.getMessage());
			return 0.0;
		}
		return ((rmsl + rmsr) / 2.0);
	}

//-------------------------------------------------------------------------------------------------
/**
 * Calculate default levels for 2D display.
 * @param gp a SpinWorks global parameter object
 */
	public static void calc_default_levels_2D(global_par gp)
	{
		double floor = Math.abs(gp.data_max_2D) / 10;
		if (floor < 10.0 * gp.data_noise)
			floor = 10.0 * gp.data_noise;

		if (gp.level_spacing_increment == SW_main.level_spacing.LINEAR)
		{
			double step = (Math.abs(gp.data_max_2D) - floor) / (gp.n_levels + 2);

			for (int i = 0; i < gp.n_levels; i++)
			{
				gp.pos_level[i].value = floor + (i * step);
				gp.neg_level[i].value = -(floor + (i * step));
			}
		}
		else    // log spacing
		{
			gp.pos_level[0].value = floor;
			gp.neg_level[0].value = - floor;

			for (int i = 1; i < gp.n_levels; i++)
			{
				gp.pos_level[i].value = gp.pos_level[i - 1].value * gp.level_spacing_factor;
				gp.neg_level[i].value = gp.neg_level[i - 1].value * gp.level_spacing_factor;
			}
		}
	}

//------------------------------------------------------------------------------------------------
/**
 * Re-scale the 2D levels floor (lowest level).
 * @param factor the scaling factor
 * @param gp a SpinWorks global parameter object
 */
	public static void rescale_floor_2D(double factor, global_par gp)
	{
		double floor = gp.pos_level[0].value * factor;
		double range = Math.abs(gp.pos_level[gp.n_levels - 1].value - floor);

		if (gp.level_spacing_increment == SW_main.level_spacing.LINEAR)
		{               
			double step = range / (double)(gp.n_levels - 1);

			for (int i = 0; i < gp.n_levels; i++)
			{
				gp.pos_level[i].value = floor + (i * step);
				gp.neg_level[i].value = -(floor + (i * step));
			}
		}
		else	// LOG
		{
			gp.pos_level[0].value = gp.pos_level[0].value * factor;
			gp.neg_level[0].value = gp.neg_level[0].value * factor;

			for (int i = 1; i < gp.n_levels; i++)
			{
				gp.pos_level[i].value = gp.pos_level[i - 1].value * gp.level_spacing_factor;
				gp.neg_level[i].value = gp.neg_level[i - 1].value * gp.level_spacing_factor;
			}
		}
	}

//------------------------------------------------------------------------------------------------	
/**
 * Re-scale the 2D levels ceiling (upper level)
 * @param factor the scaling factor
 * @param gp a SpinWorks global parameter object
 */
	public static void rescale_ceiling_2D(double factor, global_par gp)
	{
		double floor = gp.pos_level[0].value;
		double range = Math.abs(((gp.pos_level[gp.n_levels - 1].value) * factor) - floor);
		double step = range / (double)(gp.n_levels - 1);

		if (gp.level_spacing_increment == SW_main.level_spacing.LINEAR)
		{
			for (int i = 0; i < gp.n_levels; i++)
			{
				gp.pos_level[i].value = floor + (i * step);
				gp.neg_level[i].value = -(floor + (i * step));
			}
		}
		else	// LOG
		{
			if (factor > 1.0)
				gp.level_spacing_factor = gp.level_spacing_factor + 0.02;
			else if (factor < 1.0 && gp.level_spacing_factor > 1.04)
				gp.level_spacing_factor = gp.level_spacing_factor - 0.02;

			for (int i = 1; i < gp.n_levels; i++)
			{
				gp.pos_level[i].value = gp.pos_level[i - 1].value * gp.level_spacing_factor;
				gp.neg_level[i].value = gp.neg_level[i - 1].value * gp.level_spacing_factor;
			}
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * Given an intensity, return the corresponding level.
 * @param intensity intensity of point in data matrix
 * @param gp a SpinWorks global parameter object
 * @return corresponding level, -1 if below lowest level
 */
	public static int calc_level(float intensity, global_par gp)
	{
		int i;

		for (i = gp.n_levels - 1; i >= 0; i--)
			if (Math.abs(intensity) > (float)gp.pos_level[i].value) break;

		if (Math.abs(intensity) < (float)gp.pos_level[0].value)
			return -1;
		else
			return i;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Calculate default colours for 2D display.
 * @param gp a SpinWorks global parameters structure
 */
	public static void calc_default_colours(global_par gp)
	{
		int step = (int)((double)255 / (double)(gp.n_levels - 1));
		int mix_in;

		for (int i = 0; i < gp.n_levels; i++)
		{

			if (i > 8)
				mix_in = (i - 8) * (step);
			else
				mix_in = 0;

			gp.pos_level[i].red = 255;				// yellow/red, sort of
			gp.pos_level[i].green = i * step;
			gp.pos_level[i].blue = mix_in;

			gp.neg_level[i].red = mix_in;			// blue/green, sort of
			gp.neg_level[i].green = 255 - (step / 2) * i;
			gp.neg_level[i].blue = i * step;
		}
	}

//-------------------------------------------------------------------------------------------------
/**
 * Calculate colours for a positive only (magnitude) display. 
 * @param gp a SpinWorks global parameters structure
 */
	public static void calc_pos_only_colors(global_par gp)
	{
		gp.pos_level[0].red = 255;			// MAGENTA
		gp.pos_level[0].green = 0;
		gp.pos_level[0].blue = 255;

		gp.pos_level[1].red = 150;			// INDIGO
		gp.pos_level[1].green = 0;
		gp.pos_level[1].blue = 255;

		gp.pos_level[2].red = 0;			// BLUE
		gp.pos_level[2].green = 0;
		gp.pos_level[2].blue = 255;

		gp.pos_level[3].red = 0;			// CYAN
		gp.pos_level[3].green = 255;
		gp.pos_level[3].blue = 255;

		gp.pos_level[4].red = 0;			// BLUE_GREEN
		gp.pos_level[4].green = 255;
		gp.pos_level[4].blue = 150;

		gp.pos_level[5].red = 0;			// GREEN
		gp.pos_level[5].green = 255;
		gp.pos_level[5].blue = 0;

		gp.pos_level[6].red = 150;			// Bug-guts green
		gp.pos_level[6].green = 255;
		gp.pos_level[6].blue = 0;

		gp.pos_level[7].red = 255;			// YELLOW
		gp.pos_level[7].green = 255;
		gp.pos_level[7].blue = 0;

		gp.pos_level[8].red = 255;			// ORANGE
		gp.pos_level[8].green = 150;
		gp.pos_level[8].blue = 0;

		gp.pos_level[9].red = 255;			// ORANGE_RED
		gp.pos_level[9].green = 75;
		gp.pos_level[9].blue = 0;

		gp.pos_level[10].red = 255;			// RED
		gp.pos_level[10].green = 0;
		gp.pos_level[10].blue = 0;

		gp.pos_level[11].red = 255;			// LIGHT_RED
		gp.pos_level[11].green = 50;
		gp.pos_level[11].blue = 50;

		gp.pos_level[12].red = 255;			// lighter Red
		gp.pos_level[12].green = 100;
		gp.pos_level[12].blue = 100;

		gp.pos_level[13].red = 255;			//  Lighter
		gp.pos_level[13].green = 125;
		gp.pos_level[13].blue = 125;

		gp.pos_level[14].red = 255;			// even lighter
		gp.pos_level[14].green = 150;
		gp.pos_level[14].blue = 150;

		gp.pos_level[15].red = 255;			// lightest
		gp.pos_level[15].green = 175;
		gp.pos_level[15].blue = 175;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Calculate 2D scaling factors, etc.
 * @param gp a SpinWorks global parameters structure
 * @param q status data parameters, direct dimension
 * @param q_ind status data parameters, indirect dimension
 */
	public static void calc_scale_factors_2D(global_par gp, data_par q, data_par q_ind)
	{
		if (gp.rows_real != null)
		{
			gp.data_max = Math.abs(max_2D(gp.rows_real, q.SI, q_ind.SI));
			gp.data_max_2D = gp.data_max;
		}
		else
		{
			gp.data_max = Math.abs(Util.dscan(gp.data_array, q.SI, 0, q.SI));
		}

		if (gp.rows_real!= null)
			gp.data_noise = noise_2D(gp, q, q_ind);
		else
			gp.data_noise = gp.data_max / 10;

		if (gp.data_max < 1.0) gp.data_max = 10000000.0;
		gp.expt_scale = gp.dysize * 0.80 / gp.data_max;
	}

//-------------------------------------------------------------------------------------------------
/**
 * copy a row from 2D matrix to trace
 * @param row row index
 * @param gp a SpinWorks global parameters structure
 * @param q status data parameters, direct dimension
 */
	public static void copy_F2_to_trace(int row, global_par gp, data_par q)	      // direct
	{
		int i;

		if (gp.f2_trace != null)
			for (i = 0; i < q.SI; i++)
				gp.f2_trace[i] = gp.rows_real[row][i];
	}

/**
 * copy a column from 2D matrix to trace
 * @param column column index
 * @param gp a SpinWorks global parameters structure
 * @param q_ind status data parameters, indirect dimension
 */
	public static void copy_F1_to_trace(int column, global_par gp, data_par q_ind)  // indirect
	{
		int i;

		if (gp.f1_trace != null)
			for (i = 0; i < q_ind.SI; i++)
				gp.f1_trace[i] = gp.rows_real[i][column];
	}

//-------------------------------------------------------------------------------------------------
/**
 * Apply IUPAC type referencing in the current indirect dimension.  Uses value set in 
 * processing menu.
 * @param gp a SpinWorks global parameter object
 * @param p_ind set parameters in indirect dimension
 * @param q_ind status parameters in indirect dimension
 */
	public static void auto_ref_ind(global_par gp, data_par p_ind, data_par q_ind)
	{
		p_ind.SF = q_ind.SF;
		q_ind.offset = (q_ind.SFO1 - q_ind.SF) * 1.0e06 + (q_ind.SW * q_ind.SFO1) / 2.0;

		gp.freq1_indirect = q_ind.offset;
		gp.freq2_indirect = q_ind.offset - (q_ind.SW * q_ind.SFO1);

		gp.new_2D_view = true;
	}

//------------------------------------------------------------------------------------------------
/**
 * Calculate a "skyline" projection in the direct dimension.
 * @param gp a SpinWorks global parameter object
 * @param q status parameters in the direct dimension
 * @param q_ind status parameters in the indirect dimension
 */
	public static void calc_F2_projection(global_par gp, data_par q, data_par q_ind)
	{
		int i, j;

		gp.size_f2_trace = q.SI;

		gp.f2_trace = new double[gp.size_f2_trace];

		for (i = 0; i < q.SI; i++) gp.f2_trace[i] = 0.0;

		for (j = 0; j < q.SI; j++)
			for (i = 0; i < q_ind.SI; i++)             
			{
				if (Math.abs(gp.rows_real[i][j]) > Math.abs(gp.f2_trace[j]))
					gp.f2_trace[j] = gp.rows_real[i][j];
			}

		gp.hzperpt_f2_trace = q.hzperpt;
		gp.offset_f2_trace = q.offset;
		gp.f2_trace_scale = ((double)gp.ysize_2D * 0.80) / gp.data_max_2D;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Calculate a "skyline" projection in the indirect dimension.
 * @param gp a SpinWorks global parameter object
 * @param q status parameters in direct dimension
 * @param q_ind status parameters in indirect dimension
 */
	public static void calc_F1_projection(global_par gp, data_par q, data_par q_ind)
	{
		int i, j;

		gp.size_f1_trace = q_ind.SI;

		gp.f1_trace = new double[gp.size_f1_trace];

		for (i = 0; i < q_ind.SI; i++) gp.f1_trace[i] = 0;

		for (i = 0; i < q_ind.SI; i++)
			for (j = 0; j < q.SI; j++)
				if (Math.abs(gp.rows_real[i][j]) > Math.abs(gp.f1_trace[i]))
					gp.f1_trace[i] = gp.rows_real[i][j];

		gp.hzperpt_f1_trace = q_ind.hzperpt;
		gp.offset_f1_trace = q_ind.offset;
		gp.f1_trace_scale = ((double)gp.xsize_2D * 0.80) / gp.data_max_2D;
	}

//------------------------------------------------------------------------------------------------
/**
 * Calculate a "sum" projection in the direct dimension.
 * @param gp a SpinWorks global parameter object
 * @param q status parameters in the direct dimension
 * @param q_ind status parameters in the indirect dimension
 */
	public static void calc_F2_sum(global_par gp, data_par q, data_par q_ind)
	{
		int i, j;

		gp.size_f2_trace = q.SI;

		gp.f2_trace = new double[gp.size_f2_trace];

		for (i = 0; i < q.SI; i++) gp.f2_trace[i] = 0;

		for (j = 0; j < q.SI; j++)
			for (i = 0; i < q_ind.SI; i++)
				gp.f2_trace[j] 
						= (gp.f2_trace[j] 
								+ (double)gp.rows_real[i][j]/ (double)q_ind.SI);

		gp.hzperpt_f2_trace = q.hzperpt;
		gp.offset_f2_trace = q.offset;
		gp.f2_trace_scale = ((double)gp.ysize_2D * 0.80) / gp.data_max_2D;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Calculate a "sum" projection in the indirect dimension.
 * @param gp a SpinWorks global parameter object
 * @param q status parameters in direct dimension
 * @param q_ind status parameters in indirect dimension
 */
	public static void calc_F1_sum(global_par gp, data_par q, data_par q_ind)
	{
		int i, j;

		gp.size_f1_trace = q_ind.SI;

		gp.f1_trace = new double[gp.size_f1_trace];

		for (i = 0; i < q_ind.SI; i++) gp.f1_trace[i] = 0;

		for (i = 0; i < q_ind.SI; i++)
			for (j = 0; j < q.SI; j++)
				gp.f1_trace[i] 
						= (gp.f1_trace[i] 
								+ (double)gp.rows_real[i][j]/(double)q.SI);

		gp.hzperpt_f1_trace = q_ind.hzperpt;
		gp.offset_f1_trace = q_ind.offset;
		gp.f1_trace_scale = ((double)gp.xsize_2D * 0.80) / gp.data_max_2D;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Integrate a region of a 2D spectrum.
 * @param p1 start point in direct dimension
 * @param p2 end point in direct dimension
 * @param p1_ind start point in indirect dimension
 * @param p2_ind end point in indirect dimension
 * @param gp SpinWorks global parameter object
 * @param q status parameters in indirect dimension
 * @param q_ind  status parameters in indirect dimension
 * @return the integral
 */
	public static double integ_2D(int p1, int p2, int p1_ind, int p2_ind,
			global_par gp, data_par q, data_par q_ind)
	{
		int temp, i, j;
		double sum;

		if (p1 > p2)  // check ordering
		{
			temp = p1;
			p1 = p2;
			p2 = temp;
		}

		if (p1_ind > p2_ind)  // check ordering
		{
			temp = p1_ind;
			p1_ind = p2_ind;
			p2_ind = temp;
		}

		sum = 0.0;
		for (i = p1; i <= p2 && i > 0 && i < q.SI; i++)
			for (j = p1_ind; j <= p2_ind && j > 0 && j < q_ind.SI; j++)
				sum = sum + (double)gp.rows_real[j][i];

		return (sum);
	}

//-------------------------------------------------------------------------------------------------
/**
 * Scan a series of extracted rows, looking for absolute maximum, used to find pivot point for
 * 2D phasing.
 * @param gp a SpinWorks global parameter object
 * @param l the start index in the direct dimension
 * @param m the end index in the direct dimension
 * @index of the absolute maximum
 */
	public static int scan_rows(global_par gp, int l, int m)
	{
		int k = 0;
		double max = 0.0;

		for (int i = 0; i < gp.nslices; i++)
		{
			for (int j = l; j < m; j++)
			{
				if ((gp.rows_real[gp.slice_no[i]][j] * gp.rows_real[gp.slice_no[i]][j]
						+ gp.rows_imag[gp.slice_no[i]][j] * gp.rows_imag[gp.slice_no[i]][j])
						> max)
				{
					max = gp.rows_real[gp.slice_no[i]][j] * gp.rows_real[gp.slice_no[i]][j]
							+ gp.rows_imag[gp.slice_no[i]][j] * gp.rows_imag[gp.slice_no[i]][j];
					k = j;
				}
			}
		}
		return k;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Scan a series of extracted columns, looking for absolute maximum, used to find pivot point for 
 * 2D phasing.
 * @param gp a SpinWorks global parameter structure
 * @param l the start index in the indirect dimension
 * @param m the end index in the indirect dimension
 * @return index of the absolute maximum
 */
	public static int scan_columns(global_par gp, int l, int m)
	{
		int k = 0;
		double max = 0.0;

		for (int i = 0; i < gp.nslices; i++)
		{
			for (int j = l; j < m; j++)
			{
				if ((gp.rows_real[j][gp.slice_no[i]] * gp.rows_real[j][gp.slice_no[i]]
						+ gp.irows_real[j][gp.slice_no[i]] * gp.irows_real[j][gp.slice_no[i]]) 
						> max)
				{
					max = gp.rows_real[j][gp.slice_no[i]] * gp.rows_real[j][gp.slice_no[i]]
							+ gp.irows_real[j][gp.slice_no[i]] * gp.irows_real[j][gp.slice_no[i]];
					k = j;
				}
			}
		}
		return k;
	}

//-------------------------------------------------------------------------------------------------
/**
 * Scans the 2D array for the maximum in the area row +/- range, col +/- range.
 * @param array the array (matrix) to scan
 * @param row row index
 * @param col column index
 * @param range scan range (points)
 * @param q status parameters in acquisition (direct) dimension
 * @param q_ind status parameters in evolution (indirect) dimension
 * @return maximum value (+ or -)
 */
	public static double find_peak(float[][] array, int row, int col, int range,
			data_par q, data_par q_ind)
	{
		float max = 0.0F;

		if(row + range >= q_ind.SI || row - range < 0
				|| col + range >= q.SI || col - range < 0)
			return (double)array[row][col];

		for(int i = row - range; i <= row + range; i++)
			for(int j = col - range; j <= col + range; j++)
				if(Math.abs(array[i][j]) > Math.abs(max))
					max = array[i][j];

		return (double) max;
	}
	
//-------------------------------------------------------------------------------------------------	
/**
 * return the assigned 2D peak index closest to the given frequency (in Hz)
 * @param gp a SpinWorks workspace parameter object
 * @param f2 the F2 (acquisition dimension) frequency in Hz
 * @param f1 the F1 (indirect dimension) frequency in Hz
 * @param tol the search tolerance in pixels
 * @return the index of the assigned peak, -1 if not found.
 */
	public static int closestPeak(global_par gp, double f2, double f1, int tol)
	{
		double tol_f2 = gp.hzperpix * (double)tol;
		double tol_f1 = gp.hzperpix_f1 * (double)tol;

		for (int i = 0; i < gp.n_peaks; i++)
		{
			if (Math.abs(gp.peaks[i].freq - f2) < tol_f2
					&& Math.abs(gp.peaks[i].freq_f1 - f1) < tol_f1)
				return i;
		}
		return -1;
	}
	
//-------------------------------------------------------------------------------------------------	  
	
	public static int selectTrace_direct(global_par gp, data_par q, int x, int y, int tol)
	{
		if(gp.nslices == 0) return 0;		

		double hzperpix = (gp.freq1-gp.freq2)/(double)(gp.xsize_2D-1);

		for(int i = 0; i < gp.nslices; i++)
		{
			// get the index into the spectrum from the cursor x value
			int index = (int) Math.round((q.offset
					- (gp.freq1 - (double)(x - gp.left_2D) * hzperpix)) / q.hzperpt);

			if(index < 0) index = 0;
			else if(index >= q.SI) index = q.SI - 1;
						
			int datay = gp.dysize 
					- (int)(gp.rows_real[gp.slice_no[i]][index] * gp.row_scale[i]) 
					- gp.bottom_2D
					- 50 - (i * 50);
			
			if(Math.abs(y - datay) < tol)
				return i;
		}

		return -1;
	}
	
//-------------------------------------------------------------------------------------------------		
	public static int selectTrace_indirect(global_par gp, data_par q_ind, int x, int y, int tol)
	{
		if(gp.nslices == 0) return 0;		

		double hzperpix = (gp.freq1_indirect - gp.freq2_indirect)/(double)(gp.ysize_2D - 1);

		for(int i = 0; i < gp.nslices; i++)
		{
			// get the index into the spectrum from the cursor x value
			int index = (int) Math.round((q_ind.offset - (gp.freq1_indirect 
					- (double)(gp.dysize - gp.bottom_2D - y) * hzperpix)) / q_ind.hzperpt);

			if(index < 0) index = 0;
			else if(index >= q_ind.SI) index = q_ind.SI - 1;
						
			int datax =  (int)(gp.rows_real[index] [gp.slice_no[i]] * gp.row_scale[i]) 
					     + gp.left_2D
					     + 40 + (i * 80);
			
			if(Math.abs(x - datax) < tol)
				return i;
		}

		return -1;
	}
//-------------------------------------------------------------------------------------------------
}	// end class
