
/// <copyright>
///	Copyright (c) 2006-2016

///	Kirk Marat
///	The University of Manitoba
///	Winnipeg, Manitoba, Canada
///	R3T 2N2

///	All rights reserved.

///	This source may be distributed, provided that the author and the University are 
/// appropriately acknowledged.  However, incorporation of any part into any commercial product 
/// is prohibited without prior written authorization. 
///
/// see:  http://www.gnu.org/licenses/gpl.html
///
/// Note:
/// Distribution of this source or any of its components into the People's Republic of China 
/// is strictly prohibited.
/// 
///	The author disclaims all warranties with regard to this software, including all implied 
///	warranties of merchantability and fitness. In no event, shall the author or the University of 
///	Manitoba be liable for any special, indirect or consequential damages, or any damages whatsoever 
///	resulting from loss of use, data or profits, whether in an action of contract, negligence or 
///	other tortuous action, arising out of or in connection with the use or performance of this 
///	software.

///	All trademarks mentioned (e.g. XwinNMR, VNMR(J), Windows, etc.) are property of their 
///	respective owners.
/// </copyright>

package spinworks;

/**
 * Window functions for SpinWorks Java
 * @author marat
 */
public class Wfcn 
{
//--------------------------------------------------------------------------------------------------
	/**
	 * Apply an exponential window function.
	 * @param buffer array to be windowed
	 * @param td number of time domain points
	 * @param lb line broadening parameter in Hz.
	 * @param sw spectral width in Hz.
	 * @param real if true, this is real rather than complex data
	 */
	public static void do_em(double[] buffer, int td, double lb, double sw, boolean real)
	{
		double lb_factor;
		int i;

		lb_factor = (lb * Math.PI) / (2.0 * sw);


		if (real)	// sequentially acquired (real) data 
		{
			for (i = 0; i < td && i < buffer.length; i++)
			{
				buffer[i] =
						buffer[i] * Math.exp(-(double)i * lb_factor);
			}
		}
		else		// simultaneous (complex) data
		{
			for (i = 0; i < td-1 && i < buffer.length-1; i = i + 2)
			{
				buffer[i] =
						buffer[i] * Math.exp(-(double)i * lb_factor);
				buffer[i + 1] =
						buffer[i + 1] * Math.exp(-(double)i * lb_factor);
			}
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Apply a Lorentz to Gauss window function.
	 * @param buffer array to be windowed
	 * @param td number of time domain points
	 * @param lb line broadening parameter in Hz.  Must be negative!
	 * @param gf Gaussian factor (fraction)
	 * @param sw spectral width in Hz.
	 * @param real if true, this is real rather than complex data
	 */
	public static void do_gm(double[] buffer, int td, double lb, double gf, 
			double sw, boolean real)
	{
		int i;
		double a, b, t, aqtime, dwell;


		dwell = 1.0 / (2.0 * sw);		                // calculate dwell time 
		aqtime = dwell * (double)td;		            // and acquisition time  
		a = Math.PI * lb;
		b = -a / (2.0 * gf * aqtime);

		if (real)                                       // sequential data 
		{
			for (i = 0; i < td && i < buffer.length; i++)
			{
				t = (double)i * dwell;
				buffer[i] = buffer[i] * Math.exp((-a * t) - (b * t * t));
			}
		}
		else 										    // simultaneous data 
		{
			for (i = 0; i < td-1 && i < buffer.length-1; i = i + 2)
			{
				t = (double)i * dwell;
				buffer[i] = buffer[i] * Math.exp((-a * t) - (b * t * t));
				buffer[i + 1] = buffer[i + 1] * Math.exp((-a * t) - (b * t * t));
			}
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Apply a TRAF window function.
	 * @param buffer the array to be windowed
	 * @param td number of time domain data points
	 * @param lb "lb" parameter in Hz.
	 * @param sw spectral width in Hz.
	 * @param real true for real (sequential) data
	 */
	public static void do_traf(double[] buffer, int td, double lb, double sw, boolean real)
	{
		int i;
		double dwell, aqtime;
		double E, eps;								        // TRAF parameters



		dwell = 1.0 / (2.0 * sw);		                    // calculate dwell time 
		aqtime = dwell * (double)td;		                // and acquisition time                



		if (real)                                           // sequentially  
		{                                                   // acquired data
			for (i = 0; i < td && i < buffer.length; i++)
			{
				double t = (double)i * dwell;
				E = Math.exp(-t * lb * Math.PI);
				eps = Math.exp((t - aqtime) * lb* Math.PI);

				buffer[i] = buffer[i]
						* E / (E * E + eps * eps);

			}
		}
		else 										        // simultaneous (complex) data
		{
			for (i = 0; i < td-1 && i < buffer.length-1; i = i + 2)
			{
				double t = (double)i * dwell;
				E = Math.exp(-t * lb * Math.PI);
				eps = Math.exp((t - aqtime) * lb * Math.PI);

				buffer[i] = buffer[i]
						* E / (E * E + eps * eps);

				buffer[i + 1] = buffer[i + 1]
						* E / (E * E + eps * eps);

			}
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Apply a TRAFS window function.
	 * @param buffer
	 * @param td the array to be windowed
	 * @param lb "lb" parameter in Hz.
	 * @param sw spectral width in Hz.
	 * @param real true for real (sequential) data
	 */
	public static void do_trafs(double[] buffer, int td, double lb, double sw, boolean real)
	{
		int i;
		double dwell, aqtime;
		double E, eps;								    // TRAF parameters


		dwell = 1.0 / (2.0 * sw);			            // calculate dwell time 
		aqtime = dwell * (double)td;		            // and acquisition time  


		if (real)                                               // sequentially  
		{                                                       // acquired (real) data
			for (i = 0; i < td && i < buffer.length; i++)
			{
				double t = (double)i * dwell;
				E = Math.exp(-t * lb * Math.PI);
				eps = Math.exp((t - aqtime) * lb * Math.PI);

				buffer[i] = buffer[i]
						* ((E * E * (E + eps))
								/ (E * E * E + eps * eps * eps));

			}
		}
		else 										            // simultaneous (complex) data
		{
			for (i = 0; i < td-1 && i < buffer.length-1; i = i + 2)
			{
				double t = (double)i * dwell;
				E = Math.exp(-t * lb * Math.PI);
				eps = Math.exp((t - aqtime) * lb * Math.PI);

				buffer[i] = buffer[i]
						* ((E * E * (E + eps))
								/ (E * E * E + eps * eps * eps));

				buffer[i + 1] = buffer[i + 1]
						* ((E * E * (E + eps))
								/ (E * E * E + eps * eps * eps));
			}
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Sine multiplication
	 * @param buffer the array to be windowed
	 * @param td sine multiplication,
	 * @param ssb sine bell shift in degrees
	 */
	public static void do_sine(double[] buffer, int td, double ssb)
	{
		double factor; 
		double ofs;


		ofs = (180.0 - ssb)/180.0;

		for (int i = 0; i < td - 1 
				&& (td - 1 - i) < buffer.length; i = i + 2)
		{
			factor = Math.sin(((double)i/(double)td) * Math.PI * ofs);
			buffer[td-1-i] = buffer[td-1-i] * factor;
			buffer[td-2-i] = buffer[td-2-i] * factor;
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Sine squared multiplication.
	 * @param buffer array to be windowed
	 * @param td number to time domain points
	 * @param ssb sine bell shift in degrees
	 */
	public static void do_qsine(double[] buffer, int td, double ssb)
	{
		double factor, ofs;

		ofs = (180.0 - ssb) / 180.0;

		for (int i = 0; i < td - 1
				&& (td - 1 - i) < buffer.length; i = i + 2)
		{
			factor = Math.sin(((double)i / (double)td) * Math.PI * ofs)
					* Math.sin(((double)i / (double)td) * Math.PI * ofs);

			buffer[td - 1 - i] = buffer[td - 1 - i] * factor;
			buffer[td - 2 - i] = buffer[td - 2 - i] * factor;
		}
	}

//-------------------------------------------------------------------------------------------------
	/**
	 * Apply a pure Gaussian window function.
	 * @param buffer array to be windowed
	 * @param td number of time domain points
	 * @param lb line broadening parameter in Hz. not GB!
	 * @param sw spectral width in Hz.
	 * @param real if true, this is real rather than complex data
	 */
	public static void do_gaussian(double[] buffer, int td, double lb, double sw, boolean real)
	{
		int i;
		double dwell;
		double T2_squared;


		dwell = 1.0 / (2.0 * sw);		                                // calculate dwell time  
		T2_squared = (0.707107 / lb) * (0.707107 / lb);


		if (real)                                                       // sequentially acquired 
		{                                                               // (real) data 
			for (i = 0; i < td && i < buffer.length; i++)
			{
				double t = (double)i * dwell;
				buffer[i] = buffer[i]
						* Math.exp((t * t) / (-T2_squared));
			}
		}
		else 										// simultaneous (complex) data
		{
			for (i = 0; i < td-1 && i < buffer.length-1; i = i + 2)
			{
				double t = (double)i * dwell;
				buffer[i] = buffer[i]
						* Math.exp((t * t) / (-T2_squared));

				buffer[i + 1] = buffer[i + 1]
						* Math.exp((t * t) / (-T2_squared));
			}
		}
	}
	
//-------------------------------------------------------------------------------------------------
}	// end class
